#include "Bear.h"
#include <float.h>
#include <limits.h>
#include <math.h>

#define SUBSTANTIAL_ 0.1
#define MIN_COMPOSITE_US_ 5000000
#define MAX_WORSE_ 3

typedef struct {
  bool abort;
  bool *featuresHaveMissingValues;
  ulong numExamples;
  long maxTimeUs;
  Progress *p;
  SwappableDoubles **features;
} MissingPass1_;

typedef struct {
  bool abort;
  bool *featuresHaveMissingValues;
  ulong f;
  long maxTimeUs;
  double *features;
} MissingPass2_;

typedef struct {
  bool abort;
  bool checkTime;
  ulong numExamples;
  long maxTimeUs;
  Progress *p;
  SwappableDoubles **labels;
  SwappableUlongs *frequencies;
  LossFunction *lossFunction;
  Bear *this;
} EmptyPass_;

typedef struct {
  bool abort;
  bool checkTime;
  bool *featuresAreUseful;
  bool *featuresHaveMissingValues;
  byte run;
  ulong *lookups;
  long maxTimeUs;
  Lock lock;
  Progress *p;
  SwappableDoubles **features;
  Decider *decider;
  RandomWeighted **rws;
  DynamicSet **models;
  DynamicSet **useless;
  Bear *this;
} ElementaryPass_;

typedef struct {
  bool abort;
  bool modelBuilt;
  bool *featuresHaveMissingValues;
  byte construction;
  byte iteration;
  ulong numRunningProducers;
  long maxTimeUs;
  SwappableDoubles **features;
  Lock lock;
  Condition condition;
  Logger *logger;
  Progress *p;
  Decider *decider;
  DynamicSet **inProgress;
  DynamicSet **models;
  DynamicSet **useless;
  RandomWeighted **rws;
  Bear *this;
} CompositePass_;

typedef struct {
  ulong numExamples;
  SwappablePaws **labels;
  SwappablePaws *frequencies;
  LossFunction *lossFunction;
  Progress *p;
  DynamicSet **models;
  Bear *this;
} FinalPass_;

typedef struct {
  double selectedWeight;
  double testWeight;
  double *modelPredictions;
  double *selectedPredictions;
  double *testPredictions;
} FinalCombinePass_;

typedef struct {
  ulong totalFrequency;
  ulong *frequencies;
  double sum;
  double *labels;
  double *testPredictions;
  void *lossFunctionParams;
  Lock lock;
  LossFunction *lossFunction;
} FinalLossPass_;

typedef struct {
  ulong l;
  ulong numModels;
} LabelInfo_;

static void missing_values_1_(Parallel *par);

static void missing_values_2_(Parallel *par);

static void bad_frequencies_(Parallel *par);

static void unit_frequencies_(Parallel *par);

static void empty_models_(Parallel *par);

static void elementary_models_(Parallel *par);

static void composite_models_(Parallel *par);

static void final_processing_(Parallel *par);

static void final_combine_(Parallel *par);

static void final_loss_(Parallel *par);

static void rw_print_(BearModel *this, Printer *p);

static void check_vetoes_(CompositePass_ *pass);

static bool no_active_nodes_(CompositePass_ *pass);

static ulong min_models_(CompositePass_ *pass);

static bool model_built_(DynamicSet *models, BearAssembly *assembly);

static BearModel *get_model_(DynamicSet *models, BearAssembly *assembly);

static BearModel *try_get_model_(DynamicSet *models, BearAssembly *assembly);

sbyte label_info_compare_(LabelInfo_ *a, LabelInfo_ *b, void *x);

static void relink_(BearModel *model, DynamicReferenceSet *set);

static bool not_substantially_useful_(BearModel *model);

static void time_check_(bool abort, long start_time_us, long max_time_us);

static void time_check_clock_(long start_time_us, long max_time_us);

static bool about_as_good_(double total_strength, double max_total_strength);

static void add_useful_features_(Bear *this, bool *features_are_useful);

static void replace_useful_features_(
    Bear *this,
    bool **p_features_are_useful);


/* Constructor.
 */
void bear_new(Bear **p_this,
    ulong num_features,
    ulong num_labels,
    ulong num_examples,
    SwappableDoubles **features,
    SwappableDoubles **labels,
    SwappableUlongs *frequencies,
    LossFunction *loss_function,
    long start_time_us,
    long max_time_us,
    Logger *logger) {
  bool first_run_expired;
  bool either_run_expired;
  bool *_features_are_useful = null;
  bool *_features_have_missing_values = null;
  byte best_construction;
  byte construction;
  byte iteration;
  byte run;
  char *_s = null;
  ulong f;
  ulong l;
  ulong m;
  ulong num_essentials;
  ulong num_models;
  ulong num_total_elementary_models;
  ulong num_useful_elementary_models;
  ulong num_useful_features;
  ulong *_frequencies = null;
  ulong *_lookups = null;
  SwappableUlongs *_frequencies_s = null;
  long composite_start_us;
  long composite_us;
  long current_us;
  long essential_start_us;
  long essential_us;
  long final_start_us;
  long last_composite_us;
  long last_final_us;
  long max_run_time_us;
  long mean_essential_us;
  long required_us;
  long sum_essential_us;
  double max_total_strength;
  double strength;
  double total_strength;
  MissingPass1_ missing_pass;
  EmptyPass_ empty_pass;
  ElementaryPass_ elementary_pass;
  CompositePass_ composite_pass;
  FinalPass_ final_pass;
  Logger *_l;
  Progress *_p = null;
  Decider *_decider = null;
  DynamicSet **_in_progress = null;
  DynamicSet **_models = null;
  DynamicSet **_useless = null;
  RandomWeighted **_rws = null;
  Bear *_best = null;
  Bear *this;

  /* Sanity checks.
   */
  if (num_labels == 0) {
    throw("Cannot have no labels");
  }

  if (num_examples == 0) {
    throw("Cannot have no examples");
  }

  /* Use this syntactic sugar, for convenience and agreement with other code,
   * even though this is actually only a reference and not the owning pointer.
   */
  _l = logger;

  /* Initialize.
   */
  slog(_l, "Bear training");

  this = newo(p_this);
  this->numFeatures = num_features;
  this->numLabels = num_labels;
  this->lossFunction = loss_function;

  /* Determine if the features have any missing values.
   */
  if (num_features > 0) {
    slog(_l, "Check if features have missing values");
    new(&_features_have_missing_values, num_features);
    progress_new(&_p, _l, num_features);
    progress_add_callback(_p, deadline_progress, &max_time_us);
    progress_add_callback(_p, swappable_progress, null);
    missing_pass.abort = false;
    missing_pass.p = _p;
    missing_pass.maxTimeUs = max_time_us;
    missing_pass.numExamples = num_examples;
    missing_pass.features = features;
    missing_pass.featuresHaveMissingValues = _features_have_missing_values;
    parallel_for(
        num_features,
        5 * num_examples,
        missing_values_1_,
        &missing_pass);
    progress_delete(&_p);
    flog(_l);

    /* Time check.
     */
    time_check_(missing_pass.abort, start_time_us, max_time_us);
  }

  /* Check whether frequencies have been specified.
   */
  if (frequencies != null) {
    /* They have. Check them.
     */
    slog(_l, "Check frequencies");
    swappable_values_access(frequencies, &_frequencies);
    parallel_for(num_examples, 3, bad_frequencies_, _frequencies);
    swappable_values_release(frequencies, &_frequencies);
    flog(_l);
  } else {
    /* Frequencies have not been specified. Create a SwappableUlongs for them.
     * We set 'frequencies' to be a reference to it.
     */
    slog(_l, "Construct unit frequencies");
    swappable_ulongs_new_access(&_frequencies_s, num_examples, &_frequencies);
    frequencies = _frequencies_s;
    parallel_for(num_examples, 3, unit_frequencies_, _frequencies);
    swappable_values_release(frequencies, &_frequencies);
    flog(_l);
  }

  /* Time check. We didn't check in the above since the operations are so
   * low-level.
   */
  time_check_clock_(start_time_us, max_time_us);

  /* Initialize quantities for calculating the mean time for constructing the
   * "essentials" (empty models and elementary models).
   */
  sum_essential_us = 0;
  num_essentials = 0;

  /* Initialize flag that the first run or either run ran out of time before
   * minimizing loss.
   */
  first_run_expired = false;
  either_run_expired = false;

  /* Create the array of flags for features found to be useful on the first
   * run.
   */
  new(&_features_are_useful, num_features);

  /* Walk through the two runs. The second will not be done if the number of
   * used features is either zero or equal to the total number of features. We
   * can short-circuit all this if there are less than two features.
   */
  for (run = 0; run < 2; run++) {
    if (num_features >= 2) {
      slog(_l, "Run %u", run);
    }

    /* Compute the maximum time for this run.
     */
    max_run_time_us = run == 1 || num_features < 2
      ? max_time_us
      : (unixtime_current_us() + max_time_us) / 2;

    /* Walk through the constructions. We explicitly break out within the loop.
     */
    for (construction = 0;; construction++) {
      slog(_l, "Construction %u", construction);

      /* Sanity check.
       */
      if (construction == 255) {
        throw("Impossible");
      }

      /* Create the Decider for this construction.
       */
      decider_new(&_decider);

      /* Walk through the two iterations.
       */
      for (iteration = 0; iteration < 2; iteration++) {
        slog(_l, "Iteration %u", iteration);

        /* Start the clock for essential building.
         */
        essential_start_us = unixtime_current_us();

        /* Construct the empty model for each label.
         */
        slog(_l,
            num_labels == 1
            ? "Build the empty model"
            : "Build the %lu empty models",
            num_labels);
        progress_new(&_p, _l, num_labels);
        progress_add_callback(_p, deadline_progress, &max_run_time_us);
        progress_add_callback(_p, swappable_progress, null);
        new(&this->_emptyModels, num_labels);
        empty_pass.abort = false;
        empty_pass.checkTime = construction == 0;
        empty_pass.numExamples = num_examples;
        empty_pass.maxTimeUs = max_run_time_us;
        empty_pass.p = _p;
        empty_pass.labels = labels;
        empty_pass.frequencies = frequencies;
        empty_pass.lossFunction = loss_function;
        empty_pass.this = this;
        parallel_for_all(num_labels, empty_models_, &empty_pass);

        /* Time budget check.
         */
        time_check_(empty_pass.abort, start_time_us, max_run_time_us);

        /* Finish.
         */
        progress_delete(&_p);
        flog(_l);

        /* For each label we create dynamic sets for the models created, the
         * assemblies in progress, and the useless assemblies, as well as a
         * RandomWeighted for the models created.
         */
        new(&_models, num_labels);
        new(&_in_progress, num_labels);
        new(&_useless, num_labels);
        new(&_rws, num_labels);

        for (l = 0; l < num_labels; l++) {
          dynamic_set_new(
              &_models[l],
              bear_model_delete,
              bear_model_compare_assembly,
              bear_model_print);
          dynamic_set_new(
              &_in_progress[l],
              bear_assembly_delete,
              bear_assembly_compare,
              bear_assembly_print);
          dynamic_set_new(
              &_useless[l],
              bear_assembly_delete,
              bear_assembly_compare,
              bear_assembly_print);
          random_weighted_new(&_rws[l], rw_print_);
        }

        /* Compute the number of useful elementary models (the ones we will
         * actually build) and the total number of elementary models not
         * including the usefulness flag.
         */
        num_useful_elementary_models
          = (run == 1 ? num_useful_features : num_features) * num_labels;
        num_total_elementary_models = num_features * num_labels;

        /* Only build elementary models if there are any.
         */
        if (num_useful_elementary_models > 0) {
          /* Shuffle the elementary models to be built between the threads so
           * that no thread is idle if a given block of labels or features
           * takes less work to process than others. The parallel index is
           * actually across this product space of labels and features. If this
           * is the second run, we still use the total number of elementary
           * models, but skip those not useful in the callback.
           */
          shuffle_make_lookups(&_lookups, num_total_elementary_models);

          /* Build the elementary models.
           */
          slog(_l,
              num_useful_elementary_models == 1
              ? "Build the elementary model"
              : "Build the %lu elementary models",
              num_useful_elementary_models);
          progress_new(&_p, _l, num_useful_elementary_models);
          progress_add_callback(_p, deadline_progress, &max_run_time_us);
          progress_add_callback(_p, swappable_progress, null);
          elementary_pass.abort = false;
          empty_pass.checkTime = construction == 0;
          elementary_pass.p = _p;
          lock_initialize(&elementary_pass.lock);
          elementary_pass.run = run;
          elementary_pass.featuresAreUseful = _features_are_useful;
          elementary_pass.this = this;
          elementary_pass.maxTimeUs = max_run_time_us;
          elementary_pass.decider = _decider;
          elementary_pass.lookups = _lookups;
          elementary_pass.features = features;
          elementary_pass.models = _models;
          elementary_pass.useless = _useless;
          elementary_pass.rws = _rws;
          elementary_pass.featuresHaveMissingValues
            = _features_have_missing_values;
          parallel_for_all(
              num_total_elementary_models,
              elementary_models_,
              &elementary_pass);

          /* Time budget check.
           */
          time_check_(empty_pass.abort, start_time_us, max_run_time_us);

          /* Finish.
           */
          progress_delete(&_p);
          deletev(&_lookups);
          flog(_l);
        }

        /* Compute the essential build time.
         */
        essential_us = unixtime_current_us() - essential_start_us;

        /* Update totals.
         */
        sum_essential_us += essential_us;
        num_essentials++;
        mean_essential_us = (sum_essential_us + num_essentials / 2)
          / num_essentials;

        /* Don't construct composite models on the first construction, or if
         * there are no elementary models.
         */
        if (construction > 0 && num_useful_elementary_models > 0) {
          /* Construct composite models, until we either have no more possible
           * models to construct, or we run out of allotted time.
           */
          slog(_l, "Build composite models");

          /* Get the current time.
           */
          current_us = unixtime_current_us();

          /* If we are on the second iteration, start the clock for composite
           * build.
           */
          if (iteration == 1) {
            composite_start_us = current_us;
          }

          /* If we are on the second construction, set the maximum time to the
           * current time plus an arbitary lower minimum; we will not check it
           * until at least one model has been built. Otherwise, give the
           * composite build process twice the time used on the last
           * construction.
           */
          composite_us = construction == 1
            ? maximum(mean_essential_us, MIN_COMPOSITE_US_)
            : 2 * last_composite_us;
          composite_pass.maxTimeUs = current_us + composite_us;

          /* If this is the second iteration, it's possible that the first
           * iteration took a while to clean up from its composite run, and
           * overran its time budget. So bound the max time to allow us enough
           * time for final processing.
           */
          if (iteration == 1
              && composite_pass.maxTimeUs > max_run_time_us - last_final_us) {
            rlog(_l, "WARNING: First iteration overran; time is reduced.");
            composite_pass.maxTimeUs = maximum(current_us + MIN_COMPOSITE_US_,
                max_run_time_us - last_final_us);
            either_run_expired = true;
          }

          /* Provide feedback on the composite build time budget.
           */
          string_new_time_budget(&_s, time_budget_from_us(composite_us));
          rlog(_l, "Time budget for building composite models: %s", _s);
          string_delete(&_s);

          /* Create the progress object.
           */
          progress_new_unbounded(&_p, _l, "model built", "models built");
          progress_add_callback(
              _p,
              deadline_progress,
              &composite_pass.maxTimeUs);
          progress_add_callback(_p, swappable_progress, null);

          /* We create a producer-consumer system with parallel threads, where
           * every thread can actually be either. This is subtle, because if
           * there are no models left to join but other threads are potentially
           * producing new ones, then any threads waiting for a new pair of
           * models to join must wait for the condition that there are
           * potentially some available. We also need to wait if an
           * intermediate model needed for a join is in the process of being
           * built by another thread (and that thread might in fact determine
           * it to be useless). This is all handled in the callback below.
           */
          composite_pass.p = _p;
          composite_pass.construction = construction;
          composite_pass.abort = false;
          composite_pass.modelBuilt = false;
          composite_pass.iteration = iteration;
          composite_pass.numRunningProducers = 0;
          composite_pass.logger = _l;
          composite_pass.models = _models;
          composite_pass.useless = _useless;
          composite_pass.inProgress = _in_progress;
          composite_pass.rws = _rws;
          composite_pass.featuresHaveMissingValues
            = _features_have_missing_values;
          composite_pass.features = features;
          composite_pass.decider = _decider;
          composite_pass.this = this;
          lock_initialize(&composite_pass.lock);
          condition_initialize(&composite_pass.condition);
          parallel_do(composite_models_, &composite_pass);

          /* If we are on the second iteration, save the last composite build
           * time.
           */
          if (iteration == 1) {
            last_composite_us = unixtime_current_us() - composite_start_us;
          }

          /* If we aborted due to the time budget being exhausted, provide that
           * feedback. Note that the _first_ thread to notice the exhaustion of
           * the time budget will have provided feeback about that already;
           * this is to let the user know that _all_ threads have now finished.
           */
          if (composite_pass.abort) {
            rlog(_l, "Time budget exhausted: all threads have finished.");
          }

          /* Finish the feedback.
           */
          progress_delete(&_p);
          flog(_l);
        }

        /* Reset any changes that the loss function made.
         */
        for (l = 0; l < num_labels; l++) {
          this->_emptyModels[l]->lossFunction->resetParams(
              this->_emptyModels[l]->_lossFunctionParams);
        }

        /* Switch on the first or second iteration.
         */
        if (iteration == 0) {
          /* First iteration. Delete things that we will create again on the
           * second pass.
           */
          delete(&this->_emptyModels, num_labels, bear_model);
          delete(&_models, num_labels, dynamic_set);
          delete(&_in_progress, num_labels, dynamic_set);
          delete(&_useless, num_labels, dynamic_set);
          delete(&_rws, num_labels, random_weighted);
        } else {
          /* Second iteration. Count up the total number of models that we
           * built and kept.
           */
          for (l = 0, num_models = 0; l < num_labels; l++) {
            num_models += dynamic_set_cardinality(_models[l]);
          }
        }

        /* End the iteration.
         */
        decider_end_iteration(_decider);
        flog(_l);
      }

      /* Start the close for final selection.
       */
      final_start_us = unixtime_current_us();

      /* Create the array of the strength of the overall model for each label.
       */
      new(&this->_strengths, num_labels);

      /* Do final processing for each label. We run in parallel across the
       * labels. Within the parallel method, there are further layers of
       * parallelism, since there may be more processors than there are labels.
       *
       * Check if we built any models at all.
       */
      if (num_models > 0) {
        /* We did. Process.
         */
        slog(_l, "Select the final set of models to keep");
        new(&this->_models, num_labels);
        new(&this->_numsModels, num_labels);
        new(&this->_sumsWeights, num_labels);
        new(&this->_parents, num_labels);
        new(&this->_numsParents, num_labels);
        progress_new(&_p, _l, num_models);
        final_pass.p = _p;
        final_pass.lossFunction = loss_function;
        final_pass.numExamples = num_examples;
        final_pass.models = _models;
        final_pass.labels = labels;
        final_pass.frequencies = frequencies;
        final_pass.this = this;
        parallel_for_all(num_labels, final_processing_, &final_pass);
        progress_delete(&_p);
        flog(_l);
      } else {
        /* No models. Still need to create the right structures.
         */
        newm(&this->_models, num_labels, 0);
        new(&this->_numsModels, num_labels);
        new(&this->_sumsWeights, num_labels);
        newm(&this->_parents, num_labels, 0);
        new(&this->_numsParents, num_labels);
      }

      /* Clean up this construction.
       */
      slog(_l, "Clean up Bear's working memory");
      decider_delete(&_decider);
      delete(&_useless, num_labels, dynamic_set);
      delete(&_in_progress, num_labels, dynamic_set);
      delete(&_models, num_labels, dynamic_set);
      delete(&_rws, num_labels, random_weighted);
      flog(_l);

      /* Save the last final selection time.
       */
      last_final_us = unixtime_current_us() - final_start_us;

      /* We now decide if we break out and finish now.
       *
       * First, if there are no features, then we just have the empty model(s).
       */
      if (num_useful_elementary_models == 0) {
        rlog(_l, "There are no features.");
        rlog(_l,
            "I hope you find the empty model%s useful!",
            plural_s(num_labels));

        flog(_l);

        break;
      }

      /* Give some feedback about the selected models.
       */
      slog(_l, "Selected models");

      for (l = 0; l < num_labels; l++) {
        if (num_labels > 1) {
          slog(_l, "Label %lu", l);
        }

        for (m = 0; m < this->_numsModels[l]; m++) {
          string_new_bear_assembly(&_s, this->_models[l][m]->_assembly);
          rlog(_l, "%s (weight %.4g)", _s, this->_models[l][m]->weight);
          string_delete(&_s);
        }

        if (num_labels > 1) {
          flog(_l);
        }
      }

      flog(_l);

      /* There is at least one feature. Print the construction strengths, and
       * compute the total strength, simply summing the strengths for all
       * labels.
       */
      if (num_labels == 1) {
        rlog(_l,
            "Construction weight: %.4g",
            total_strength = this->_strengths[0]);
      } else {
        slog(_l, "Construction weights");

        for (l = 0, total_strength = 0; l < num_labels; l++) {
          total_strength += strength = this->_strengths[l];
          rlog(_l, "Label %lu: %.4g", l, strength);
        }

        rlog(_l, "Total: %.4g", total_strength);
        flog(_l);
      }

      /* Next, check if we're not on the first construction.
       */
      if (construction > 0) {
        /* We're not on the firset construction. Check if we're on the first
         * run.
         */
        if (run == 0) {
          /* We are. Check if this last construction is about as good as the
           * best one.
           */
          if (about_as_good_(total_strength, max_total_strength)) {
            /* About as good. Add in the useful features.
            */
            add_useful_features_(this, _features_are_useful);
          } else if (total_strength > max_total_strength) {
            /* Significantly better. Replace the useful features.
             */
            replace_useful_features_(this, &_features_are_useful);
          }
        }

        /* Check if we didn't run out of time.
         */
        if (!composite_pass.abort) {
          /* We didn't run out of time. Check if the first run ran out of time.
           */
          if (first_run_expired) {
            rlog(_l, "WARNING: The first run ran out of time before "
                "minimizing the loss.");
            rlog(_l, "WARNING: All useful models in the second run were "
                "built, but this is suboptimal.");
          } else {
            /* It didn't. Provide good feedback.
             */
            rlog(_l, "Congratulations! All potentially useful models "
                "were built.");
          }

          /* Check if this final construction is better than the best previous
           * one.
           */
          if (total_strength > max_total_strength) {
            /* It is. Provide that feedback.
             */
            rlog(_l,
                "This final construction is lowest-loss. "
                "I hope it is useful!");

            /* Delete the previous best one.
             */
            bear_delete(&_best);
          } else {
            /* Best previous was better. Provide that feedback.
             */
            rlog(_l,
                "This construction is not better than the previous best.");
            rlog(_l,
                "Using the previous best construction. I hope it is useful!");
            rlog(_l, "Best construction strength: %.4g", max_total_strength);

            /* Use the previous best.
             */
            bear_delete(p_this);
            this = donate(&_best, p_this);
          }

          /* Break out.
          */
          flog(_l);

          break;
        }

        /* We didn't run out of time. Check if this construction was about as
         * good as the last one.
         */
        if (about_as_good_(total_strength, max_total_strength)) {
          /* About as good. Fine-tune the message, and decide which to keep.
           */
          if (total_strength < max_total_strength) {
            rlog(_l,
                "This is almost as good as the best previous construction.");
            bear_delete(p_this);
            this = donate(&_best, p_this);
          } else if (total_strength > max_total_strength) {
            rlog(_l,
                "This is slightly better than the best "
                "previous construction.");
            bear_delete(&_best);
            max_total_strength = total_strength;
          } else {
            rlog(_l, "This is as good as the best previous construction.");
            bear_delete(p_this);
            this = donate(&_best, p_this);

            /* If we still have not found a nontrivial model, make this
             * construction the best, so that the following test does not cause
             * Bear to give up prematurely.
             */
            if (max_total_strength == 0) {
              best_construction = construction;
            }
          }

          /* If the best is receding into the distance, cut out here.
           */
          if (construction - best_construction >= MAX_WORSE_) {
            rlog(_l,
                "But that is %u constructions without substantial "
                "improvement. ",
                construction - best_construction);
            rlog(_l, "Congratulations! Giving you the best construction.");
            rlog(_l, "Best construction strength: %.4g", max_total_strength);
            flog(_l);

            break;
          }
        } else if (total_strength < max_total_strength) {
          /* We are substantially worse than the best one. Provide that
           * feedback, and keep the previous best.
           */
          rlog(_l,
              "A previous construction had substantially lower loss. "
              "I hope it is useful!");
          rlog(_l, "Best construction strength: %.4g", max_total_strength);
          bear_delete(p_this);
          this = donate(&_best, p_this);

          /* Break out.
           */
          flog(_l);

          break;
        } else {
          /* This is the best so far, significantly. Give appropriate feedback.
           */
          rlog(_l, "This is the best construction so far, substantially.");
          max_total_strength = total_strength;
          best_construction = construction;
          bear_delete(&_best);
        }
      } else {
        /* First construction. Store the best total strength.
         */
        max_total_strength = total_strength;
        best_construction = 0;
      }

      /* Now check if there is not enough time for another construction. We
       * have two iterations, each of which has the essential builds, plus the
       * composite builds that we will make twice as long as the last one for
       * this construction. There is then a single final selection phase, which
       * we can't completely predict, but use this last one as a best guess.
       */
      required_us = 2 * mean_essential_us
        + (construction > 0
            ? 4 * last_composite_us
            : 2 * maximum(mean_essential_us, MIN_COMPOSITE_US_))
        + last_final_us;

      /* Check if we have enough time left.
       */
      if (unixtime_current_us() + required_us > max_run_time_us) {
        /* We don't. Create a string for the inferred time budget.
         */
        string_new_time_budget(
            &_s,
            time_budget_from_us(max_time_us - start_time_us));

        /* Provide a warning.
         */
        rlog(_l, "WARNING: Insufficient time left for another construction.");
        rlog(_l, "WARNING: %s budget exhausted before minimizing loss.", _s);
        rlog(_l, "Best construction strength: %.4g", max_total_strength);
        string_delete(&_s);

        /* Set flag.
         */
        either_run_expired = true;

        /* If this is the first run, set flag so that we can modify our
         * messaging at the end of the second run.
         */
        if (run == 0) {
          first_run_expired = true;
        }

        /* Break out.
         */
        flog(_l);

        break;
      }

      /* Roll whatever we have put in 'this' to be the best.
       */
      donate(p_this, &_best);

      /* Create a new object.
       */
      this = newo(p_this);
      this->numFeatures = num_features;
      this->numLabels = num_labels;
      this->lossFunction = loss_function;

      /* Finish the construction.
       */
      flog(_l);
    }

    /* If there are less than two features, just break out now.
     */
    if (num_features < 2) {
      break;
    }

    /* At least two features. Check if we're on the first run.
     */
    if (run == 0) {
      /* We are. Count up the number of useful features.
       */
      for (f = 0, num_useful_features = 0; f < num_features; f++) {
        if (_features_are_useful[f]) {
          num_useful_features++;
        }
      }

      /* If the number of useful features is either zero or equal to the total
       * number of features, then we don't need a second run.
       */
      if (num_useful_features == 0) {
        rlog(_l, "No features found useful; no need for a second run.");
        flog(_l);

        break;
      } else if (num_useful_features == num_features) {
        rlog(_l,
            "%s feature%s found useful; no need for a second run.",
            num_features == 1 ? "The" : "All",
            num_features == 1 ? " was" : "s");
        flog(_l);

        break;
      }

      /* We want a second run. Provide feedback.
       */
      rlog(_l,
          "Found %lu of %lu features to be useful.",
          num_useful_features,
          num_features);
      rlog(_l,
          "Will do a second run with just th%s %lu useful feature%s.",
          num_useful_features == 1 ? "at" : "ose",
          num_useful_features,
          plural_s(num_useful_features));

      /* Delete things that we will create again on that
       * run.
       */
      delete(&_models, num_labels, dynamic_set);
      delete(&_in_progress, num_labels, dynamic_set);
      delete(&_useless, num_labels, dynamic_set);
      delete(&_rws, num_labels, random_weighted);

      /* Delete and recreate a bare object.
       */
      bear_delete(p_this);
      this = newo(p_this);
      this->numFeatures = num_features;
      this->numLabels = num_labels;
      this->lossFunction = loss_function;
    }

    /* Finish the run.
     */
    flog(_l);
  }

  /* Give another warning if the time budget expired for either run.
   */
  if (either_run_expired) {
    string_new_time_budget(
        &_s,
        time_budget_from_us(max_time_us - start_time_us));

    rlog(_l,
        "WARNING: %s budget was insufficient; solution is suboptimal.",
        _s);
    string_delete(&_s);
  }

  /* Clean up.
   */
  deletev(&_features_are_useful);
  deletev(&_features_have_missing_values);
  swappable_values_delete(&_frequencies_s);
  flog(_l);
}

/* Load.
 */
void bear_new_load(Bear **p_this, File *f) {
  char *_loss_function_name = null;
  ulong l;
  ulong m;
  ulong p;
  DynamicReferenceSet *_set = null;
  BearModel *model;
  Bear *this;

  this = newo(p_this);
  load_label("[Bear]", f);
  load_value(this->numFeatures, f);
  load_value(this->numLabels, f);
  load(_loss_function_name, string, f);
  this->lossFunction = loss_function_get(_loss_function_name);
  string_delete(&_loss_function_name);
  load_new_array(this->_emptyModels, this->numLabels, bear_model, f);
  load_new_values(this->_numsModels, this->numLabels, f);
  load_new_values(this->_numsParents, this->numLabels, f);
  load_new_values(this->_sumsWeights, this->numLabels, f);
  load_new_values(this->_strengths, this->numLabels, f);
  load_new_imatrix(
      this->_models,
      this->numLabels,
      this->_numsModels,
      bear_model,
      f);
  load_new_imatrix(
      this->_parents,
      this->numLabels,
      this->_numsParents,
      bear_model,
      f);
  load_label("[/Bear]", f);

  /* Relink all the logical parent pointers for all models that were saved.
   * Create a dynamic reference set of all the models, sorted by ID.
   */
  dynamic_reference_set_new_specify(
      &_set,
      bear_model_compare_id,
      bear_model_print);

  /* Add all the models to this reference set, and register every ID.
   */
  for (l = 0; l < this->numLabels; l++) {
    dynamic_reference_set_add(_set, model = this->_emptyModels[l]);
    bear_model_register_id(model->id);

    for (m = 0; m < this->_numsModels[l]; m++) {
      dynamic_reference_set_add(_set, model = this->_models[l][m]);
      bear_model_register_id(model->id);
    }

    for (p = 0; p < this->_numsParents[l]; p++) {
      dynamic_reference_set_add(_set, model = this->_parents[l][p]);
      bear_model_register_id(model->id);
    }
  }

  /* Now go through and populate all the logical parent pointers.
   */
  for (l = 0; l < this->numLabels; l++) {
    for (m = 0; m < this->_numsModels[l]; m++) {
      relink_(this->_models[l][m], _set);
    }

    for (p = 0; p < this->_numsParents[l]; p++) {
      relink_(this->_parents[l][p], _set);
    }
  }

  /* Clean up.
   */
  dynamic_reference_set_delete(&_set);
}

/* Destructor.
 */
void bear_delete(Bear **p_this) {
  Bear *this;

  if ((this = *p_this) == null) {
    return;
  }

  deleteim(&this->_parents, this->numLabels, this->_numsParents, bear_model);
  deleteim(&this->_models, this->numLabels, this->_numsModels, bear_model);
  delete(&this->_emptyModels, this->numLabels, bear_model);
  deletev(&this->_strengths);
  deletev(&this->_sumsWeights);
  deletev(&this->_numsParents);
  deletev(&this->_numsModels);
  deleteo(p_this);
}

/* Print.
 */
void bear_print(Bear *this, Printer *p) {
  ulong l;
  ulong m;

  sprint(p, "Bear");
  rprint(p, "Number of features: %lu", this->numFeatures);
  rprint(p, "Number of labels: %lu", this->numLabels);

  for (l = 0; l < this->numLabels; l++) {
    if (this->numLabels > 1) {
      sprint(p, "Label %lu", l);
    }

    if (this->_numsModels[l] > 0) {
      rprint(p, "Number of models: %lu", this->_numsModels[l]);

      for (m = 0; m < this->_numsModels[l]; m++) {
        sprint(p, "Model %lu", m);
        bear_model_print(this->_models[l][m], p);
        fprint(p);
      }
    } else {
      sprint(p, "Empty model");
      bear_model_print(this->_emptyModels[l], p);
      fprint(p);
    }

    if (this->numLabels > 1) {
      fprint(p);
    }
  }

  fprint(p);
}

/* Save.
 */
void bear_save(Bear *this, File *f) {
  save_label("[Bear]", f);
  save_value(this->numFeatures, f);
  save_value(this->numLabels, f);
  save(this->lossFunction->name, string, f);
  save_array(this->_emptyModels, this->numLabels, bear_model, f);
  save_values(this->_numsModels, this->numLabels, f);
  save_values(this->_numsParents, this->numLabels, f);
  save_values(this->_sumsWeights, this->numLabels, f);
  save_values(this->_strengths, this->numLabels, f);
  save_imatrix(
      this->_models,
      this->numLabels,
      this->_numsModels,
      bear_model,
      f);
  save_imatrix(
      this->_parents,
      this->numLabels,
      this->_numsParents,
      bear_model,
      f);
  save_label("[/Bear]", f);
}

/* Equals.
 */
bool bear_equals(Bear *this, Bear *other) {
  ulong l;
  ulong m;
  ulong p;

  if (other->numFeatures != this->numFeatures
      || other->numLabels != this->numLabels) {
    return false;
  }

  for (l = 0; l < this->numLabels; l++) {
    if (other->_numsModels[l] != this->_numsModels[l]
        || other->_numsParents[l] != this->_numsParents[l]
        || other->_sumsWeights[l] != this->_sumsWeights[l]
        || other->_strengths[l] != this->_strengths[l]
        || !bear_model_equals(other->_emptyModels[l], this->_emptyModels[l])) {
      return false;
    }

    for (m = 0; m < this->_numsModels[l]; m++) {
      if (!bear_model_equals(other->_models[l][m], this->_models[l][m])) {
        return false;
      }
    }

    for (p = 0; p < this->_numsParents[l]; p++) {
      if (!bear_model_equals(other->_parents[l][p], this->_parents[l][p])) {
        return false;
      }
    }
  }

  return true;
}

/* Predict the array of label values from the given array of feature values.
 */
void bear_predict(Bear *this, double *features, double *labels) {
  ulong l;
  ulong m;
  ulong num_models;
  double prediction;
  BearModel *model;

  /* Walk through the labels.
   */
  for (l = 0; l < this->numLabels; l++) {
    /* Check if we have any models.
     */
    if (this->numFeatures > 0 && (num_models = this->_numsModels[l]) > 0) {
      /* We do. Walk through the models for this label and compute the weighted
       * prediction.
       */
      for (m = 0, prediction = 0; m < num_models; m++) {
        /* Get a pointer to this model.
         */
        model = this->_models[l][m];

        /* Add in its weighted prediction.
         */
        prediction += model->weight * bear_model_predict(model, features);
      }

      /* Normalize, compute the estimate, and store.
       */
      labels[l] = this->_emptyModels[l]->lossFunction->computeLabelEstimate(
          prediction / this->_sumsWeights[l],
          this->_emptyModels[l]->_lossFunctionParams);
    } else {
      /* We don't have any built models. Use the empty model estimate.
       */
      labels[l] = this->_emptyModels[l]->emptyLabelEstimate;
    }
  }
}

/* Write out a summary CSV file of Bear's final model selection.
 */
void bear_write_model_summary(Bear *this, char *filename) {
  char *_s = null;
  ulong l;
  ulong m;
  TextFile *_out = null;
  BearModel *model;

  /* Open the output file.
   */
  if (string_equals(filename, "-")) {
    text_file_new_type(&_out, filename, "w", "csv");
  } else {
    text_file_new(&_out, filename, "w");
  }

  /* Write a header row.
   */
  if (this->numLabels > 1) {
    text_file_write(_out, "label");
  }

  text_file_write(_out, "model");
  text_file_write(_out, "weight");
  text_file_write(_out, "assembly");
  text_file_end_record(_out);

  /* Walk through the labels.
   */
  for (l = 0; l < this->numLabels; l++) {
    /* Check if we have no prediction models.
     */
    if (this->_numsModels[l] == 0) {
      /* We have no models. Write an empty model row.
       */
      if (this->numLabels > 1) {
        text_file_write_ulong(_out, l);
      }

      text_file_write_ulong(_out, 0);
      text_file_write_double(_out, 0);
      text_file_write(_out, "e");

      /* End the record.
       */
      text_file_end_record(_out);
    } else {
      /* Walk through the selected models for this label.
       */
      for (m = 0; m < this->_numsModels[l]; m++) {
        /* Extract reference to the model.
         */
        model = this->_models[l][m];

        /* If there is more than one label, write the label number.
         */
        if (this->numLabels > 1) {
          text_file_write_ulong(_out, l);
        }

        /* Write the model number.
         */
        text_file_write_ulong(_out, m);

        /* Write out the weight.
         */
        text_file_write_double(_out, model->weight);

        /* Write out the abbreviated form of the assembly.
         */
        string_new_bear_assembly(&_s, model->_assembly);
        text_file_write(_out, _s);
        string_delete(&_s);

        /* End the record.
         */
        text_file_end_record(_out);
      }
    }
  }

  /* Close the output file.
   */
  text_file_delete(&_out);
}

/* Get an array of flags for each feature indicating if it actually used in
 * Bear's final models. Returns the number of used features.
 */
ulong bear_get_features_are_used(Bear *this, bool **p_features_are_used) {
  bool *features_are_used;
  ulong f;
  ulong l;
  ulong m;
  ulong num_used_features;
  ulong p;
  BearPart *part;
  BearAssembly *assembly;
  BearModel *model;

  /* Create the array.
   */
  features_are_used = new(p_features_are_used, this->numFeatures);

  /* Walk through the labels.
   */
  for (l = 0; l < this->numLabels; l++) {
    /* Walk through the selected models for this label.
     */
    for (m = 0; m < this->_numsModels[l]; m++) {
      /* Get a reference to the model.
       */
      model = this->_models[l][m];

      /* Get a reference to the assembly for this model.
       */
      assembly = model->_assembly;

      /* Walk through the parts of the assembly.
       */
      for (p = 0; p < assembly->numParts; p++) {
        /* Extract reference to this part.
         */
        part = assembly->_parts[p];

        /* Walk through the features of this part.
         */
        for (f = 0; f < part->numFeatures; f++) {
          /* Set the feature to be used.
           */
          features_are_used[part->_features[f]] = true;
        }
      }
    }
  }

  /* Count up the number of used features.
   */
  for (f = 0, num_used_features = 0; f < this->numFeatures; f++) {
    if (features_are_used[f]) {
      num_used_features++;
    }
  }

  /* Return it.
   */
  return num_used_features;
}

/* Get an array of all the features actually used in Bear's final models.
 */
ulong bear_get_used_features(Bear *this, ulong **p_used_features) {
  bool *_features_are_used = null;
  ulong f;
  ulong num_used_features;
  ulong u;
  ulong *used_features;

  /* Get the array of flags of whether each feature is used.
   */
  num_used_features = bear_get_features_are_used(this, &_features_are_used);

  /* Create the array.
   */
  used_features = new(p_used_features, num_used_features);

  /* Populate it.
   */
  for (f = 0, u = 0; f < this->numFeatures; f++) {
    if (_features_are_used[f]) {
      used_features[u++] = f;
    }
  }

  /* Sanity check.
   */
  if (u != num_used_features) {
    throw("Impossible");
  }

  /* Clean up.
   */
  deletev(&_features_are_used);

  /* Return the number of used features.
   */
  return num_used_features;
}


/* Top-level parallel method for determining if the features have missing
 * values.
 */
static void missing_values_1_(Parallel *par) {
  ulong f;
  double *_features = null;
  MissingPass1_ *pass;
  MissingPass2_ pass2;

  /* Initialize.
   */
  pass = par->x;

  /* Walk through the features for this parallel thread.
   */
  for (f = par->start; f < par->end; f++) {
    /* Time budget check.
     */
    if (pass->abort || unixtime_current_us() > pass->maxTimeUs) {
      pass->abort = true;
      progress_bump_multi(pass->p, par->end - f);

      return;
    }

    /* Access the values of this feature.
     */
    swappable_values_access(pass->features[f], &_features);

    /* Run through the examples in another parallel run.
     */
    pass2.abort = false;
    pass2.maxTimeUs = pass->maxTimeUs;
    pass2.featuresHaveMissingValues = pass->featuresHaveMissingValues;
    pass2.f = f;
    pass2.features = _features;
    parallel_for(pass->numExamples, 5, missing_values_2_, &pass2);

    /* Release the values of this feature.
     */
    swappable_values_release(pass->features[f], &_features);

    /* Bump progress.
     */
    progress_bump(pass->p);
  }
}

/* Second-level parallel method for determining if the features have missing
 * values.
 */
static void missing_values_2_(Parallel *par) {
  ulong e;
  ulong f;
  double *features;
  MissingPass2_ *pass;

  /* Initialize.
   */
  pass = par->x;
  features = pass->features;
  f = pass->f;

  /* Walk through the examples for this parallel thread. If any thread
   * (including this one) has marked missing values, we don't need to continue
   * checking.
   */
  for (e = par->start;
      !pass->featuresHaveMissingValues[f] && e < par->end;
      e++) {
    /* Check if this example for this feature is missing.
     */
    if (double_pawized_is_infinite_or_nan(features[e])) {
      /* It is. Set the flag.
       */
      pass->featuresHaveMissingValues[f] = true;
    }
  }
}

/* Parallel method for checking for bad frequency values.
 */
static void bad_frequencies_(Parallel *par) {
  ulong e;
  ulong *frequencies;

  /* Initialize.
   */
  frequencies = par->x;

  /* Walk through the examples for this parallel thread.
   */
  for (e = par->start; e < par->end; e++) {
    /* Check the frequency for this example.
     */
    if (frequencies[e] == 0) {
      throw("Frequency for example %lu is zero, which is not allowed", e);
    }
  }
}

/* Parallel method for settiong unit frequency values.
 */
static void unit_frequencies_(Parallel *par) {
  ulong e;
  ulong *frequencies;

  /* Initialize.
   */
  frequencies = par->x;

  /* Walk through the examples for this parallel thread.
   */
  for (e = par->start; e < par->end; e++) {
    /* Set the frequency.
     */
    frequencies[e] = 1;
  }
}

/* Parallel method for constructing the empty model for each label.
 */
static void empty_models_(Parallel *par) {
  ulong l;
  EmptyPass_ *pass;
  Bear *this;

  /* Initialize.
   */
  pass = par->x;
  this = pass->this;

  /* Walk through the labels for this parallel thread.
   */
  for (l = par->start; l < par->end; l++) {
    /* Time budget check.
     */
    if (pass->checkTime
        && (pass->abort || unixtime_current_us() > pass->maxTimeUs)) {
      pass->abort = true;
      progress_bump_multi(pass->p, par->end - l);

      return;
    }

    /* Create the empty model for this label.
     */
    bear_model_new_empty(
        &this->_emptyModels[l],
        pass->numExamples,
        pass->labels[l],
        pass->frequencies,
        pass->lossFunction);

    /* Bump progress.
     */
    progress_bump(pass->p);
  }
}

/* Parallel method for constructing the elementary model for each feature for
 * each label.
 */
static void elementary_models_(Parallel *par) {
  ulong f;
  ulong l;
  ulong serial;
  ulong u;
  ElementaryPass_ *pass;
  BearAssembly *_assembly = null;
  BearModel *_model = null;
  Bear *this;

  /* Initialize.
   */
  pass = par->x;
  this = pass->this;

  /* Walk through the parallel indexes for this thread.
   */
  for (u = par->start; u < par->end; u++) {
    /* Time budget check.
     */
    if (pass->checkTime
        && (pass->abort || unixtime_current_us() > pass->maxTimeUs)) {
      pass->abort = true;
      progress_bump_multi(pass->p, par->end - u);

      return;
    }

    /* The serial value in the label-feature product space is the shuffled
     * lookup from this parallel index.
     */
    serial = pass->lookups[u];

    /* Compute the label and feature values from the product space for this
     * shuffled parallel index.
     */
    l = serial % this->numLabels;
    f = serial / this->numLabels;

    /* Check if this is the second run.
     */
    if (pass->run == 1) {
      /* It is. Check if this feature is a useful one.
       */
      if (!pass->featuresAreUseful[f]) {
        /* It's not. Skip it.
         */
        continue;
      }
    }

    /* Create the elementary assembly for this feature.
     */
    bear_assembly_new_elementary(&_assembly, f);

    /* Create the elementary model for this feature for this label, based on
     * the empty model for this label, which is also the parent.
     */
    bear_model_new(
        &_model,
        _assembly,
        this->_emptyModels[l],
        this->_emptyModels[l],
        pass->features,
        pass->featuresHaveMissingValues,
        pass->decider);

    /* Under a lock, add this model to the set of models constructed, and also
     * to the RandomWeighted of models. Note that we always keep elementary
     * models, even if they are not by themselves useful, because they cannot
     * have a repeated last part, so we don't add its assembly to the list of
     * useless assemblies.
     */
    lock(&pass->lock);
    random_weighted_add(pass->rws[l], _model, _model->selectionWeight);
    dynamic_set_insert(pass->models[l], &_model);
    unlock(&pass->lock);

    /* Clean up.
     */
    bear_assembly_delete(&_assembly);
    progress_bump(pass->p);
  }
}

/* Parallel method for creating composite models.
 */
static void composite_models_(Parallel *par) {
  bool abort;
  bool join_rev;
  byte join;
  byte join_it;
  ulong l;
  ulong attachment_p;
  double bonus_weight;
  CompositePass_ *pass;
  Random *_random = null;
  BearPart *attachment_part;
  BearAssembly *attachment_assembly;
  BearAssembly *base_assembly;
  BearAssembly *_assembly = null;
  BearAssembly *_assembly_copy = null;
  BearAssembly *_parent_assembly = null;
  RandomWeightedNode *attachment_node;
  RandomWeightedNode *base_node;
  BearModel *parent_model;
  BearModel *attachment_model;
  BearModel *base_model;
  BearModel *_model = null;
  Bear *this;

  /* Initialize.
   */
  pass = par->x;
  this = pass->this;

  /* Create a pseudorandom number generator for use below. Use the offset
   * version since we are in a parallel method.
   */
  random_new_offset(&_random, par->t);

  /* Sanity check.
   */
  if (pass->construction == 0) {
    throw("Impossible");
  }

  /* Main loop. We explicitly break out if the time budget is hit or there is
   * nothing left to do.
   */
  for (abort = false;;) {
    /* Time budget check.
     */
    if ((pass->construction > 1 || pass->modelBuilt)
        && (abort || pass->abort || unixtime_current_us() > pass->maxTimeUs)) {
      /* Check if any other thread has signaled the abort yet.
       */
      if (!pass->abort) {
        /* None has. Grab the lock.
         */
        lock(&pass->lock);

        /* Check if any other thread got here first.
         */
        if (!pass->abort) {
          /* They didn't. Print a message and flag the abort.
           */
          rlog(pass->logger,
              "Time budget exhausted: signaling to all threads to finish.");
          pass->abort = true;
        }

        /* Release the lock.
         */
        unlock(&pass->lock);
      }

      /* Clean up.
       */
      random_delete(&_random);

      /* Signal all waiting threads.
       */
      condition_signal(&pass->condition);

      return;
    }

    /* Start the "consumer" phase. Take the lock.
     */
    lock(&pass->lock);

    /* Sanity check: all vetoes should have been cleared.
     */
    check_vetoes_(pass);

    /* If there are no active nodes for any label, loop until that is no longer
     * true, or we return because there are no active producers either.
     */
    while (no_active_nodes_(pass)) {
      /* Time budget check, since we might do some waiting in this loop.
       */
      if ((pass->construction > 1 || pass->modelBuilt)
          && (abort || pass->abort
            || unixtime_current_us() > pass->maxTimeUs)) {
        abort = true;

        break;
      }

      /* There are no active nodes for any label. Check if there are any
       * producers running.
       */
      if (pass->numRunningProducers > 0) {
        /* There are producers running. We need to wait until one of them
         * signals the condition that they have finished a producer run, or
         * that the time budget is exhausted. This yields the lock to any other
         * thread that wants it.
         */
        condition_wait(&pass->condition, &pass->lock);

        /* We've now received the broadcast signal that one of the running
         * producers has finished its producer phase. It may or may not have
         * created any new nodes. We have the lock back again.
         */
      } else {
        /* There are no active nodes for any label, there are no producers
         * running, and we have the lock, so we're done. Release the lock,
         * clean up our local objects, and return. All other threads that are
         * still running must be in the consumer phase, so they will also find
         * this to be the case, and will likewise return.
         */
        condition_signal(&pass->condition);
        unlock(&pass->lock);

        /* Clean up.
         */
        random_delete(&_random);

        return;
      }
    }

    /* If we flagged an abort, skip to the next iteration of the main loop,
     * where the condition is checked.
     */
    if (abort) {
      unlock(&pass->lock);

      continue;
    }

    /* We now have at least one label with at least one active node, and we
     * have the lock. Find the label that has at least one active node that has
     * the least number of models built.
     */
    l = min_models_(pass);

    /* First compute the "bonus," per-model weight for this label at this time,
     * which is 1 / n, where n is the total number of models created for this
     * label (not just the number of active models in the RandomWeighted).
     */
    bonus_weight = 1.0 / dynamic_set_cardinality(pass->models[l]);

    /* Select a base node at random (weighted) from the active nodes.
     */
    base_node = random_weighted_get(pass->rws[l], bonus_weight);

    /* Extract a reference to the base model from this base node.
     */
    base_model = base_node->object;

    /* Veto all of the models listed in the base model's set of attachments
     * that are vetoed (used up).
     */
    bear_model_apply_vetoes(base_model, pass->rws[l]);

    /* If the RandomWeighted is now empty, including these vetoes, then the
     * base model is done unless and until we add another model; remove it
     * for now.
     */
    if (random_weighted_is_empty(pass->rws[l])) {
      /* Remove the base node from the RandomWeighted.
       */
      random_weighted_remove(pass->rws[l], base_node);

      /* Clear the vetoes.
       */
      random_weighted_clear_vetoes(pass->rws[l]);

      /* Skip to the next iteration of the main loop.
       */
      unlock(&pass->lock);

      continue;
    }

    /* The RandomWeighted is not empty, even taking into account the vetoes for
     * this base. Select an attachment node at random (weighted) from the
     * non-vetoed model nodes.
     */
    attachment_node = random_weighted_get(pass->rws[l], bonus_weight);

    /* Get the next part to use for this attachment to this base.
     */
    attachment_p = bear_model_get_next_attachment_part(
        base_model,
        attachment_node);

    /* We now have the attachment node and part, so we can clear the vetoes
     * from the base in the RandomWeighted.
     */
    random_weighted_clear_vetoes(pass->rws[l]);

    /* Signal the other threads and release the lock. We will take it back
     * shortly.
     */
    condition_signal(&pass->condition);
    unlock(&pass->lock);

    /* Extract a reference to the attachment model.
     */
    attachment_model = attachment_node->object;

    /* Extract references to the assemblies for the base and the attachment.
     */
    base_assembly = base_model->_assembly;
    attachment_assembly = attachment_model->_assembly;

    /* Get a reference to the attachment part we are going to attempt to attach
     * to the base.
     */
    attachment_part = attachment_assembly->_parts[attachment_p];

    /* Walk through the two different join types (glue and melt). Do them in
     * random order.
     */
    join_rev = random_modulo(_random, 2) == 0;

    for (join_it = 0; join_it < 2; join_it++) {
      join = join_rev ? 1 - join_it : join_it;

      /* Do a time budget check, since we yield the lock in this loop.
       */
      if ((pass->construction > 1 || pass->modelBuilt)
          && (abort || pass->abort
            || unixtime_current_us() > pass->maxTimeUs)) {
        abort = true;

        break;
      }

      /* If we're doing the glue join type (0) and the base model is not
       * substantially useful, then skip this join type.
       */
      if (join == 0 && not_substantially_useful_(base_model)) {
        continue;
      }

      /* If we're doing the melt join type (1) and the attachment part cannot
       * be melted to the base, skip this join type.
       */
      if (join == 1
          && !bear_assembly_can_melt(base_assembly, attachment_part)) {
        continue;
      }

      /* Create the new assembly.
       */
      (join == 0 ? bear_assembly_new_glue : bear_assembly_new_melt)(
          &_assembly,
          base_assembly,
          attachment_part);

      /* Create the parent assembly.
       */
      bear_assembly_new_parent(&_parent_assembly, _assembly);

      /* Get the lock back.
       */
      lock(&pass->lock);

      /* If the assembly or its parent have been found to be useless, or the
       * assembly has been built or is in progress, then skip this assembly.
       */
      if (dynamic_set_contains(pass->useless[l], _assembly)
          || dynamic_set_contains(pass->useless[l], _parent_assembly)
          || dynamic_set_contains(pass->inProgress[l], _assembly)
          || model_built_(pass->models[l], _assembly)) {
        bear_assembly_delete(&_assembly);
        bear_assembly_delete(&_parent_assembly);

        unlock(&pass->lock);

        continue;
      }

      /* We have an assembly to build. We are now a producer.
       */
      pass->numRunningProducers++;

      /* Get references to the base model and parent model before we release
       * the lock.
       */
      if (bear_assembly_is_empty(base_assembly)) {
        base_model = this->_emptyModels[l];
      } else {
        base_model = get_model_(pass->models[l], base_assembly);
      }

      if (bear_assembly_is_empty(_parent_assembly)) {
        parent_model = this->_emptyModels[l];
      } else {
        parent_model = get_model_(pass->models[l], _parent_assembly);
      }

      /* We don't need the parent assembly any more.
       */
      bear_assembly_delete(&_parent_assembly);

      /* Copy the assembly and insert it into the set of those in progress.
       */
      bear_assembly_new_copy(&_assembly_copy, _assembly);
      dynamic_set_insert(pass->inProgress[l], &_assembly_copy);

      /* Signal the other threads, and release the lock.
       */
      condition_signal(&pass->condition);
      unlock(&pass->lock);

      /* Create the model.
       */
      bear_model_new(
          &_model,
          _assembly,
          parent_model,
          base_model,
          pass->features,
          pass->featuresHaveMissingValues,
          pass->decider);

      /* Bump progress, to add to the count of models constructed, regardless
       * of whether we find this model to be useful or useless.
       */
      progress_bump(pass->p);

      /* Grab the lock so that we can process our constructed model.
       */
      lock(&pass->lock);

      /* Flag that we have built a model, which is useful for the second
       * construction.
       */
      pass->modelBuilt = true;

      /* Remove this assembly from the set of those in progress.
       */
      dynamic_set_remove(pass->inProgress[l], _assembly);

      /* By the time we release the lock, we will no longer be a running
       * producer.
       */
      pass->numRunningProducers--;

      /* Check if this model is useless, which is only true if it is not
       * substantially useful and the last part of its assembly is a repeat of
       * an earlier part. Note that a model that is not substantially useful
       * (including a trivial model) is not useless if the second criterion is
       * not met: it cannot be used as a parent model, but it can be used as a
       * base model for melting.
       */
      if (not_substantially_useful_(_model)
          && bear_assembly_last_part_is_repeat(_assembly)) {
        /* This model is useless. Delete it.
         */
        bear_model_delete(&_model);

        /* Insert this assembly to the set of useless assemblies.
         */
        dynamic_set_insert(pass->useless[l], &_assembly);

        /* Signal and release the lock and move to the next join type.
         */
        condition_signal(&pass->condition);
        unlock(&pass->lock);

        continue;
      }

      /* This model is potentially useful, even if only as a base for melting.
       * We are now in "producer update" phase; we still have the lock. Add a
       * reference to this model to the RandomWeighted for this label, getting
       * back a reference to the RandomWeightedNode that it was put into.
       */
      random_weighted_add(
          pass->rws[l],
          _model,
          _model->selectionWeight);

      /* Clear any removes in the RandomWeighted, since there is now a new
       * model that can potentially be joined with all of those already there.
       */
      random_weighted_clear_removes(pass->rws[l]);

      /* Insert this model into the set of built models.
       */
      dynamic_set_insert(pass->models[l], &_model);

      /* We don't need a separate copy of the assembly any more.
       */
      bear_assembly_delete(&_assembly);

      /* Signal and release the lock to exit the producer update phase.
       */
      condition_signal(&pass->condition);
      unlock(&pass->lock);
    }

    /* End of outer loop.
     */
  }

  /* We should never get here. We must have broken out of the outer loop by
   * mistake.
   */
  throw("Impossible");
}

/* Parallel method for final processing.
 */
static void final_processing_(Parallel *par) {
  ulong l;
  ulong m;
  ulong num_examples;
  ulong num_models;
  ulong num_owned;
  ulong o;
  ulong p;
  ulong *_frequencies = null;
  ulong *_ms = null;
  double last_weight;
  double loss;
  double selected_weight;
  double test_loss;
  double test_weight;
  double weight;
  double *model_predictions;
  double *_labels = null;
  double *_selected_predictions = null;
  double *_temp = null;
  double *_test_predictions = null;
  void *loss_function_params;
  FinalPass_ *pass;
  FinalCombinePass_ combine_pass;
  FinalLossPass_ loss_pass;
  LossFunction *loss_function;
  DynamicReferenceSet *_owned = null;
  BearModel *model;
  BearModel *parent;
  BearModel **_models = null;
  Bear *this;

  /* Initialize.
   */
  pass = par->x;
  this = pass->this;
  num_examples = pass->numExamples;
  loss_function = pass->lossFunction;

  /* Access the frequencies.
   */
  swappable_values_access(pass->frequencies, &_frequencies);

  /* Walk through the labels for this thread.
   */
  for (l = par->start; l < par->end; l++) {
    /* Initialize the array of final prediction models for this label, owned by
     * the object. Also set the loss to the empty model loss, in case we have
     * no models.
     */
    new(&this->_models[l], 0);
    loss = this->_emptyModels[l]->loss;

    /* Create a local array for the current prediction for each example for the
     * set of all models currently selected, and likewise the "test"
     * predictions that include those of the next model that we are considering
     * adding to our set of selected models.
     */
    new(&_selected_predictions, num_examples);
    new(&_test_predictions, num_examples);

    /* Extract all the models from the dynamic set into an array.
     */
    num_models = dynamic_set_extract_all(pass->models[l], &_models);

    /* We now determine the set of "final" models that we will use for
     * predictions, for this label.
     *
     * We do this using a greedy algorithm. We first sort all the models
     * for this label by decreasing weight.
     */
    qsortx(_models, num_models, bear_model_pointer_compare_weight, true, null);

    /* Access the label values for this label.
     */
    swappable_values_access(pass->labels[l], &_labels);

    /* Get a reference to the loss function params from the empty model for
     * this label.
     */
    loss_function_params = this->_emptyModels[l]->_lossFunctionParams;

    /* We now walk through all the models in order of decreasing weight, and
     * decide which to select.
     *
     * We do this by computing the loss for the overall predictions after we
     * add each given model's predictions to those of all previous selected
     * models, weighted appropriately. If the given model lowers the loss, we
     * select it into our list of prediction models and recompute the overall
     * predictions; otherwise, we do not select it, and we stick with the
     * predictions we had for the previously selected models.
     *
     * Walk through all the models in order of decreasing weight.
     */
    for (m = 0, selected_weight = 0, last_weight = DBL_MAX;
        m < num_models;
        m++, last_weight = weight) {
      /* Get a reference to the model.
       */
      model = _models[m];

      /* Extract its weight.
       */
      weight = model->weight;

      /* Sanity check.
       */
      if (weight > last_weight) {
        throw("Impossible");
      }

      /* Check what the new total selected weight would be.
       */
      test_weight = selected_weight + weight;

      /* If we have gotten down to models where their weights are either zero
       * or wouldn't individually change the sum of weights at all, to within
       * precision, then we can break out.
       */
      if (test_weight == selected_weight) {
        progress_bump_multi(pass->p, num_models - m);

        break;
      }

      /* Reacquire the working for this model.
       */
      swappable_reacquire(model->_wS);

      /* Extract reference to the label predictions for this model.
       */
      model_predictions = model->_w->_labelPredictions;

      /* Switch on whether this is the first model or not.
       */
      if (m == 0) {
        /* First model. Just copy the model predictions across.
         */
        copy_values(model_predictions, _test_predictions, num_examples);
      } else {
        /* Not the first model. Combine the predictions.
         */
        combine_pass.testWeight = test_weight;
        combine_pass.selectedWeight = selected_weight;
        combine_pass.testPredictions = _test_predictions;
        combine_pass.selectedPredictions = _selected_predictions;
        combine_pass.modelPredictions = model_predictions;
        parallel_for(num_examples, 6, final_combine_, &combine_pass);
      }

      /* Release the working for this model.
       */
      swappable_release(model->_wS);

      /* Compute the loss including this model.
       */
      lock_initialize(&loss_pass.lock);
      loss_pass.sum = 0;
      loss_pass.totalFrequency = 0;
      loss_pass.frequencies = _frequencies;
      loss_pass.labels = _labels;
      loss_pass.testPredictions = _test_predictions;
      loss_pass.lossFunction = loss_function;
      loss_pass.lossFunctionParams = loss_function_params;
      parallel_for(num_examples, 9, final_loss_, &loss_pass);

      /* Compute the test loss.
       */
      test_loss = loss_pass.sum / loss_pass.totalFrequency;

      /* If we're not on the first model and loss has not decreased, then skip
       * this model.
       */
      if (m > 0 && test_loss >= loss) {
        progress_bump(pass->p);

        continue;
      }

      /* Append a reference (only) to this model to the array of selected
       * models in the object. (We will change this to an owning pointer
       * below.)
       */
      append_copy(&this->_models[l], this->_numsModels[l], model);

      /* Use the new weight and loss.
       */
      selected_weight = test_weight;
      loss = test_loss;

      /* Switch the selected and test prediction arrays.
       */
      donate(&_selected_predictions, &_temp);
      donate(&_test_predictions, &_selected_predictions);
      donate(&_temp, &_test_predictions);

      /* Bump progress.
       */
      progress_bump(pass->p);
    }

    /* Release the label values for this label.
     */
    swappable_values_release(pass->labels[l], &_labels);

    /* Store the sum of weights for this label's selected prediction models.
     */
    this->_sumsWeights[l] = selected_weight;

    /* Compute the "strength" of the overall combined model as the weight that
     * would attach to the overall loss.
     */
    this->_strengths[l] = bear_model_weight(
        loss,
        this->_emptyModels[l]->loss);

    /* Clean up.
     */
    deletev(&_selected_predictions);
    deletev(&_test_predictions);

    /* We now need to transfer ownership of all of the final selected models
     * from the local array of models built in this constructor to the array of
     * models stored in the object, so that they are not deleted when we delete
     * the local array before the end of the constructor. We also need to find
     * all those parent models of the final models that are not themselves
     * final models, since they are also needed by the final models, and donate
     * those to the object as well. Bear in mind that models may appear in
     * multiple parent-child chains, but ownership of each model must be in
     * just one definite place.
     *
     * The simplest way to transfer ownership _from_ the local array of models
     * is to re-sort the local array of models by pointer, so that we can do a
     * binary search on them and find their indexes, which is as efficient as
     * any other scalable method available in this codebase (since we don't use
     * hashsets). So do that now.
     */
    qsortx(_models, num_models, pointer_pointer_compare, false, null);

    /* This approach, however, means that we can't set those array entries to
     * null until we have finished, or else we would break the binary search.
     * So we keep a simple array of index values of those models, and will use
     * this array later to actually set the entries to null.
     *
     * Start with the number of owned models being the number of final models;
     * we will find and append any parents not in the set of final models
     * shortly.
     */
    new(&_ms, num_owned = this->_numsModels[l]);

    /* The easiest way to keep track of the models whose ownership we have
     * transferred _to_ the object (except for the empty model for each label,
     * which is already owned by the object) is through a dynamic reference
     * set. So create that now.
     */
    dynamic_reference_set_new(&_owned);

    /* Walk through all of the final selected prediction models for this label.
     */
    for (p = 0; p < this->_numsModels[l]; p++) {
      /* Extract a reference to this model.
       */
      model = this->_models[l][p];

      /* Find its index in the local array of models.
       */
      m = bsearchx_find_index(
          &model,
          _models,
          num_models,
          pointer_pointer_compare,
          null);

      /* Save that index.
       */
      _ms[p] = m;

      /* Add the model to the set of those owned.
       */
      dynamic_reference_set_add(_owned, model);
    }

    /* Now do all of the parents of the final models, except for the empty
     * model. Create an empty array of parents for this label, in the object.
     */
    new(&this->_parents[l], this->_numsParents[l] = 0);

    /* Walk back through the final models for this label.
     */
    for (p = 0; p < this->_numsModels[l]; p++) {
      /* Extract a reference to this model.
       */
      model = this->_models[l][p];

      /* Walk through its parents, stopping when we get to the empty model.
       */
      for (parent = model->parent; !parent->empty; parent = parent->parent) {
        /* Check if we already have this parent in our set of owned models.
         */
        if (dynamic_reference_set_contains(_owned, parent)) {
          /* We do. Skip it.
           */
          continue;
        }

        /* We need to (later) take ownership of this parent. Find its index.
         */
        m = bsearchx_find_index(
            &parent,
            _models,
            num_models,
            pointer_pointer_compare,
            null);

        /* Append this index to our array of owned model indexes.
         */
        append_copy(&_ms, num_owned, m);

        /* Append a reference to this parent to our array of parents.
         */
        append_copy(&this->_parents[l], this->_numsParents[l], parent);

        /* Add this parent to the set of owned models.
         */
        dynamic_reference_set_add(_owned, parent);
      }
    }

    /* Now go through our list of indexes, and set each corresponding entry in
     * the local array of models to null, effectively transferring ownership to
     * the object.
     */
    for (o = 0; o < num_owned; o++) {
      _models[_ms[o]] = null;
    }

    /* Delete the models not taken over.
     */
    delete(&_models, num_models, bear_model);

    /* Clean up.
     */
    dynamic_reference_set_delete(&_owned);
    deletev(&_ms);
  }

  /* Release the frequencies.
   */
  swappable_values_release(pass->frequencies, &_frequencies);

  /* Withdraw the hypercubes for all our selected models and any required
   * parents, and delete the corresponding Swappable objects. We do this
   * because we will now make use of every one of these models, so it makes no
   * sense to swap any of their hypercubes out any more. We must not do this
   * for any trivial models.
   */
  for (l = 0; l < this->numLabels; l++) {
    for (m = 0; m < this->_numsModels[l]; m++) {
      model = this->_models[l][m];

      if (!model->completeTrivial) {
        swappable_withdraw_delete(&model->_completeHypercubeS);
      }
    }

    for (p = 0; p < this->_numsParents[l]; p++) {
      parent = this->_parents[l][p];

      if (!parent->completeTrivial) {
        swappable_withdraw_delete(&parent->_completeHypercubeS);
      }
    }
  }
}

/* Parallel method for combining a model with the current selection.
 */
static void final_combine_(Parallel *par) {
  ulong e;
  double selected_weight;
  double test_weight;
  double *model_predictions;
  double *selected_predictions;
  double *test_predictions;
  FinalCombinePass_ *pass;

  /* Initialize.
   */
  pass = par->x;
  selected_weight = pass->selectedWeight;
  test_weight = pass->testWeight;
  selected_predictions = pass->selectedPredictions;
  test_predictions = pass->testPredictions;
  model_predictions = pass->modelPredictions;

  /* Walk through the examples for this parallel thread.
   */
  for (e = par->start; e < par->end; e++) {
    /* Combine the predictions, weighted.
     */
    test_predictions[e]
      = (selected_weight * selected_predictions[e]
          + test_weight * model_predictions[e])
      / (selected_weight + test_weight);
  }
}

/* Parallel method for computing the loss of a combined model.
 */
static void final_loss_(Parallel *par) {
  ulong e;
  ulong frequency;
  ulong total_frequency;
  ulong *frequencies;
  double sum;
  double test_estimate;
  double test_prediction;
  double *labels;
  double *test_predictions;
  void *loss_function_params;
  LossFunction *loss_function;
  FinalLossPass_ *pass;

  /* Initialize.
   */
  pass = par->x;
  frequencies = pass->frequencies;
  test_predictions = pass->testPredictions;
  labels = pass->labels;
  loss_function = pass->lossFunction;
  loss_function_params = pass->lossFunctionParams;

  /* Walk through the examples for this parallel thread.
   */
  for (e = par->start, sum = 0, total_frequency = 0; e < par->end; e++) {
    /* Extract the frequency for this example and add it to the total.
     */
    total_frequency += (frequency = frequencies[e]);

    /* Extract the prediction for this example.
     */
    test_prediction = test_predictions[e];

    /* Compute the corresponding estimate.
     */
    test_estimate = loss_function->computeLabelEstimate(
        test_prediction,
        loss_function_params);

    /* Compute the loss for this example, and add it to our sum.
     */
    sum += frequency * loss_function->computeLoss(
        labels[e],
        test_estimate,
        loss_function_params);
  }

  /* Update the totals, under the lock.
   */
  lock(&pass->lock);
  pass->sum += sum;
  pass->totalFrequency += total_frequency;
  unlock(&pass->lock);
}

/* Print method for the models in the RandomWeighted.
 */
static void rw_print_(BearModel *this, Printer *p) {
  sprint(p, "BearModel");
  bear_assembly_print(this->_assembly, p);
  fprint(p);
}

/* Check that all vetoes are cleared for all labels.
 */
static void check_vetoes_(CompositePass_ *pass) {
  ulong l;

  /* Walk through all the labels.
   */
  for (l = 0; l < pass->this->numLabels; l++) {
    /* Sanity check.
     */
    if (pass->rws[l]->vetoedNum != 0) {
      throw("Impossible: label %lu has %lu vetoes",
          l,
          pass->rws[l]->vetoedNum);
    }
  }
}

/* Check if there are no active nodes for any label.
 */
static bool no_active_nodes_(CompositePass_ *pass) {
  ulong l;

  /* Walk through all the labels.
   */
  for (l = 0; l < pass->this->numLabels; l++) {
    /* If the RandomWeighted is not empty, then we have at least one active
     * node.
     */
    if (!random_weighted_is_empty(pass->rws[l])) {
      return false;
    }
  }

  /* We didn't find any labels with anything in their RandomWeighted. There are
   * no active nodes.
   */
  return true;
}

/* Find the label that has nodes available that has the least number of models
 * already built, randomly tie-breaking if there is more than one.
 */
static ulong min_models_(CompositePass_ *pass) {
  ulong l;
  ulong num_labels;
  LabelInfo_ *_label_infos = null;
  RandomWeighted **rws;
  DynamicSet **models;
  Bear *this;

  /* Extract information.
   */
  this = pass->this;
  num_labels = this->numLabels;
  rws = pass->rws;
  models = pass->models;

  /* Create an array of LabelInfo_ structure objects.
   */
  new(&_label_infos, num_labels);

  /* Walk through the labels.
   */
  for (l = 0; l < num_labels; l++) {
    /* Set the label number.
     */
    _label_infos[l].l = l;

    /* If this label has nodes available, set the number of models already
     * built; otherwise, set this number to ULONG_MAX so that it never wins.
     */
    _label_infos[l].numModels
      = random_weighted_is_empty(rws[l])
      ? ULONG_MAX
      : dynamic_set_cardinality(models[l]);
  }

  /* Sort the LabelInfo_ array in order of the number of models already built.
   */
  qsortx(_label_infos, num_labels, label_info_compare_, true, null);

  /* Sanity check.
   */
  if (_label_infos[0].numModels == ULONG_MAX) {
    throw("Impossible");
  }

  /* Extract the label index of the label with the smallest number of models
   * already built.
   */
  l = _label_infos[0].l;

  /* Clean up.
   */
  deletev(&_label_infos);

  /* Return our desired label value.
   */
  return l;
}

/* Whether a model has been built.
 */
static bool model_built_(DynamicSet *models, BearAssembly *assembly) {
  return try_get_model_(models, assembly) != null;
}

/* Get a built model from the set of models. Throws if the model is not found.
 */
static BearModel *get_model_(DynamicSet *models, BearAssembly *assembly) {
  char *_s = null;
  BearModel *model;

  if ((model = try_get_model_(models, assembly)) == null) {
    aprint(&_s, 1, bear_assembly, assembly);

    throw("Impossible: Could not find model with this assembly:\n%s", _s);
  }

  return model;
}

/* Try to get a built model from the set of models.
 */
static BearModel *try_get_model_(DynamicSet *models, BearAssembly *assembly) {
  BearModel needle;

  needle._assembly = assembly;
  return dynamic_set_find(models, &needle);
}

/* Comparator for LabelInfo_ structures.
 */
sbyte label_info_compare_(LabelInfo_ *a, LabelInfo_ *b, void *x) {
  if (a->numModels < b->numModels) {
    return -1;
  }

  if (a->numModels > b->numModels) {
    return +1;
  }

  return 0;
}

/* Relink the logical parent of a model from its ID and the set of models.
 */
static void relink_(BearModel *model, DynamicReferenceSet *set) {
  ulong parent_id;
  BearModel needle;
  BearModel *parent;

  /* Get the parent ID from the model. If it is zero, return.
   */
  if ((parent_id = model->parentId) == 0) {
    return;
  }

  /* Find the model with this ID.
   */
  needle.id = parent_id;
  parent = dynamic_reference_set_find(set, &needle);

  /* Sanity check.
   */
  if (parent == null) {
    eprint(0, dynamic_reference_set, set);

    throw("Impossible: could not find model ID %lu", parent_id);
  }

  /* Set the parent pointer in the model.
   */
  model->parent = parent;

  /* Sanity check.
   */
  if (model->parent->id != parent_id) {
    throw("Impossible");
  }
}

/* Whether a model is not "substantially useful," i.e., its selection weight is
 * not at least the specified fraction of its weight, or its weight is zero.
 */
static bool not_substantially_useful_(BearModel *model) {
  return model->weight == 0
    || model->selectionWeight / model->weight < SUBSTANTIAL_;
}

/* Throw with a message if an insufficient time budget has been set for even a
 * minimum viable run.
 */
static void time_check_(bool abort, long start_time_us, long max_time_us) {
  char *_s = null;
  TimeBudget tb;

  /* Check if we aborted.
   */
  if (!abort) {
    return;
  }

  /* Reconstruct the time budget.
   */
  tb = time_budget_from_us(max_time_us - start_time_us);

  /* Convert to a string.
   */
  string_new_time_budget(&_s, tb);

  /* Throw.
   */
  throw("Time budget of %s wasn't enough for even a minimal construction; "
      "please increase it",
      _s);
}

/* Same, explicitly checking the clock.
 */
static void time_check_clock_(long start_time_us, long max_time_us) {
  time_check_(unixtime_current_us() > max_time_us,
      start_time_us,
      max_time_us);
}

/* Determine if the total strength is about as good as the best so far.
 */
static bool about_as_good_(double total_strength, double max_total_strength) {
  return total_strength + max_total_strength < SUBSTANTIAL_
    || fabs(total_strength - max_total_strength) * 2
    / (total_strength + max_total_strength) < SUBSTANTIAL_;
}

/* Add the features used by this model to the set of useful features.
 */
static void add_useful_features_(Bear *this, bool *features_are_useful) {
  bool *_features_are_used = null;
  ulong f;

  /* Get the set of flags of used features in this model.
   */
  bear_get_features_are_used(this, &_features_are_used);

  /* Add all the true ones to the set.
   */
  for (f = 0; f < this->numFeatures; f++) {
    if (_features_are_used[f]) {
      features_are_useful[f] = true;
    }
  }

  /* Clean up.
   */
  deletev(&_features_are_used);
}

/* Replace the set of useful features by the features used by this model.
 */
static void replace_useful_features_(
    Bear *this,
    bool **p_features_are_useful) {
  /* Delete the current set.
   */
  deletev(p_features_are_useful);

  /* Create the set as those used in this model.
   */
  bear_get_features_are_used(this, p_features_are_useful);
}
