/* Object class for an in-memory implementation of the Bear machine learning
 * engine.
 *
 * Bear's constructor takes in the training data and a time limit. It attempts
 * a number of "constructions" that spend an exponentially increasing amount of
 * time trying to create models, until one of three things happens: the time
 * budget is exhausted, all possibly useful models have been built, or the loss
 * of a given construction is actually worse than the previous one. In all
 * cases the best overall model is returned; in the first case, a warning is
 * printed that the time budget was exhausted and that a greater budget may
 * result in a better model.
 *
 * Within each construction, two iterations of the model-building process are
 * performed. The first iteration simply provides the Decider with an
 * approximate order of magnitude of the number of decisions that are needed;
 * the models from this first iteration are then thrown away, and constructed
 * anew in the second iteration, where the Decider now knows an appropriate
 * p-value for the approximate number of decisions that will be made. This
 * might seem wasteful, but is the only general and reliable way to avoid
 * multiple-comparisons overfitting. Bear then determines the best combination
 * of these second-iteration models, with the weight of each model taken to be
 * the reciprocal of its loss, that minimizes the loss. The loss function is
 * provided to the constructor as a reference to one of the LossFunction
 * objects, which knows how to perform the operations needed by Bear and its
 * subclasses.
 *
 * The training data 'features' and 'labels' must be arrays of SwappableDoubles
 * objects, indexed by the feature or label respectively; each SwappableDoubles
 * must supply an array of doubles for all examples for that feature or label.
 * Any feature value that is infinite or not a number is considered to be
 * missing. No label may be missing. An optional SwappableUlongs 'frequencies'
 * may be supplied; otherwise, each example will be assumed to have unit
 * frequency.
 *
 * A non-empty model is deemed to be "potentially useful" if it either has a
 * lower loss than the "base" model on which its "attachment" is being attached
 * (see BearAssembly for the "gluing" and "melting" operations), or if it fails
 * this test but its last part is not a repeated part. This allows combinations
 * of features to prove themselves to be useful even if none alone (or in
 * simpler combinations) are immediately useful, while also preventing endless
 * attempts to build with the same parts (which also allows the model-building
 * process to finish, if the combinatorics of the number of features is
 * feasible within the time available). A non-empty model that is not
 * potentially useful is deemed "useless," and is not kept.
 *
 * Any non-empty model that is trivial, i.e., its BearHypercube reduces down to
 * a single cell, is not useful in itself, and is never used as a parent model
 * for any other model. However, if its last part is not a repeated part, it is
 * kept around, as its assembly may be potentially useful as a base for melting
 * other models onto.
 *
 * If the time limit specified is not sufficient for Bear to even set up its
 * data structures, create the empty and elementary models (but no composite
 * models) for each of the two iterations of the first construction, and
 * determine which models to keep, it throws with an error that says that a
 * larger time budget is required. On the other hand, if it gets past this
 * "minimum viable" point, it uses the time expended on each of these
 * operations in its time estimates for further constructions.
 *
 * Bear will report progress through the supplied Logger (which may be created
 * in quiet mode if desired).
 */
#ifndef BEAR_H
#define BEAR_H

#ifdef __cplusplus
extern "C" {
#endif

#include "BearModel.h"
#include "Condition.h"
#include "TimeBudget.h"

#define BEAR_VERSION "0.4.9"

typedef struct {
  ulong numFeatures;
  ulong numLabels;
  ulong *_numsModels;
  ulong *_numsParents;
  double *_strengths;
  double *_sumsWeights;
  LossFunction *lossFunction;
  BearModel **_emptyModels;
  BearModel ***_models;
  BearModel ***_parents;
} Bear;

/* Constructor.
 */
void bear_new(Bear **p_this,
    ulong num_features,
    ulong num_labels,
    ulong num_examples,
    SwappableDoubles **features,
    SwappableDoubles **labels,
    SwappableUlongs *frequencies,
    LossFunction *loss_function,
    long start_time_us,
    long max_time_us,
    Logger *logger);

/* Load.
 */
void bear_new_load(Bear **p_this, File *f);

/* Destructor.
 */
void bear_delete(Bear **p_this);

/* Print.
 */
void bear_print(Bear *this, Printer *p);

/* Save.
 */
void bear_save(Bear *this, File *f);

/* Equals.
 */
bool bear_equals(Bear *this, Bear *other);

/* Predict the array of label values from the given array of feature values.
 */
void bear_predict(Bear *this, double *features, double *labels);

/* Write out a summary CSV file of Bear's final model selection.
 */
void bear_write_model_summary(Bear *this, char *filename);

/* Get an array of flags for each feature indicating if it actually used in
 * Bear's final models. Returns the number of used features.
 */
ulong bear_get_features_are_used(Bear *this, bool **p_features_are_used);

/* Get an array of all the features actually used in Bear's final models.
 * Returns the number ofg used features.
 */
ulong bear_get_used_features(Bear *this, ulong **p_used_features);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _H */
