/* Object class for a Bear assembly, which is the symbolic representation of a
 * sequence of parts "glued together," i.e., each successive part models the
 * quantile residuals of its predecessor, for a given label field.
 *
 * The only restriction on an assembly is that no part in the assembly may be
 * repeated anywhere else in the assembly. This ensures that the number of
 * possible assemblies for a given number of features is finite.
 *
 * Immutable class.
 */
#ifndef BEAR_ASSEMBLY_H
#define BEAR_ASSEMBLY_H

#ifdef __cplusplus
extern "C" {
#endif

#include "BearPart.h"
#include "qsortx.h"

typedef struct BearAssembly {
  ulong numParts;
  BearPart **_parts;
} BearAssembly;

/* Constructor for an empty assembly, i.e., an assembly with no parts. Note
 * that this assembly cannnot be joined with any other.
 */
void bear_assembly_new_empty(BearAssembly **p_this);

/* Constructor for an elementary assembly, i.e., an assembly with just one
 * elementary part with just one feature.
 */
void bear_assembly_new_elementary(BearAssembly **p_this, ulong feature);

/* Copy constructor.
 */
void bear_assembly_new_copy(BearAssembly **p_this, BearAssembly *other);

/* Constructor for the parent of the given assembly.
 */
void bear_assembly_new_parent(BearAssembly **p_this, BearAssembly *other);

/* Constructor for "gluing" an "attachment" part to the end of a "base"
 * assembly.
 */
void bear_assembly_new_glue(
    BearAssembly **p_this,
    BearAssembly *base,
    BearPart *attachment);

/* Constructor for "melting" an "attachment" part to the last part of a "base"
 * assembly. Throws if can't be so melted.
 */
void bear_assembly_new_melt(
    BearAssembly **p_this,
    BearAssembly *base,
    BearPart *attachment);

/* Load.
 */
void bear_assembly_new_load(BearAssembly **p_this, File *f);

/* Destructor.
 */
void bear_assembly_delete(BearAssembly **p_this);

/* Print.
 */
void bear_assembly_print(BearAssembly *this, Printer *p);

/* Save.
 */
void bear_assembly_save(BearAssembly *this, File *f);

/* Equals.
 */
bool bear_assembly_equals(BearAssembly *this, BearAssembly *other);

/* Create a string with a single-line representation of the assembly.
 */
void string_new_bear_assembly(char **p_this, BearAssembly *assembly);

/* Whether an assembly is empty.
 */
bool bear_assembly_is_empty(BearAssembly *this);

/* Whether the last part of an assembly is a repeat.
 */
bool bear_assembly_last_part_is_repeat(BearAssembly *this);

/* Whether an assembly is the parent of another; i.e., the other assembly is
 * this assembly with an extra part glued to it.
 */
bool bear_assembly_is_parent(BearAssembly *this, BearAssembly *other);

/* Whether an assembly is a sibling of another; i.e., the parent of the other
 * assembly is the same as the parent of this assembly, but the other assembly
 * is not equal to this one.
 */
bool bear_assembly_is_sibling(BearAssembly *this, BearAssembly *other);

/* Whether a base assembly could have the given attachment part melted to its
 * last part.
 */
bool bear_assembly_can_melt(BearAssembly *base, BearPart *attachment);

/* Return a pointer to the first part of an assembly. Throws if it is an empty
 * assembly.
 */
BearPart *bear_assembly_first_part(BearAssembly *this);

/* Return a pointer to the last part of an assembly. Throws if it is an empty
 * assembly.
 */
BearPart *bear_assembly_last_part(BearAssembly *this);

/* Comparator. Prefers a lower number of parts, and then compares the parts.
 */
sbyte bear_assembly_compare(BearAssembly *a, BearAssembly *b, void *x);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _H */
