#include "BearField.h"
#include <limits.h>

#define NUM_US_ 0x10000
#define NUM_NORMALS_ (2 * 126 * 256 + 1)
#define NUM_ABNORMALS_ (NUM_US_ - NUM_NORMALS_)

static bool init_ = false;
static ushort *_abnormals_ = null;
static paw *_normals_ = null;
static Lock lock_ = LOCK_INITIALIZER;

static void ensure_init_(void);

static paw right_(paw *values, ushort c, ushort i);

static void check_binned_(BearField *this, char *verb);


void bear_field_new(BearField **p_this, bool feature) {
  BearField *this;

  this = newo(p_this);
  this->feature = feature;
  new(&this->_rawFrequencies, NUM_US_);
  new(&this->_valueSums, NUM_US_);

  if (feature) {
    new(&this->_residualSums, NUM_US_);
  }
}

void bear_field_new_copy(BearField **p_this, BearField *other) {
  BearField *this;

  this = newo(p_this);
  this->feature = other->feature;
  this->cardinality = other->cardinality;
  this->totalFrequency = other->totalFrequency;

  if (other->_lookups != null) {
    new_copy(&this->_lookups, NUM_US_, other->_lookups);
  }

  if (other->_values != null) {
    new_copy(&this->_values, this->cardinality, other->_values);
  }

  if (other->_frequencies != null) {
    new_copy(&this->_frequencies, this->cardinality, other->_frequencies);
  }

  if (other->_rawFrequencies != null) {
    new_copy(&this->_rawFrequencies, NUM_US_, other->_rawFrequencies);
  }

  if (other->_meanResiduals != null) {
    new_copy(&this->_meanResiduals, this->cardinality, other->_meanResiduals);
  }

  if (other->_meanValues != null) {
    new_copy(&this->_meanValues, this->cardinality, other->_meanValues);
  }

  if (other->_residualSums != null) {
    new_copy(&this->_residualSums, NUM_US_, other->_residualSums);
  }

  if (other->_valueSums != null) {
    new_copy(&this->_valueSums, NUM_US_, other->_valueSums);
  }
}

void bear_field_new_load(BearField **p_this, File *f) {
  ushort a;
  ushort i;
  ushort n;
  ushort num_transitions;
  ushort *lookups;
  ushort *_transitions = null;
  BearField *this;

  this = newo(p_this);
  this->feature = true;
  load_label("[BearField]", f);
  load_value(num_transitions, f);
  load_new_values(_transitions, num_transitions, f);
  lookups = new(&this->_lookups, NUM_US_);
  ensure_init_();

  for (n = 0, i = 0; n < NUM_NORMALS_; n++) {
    if (i < num_transitions && n == _transitions[i]) {
      i++;
    }

    lookups[_normals_[n].u] = i;
  }

  deletev(&_transitions);

  for (a = 0; a < NUM_ABNORMALS_; a++) {
    lookups[_abnormals_[a]] = USHRT_MAX;
  }

  load_label("[/BearField]", f);
}

void bear_field_delete(BearField **p_this) {
  BearField *this;

  if ((this = *p_this) == null) {
    return;
  }

  deletev(&this->_lookups);
  deletev(&this->_meanResiduals);
  deletev(&this->_meanValues);
  deletev(&this->_frequencies);
  deletev(&this->_values);
  deletev(&this->_residualSums);
  deletev(&this->_valueSums);
  deletev(&this->_rawFrequencies);
  deleteo(p_this);
}

void bear_field_print(BearField *this, Printer *p) {
  ushort a;
  ushort i;
  ushort l;
  ushort n;
  ushort u;
  ulong frequency;
  paw v;

  sprint(p, "BearField");
  rprint(p, "Type of field: %s", this->feature ? "feature" : "residual");

  if (this->_rawFrequencies != null) {
    rprint(p, "Status: building");
    sprint(p, "Current entries");
    ensure_init_();

    for (n = 0, i = 0; n < NUM_NORMALS_; n++) {
      if ((frequency = this->_rawFrequencies[u = (v = _normals_[n]).u]) > 0) {
        sprint(p, "Index %u", i++);
        rprint(p, "Paw value: %.4g", double_from_paw(v));
        rprint(p, "Frequency: %lu", frequency);
        rprint(p, "Mean value: %.16g", this->_valueSums[u] / frequency);

        if (this->feature) {
          rprint(p,
              "Mean residual: %.16g",
              this->_residualSums[u] / frequency);
        }

        fprint(p);
      }
    }

    fprint(p);
  } else if (this->_values != null) {
    rprint(p, "Status: finalized, not binned");
    rprint(p, "Cardinality: %u", this->cardinality);
    sprint(p, "Entries");

    for (i = 0; i < this->cardinality; i++) {
      sprint(p, "Index %u", i);
      rprint(p, "Paw value: %.4g", double_from_paw(this->_values[i]));
      rprint(p, "Frequency: %lu", this->_frequencies[i]);
      rprint(p, "Mean value: %.16g", this->_meanValues[i]);

      if (this->feature) {
        rprint(p, "Mean residual: %.16g", this->_meanResiduals[i]);
      }

      fprint(p);
    }

    fprint(p);
  } else {
    rprint(p, "Status: binned");
    sprint(p, "Lookup transitions");

    for (n = 0, i = 0; n < NUM_NORMALS_; n++) {
      if ((l = this->_lookups[(v = _normals_[n]).u]) != i) {
        if (l != i + 1) {
          throw("Impossible: %u -> %u", i, l);
        }

        rprint(p, "%u -> %u: %.4g", i, l, double_from_paw(v));
        i++;
      }
    }

    for (a = 0; a < NUM_ABNORMALS_; a++) {
      if (this->_lookups[_abnormals_[a]] != USHRT_MAX) {
        throw("Impossible: %u -> %u -> %u",
            a,
            _abnormals_[a],
            this->_lookups[_abnormals_[a]]);
      }
    }

    fprint(p);
  }

  rprint(p, "Total frequency: %lu", this->totalFrequency);
  fprint(p);
}

void bear_field_save(BearField *this, File *f) {
  ushort i;
  ushort l;
  ushort n;
  ushort num_transitions;
  ushort *lookups;
  ushort *_transitions = null;

  check_binned_(this, "save");
  save_label("[BearField]", f);
  num_transitions = this->_lookups[_normals_[NUM_NORMALS_ - 1].u];
  save_value(num_transitions, f);
  new(&_transitions, num_transitions);
  lookups = this->_lookups;

  for (n = 0, i = 0; n < NUM_NORMALS_; n++) {
    if ((l = lookups[_normals_[n].u]) > i) {
      if (l != i + 1) {
        throw("Impossible: %u -> %u -> %u vs %u",
            n,
            _normals_[n].u,
            l,
            i);
      }

      _transitions[i++] = n;
    }
  }

  if (i != num_transitions) {
    throw("Impossible: %u v %u", i, num_transitions);
  }

  save_values(_transitions, num_transitions, f);
  deletev(&_transitions);
  save_label("[/BearField]", f);

  this->feature = true;
  this->cardinality = 0;
  this->totalFrequency = 0;
  deletev(&this->_values);
  deletev(&this->_frequencies);
  deletev(&this->_rawFrequencies);
  deletev(&this->_meanResiduals);
  deletev(&this->_meanValues);
  deletev(&this->_residualSums);
  deletev(&this->_valueSums);
}

bool bear_field_equals(BearField *this, BearField *other) {
  ushort c;
  uint u;

  if (other->feature != this->feature
      || (c = other->cardinality) != this->cardinality
      || other->totalFrequency != this->totalFrequency) {
    return false;
  }

  u = NUM_US_;

  if (other->_lookups != null
      && !values_equal(other->_lookups, this->_lookups, u)) {
    return false;
  }

  if (other->_values != null
      && !values_equal(other->_values, this->_values, c)) {
    return false;
  }

  if (other->_frequencies != null
      && !values_equal(other->_frequencies, this->_frequencies, c)) {
    return false;
  }

  if (other->_rawFrequencies != null
      && !values_equal(other->_rawFrequencies, this->_rawFrequencies, u)) {
    return false;
  }

  if (other->_meanResiduals != null
      && !values_equal(other->_meanResiduals, this->_meanResiduals, c)) {
    return false;
  }

  if (other->_meanValues != null
      && !values_equal(other->_meanValues, this->_meanValues, c)) {
    return false;
  }

  if (other->_residualSums != null
      && !values_equal(other->_residualSums, this->_residualSums, u)) {
    return false;
  }

  if (other->_valueSums != null
      && !values_equal(other->_valueSums, this->_valueSums, u)) {
    return false;
  }

  return true;
}


void bear_field_merge(BearField *this, BearField **p_other) {
  uint u;
  BearField *other;

  other = *p_other;
  this->totalFrequency += other->totalFrequency;

  if (this->_rawFrequencies != null) {
    for (u = 0; u < 0x10000; u++) {
      this->_rawFrequencies[u] += other->_rawFrequencies[u];
      this->_valueSums[u] += other->_valueSums[u];

      if (this->feature) {
        this->_residualSums[u] += other->_residualSums[u];
      }
    }
  }

  bear_field_delete(p_other);
}

void bear_field_add(
    BearField *this,
    double value,
    ulong frequency,
    double residual) {
  ushort u;
  paw p;

  if (this->_rawFrequencies == null) {
    throw("Already finalized");
  }

  if (frequency == 0) {
    throw("Frequency cannot be zero");
  }

  if (double_pawized_is_infinite_or_nan(residual)) {
    throw("Cannot add the paw-infinite or NaN residual %.16g", residual);
  }

  if (!this->feature && residual != value) {
    throw("Residual (%.16g) must equal value (%.16g) for a residual field",
        residual,
        value);
  }

  this->totalFrequency += frequency;
  p = paw_make(value);

  if (paw_is_infinite_or_nan(p)) {
    throw("Cannot add the paw-infinite or NaN value %.16g", value);
  }

  u = p.u;

  this->_rawFrequencies[u] += frequency;
  this->_valueSums[u] += value * frequency;

  if (this->feature) {
    this->_residualSums[u] += residual * frequency;
  }
}

void bear_field_finalize(BearField *this) {
  bool feature;
  ushort c;
  ushort i;
  ushort n;
  ushort u;
  ushort *lookups;
  ulong frequency;
  ulong *frequencies;
  ulong *raw_frequencies;
  paw v;
  paw *values;
  double *mean_residuals;
  double *mean_values;
  double *residual_sums;
  double *value_sums;

  if ((raw_frequencies = this->_rawFrequencies) == null) {
    throw("Already finalized");
  }

  value_sums = this->_valueSums;
  values = new(&this->_values, NUM_US_);
  frequencies = new(&this->_frequencies, NUM_US_);
  mean_values = new(&this->_meanValues, NUM_US_);

  if ((feature = this->feature)) {
    residual_sums = this->_residualSums;
    mean_residuals = new(&this->_meanResiduals, NUM_US_);
  }

  ensure_init_();

  for (n = 0, c = 0; n < NUM_NORMALS_; n++) {
    if ((frequency = raw_frequencies[u = (v = _normals_[n]).u]) > 0) {
      values[c] = v;
      frequencies[c] = frequency;
      mean_values[c] = value_sums[u] / frequency;

      if (feature) {
        mean_residuals[c] = residual_sums[u] / frequency;
      }

      c++;
    }
  }

  if ((this->cardinality = c) == 0) {
    throw("No values added");
  }

  values = renewv(&this->_values, NUM_US_, c);
  renewv(&this->_frequencies, NUM_US_, c);
  renewv(&this->_meanValues, NUM_US_, c);
  deletev(&this->_rawFrequencies);
  deletev(&this->_valueSums);

  if (feature) {
    renewv(&this->_meanResiduals, NUM_US_, c);
    deletev(&this->_residualSums);
  }

  lookups = new(&this->_lookups, NUM_US_);

  for (i = 0; i < c; i++) {
    lookups[values[i].u] = i;
  }
}

void bear_field_bin(BearField *this, ushort *bins_from_indexes) {
  ushort a;
  ushort c;
  ushort i;
  ushort n;
  ushort *lookups;
  paw right;
  paw v;
  paw *values;

  if (!this->feature) {
    throw("Can only bin feature fields");
  }

  if ((lookups = this->_lookups) == null) {
    throw("Not finalized");
  }

  if ((values = this->_values) == null) {
    throw("Already binned");
  }

  if ((c = this->cardinality) == 0) {
    throw("Impossible");
  }

  for (n = 0, i = 0, right = right_(values, c, i); n < NUM_NORMALS_; n++) {
    if (paw_ge(v = _normals_[n], right)) {
      if (paw_ge(v, right = right_(values, c, ++i))) {
        throw("Impossible");
      }
    }

    lookups[v.u] = bins_from_indexes[i];
  }

  for (a = 0; a < NUM_ABNORMALS_; a++) {
    lookups[_abnormals_[a]] = USHRT_MAX;
  }

  deletev(&this->_values);
  deletev(&this->_frequencies);
  deletev(&this->_meanValues);
  deletev(&this->_meanResiduals);
  this->cardinality = 0;
}

ushort bear_field_lookup(BearField *this, double value) {
  return this->_lookups[paw_make(value).u];
}


static void ensure_init_(void) {
  ushort a;
  ushort n;
  uint u;

  if (!init_) {
    lock(&lock_);

    if (!init_) {
      new(&_normals_, NUM_NORMALS_);
      new(&_abnormals_, NUM_ABNORMALS_);

      for (u = 0, n = 0, a = 0; u < NUM_US_; u++) {
        if (u == 0 || ((u & 0x7f00) != 0 && (u & 0x7f00) != 0x7f00)) {
          _normals_[n++].u = u;
        } else {
          _abnormals_[a++] = u;
        }
      }

      if (n != NUM_NORMALS_ || a != NUM_ABNORMALS_) {
        throw("Impossible");
      }

      qsortx(_normals_, NUM_NORMALS_, paw_compare, false, null);
      init_ = true;
    }

    unlock(&lock_);
  }
}

static paw right_(paw *values, ushort c, ushort i) {
  paw n;
  paw right;
  paw v;

  if (i == c - 1) {
    return PAW_POSINF;
  }

  v = values[i];
  n = values[i + 1];

  if (paw_ge(paw_max(paw_abs(v), paw_abs(n)), PAW_ONE)) {
    right = paw_add(paw_multiply(v, PAW_HALF), paw_multiply(n, PAW_HALF));
  } else {
    right = paw_multiply(paw_add(v, n), PAW_HALF);
  }

  if (paw_equals(right, v)) {
    return n;
  }

  return right;
}

static void check_binned_(BearField *this, char *verb) {
  if (this->_lookups == null || this->_values != null) {
    throw("Can only %s when binned", verb);
  }
}
