/* Object class for a Bear field, which quantizes the double values supplied to
 * it into paw values, keeping track of the frequency of each quantized paw
 * value added, and the frequency-weighted mean (double) value added to each
 * paw value. For feature fields it also keeps track of the frequency-weighted
 * mean (double) residual for each paw value.
 *
 * When finalized, ths class builds a lookup table from paw values to indexes,
 * stores an array of just these paw values (effectively, a lookup table from
 * index back to paw value), and a corresponding array of the mean (double)
 * value added to each paw value. For feature fields it likewise stores a
 * corresponding array of the mean (double) residual for each paw value.
 *
 * For a feature field, a final "binning" stage can be applied, where a mapping
 * of paw value indexes to bins is supplied. First, the lookup table of paw
 * values to indexes is extended to all possible paw values, in intervals
 * extending half-way in paw value space to the adjacent used paw values. Next,
 * all indexes are then replaced by the bins they map to according to the
 * supplied mapping, and the frequency, mean (double) value, and, for a feature
 * field, mean (double) residual for each bin is computed.
 *
 * Because paws are 16-bit, these algorithms are efficient, since lookup tables
 * never need to have more than 64K entries. This class encpasulates the use of
 * paws completely; all interfaces to the class pass in or return doubles.
 *
 * This class is not set up for values to be added in parallel; if that is
 * needed in future, then individual BearField objects would need to be
 * constructed serially, and then merged together with new methods.
 */
#ifndef BEAR_FIELD_H
#define BEAR_FIELD_H

#ifdef __cplusplus
extern "C" {
#endif

#include "paw.h"

typedef struct {
  bool feature;
  ushort cardinality;
  ushort *_lookups;
  paw *_values;
  ulong totalFrequency;
  ulong *_frequencies;
  ulong *_rawFrequencies;
  double *_meanResiduals;
  double *_meanValues;
  double *_residualSums;
  double *_valueSums;
} BearField;

/* Constructor.
 */
void bear_field_new(BearField **p_this, bool feature);

/* Copy constructor.
 */
void bear_field_new_copy(BearField **p_this, BearField *other);

/* Load.
 */
void bear_field_new_load(BearField **p_this, File *f);

/* Destructor.
 */
void bear_field_delete(BearField **p_this);

/* Print.
 */
void bear_field_print(BearField *this, Printer *p);

/* Save.
 */
void bear_field_save(BearField *this, File *f);

/* Equals.
 */
bool bear_field_equals(BearField *this, BearField *other);

/* Merge another field into this one, deleting the other afterwards.
 */
void bear_field_merge(BearField *this, BearField **p_other);

/* Add a value with a given frequency, and the corresponding residual value.
 */
void bear_field_add(
    BearField *this,
    double value,
    ulong frequency,
    double residual);

/* Finalize.
 */
void bear_field_finalize(BearField *this);

/* Bin the lookups, as described above.
 */
void bear_field_bin(BearField *this, ushort *bins_from_indexes);

/* Look up the index or bin corresponding to a given value.
 */
ushort bear_field_lookup(BearField *this, double value);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _H */
