#include "BearHypercube.h"
#include <float.h>
#include <limits.h>
#include <math.h>

#define P_VALUE_ 0.05

typedef struct {
  ulong frequency;
  ulong numPairs;
  ulong pairOffset;
  double meanResidual;
  double meanValue;
} IndexInfo_;

typedef struct {
  bool feature;
  ushort cardinality;
  ushort origCardinality;
  ulong f;
  IndexInfo_ *_indexInfos;
} FieldInfo_;

typedef struct {
  bool feature;
  ulong leftFrequency;
  ulong leftIndex;
  ulong random;
  ulong rightFrequency;
  ulong rightIndex;
  double chiSquared;
  double deltaResidual;
  double deltaValue;
} Candidate_;

typedef struct {
  ulong numThreadsInitialized;
  ulong *frequencies;
  double **values;
  Lock lock;
  Condition condition;
  BearHypercube *this;
} AddPass_;

typedef struct {
  ushort indexL;
  ushort indexR;
  ulong f;
  KeyValueArray *hyperslice;
  KeyValueInfo *info;
} RekeyPass_;

typedef struct {
  ushort **binsFromIndexes;
  KeyValueArray *reduced;
  BearHypercube *this;
} BinPass_;

typedef struct {
  ushort numResidualBins;
  UlongHypercube *reduced;
} FrequenciesPass_;

typedef struct {
  double meanResidual;
  UlongHypercube *frequencies;
} PredictionsPass_;

static sbyte dominant_compare_(
    KeyValuePair *a,
    KeyValuePair *b,
    KeyValueInfo *info);

static sbyte non_dominant_compare_(
    KeyValuePair *a,
    KeyValuePair *b,
    KeyValueInfo *info);

static void working_delete_(BearHypercubeWorking **p_this);

static sbyte field_info_compare_cardinality_(
    FieldInfo_ *a,
    FieldInfo_ *b,
    void *x);

static sbyte field_info_compare_f_(
    FieldInfo_ *a,
    FieldInfo_ *b,
    void *x);

static sbyte candidate_compare_(Candidate_ *a, Candidate_ *b, void *x);

static void candidate_print_(Candidate_ *this, Printer *p);

static void candidate_heap_print_(HeapObjectPrintInfo *hopi, Printer *p);

static double chi_squared_(
    KeyValueArray *cube,
    IndexInfo_ *index_infos,
    ushort i_l,
    ushort i_r,
    ulong f);

static void rekey_(Parallel *par);

static void bin_(Parallel *par);

static void add_(Parallel *par);

static void compute_frequencies_(
    Hypercube *this,
    bool *walk_dimensions,
    ulong *indexes,
    ulong o,
    void *x);

static void scale_predictions_(
    Hypercube *this,
    bool *walk_dimensions,
    ulong *indexes,
    ulong o,
    void *x);

static void merge_predictions_(
    Hypercube *this,
    bool *walk_dimensions,
    ulong *indexes,
    ulong o,
    void *x);

static void field_info_print_(FieldInfo_ *this, Printer *p);

static void create_candidate_(
    KeyValueArray *reduced,
    IndexInfo_ *index_infos,
    ushort i_l,
    ushort i_r,
    ulong f,
    double critical_chi_squared,
    bool feature,
    Heap *candidates,
    Random *random);

static ulong offset_(BearHypercube *this, double *features);

static byte iteration_(BearHypercube *this);

/* General constructor allowing streaming in an arbitrary amount of data, which
 * needs to be done in three iterations, with bear_hypercube_end_iteration()
 * being called at the end of each iteration. If bear_hypercube_is_trivial()
 * returns true after the second iteration, the object can no longer be used,
 * and a third iteration cannot be attempted.
 */
void bear_hypercube_new(
    BearHypercube **p_this,
    ulong num_features,
    Decider *decider) {
  ulong f;
  ulong num_fields;
  BearHypercubeWorking *w;
  BearHypercube *this;

  /* Sanity check.
   */
  if (num_features == 0) {
    throw("Must have at least one feature");
  }

  /* Initialize.
   */
  this = newo(p_this);
  num_fields = (this->numFeatures = num_features) + 1;
  w = newo(&this->_w);
  w->decider = decider;

  /* Create the array for the fields.
   */
  new(&this->_fields, num_fields);

  /* Create the fields.
   */
  for (f = 0; f < num_features + 1; f++) {
    bear_field_new(&this->_fields[f], f < num_features);
  }
}

/* Simplified constructor when the input data fits in memory. The 'values'
 * matrix must have the example index first and the field index second. The
 * residual field must be the last field, after the features.
 */
void bear_hypercube_new_memory(
    BearHypercube **p_this,
    ulong num_features,
    Decider *decider,
    ulong num_examples,
    double **values,
    ulong *frequencies) {
  byte iteration;
  AddPass_ pass;
  BearHypercube *this;

  /* Call the general constructor.
   */
  bear_hypercube_new(p_this, num_features, decider);
  this = *p_this;

  /* Initialize the pass object.
   */
  lock_initialize(&pass.lock);
  condition_initialize(&pass.condition);

  /* Walk through the iterations.
   */
  for (iteration = 0; iteration < 3; iteration++) {
    /* Add the data in parallel. The parallel method does all the magic that
     * makes it seem that our hypercube is being added to in parallel.
     */
    pass.numThreadsInitialized = 0;
    pass.this = this;
    pass.values = values;
    pass.frequencies = frequencies;
    parallel_for(
        num_examples,
        iteration == 0 ? 1 : iteration == 1 ? 1 : 1,
        add_,
        &pass);

    /* End the iteration.
     */
    bear_hypercube_end_iteration(this);

    /* If the hypercube is trivial, don't attempt another iteration.
     */
    if (bear_hypercube_is_trivial(this)) {
      break;
    }
  }
}

/* Load.
 */
void bear_hypercube_new_load(BearHypercube **p_this, File *f) {
  BearHypercube *this;

  this = newo(p_this);
  load_label("[BearHypercube]", f);
  load_value(this->numFeatures, f);
  new(&this->_fields, this->numFeatures + 1);
  load_array(this->_fields, this->numFeatures, bear_field, f);
  load(this->_predictions, hypercube, f);
  load_label("[/BearHypercube]", f);
}

/* Copy.
 */
void bear_hypercube_new_copy(BearHypercube **p_this, BearHypercube *other) {
  char *_e = null;
  char *_s = null;
  ulong f;
  ulong num_fields;
  KeyValueArray *_copy_data = null;
  KeyValueArray *_temp_data = null;
  BearHypercubeWorking *other_w;
  BearHypercubeWorking *w;
  BearHypercube *this;

  this = newo(p_this);
  num_fields = (this->numFeatures = other->numFeatures) + 1;

  if (other->_fields != null) {
    new(&this->_fields, num_fields);

    for (f = 0; f < num_fields; f++) {
      bear_field_new_copy(&this->_fields[f], other->_fields[f]);
    }
  }

  if (other->_predictions != null) {
    hypercube_new_copy(&this->_predictions, other->_predictions);
  }

  if ((other_w = other->_w) != null) {
    w = newo(&this->_w);
    w->totalFrequency = other_w->totalFrequency;
    w->totalFrequencyCheck = other_w->totalFrequencyCheck;
    w->meanResidual = other_w->meanResidual;
    w->sumResidualFrequency = other_w->sumResidualFrequency;
    w->decider = other_w->decider;

    if (other_w->_info != null) {
      key_value_info_new_copy(&w->_info, other_w->_info);

      if (other_w->_pair != null) {
        key_value_pair_new(&w->_pair, other_w->_info);
      }
    }

    if (other_w->_data != null) {
      key_value_array_new_from_collection(&_temp_data, &other_w->_data);
      key_value_array_new_copy(&_copy_data, _temp_data);
      key_value_collection_new_from_array(&w->_data, &_copy_data);
      key_value_collection_new_from_array(&other_w->_data, &_temp_data);
    }

    if (other_w->_frequencies != null) {
      hypercube_new_copy(&w->_frequencies, other_w->_frequencies);
    }
  }

  if (!bear_hypercube_equals(this, other)) {
    string_new_copy(&_e, "Impossible: this != other:\n  This:\n");
    aprint(&_s, 2, bear_hypercube, this);
    string_renew_append(&_e, _s);
    string_delete(&_s);
    string_renew_append(&_e, "\n  Other:\n");
    aprint(&_s, 2, bear_hypercube, other);
    string_renew_append(&_e, _s);
    string_delete(&_s);

    throw(_e);
  }
}

/* Destructor.
 */
void bear_hypercube_delete(BearHypercube **p_this) {
  BearHypercube *this;

  if ((this = *p_this) == null) {
    return;
  }

  working_delete_(&this->_w);
  hypercube_delete(&this->_predictions);
  delete(&this->_fields, this->numFeatures + 1, bear_field);
  deleteo(p_this);
}

/* Print.
 */
void bear_hypercube_print(BearHypercube *this, Printer *p) {
  byte iteration;
  ulong f;
  BearHypercubeWorking *w;

  sprint(p, "BearHypercube");
  rprint(p, "Number of features: %lu", this->numFeatures);

  if (bear_hypercube_is_trivial(this)) {
    rprint(p, "This hypercube is trivial.");
  }

  rprint(p, "Iteration: %u", iteration = iteration_(this));

  if (this->_fields != null) {
    for (f = 0; f <= this->numFeatures; f++) {
      if (f < this->numFeatures || this->_fields[f] != null) {
        sprint(p, "Field %lu", f);
        bear_field_print(this->_fields[f], p);
        fprint(p);
      }
    }
  }

  if ((w = this->_w) != null) {
    rprint(p, "Total frequency: %lu", w->totalFrequency);
    rprint(p, "Total frequency check: %lu", w->totalFrequencyCheck);
    rprint(p, "Mean residual: %.16g", w->meanResidual);
    rprint(p,
        "Sum of residuals x frequencies: %.16g",
        w->sumResidualFrequency);

    if (w->_data != null) {
      key_value_collection_print(w->_data, p);
    }

    if (w->_frequencies != null) {
      rprint(p, "Frequencies:");
      hypercube_print(w->_frequencies, p);
    }

  }

  if (this->_predictions != null) {
    sprint(p, "Predictions");
    hypercube_print(this->_predictions, p);
    fprint(p);
  }

  fprint(p);
}

/* Save.
 */
void bear_hypercube_save(BearHypercube *this, File *f) {
  if (iteration_(this) != 3 || bear_hypercube_is_trivial(this)) {
    if (iteration_(this) != 3) {
      throw("Impossible: iteration is %u, not 3", iteration_(this));
    }

    throw("Impossible: hypercube is trivial");
  }

  save_label("[BearHypercube]", f);
  save_value(this->numFeatures, f);
  save_array(this->_fields, this->numFeatures, bear_field, f);
  save(this->_predictions, hypercube, f);
  save_label("[/BearHypercube]", f);
  working_delete_(&this->_w);
}

/* Equals.
 */
bool bear_hypercube_equals(BearHypercube *this, BearHypercube *other) {
  ulong f;

  if (other->numFeatures != this->numFeatures) {
    return false;
  }

  for (f = 0; f < this->numFeatures; f++) {
    if (!bear_field_equals(other->_fields[f], this->_fields[f])) {
      return false;
    }
  }

  if (this->_predictions != null
      && !hypercube_equals(other->_predictions, this->_predictions)) {
    return false;
  }

  return true;
}

/* Add an example, for any iteration.
 */
void bear_hypercube_add(BearHypercube *this, double *values, ulong frequency) {
  byte iteration;
  ushort i;
  ulong f;
  ulong num_features;
  ulong o;
  double residual;
  double sum;
  BearField *field;
  BearHypercubeWorking *w;

  /* Figure out which iteration we are on.
   */
  iteration = iteration_(this);

  /* Sanity checks.
   */
  if (frequency == 0) {
    throw("Cannot add zero frequency");
  }

  if (bear_hypercube_is_trivial(this)) {
    throw("Hypercube was already determined to be trivial");
  }

  if (iteration == 3) {
    throw("Cannot add after the third iteration has been finalized");
  }

  /* Extract information.
   */
  w = this->_w;
  residual = values[num_features = this->numFeatures];

  /* Switch on the iteration.
   */
  switch(iteration) {
    case 0:
      /* First iteration. Add to each field the corresponding value.
       */
      for (f = 0; f <= num_features; f++) {
        bear_field_add(this->_fields[f], values[f], frequency, residual);
      }

      /* Update the total frequency.
       */
      w->totalFrequency += frequency;

      /* Update the sum of residual x frequency.
       */
      w->sumResidualFrequency += residual * frequency;

      break;

    case 1:
      /* Second iteration. Walk through the fields.
       */
      for (f = 0; f <= num_features; f++) {
        /* Get reference to this field.
         */
        field = this->_fields[f];

        /* Look up the index for this field from its value.
         */
        i = bear_field_lookup(field, values[f]);

        /* Set the key index.
         */
        key_value_pair_set_key(w->_pair, w->_info, f, i);
      }

      /* Set the frequency for the data pair in the object.
       */
      sparse_key_frequency_pair_set_frequency(w->_pair, w->_info, frequency);

      /* Add this data pair to the collection.
       */
      key_value_collection_add(w->_data, w->_pair);

      /* Update the total frequency check.
       */
      w->totalFrequencyCheck += frequency;

      break;

    case 2:
      /* Third iteration. Get the offset for this set of features.
       */
      o = offset_(this, values);

      /* Extract the current sum of frequency x residual.
       */
      hypercube_get(this->_predictions, o, &sum);

      /* Add in this frequency x residual.
       */
      sum += frequency * residual;

      /* Store it back.
       */
      hypercube_set(this->_predictions, o, &sum);

      /* Update the total frequency check.
       */
      w->totalFrequencyCheck += frequency;

      break;

    default:
      throw("Impossible");
  }
}

/* End an iteration. If the hypercube is found to be trivial, i.e., reduces to
 * or will reduce to a single cell, then the working object is destroyed, and
 * the hypercube should be destroyed by the calling code; it cannot be used
 * further.
 */
void bear_hypercube_end_iteration(BearHypercube *this) {
  bool changed;
  bool done_merge;
  bool feature;
  bool trivial;
  byte iteration;
  ushort b;
  ushort cardinality;
  ushort current_i;
  ushort i;
  ushort i_l;
  ushort i_l_l;
  ushort i_l_l_rev;
  ushort i_r;
  ushort num_residual_bins;
  ushort orig_cardinality;
  ushort **_bins_from_indexes = null;
  ulong combined_frequency;
  ulong dominant_f;
  ulong f;
  ulong fi;
  ulong frequency_l;
  ulong frequency_r;
  ulong num_decisions;
  ulong num_features;
  ulong num_fields;
  ulong num_pairs;
  ulong other_fi;
  ulong p;
  ulong pair_offset;
  ulong threshold;
  ulong total_frequency;
  ulong *_dimensions = null;
  ulong *_max_indexes = null;
  double critical_chi_squared;
  double dof;
  double p_value;
  RekeyPass_ rekey_pass;
  BinPass_ bin_pass;
  FrequenciesPass_ frequencies_pass;
  PredictionsPass_ predictions_pass;
  Candidate_ candidate;
  Candidate_ candidate_max;
  IndexInfo_ *index_info;
  IndexInfo_ *index_info_l;
  IndexInfo_ *index_info_r;
  IndexInfo_ *index_infos;
  FieldInfo_ *field_info;
  FieldInfo_ *_field_infos = null;
  KeyValueInfo *info;
  KeyValuePair *pair;
  KeyValueArray *_hyperslice = null;
  KeyValueArray *_reduced = null;
  UlongHypercube *_dense = null;
  Random *_random = null;
  BearField *field;
  BearField **fields;
  Heap *_candidates = null;
  BearHypercubeWorking *w;

  /* Figure out which iteration we are on.
   */
  iteration = iteration_(this);

  /* Sanity checks.
   */
  if (bear_hypercube_is_trivial(this)) {
    throw("Hypercube was already determined to be trivial");
  }

  if (iteration == 3) {
    throw("Have already ended the third iteration");
  }

  /* Extract data.
   */
  num_fields = (num_features = this->numFeatures) + 1;
  fields = this->_fields;
  w = this->_w;
  total_frequency = w->totalFrequency;

  /* Check if we are on the first iteration.
   */
  if (iteration == 0) {
    /* We are. Walk through the fields.
     */
    for (f = 0; f < num_fields; f++) {
      /* Finalize the field. This will throw if no data has been added.
       */
      bear_field_finalize(fields[f]);

      /* If this field has unit cardinality, then the hypercube will be
       * trivial: it is or will eventually reduce to a single cell. Delete the
       * working and fields, which (together with the fact that the predictions
       * hypercube has not yet been created) is the signal of triviality, and
       * return.
       */
      if (fields[f]->cardinality == 1) {
        delete(&this->_fields, num_fields, bear_field);
        working_delete_(&this->_w);

        return;
      }
    }

    /* Create the array of maximum indexes from the cardinalities of the
     * fields. By the above, all of these will be nonzero.
     */
    new(&_max_indexes, num_fields);

    for (f = 0; f < num_fields; f++) {
      _max_indexes[f] = fields[f]->cardinality - 1;
    }

    /* Create the info object. The key-frequency pairs will in general be
     * "sparse," i.e., the frequency can be zero for pairs that have been
     * merged into other pairs.
     */
    sparse_key_frequency_info_new(
        &w->_info,
        num_fields,
        _max_indexes,
        total_frequency);
    deletev(&_max_indexes);

    /* Create a single pair inside the object itself that we can use for
     * populating the array. Note that this cannot be used in parallel threads;
     * we stipulate that the add method must be called serially, as it is not
     * efficient to parallelize it.
     */
    key_value_pair_new(&w->_pair, w->_info);

    /* Create the collection.
     */
    key_value_collection_new(&w->_data, w->_info);

    /* Compute the mean residual across the entire dataset. This is used as the
     * prediction for any combination of feature bins that has no frequency at
     * all.
     */
    w->meanResidual = w->sumResidualFrequency / total_frequency;

    /* End of the first iteration: return.
     */
    return;
  }

  /* Check if we are ending the second iteration.
   */
  if (iteration == 1) {
    /* We are. Check that the same total frequency was added on the second
     * iteration as on the first.
     */
    if (w->totalFrequencyCheck != w->totalFrequency) {
      throw("A different total frequency was added on the second iteration "
          "(%lu) than on the first (%lu)",
          w->totalFrequencyCheck,
          w->totalFrequency);
    }

    /* Reset for the third iteration.
     */
    w->totalFrequencyCheck = 0;

    /* Extract the collection into a local KeyValueArray, which we call
     * _reduced because we will reduce it (but it isn't reduced yet). This
     * removes all data from the collection and destroys it.
     */
    key_value_array_new_from_collection(&_reduced, &w->_data);

    /* Extract reference to the info object in the array to be reduced.
     */
    info = _reduced->_info;

    /* We patch in a modified comparator for this to-be-reduced hypercube so
     * that it compares the "dominant_field" first, and then all the others in
     * order. This allows us to re-sort the data hypercube so that any given
     * field is "dominant."
     */
    dominant_f = 0;
    info->compare = dominant_compare_;
    info->x = &dominant_f;

    /* Compute the number of decisions that we will need for this reduce phase.
     * To determine this, consider each field separately, ignoring the rotation
     * between fields actually performed by below. With c the cardinality of a
     * given field, the reduce algorithm determines whether each of those
     * boundaries can be removed. Thus there are c - 1 decisions for that
     * field, which is the same as the maximum index for that field.
     */
    for (f = 0, num_decisions = 0; f < num_fields; f++) {
      num_decisions += w->_info->_maxFields[f];
    }

    /* Get the p-value from the Decider, adding this number of decisions.
     */
    p_value = decider_get_p_value(w->decider, num_decisions);

    /* Create the local array of FieldInfo_ objects.
     */
    new(&_field_infos, num_fields);

    /* Create a pseudorandom number generator for tie-breaking decisions.
     */
    random_new(&_random);

    /* We now set up the metadata that will track the reduce phase.
     *
     * We create an array of FieldInfo_ structure objects, one for each field.
     * During the reduce phase we will continually re-sort these FieldInfo_
     * objects, so we need to store the actual field index 'f' in each.
     *
     * For each field, we repeatedly attempt to merge neighboring indexes if
     * the densities of adjacent cells in the hyperslices corresponding to the
     * given two index values are, overall, not statistically significantly
     * different. For each merge, the index of every pair in the hyperslice
     * corresponding to the "right" (larger) index is changed to that of the
     * "left" (smaller) index, and the hyperslices are then merged and sorted
     * (leaving, in general, empty invalid pairs after the valid pairs, if any
     * pairs have equal sets of keys after this "rekeying" and are hence
     * merged).
     *
     * To track this process, an IndexInfo_ structure object is created for
     * each index for each field and stored in the corresponding FieldInfo_
     * object. When two indexes are merged, the right index is changed to that
     * of the left index. The IndexInfo_ objects are updated, with the right
     * index being made invalid.
     *
     * Walk through the fields.
     */
    for (f = 0; f < num_fields; f++) {
      /* Extract references to the BearField and FieldInfo_ structure objects
       * for this field.
       */
      field = fields[f];
      field_info = &_field_infos[f];

      /* Because we will sort the array of FieldInfo_ objects, we need to store
       * the actual field index 'f' in the object.
       */
      field_info->f = f;

      /* Extract and copy the feature flag.
       */
      feature = field_info->feature = field->feature;

      /* Store the original cardinality of this field in the Field_ object,
       * which is also the initial value of the reduced cardinality of this
       * field.
       */
      field_info->origCardinality = field_info->cardinality
        = field->cardinality;

      /* Create the array of IndexInfo_ structure objects in this FieldInfo_
       * object. The members 'pairOffset' and 'numPairs' will be populated in
       * each IndexInfo_ object in the main loop below, because they can change
       * when the hypercube is rotated to make another field dominant and then
       * later rotated back, and so need to be recomputed for every such
       * iteration. The frequency, mean value, and mean residual, on the other
       * hand, are properties of the indexes themselves, updated only by
       * merges, but not dependent on the details of the hypercube's array
       * storage.
       */
      index_infos = new(&field_info->_indexInfos, field_info->origCardinality);

      /* Walk through the indexes for this field.
       */
      for (i = 0; i < field->cardinality; i++) {
        /* Extract a pointer to the IndexInfo_ structure object for this index.
         */
        index_info = &index_infos[i];

        /* Extract and store information. For residual fields we move the value
         * to the 'meanResdual' member to simplify the comparator and the
         * creation of the maximum Heap element.
         */
        index_info->frequency = field->_frequencies[i];

        if (feature) {
          index_info->meanResidual = field->_meanResiduals[i];
          index_info->meanValue = field->_meanValues[i];
        } else {
          index_info->meanResidual = field->_meanValues[i];
          index_info->meanValue = DBL_NAN;
        }
      }
    }

    /* Our job now is to now merge neighboring indexes in all of the fields
     * until the density in the hyperslice of data for each index of each field
     * is statistically significantly different, overall, from those of its
     * neighbors.
     *
     * We start with the field with the greatest cardinality, as this is most
     * likely to have indexes that need to be merged in order to obtain
     * statistical significance. We prefer the residual field itself to any
     * feature field, in the case of equal cardinalities.
     *
     * For the chosen field, we compute the chi-squared value of density
     * differences between all corresponding cells of the hyperslices
     * corresponding to all possible pairs of neighboring indexes. For all
     * pairs of neighboring indexes with a chi-squared value that does not
     * indicate a statistically significant overall difference in densities, we
     * create a "candidate" and put it into a min-heap. The comparator for this
     * min-heap uses the following decision waterfall:
     *
     * - It prefers the candidate with the lower absolute value difference in
     *   mean residual values of the two indexes, so that similar mean residual
     *   values tend to be clustered.
     *
     * - If the field is a feature, it then prefers the candidate with the
     *   lower difference in mean values of the two indexes, so that similar
     *   mean values tend to be clustered.
     *
     * - It then prefers the candidate with the lower total frequency of the
     *   two neighbor indexes, so that merged indexes that are the "thinnest"
     *   are preferred.
     *
     * - It then prefers the candidate with the smaller absolute difference in
     *   frequencies between the two indexes, so that merges between indexes
     *   more equal in "width" are preferred.
     *
     * - It then prefers the candidate with the smaller chi-squared value, so
     *   that index pairs with less statistically different densities are
     *   preferred.
     *
     * - Finally, if all of these compare equal, it uses a random number stored
     *   in each candidate to randomly (but consistently) select one or the
     *   other.
     *
     * The chi-squared values are demoted to the last criterion (other than a
     * random tie-breaker) because by definition the chi-squared values of
     * candidates signify no statistical significance, so it makes no sense to
     * place any significant weight on what is essentially noise.
     *
     * The absolute delta in mean residual values is taken as the first
     * criterion because our overall goal is to minimize loss, which with a
     * piecewise constant model will be minimized when similar residual values
     * are clustered, so that their deviations from a constant prediction for
     * that bin are minimized.
     *
     * If this criterion is neutral, and the field is a feature, we use the
     * same logic on the mean feature values, to cluster feature bins closest
     * in mean value.
     *
     * We start this process with the field with highest cardinality, but we
     * need to give other fields a chance to perform merges once (and if) that
     * field's cardinality is reduced to be comparable to the others. But we
     * just don't switch to the field with the second-highest cardinality as
     * soon as the cardinality of the field with initially highest cardinality
     * drops below that of that second-place field, because that sort of logic
     * would imply that we would like to keep switching fields as they drop
     * below each other (and others that are initially even lower), and (as
     * will become clear) switching to a different field is an expensive
     * operation, requiring the re-sorting of the entire hypercube. So we set a
     * different threshold: half the cardinality of the field with the
     * second-highest cardinality. We use this threshold for all of the fields,
     * not just the first. This ensures rotation between fields, but guarantees
     * no more than a logarithmic number of hypercube re-sortings. It does
     * mean, however, that we need to have an outer loop of iterations (within
     * which we iterate through the entire set of fields), where at the start
     * of each iteration we re-sort the list of fields by decreasing
     * cardinality, and reset the flag as to whether any merges have been
     * successful. Only if our outer loop tries every field with no successful
     * merges do we abort the outer loop.
     *
     * Start this outer loop.
     */
    do {
      /* If we have an iteration of this loop that does not change anything,
       * then we are finished. Initialize that flag.
       */
      changed = false;

      /* Sort the fields by decreasing cardinality, preferring the residual
       * field over a feature field, but otherwise randomly shuffling any
       * tie-breakers. Note that the array _field_infos will now NOT be in
       * ascending order of the 'f' value after this point, so that anything
       * that needs the actual 'f' value must extract it from the FieldInfo_
       * object. To defensively guard against confusion on this, the index for
       * the array _field_infos of ordered FieldInfo_ objects will be denoted
       * 'fi' in the following.
         */
      qsortx(_field_infos,
          num_fields,
          field_info_compare_cardinality_,
          true,
          null);

      /* Per the above, set the threshold number of bins to be half the number
       * of active bins of the second field in this sort order. (We have at
       * least two fields: a residual field and at least one feature field.)
       */
      threshold = _field_infos[1].cardinality / 2;

      /* Walk through the FieldInfo_ objects. We watch out for any field
       * dropping to a cardinality of 1, since that means that the hypercube
       * will be trivial, and we can break out and return early.
       */
      for (fi = 0, trivial = false; fi < num_fields; fi++) {
        /* Get a reference to the FieldInfo_ object and extract data from it.
         */
        field_info = &_field_infos[fi];
        f = field_info->f;
        feature = field_info->feature;
        orig_cardinality = field_info->origCardinality;
        cardinality = field_info->cardinality;
        index_infos = field_info->_indexInfos;

        /* If this cardinality of this field has been reduced to 1, then the
         * hypercube will reduce to a single cell and hence will be trivial.
         * Break out early (we clean up and return below).
         */
        if (cardinality == 1) {
          trivial = true;

          break;
        }

        /* The number of degrees of freedom for our comparison is the number of
         * hypercubes in each hyperslice for this field, which is the same for
         * every index of this field. Compute it. Note that this changes as we
         * reduce each field in this overall iteration, so needs to be computed
         * here in this inner loop. Finally, note that 'dof' can easily
         * overflow a ulong value, because it is a virtual number of
         * hypercells, not a physical number of hypercells, so we must use a
         * double type for it.
         */
        for (other_fi = 0, dof = 1; other_fi < num_fields; other_fi++) {
          if (other_fi != fi) {
            dof *= _field_infos[other_fi].cardinality;
          }
        }

        /* Compute the approximate critical chi-squared value for this number
         * of degrees of freedom and the p-value reduced by the number of
         * decisions. The approximation uses the normal approximation for more
         * than 50 degrees of freedom. It also takes in a double 'dof', for the
         * above reasons; for values 50 and under it checks that the value is
         * actually integral.
         */
        critical_chi_squared = approx_critchisqq(p_value, dof);

        /* Rotate the reduced hypercube so that the field 'f' that we are
         * working on is dominant, if it wasn't already.
         */
        if (f != dominant_f) {
          dominant_f = f;
          key_value_array_refinalize(_reduced);
        }

        /* Compute the range of pairs for each index of this field 'f'. The
         * index 'current_i' is the current index that we are computing the
         * offset and number of pairs for.
         */
        current_i = 0;
        pair_offset = 0;
        num_pairs = 0;

        /* Walk through the pairs of the reduced hypercube.
         */
        for (p = 0; p < _reduced->numPairs; p++) {
          /* Extract reference to this pair.
           */
          pair = key_value_array_get(_reduced, p);

          /* Get the index value for this field for this pair.
           */
          i = key_value_pair_get_key(pair, info, f);

          /* Sanity check: we sorted the hypercube's pairs on this field, so
           * its index values can never go backwards.
           */
          if (i < current_i) {
            throw("Impossible: Index decreased from %lu to %lu", current_i, i);
          }

          /* If the index of the pair is greater than our current index, then
           * we have found the first pair of the next index that is still used
           * in the hypercube for this field. Compute and save quantities for
           * the current index that we have been working on.
           */
          if (i > current_i) {
            /* Walk the current_i until it gets up to this new 'i'. For the
             * initial current_i that we have been working on, we store the
             * actual information we found. For all of the others we mark them
             * as now being unused, since there were no pairs with such indexes
             * between current_i and this new 'i'.
             */
            for (; current_i < i; current_i++) {
              /* Sanity check: we can't hit the end, since 'i' must be less
               * than the original cardinality.
               */
              if (current_i >= orig_cardinality) {
                throw("Impossible index %lu (max %lu)",
                    current_i,
                    orig_cardinality - 1);
              }

              /* Store what we found for this current index that we have been
               * working on in the corresponding IndexInfo_ object.
               */
              index_info = &index_infos[current_i];
              index_info->pairOffset = pair_offset;
              index_info->numPairs = num_pairs;

              /* Reset the values to indicate an unused index for any other
               * indexes between the initial current_i and the new 'i'.
               */
              pair_offset = ULONG_MAX;
              num_pairs = 0;
            }

            /* Now that we have done all of the indexes from current_i up to
             * but not including this new 'i', reinitialize the pair offset to
             * the offset of the pair that we have just read, which is the
             * first pair belonging to the new index 'i'. The number of pairs
             * is still zero, though, at this point.
             */
            pair_offset = p;
          }

          /* If we just finished off the last index, then num_pairs was set to
           * zero, but we have just read the first pair of the new index 'i',
           * so we need to increment num_pairs to 1. Alternatively, if we are
           * still on the same index, i.e., i == current_i, then we just need
           * to increment the number of pairs for this index. Either way, we
           * just need to increment.
           */
          num_pairs++;
        }

        /* We have now gone through all the pairs. We still need to finish off
         * the current_i that we were working on by storing the information in
         * its corresponding Index_ object.
         */
        index_info = &index_infos[current_i];
        index_info->pairOffset = pair_offset;
        index_info->numPairs = num_pairs;

        /* Walk through any other indexes after this last one, which are
         * therefore now unused in the reduced hypercube, and denote in their
         * corresponding Index_ objects that they are unused.
         */
        for (current_i++; current_i < orig_cardinality; current_i++) {
          index_info = &index_infos[current_i];
          index_info->pairOffset = ULONG_MAX;
          index_info->numPairs = 0;
        }

        /* We now initialize the heap of candidates for merging. Each candidate
         * has a chi-squared value that is less than the critical value, so is
         * eligible for merging. However, any candidate in the heap can "go
         * stale" as a result of another merge. The code below detects these
         * stale candidates and ignores them, which is more efficient than
         * trying to find them in the heap or keeping parallel data structures
         * to enable them to be found.
         *
         * First, set up the max candidate for the heap. This works because the
         * delta in mean residual values is the first criterion in the
         * comparator.
         */
        candidate_max.deltaResidual = DBL_MAX;

        /* Create the heap.
         */
        heap_new(&_candidates,
            &candidate_max,
            candidate_compare_,
            candidate_heap_print_,
            null);

        /* Walk through all possible left indexes for candidate pairs of used
         * indexes. Obviously the last index can't be a possible left index,
         * since we need a right index.
         */
        for (i_l = 0; i_l < orig_cardinality - 1; i_l++) {
          /* If this left index has no frequency, it's unused. Skip it.
           */
          if (index_infos[i_l].frequency == 0) {
            continue;
          }

          /* Look for the first index to the right of this left index that is
           * used.
           */
          for (i_r = i_l + 1;
              i_r < orig_cardinality && index_infos[i_r].frequency == 0;
              i_r++) {
          }

          /* If we found no used index to the right of the used left index,
           * then skip this possible left index.
           */
          if (i_r == orig_cardinality) {
            continue;
          }

          /* Create a candidate and add it to the heap if justified.
           */
          create_candidate_(
              _reduced,
              index_infos,
              i_l,
              i_r,
              f,
              critical_chi_squared,
              feature,
              _candidates,
              _random);
        }

        /* Build the heap.
         */
        heap_build(_candidates);

        /* Initialize the flag of whether we have done any merges for this
         * field.
         */
        done_merge = false;

        /* We now work through candidates, until either there are no candidates
         * left, or until the cardinality drops to our threshold.
         */
        while (heap_num_elements(_candidates) > 0 && cardinality > threshold) {
          /* Extract the best candidate. The 'continue' statements below are
           * triggered if the candidate is discovered to be stale (i.e.,
           * superseded by a merge made subsequent to the candidate's insertion
           * into the heap).
           */
          heap_extract_min(_candidates, &candidate);

          /* Extract the left index of this candidate.
           */
          i_l = candidate.leftIndex;

          /* If the left index's frequency is wrong, then the candidate is
           * stale.
           */
          if (index_infos[i_l].frequency != candidate.leftFrequency) {
            continue;
          }

          /* Find the right index, which is the next used index to the right of
           * the left index.
           */
          for (i_r = i_l + 1;
              i_r < orig_cardinality && index_infos[i_r].frequency == 0;
              i_r++) {
          }

          /* If we didn't find any right index, then this candidate is stale.
           */
          if (i_r == orig_cardinality) {
            continue;
          }

          /* If the right index is wrong, or its frequency is wrong, then this
           * candidate is stale.
           */
          if (i_r != candidate.rightIndex
              || index_infos[i_r].frequency != candidate.rightFrequency) {
            continue;
          }

          /* The candidate is not stale. We will perform this merge.
           *
           * Flag that we have done at least one merge for this field, and that
           * we will have performed at least one change in this overall
           * do-loop.
           */
          done_merge = true;
          changed = true;

          /* Extract references.
           */
          index_info_l = &index_infos[i_l];
          index_info_r = &index_infos[i_r];

          /* We now set the key index value for field 'f' of every pair in the
           * right index's hyperslice to that of the left index. Create a
           * hyperslice of the right index, which is just a shallow wrapper
           * around the full _reduced array.
           */
          key_value_array_new_hyperslice(
              &_hyperslice,
              _reduced,
              index_info_r->pairOffset,
              index_info_r->numPairs);

          /* Rekey all the pairs in that hyperslice in parallel.
           */
          rekey_pass.indexL = i_l;
          rekey_pass.indexR = i_r;
          rekey_pass.f = f;
          rekey_pass.hyperslice = _hyperslice;
          rekey_pass.info = info;
          parallel_for(_hyperslice->numPairs, 1, rekey_, &rekey_pass);

          /* Delete the hyperslice, which only deletes the shallow wrapper.
           */
          key_value_array_delete(&_hyperslice);

          /* Now create a hyperslice that encompasses the left and right
           * indexes and the possibly empty space between them (which is why
           * the number of pairs is set to the full distance from the pair
           * offset of the left index to the pair offset of the right index
           * plus the number of pairs in the right index).
           */
          key_value_array_new_hyperslice(
              &_hyperslice,
              _reduced,
              index_info_l->pairOffset,
              index_info_r->pairOffset + index_info_r->numPairs
              - index_info_l->pairOffset);

          /* Sort this merged hyperslice. This can leave empty space after the
           * merged pairs.
           */
          key_value_array_refinalize(_hyperslice);

          /* Update the information for the two indexes. Extract frequencies.
           */
          frequency_l = index_info_l->frequency;
          frequency_r = index_info_r->frequency;

          /* Compute the combined frequency.
           */
          combined_frequency = frequency_l + frequency_r;

          /* Compute the new mean residual of the left index, and, if this is a
           * feature field, the new mean value of the left index.
           */
          index_info_l->meanResidual
            = (index_info_l->meanResidual * frequency_l
                + index_info_r->meanResidual * frequency_r)
            / combined_frequency;

          if (feature) {
            index_info_l->meanValue
              = (index_info_l->meanValue * frequency_l
                  + index_info_r->meanValue * frequency_r)
              / combined_frequency;
          }

          /* The left index has the combined frequency and the number of pairs
           * of the merged hyperslice.
           */
          index_info_l->frequency = combined_frequency;
          index_info_l->numPairs = _hyperslice->numPairs;

          /* The right index is now empty and unused, so invalidate it.
           */
          index_info_r->frequency = 0;
          index_info_r->numPairs = 0;
          index_info_r->pairOffset = ULONG_MAX;
          index_info_r->meanResidual = DBL_NAN;

          if (feature) {
            index_info_r->meanValue = DBL_NAN;
          }

          /* Delete the merged hyperslice, which just deletes the shallow
           * wrapper.
           */
          key_value_array_delete(&_hyperslice);

          /* The cardinality is now one less than it was.
           */
          cardinality--;

          /* Create new candidates from this new merged index and the next and
           * previous used indexes, if each of those exist, and insert these
           * candidates into the heap. Note that the above code automatically
           * invalidates any candidates remaining in the heap based on the old
           * left and right indexes, and simply discards them after pulling
           * them from the heap.
           *
           * Start by looking for the next used index to the right.
           */
          for (i_r++;
              i_r < orig_cardinality && index_infos[i_r].frequency == 0;
              i_r++) {
          }

          /* Check if we found a next used index to the right.
           */
          if (i_r != orig_cardinality) {
            /* We found a new right index. Create a candidate and add it to the
             * heap if justified.
             */
            create_candidate_(
                _reduced,
                index_infos,
                i_l,
                i_r,
                f,
                critical_chi_squared,
                feature,
                _candidates,
                _random);
          }

          /* Try to find a used index to the left of our merged index 'i_l'.
           */
          for (i_l_l_rev = 0; i_l_l_rev < i_l; i_l_l_rev++) {
            i_l_l = i_l - 1 - i_l_l_rev;

            if (index_infos[i_l_l].frequency > 0) {
              break;
            }
          }

          /* Check if we found a used index to the left.
           */
          if (i_l_l_rev < i_l) {
            /* We did. Create a candidate and add it to the heap if justified.
             * Note that our "left" index is 'i_l_l' and our "right" index is
             * actually 'i_l'.
             */
            create_candidate_(
                _reduced,
                index_infos,
                i_l_l,
                i_l,
                f,
                critical_chi_squared,
                feature,
                _candidates,
                _random);
          }
        }

        /* We've either got no candidates left, or we hit the threshold
         * cardinality. Delete the heap of candidates; we will create a new
         * heap on the next iteration.
         */
        heap_delete(&_candidates);

        /* Check if we did any merges.
         */
        if (done_merge) {
          /* We did. Update the cardinality in the FieldInfo_ object.
           */
          field_info->cardinality = cardinality;

          /* Re-sort the reduced hypercube to reflect all of our merges. This
           * "defragments" the pairs in the hypercube, which is why we need to
           * recompute the pair offsets at the start of this inner loop. The
           * potential gain in efficiency in doing this far outweighs the cost
           * of recomputing these offsets.
           */
          key_value_array_refinalize(_reduced);
        }
      }

      /* We've done all of the fields for this iteration. Do it all again only
       * if we did not break out as trivial and we made some change in this
       * iteration.
       */
    } while (!trivial && changed);

    /* Clean up.
     */
    random_delete(&_random);

    /* If the hypercube is trivial, clean up and return.
     */
    if (trivial) {
      key_value_array_delete(&_reduced);

      for (f = 0; f < num_fields; f++) {
        field_info = &_field_infos[f];
        deletev(&field_info->_indexInfos);
      }

      deletev(&_field_infos);
      delete(&this->_fields, num_fields, bear_field);
      working_delete_(&this->_w);

      return;
    }

    /* The hypercube is now fully reduced, but the indexes used for each field
     * are the original indexes, which in general will now have gaps between
     * them due to the merges. We now compute the "defragmented" indexes for
     * this fully reduced hypercube, or "bins," where for each field 'f' each
     * bin will be in the range [0, cardinalities[f]), whereas the
     * corresponding index is in the range [0, orig_cardinalities[f]).
     *
     * Start by sorting the array of FieldInfo_ objects back into field index
     * order.
     */
    qsortx(_field_infos, num_fields, field_info_compare_f_, false, null);

    /* Create the top level of the matrix for index-to-bin mappings.
     */
    new(&_bins_from_indexes, num_fields);

    /* Walk through the fields.
     */
    for (f = 0; f < num_fields; f++) {
      /* Extract information for this field.
       */
      field_info = &_field_infos[f];
      feature = field_info->feature;
      orig_cardinality = field_info->origCardinality;
      cardinality = field_info->cardinality;
      index_infos = field_info->_indexInfos;

      /* The mapping from index to bin has the original cardinality.
       */
      new(&_bins_from_indexes[f], orig_cardinality);

      /* Populate these arrays. Walk through the indexes.
       */
      for (i = 0, b = 0; i < orig_cardinality; i++) {
        /* Extract reference.
         */
        index_info = &index_infos[i];

        /* Check if this index has frequency, i.e., is the first index of a
         * bin.
         */
        if (index_info->frequency > 0) {
          /* It is. If this is not the first index, move to the next bin.
           */
          if (i > 0) {
            b++;
          }
        } else if (i == 0) {
          /* The first index must always have frequency, since we always merge
           * to the left.
           */
          throw("Impossible");
        }

        /* Store the bin for this index.
         */
        _bins_from_indexes[f][i] = b;

      }

      /* Sanity check.
       */
      if (b != cardinality - 1) {
        throw("Impossible: %lu v %lu", b, cardinality - 1);
      }

      /* If this is a feature field, bin it.
       */
      if (feature) {
        bear_field_bin(fields[f], _bins_from_indexes[f]);
      }
    }

    /* Now replace all index values with bin values, in parallel.
     */
    bin_pass.reduced = _reduced;
    bin_pass.this = this;
    bin_pass.binsFromIndexes = _bins_from_indexes;
    parallel_for(_reduced->numPairs, 2 * num_fields, bin_, &bin_pass);

    /* Clean up.
     */
    deletemv(&_bins_from_indexes, num_fields);

    /* We now extract the _reduced KeyValueArray out into a UlongHypercube.
     * Extract its dimensions from the reduced cardinlities.
     */
    new(&_dimensions, num_fields);

    for (f = 0; f < num_fields; f++) {
      _dimensions[f] = _field_infos[f].cardinality;
    }

    /* Extract the reduced data as a dense UlongHypercube, destroying the
     * reduced KeyValueArray.
     */
    ulong_hypercube_new_from_sparse_key_frequency_array(
        &_dense,
        &_reduced,
        _dimensions);

    /* We now set up for the third iteration.
     *
     * Our first step is to create a UlongHypercube of frequencies for all
     * combinations of feature bins. This is just the same as the _reduced
     * hypercube, except leaving out and summing over the final residual
     * dimension. We store this away for use at the end of the third iteration.
     *
     * We then create a DoubleHypercube for the prediction for each of these
     * feature hyperbins. In the third iteration we actually accumulate the sum
     * of frequency x residual for all examples, and at the end of the third
     * iteration we make use of the frequencies hypercube above to normalize
     * the predictions. (We could in principle do this as we go, but this would
     * be less computationally efficient.) Those hyperbins with no frequency
     * are allocated a prediction that simply is the mean residual of all
     * examples, since those hyperbins do not appear in the training data and
     * so we have no idea of how they may vary from the average.
     *
     * Create the hypercubes. We use the same array of dimensions as above, but
     * simply omitting the final residual field.
     */
    ulong_hypercube_new(&w->_frequencies, num_features, _dimensions);
    double_hypercube_new(&this->_predictions, num_features, _dimensions);

    /* Extract the number of indexes and bins for the residual field.
     */
    num_residual_bins = _dimensions[num_features];

    /* Compute the frequencies for all the feature hyperbins.
     */
    frequencies_pass.reduced = _dense;
    frequencies_pass.numResidualBins = num_residual_bins;

    hypercube_foreach(
        w->_frequencies,
        null,
        null,
        compute_frequencies_,
        &frequencies_pass,
        2 * num_residual_bins);

    /* Clean up.
     */
    hypercube_delete(&_dense);
    deletev(&_dimensions);

    for (f = 0; f < num_fields; f++) {
      field_info = &_field_infos[f];
      deletev(&field_info->_indexInfos);
    }

    deletev(&_field_infos);

    /* We've now finished the second iteration.
     */
    return;
  }

  /* Third iteration. Check that the same total frequency was added on the
   * second iteration as on the first.
   */
  if (w->totalFrequencyCheck != w->totalFrequency) {
    throw("A different total frequency was added on the third iteration "
        "(%lu) than on the first two (%lu)",
        w->totalFrequencyCheck,
        w->totalFrequency);
  }

  /* We now just need to scale the predictions.
   */
  predictions_pass.frequencies = w->_frequencies;
  predictions_pass.meanResidual = w->meanResidual;
  hypercube_foreach(
      this->_predictions,
      null,
      null,
      scale_predictions_,
      &predictions_pass,
      2);

  /* Clean up.
   */
  working_delete_(&this->_w);

  /* For keeping this internal debug method available.
   */
  preserve(field_info_print_);
}

/* Merge another hypercube into this one, deleting the other afterwards.
 */
void bear_hypercube_merge(BearHypercube *this, BearHypercube **p_other) {
  byte iteration;
  ulong f;
  ulong num_fields;
  KeyValueArray *_other_array = null;
  KeyValueArray *_this_array = null;
  BearHypercubeWorking *other_w;
  BearHypercubeWorking *w;
  BearHypercube *other;

  /* Extract a reference to the other hypercube.
   */
  other = *p_other;

  /* Sanity check.
   */
  if (other->numFeatures != this->numFeatures) {
    throw("Impossible");
  }

  /* Extract information.
   */
  num_fields = this->numFeatures + 1;
  iteration = iteration_(this);

  /* Merge in.
   */
  if (this->_fields != null) {
    for (f = 0; f < num_fields; f++) {
      bear_field_merge(this->_fields[f], &other->_fields[f]);
    }
  }

  if ((w = this->_w) != null) {
    other_w = other->_w;

    if (iteration == 0) {
      w->totalFrequency += other_w->totalFrequency;
    } else {
      w->totalFrequencyCheck += other_w->totalFrequencyCheck;
    }

    w->sumResidualFrequency += other_w->sumResidualFrequency;

    if (w->_data != null) {
      key_value_array_new_from_collection(&_this_array, &w->_data);
      key_value_array_new_from_collection(&_other_array, &other_w->_data);
      key_value_array_merge(_this_array, _other_array);
      key_value_collection_new_from_array(&w->_data, &_this_array);
      key_value_array_delete(&_other_array);
    }

    if (this->_predictions != null) {
      hypercube_foreach(
          this->_predictions,
          null,
          null,
          merge_predictions_,
          other->_predictions,
          1);
    }
  }

  /* Delete the other hypercube.
   */
  bear_hypercube_delete(p_other);
}

/* Create a candidate.
 */
static void create_candidate_(
    KeyValueArray *reduced,
    IndexInfo_ *index_infos,
    ushort i_l,
    ushort i_r,
    ulong f,
    double critical_chi_squared,
    bool feature,
    Heap *candidates,
    Random *random) {
  double chi_squared;
  IndexInfo_ *index_info_l;
  IndexInfo_ *index_info_r;
  Candidate_ candidate;

  /* Compute the chi-squared value between the densities of the left and right
   * hyperslices.
   */
  chi_squared = chi_squared_(reduced, index_infos, i_l, i_r, f);

  /* If it is over the critical value, there's nothing else to do.
   */
  if (chi_squared > critical_chi_squared) {
    return;
  }

  /* We're still here, so we have a candidate. Extract references.
   */
  index_info_l = &index_infos[i_l];
  index_info_r = &index_infos[i_r];

  /* Create the candidate.
   */
  candidate.feature = feature;
  candidate.deltaResidual
    = fabs(index_info_r->meanResidual - index_info_l->meanResidual);
  candidate.deltaValue = feature
    ? index_info_r->meanValue - index_info_l->meanValue
    : DBL_NAN;
  candidate.leftIndex = i_l;
  candidate.rightIndex = i_r;
  candidate.leftFrequency = index_info_l->frequency;
  candidate.rightFrequency = index_info_r->frequency;
  candidate.chiSquared = chi_squared;
  candidate.random = random_ulong(random);

  /* Sanity checks.
   */
  if (feature && (candidate.deltaValue <= 0
        || double_pawized_is_infinite_or_nan(candidate.deltaValue))) {
    throw("Impossible: delta value %.16g", candidate.deltaValue);
  }

  /* Insert it into the heap.
   */
  heap_insert(candidates, &candidate);
}

/* Whether the hypercube is trivial. If so, no predictions can be extracted
 * from it; the calling code should abort any attempt to use this set of
 * features for this label's residuals.
 */
bool bear_hypercube_is_trivial(BearHypercube *this) {
  return this->_w == null && this->_predictions == null;
}

/* Get the residual prediction for the given set of feature values.
 */
double bear_hypercube_predict(BearHypercube *this, double *features) {
  ulong o;
  double prediction;

  /* Sanity checks.
   */
  if (bear_hypercube_is_trivial(this)) {
    throw("Cannot get predictions from a trivial hypercube");
  }

  if (this->_w != null) {
    throw("Third iteration has not been completed");
  }

  /* Get the offset of this set of features.
   */
  o = offset_(this, features);

  /* Get the prediction for this offset.
   */
  hypercube_get(this->_predictions, o, &prediction);

  /* Return it.
   */
  return prediction;
}

/* Destructor for the working.
 */
static void working_delete_(BearHypercubeWorking **p_this) {
  BearHypercubeWorking *this;

  if ((this = *p_this) == null) {
    return;
  }

  hypercube_delete(&this->_frequencies);
  key_value_collection_delete(&this->_data);
  key_value_pair_delete(&this->_pair, this->_info);
  key_value_info_delete(&this->_info);
  deleteo(p_this);
}

/* KeyValueArray comparator that sorts on the dominant key first, and then the
 * remaining "non-dominant" keys (in field order).
 */
static sbyte dominant_compare_(
    KeyValuePair *a,
    KeyValuePair *b,
    KeyValueInfo *info) {
  ulong dominant_key;
  ulong key_a;
  ulong key_b;
  ulong *p_dominant_key;

  p_dominant_key = info->x;
  dominant_key = *p_dominant_key;

  key_a = key_value_pair_get_key(a, info, dominant_key);
  key_b = key_value_pair_get_key(b, info, dominant_key);

  if (key_a < key_b) {
    return -1;
  }

  if (key_a > key_b) {
    return +1;
  }

  return non_dominant_compare_(a, b, info);
}

/* KeyValueArray comparator that sorts on the "non-dominant" keys, i.e., all
 * the keys other than that of the dominant field, in field order. It is worth
 * splitting this out because it is not only used in the above comparator, but
 * also directly in the code.
 */
static sbyte non_dominant_compare_(
    KeyValuePair *a,
    KeyValuePair *b,
    KeyValueInfo *info) {
  ulong dominant_key;
  ulong k;
  ulong key_a;
  ulong key_b;
  ulong *p_dominant_key;

  p_dominant_key = info->x;
  dominant_key = *p_dominant_key;

  for (k = 0; k < info->numKeys; k++) {
    if (k == dominant_key) {
      continue;
    }

    key_a = key_value_pair_get_key(a, info, k);
    key_b = key_value_pair_get_key(b, info, k);

    if (key_a < key_b) {
      return -1;
    }

    if (key_a > key_b) {
      return +1;
    }
  }

  return 0;
}

/* Comparator for FieldInfo_ objects that sorts into order of decreasing
 * cardinality, preferring the residual field where cardinalities are equal.
 */
static sbyte field_info_compare_cardinality_(
    FieldInfo_ *a,
    FieldInfo_ *b,
    void *x) {
  ushort cardinality_a;
  ushort cardinality_b;

  cardinality_a = a->cardinality;
  cardinality_b = b->cardinality;

  if (cardinality_a > cardinality_b) {
    return -1;
  }

  if (cardinality_a < cardinality_b) {
    return +1;
  }

  if (!a->feature) {
    return -1;
  }

  if (!b->feature) {
    return +1;
  }

  return 0;
}

/* Comparator for FieldInfo_ objects that sorts into order of ascending field
 * index 'f'.
 */
static sbyte field_info_compare_f_(FieldInfo_ *a, FieldInfo_ *b, void *x) {
  ulong fa;
  ulong fb;

  fa = a->f;
  fb = b->f;

  return fa < fb ? -1 : fa > fb ? +1 : 0;
}

/* Comparator for Candidate_ objects.
 */
static sbyte candidate_compare_(Candidate_ *a, Candidate_ *b, void *x) {
  ulong combined_frequency_a;
  ulong combined_frequency_b;
  ulong random_a;
  ulong random_b;
  long delta_frequency_a;
  long delta_frequency_b;
  double chi_squared_a;
  double chi_squared_b;
  double delta_residual_a;
  double delta_residual_b;
  double delta_value_a;
  double delta_value_b;

  /* Prefer the lower delta in residual values.
   */
  delta_residual_a = a->deltaResidual;
  delta_residual_b = b->deltaResidual;

  if (delta_residual_a < delta_residual_b) {
    return -1;
  }

  if (delta_residual_a > delta_residual_b) {
    return +1;
  }

  /* Otherwise, if this is a feature field, prefer the lower delta in feature
   * values.
   */
  if (a->feature) {
    delta_value_a = a->deltaValue;
    delta_value_b = b->deltaValue;

    if (delta_value_a < delta_value_b) {
      return -1;
    }

    if (delta_value_a > delta_value_b) {
      return +1;
    }
  }

  /* Otherwise, prefer the lower combined frequency of the two bins.
   */
  combined_frequency_a = a->leftFrequency + a->rightFrequency;
  combined_frequency_b = b->leftFrequency + b->rightFrequency;

  if (combined_frequency_a < combined_frequency_b) {
    return -1;
  }

  if (combined_frequency_a > combined_frequency_b) {
    return +1;
  }

  /* Otherwise, prefer the two bins having more-equal frequancies.
   */
  delta_frequency_a = labs((long)a->leftFrequency - (long)a->rightFrequency);
  delta_frequency_b = labs((long)b->leftFrequency - (long)b->rightFrequency);

  if (delta_frequency_a < delta_frequency_b) {
    return -1;
  }

  if (delta_frequency_a > delta_frequency_b) {
    return +1;
  }

  /* Otherwise, prefer the lower chi-squared value.
   */
  chi_squared_a = a->chiSquared;
  chi_squared_b = b->chiSquared;

  if (chi_squared_a < chi_squared_b) {
    return -1;
  }

  if (chi_squared_a > chi_squared_b) {
    return +1;
  }

  /* Otherwise, use the random members to tie-break.
   */
  random_a = a->random;
  random_b = b->random;

  if (random_a < random_b) {
    return -1;
  }

  if (random_a > random_b) {
    return +1;
  }

  /* Otherwise (this should be close do impossible unless a candidate is being
   * compared to itself) they sort equal.
   */
  return 0;
}

/* Print a Candidate_ object.
 */
static void candidate_print_(Candidate_ *this, Printer *p) {
  sprint(p, "Candidate_");
  rprint(p, "Left index: %lu", this->leftIndex);
  rprint(p, "Right index: %lu", this->rightIndex);
  rprint(p, "Field type: %s", this->feature ? "feature" : "residual");
  rprint(p, "Delta residual: %.16g", this->deltaResidual);

  if (this->feature) {
    rprint(p, "Delta value: %.16g", this->deltaValue);
  }

  sprint(p, "Frequencies");
  rprint(p, "Left: %lu", this->leftFrequency);
  rprint(p, "Right: %lu", this->rightFrequency);
  rprint(p, "Combined: %lu", this->leftFrequency + this->rightFrequency);
  rprint(p,
      "Difference: %ld",
      labs((long)this->leftFrequency - (long)this->rightFrequency));
  fprint(p);

  rprint(p, "Chi-squared: %.16g", this->chiSquared);
  rprint(p, "Random value: 0x%016lx", this->random);

  fprint(p);
}

/* Print a Candidate_ object that is on the heap.
 */
static void candidate_heap_print_(HeapObjectPrintInfo *hopi, Printer *p) {
  Candidate_ *this;

  this = hopi->this;
  candidate_print_(this, p);
}

/* Internal method for computing the chi-squared value for the differences
 * between the densities of adjacent cells for the hyperslices corresponding to
 * the two given indexes of the dominant field.
 */
static double chi_squared_(
    KeyValueArray *cube,
    IndexInfo_ *index_infos,
    ushort i_l,
    ushort i_r,
    ulong f) {
  ushort i;
  int c;
  ulong frequency_l;
  ulong frequency_r;
  ulong n_l;
  ulong n_r;
  ulong p_l;
  ulong p_r;
  ulong total_frequency_l;
  ulong total_frequency_r;
  double chi_squared;
  double diff;
  double diff_sq;
  double sum_chi_squared;
  double t_f_l_sq;
  double t_f_r_sq;
  double variance;
  IndexInfo_ *index_info_l;
  IndexInfo_ *index_info_r;
  KeyValueArray *hyperslice_l = null;
  KeyValueArray *hyperslice_r = null;
  KeyValuePair *pair_l;
  KeyValuePair *pair_r;
  KeyValueInfo *info;

  /* Extract references.
   */
  index_info_l = &index_infos[i_l];
  index_info_r = &index_infos[i_r];

  /* Extract the total frequency of each of the left and right bins.
   */
  total_frequency_l = index_info_l->frequency;
  total_frequency_r = index_info_r->frequency;

  /* Get the hyperslices of the cube corresponding to the left and right bins.
   */
  key_value_array_new_hyperslice(
      &hyperslice_l,
      cube,
      index_info_l->pairOffset,
      index_info_l->numPairs);
  key_value_array_new_hyperslice(
      &hyperslice_r,
      cube,
      index_info_r->pairOffset,
      index_info_r->numPairs);

  info = hyperslice_l->_info;

  n_l = hyperslice_l->numPairs;
  n_r = hyperslice_r->numPairs;

  /* Convenient factors for the variance calculation.
   */
  t_f_l_sq = total_frequency_l * total_frequency_l;
  t_f_r_sq = total_frequency_r * total_frequency_r;

  /* Initialize sum.
   */
  sum_chi_squared = 0;

  /* Walk through both hyperslices, merge-sorting them together for comparison
   * purposes.
   */
  for (p_l = 0, p_r = 0; p_l < n_l || p_r < n_r;) {
    /* Get the next pair of each hyperslice.
     */
    pair_l = p_l < n_l ? key_value_array_get(hyperslice_l, p_l) : null;
    pair_r = p_r < n_r ? key_value_array_get(hyperslice_r, p_r) : null;

    /* Check the bin for the left pair, if it is not null.
     */
    if (pair_l != null) {
      /* Get the index for the left pair.
       */
      i = key_value_pair_get_key(pair_l, info, f);

      /* Check it.
       */
      if (i != i_l) {
        throw("Impossible: Field %lu from pair_l incorrect: "
            "expected index %lu, actual %lu",
            f,
            i_l,
            i);
      }
    }

    /* Check the index for the right pair, if it is not null.
     */
    if (pair_r != null) {
      /* Get the index for the right pair.
       */
      i = key_value_pair_get_key(pair_r, info, f);

      /* Check it.
       */
      if (i != i_r) {
        throw("Impossible: Field %lu from pair_r incorrect: "
            "expected index %lu, actual %lu",
            f,
            i_r,
            i);
      }
    }

    /* If each hyperslice has a pair, keep the one that sorts lower on the
     * non-dominant keys, or both if this comparison comes out equal.
     */
    if (pair_l != null && pair_r != null) {
      c = non_dominant_compare_(pair_l, pair_r, info);

      if (c < 0) {
        pair_r = null;
      } else if (c > 0) {
        pair_l = null;
      }
    }

    /* Sanity check.
     */
    if (pair_l == null && pair_r == null) {
      throw("Impossible: Both pair_l and pair_r null");
    }

    /* Extract the left and right frequencies, and move to the next pair
     * whenever we use one.
     */
    if (pair_l != null) {
      frequency_l = sparse_key_frequency_pair_get_frequency(pair_l, info);
      p_l++;
    } else {
      frequency_l = 0;
    }

    if (pair_r != null) {
      frequency_r = sparse_key_frequency_pair_get_frequency(pair_r, info);
      p_r++;
    } else {
      frequency_r = 0;
    }

    /* Compute the difference in density between these neighboring cells.
     * Computing it here with divisions rather than multiplications by
     * inverses avoids roundoff and can yield an exact zero when the densities
     * are in fact equal.
     */
    diff = (double)frequency_l / total_frequency_l
      - (double)frequency_r / total_frequency_r;

    /* Short-circuit if this difference in density is zero.
     */
    if (diff != 0) {
      /* Difference in density is nonzero. Square it.
       */
      diff_sq = diff * diff;

      /* Approximate the variance of the Poisson process as the observed
       * frequency plus the "spontaneous correction" (justified by Bayesian).
       */
      variance = (frequency_l + 1) / t_f_l_sq + (frequency_r + 1) / t_f_r_sq;

      /* Approximate the difference of Poissons as a difference of Normals, and
       * compute a chi-squared value. Good enough approximation in practice.
       */
      chi_squared = diff_sq / variance;

      /* Update totals.
       */
      sum_chi_squared += chi_squared;
    }
  }

  /* Clean up.
   */
  key_value_array_delete(&hyperslice_l);
  key_value_array_delete(&hyperslice_r);

  /* Return the chi-squared value.
   */
  return sum_chi_squared;
}

/* Parallel method for rekeying all of the pairs of the hyperslice of the right
 * index of a candidate.
 */
static void rekey_(Parallel *par) {
  ushort i;
  ushort i_l;
  ushort i_r;
  ulong f;
  ulong p;
  RekeyPass_ *pass;
  KeyValueInfo *info;
  KeyValuePair *pair;
  KeyValueArray *hyperslice;

  /* Extract information.
   */
  pass = par->x;
  f = pass->f;
  i_l = pass->indexL;
  i_r = pass->indexR;
  hyperslice = pass->hyperslice;
  info = hyperslice->_info;

  /* Walk through the pairs for this thread.
   */
  for (p = par->start; p < par->end; p++) {
    /* Get the pair from the hypersclice.
     */
    pair = key_value_array_get(hyperslice, p);

    /* Extract the index for the dominant field.
     */
    i = key_value_pair_get_key(pair, info, pass->f);

    /* Check that it is correct.
     */
    if (i != i_r) {
      throw("Impossible: br wrong: expected %lu, got %lu", i_r, i);
    }

    /* Change it to the left index.
     */
    key_value_pair_set_key(pair, info, f, i_l);
  }
}

/* Parallel method for replacing all index values in all pairs of the reduced
 * hypercube with their corresponding bin values.
 */
static void bin_(Parallel *par) {
  ushort b;
  ushort i;
  ushort **bins_from_indexes;
  ulong f;
  ulong num_fields;
  ulong p;
  BinPass_ *pass;
  KeyValueInfo *info;
  KeyValuePair *pair;
  KeyValueArray *reduced;
  BearHypercube *this;

  /* Extract information.
   */
  pass = par->x;
  reduced = pass->reduced;
  info = reduced->_info;
  this = pass->this;
  num_fields = this->numFeatures + 1;
  bins_from_indexes = pass->binsFromIndexes;

  /* Walk through the pairs for this thread.
   */
  for (p = par->start; p < par->end; p++) {
    /* Get reference to the pair.
     */
    pair = key_value_array_get(reduced, p);

    /* Walk through all the fields.
     */
    for (f = 0; f < num_fields; f++) {
      /* Extract the index for this field for this pair.
       */
      i = key_value_pair_get_key(pair, info, f);

      /* Look up its bin value.
       */
      b = bins_from_indexes[f][i];

      /* Set that bin value back in the pair.
       */
      key_value_pair_set_key(pair, info, f, b);
    }
  }
}

/* Parallel method for adding examples for bear_hypercube_new_memory().
 */
static void add_(Parallel *par) {
  ulong e;
  ulong *frequencies;
  double **values;
  AddPass_ *pass;
  BearHypercube *this = null;

  /* Initialize.
   */
  pass = par->x;
  values = pass->values;
  frequencies = pass->frequencies;

  /* Check if we have more than one thread.
   */
  if (par->numThreads > 1) {
    /* We do. Under the lock, copy the master hypercube in this pass object,
     * and update the number of threads initialized.
     */
    lock(&pass->lock);
    bear_hypercube_new_copy(&this, pass->this);
    pass->numThreadsInitialized++;
    condition_signal(&pass->condition);
    unlock(&pass->lock);
  } else {
    /* Only one thread. Make 'this' a reference to the pass object.
     */
    this = pass->this;
  }

  /* Walk through the examples for this thread and add them to our local
   * object.
   */
  for (e = par->start; e < par->end; e++) {
    bear_hypercube_add(this, values[e], frequencies[e]);
  }

  /* Check again if we have more than one thread.
   */
  if (par->numThreads > 1) {
    /* We do. Grab the lock.
     */
    lock(&pass->lock);

    /* Wait until all threads have been initialized.
     */
    while (pass->numThreadsInitialized != par->numThreads) {
      condition_wait(&pass->condition, &pass->lock);
    }

    /* Merge this hypercube in with the master one in the pass object. This
     * deletes our local hypercube.
     */
    bear_hypercube_merge(pass->this, &this);

    /* Release the lock.
     */
    unlock(&pass->lock);
  }
}

/* Foreach method for the frequencies UlongHypercube for computing the frquency
 * for each feature hyperbin.
 */
static void compute_frequencies_(
    Hypercube *this,
    bool *walk_dimensions,
    ulong *indexes,
    ulong o,
    void *x) {
  ushort b;
  ushort num_residual_bins;
  ulong bin_frequency;
  ulong o_r;
  ulong total_frequency;
  FrequenciesPass_ *pass;
  UlongHypercube *reduced;

  /* Extract information.
   */
  pass = x;
  reduced = pass->reduced;
  num_residual_bins = pass->numResidualBins;

  /* Since the residual field is always the last field, the offset into the
   * reduced hypercube is just the predictions offset times the number of
   * residual bins, and we just read consecutive frequencies for residual bins.
   */
  o_r = o * num_residual_bins;

  /* Walk through the residual bins.
   */
  for (b = 0, total_frequency = 0; b < num_residual_bins; b++, o_r++) {
    /* Get the frequency of this residual bin from the reduced hypercube.
     */
    hypercube_get(reduced, o_r, &bin_frequency);

    /* Add it to the total frequency.
     */
    total_frequency += bin_frequency;
  }

  /* Store it.
   */
  hypercube_set(this, o, &total_frequency);
}

/* Foreach method for scaling the predictions DoubleHypercube.
 */
static void scale_predictions_(
    Hypercube *this,
    bool *walk_dimensions,
    ulong *indexes,
    ulong o,
    void *x) {
  ulong frequency;
  double prediction;
  PredictionsPass_ *pass;
  UlongHypercube *frequencies;

  /* Extract information.
   */
  pass = x;
  frequencies = pass->frequencies;

  /* Get the frequency for this features hyperbin.
   */
  hypercube_get(frequencies, o, &frequency);

  /* Check if it is zero.
   */
  if (frequency == 0) {
    /* It is. The prediction is the mean residual.
     */
    prediction = pass->meanResidual;
  } else {
    /* Frequency is nonzero. Get the sum of frequency x residual.
     */
    hypercube_get(this, o, &prediction);

    /* Scale it by the total frequency of this hyperbin.
     */
    prediction /= frequency;
  }

  /* Store the prediction.
   */
  hypercube_set(this, o, &prediction);
}

/* Foreach method for merging the predictions DoubleHypercube into the master
 * one.
 */
static void merge_predictions_(
    Hypercube *this,
    bool *walk_dimensions,
    ulong *indexes,
    ulong o,
    void *x) {
  double other_sum;
  double this_sum;
  DoubleHypercube *other;

  /* Extract information.
   */
  other = x;

  /* Get the frequencies from this hyperbin and the other hyperbin.
   */
  hypercube_get(this, o, &this_sum);
  hypercube_get(other, o, &other_sum);

  /* Sum them.
   */
  this_sum += other_sum;

  /* Store back in this hypercube.
   */
  hypercube_set(this, o, &this_sum);
}

/* Print a FieldInfo_ structure object.
 */
static void field_info_print_(FieldInfo_ *this, Printer *p) {
  ulong i;
  IndexInfo_ *index_info;

  sprint(p, "FieldInfo_");
  rprint(p, "Field type: %s", this->feature ? "feature" : "residual");
  rprint(p, "f: %lu", this->f);
  rprint(p, "Original cardinality: %lu", this->origCardinality);
  rprint(p, "Cardinality: %lu", this->cardinality);

  sprint(p, "IndexInfo_ objects");

  for (i = 0; i < this->origCardinality; i++) {
    sprint(p, "Index_ %lu", i);
    index_info = &this->_indexInfos[i];
    rprint(p, "Frequency: %lu", index_info->frequency);
    rprint(p, "Number of pairs: %lu", index_info->numPairs);
    rprint(p, "Pair offset: %lu", index_info->pairOffset);
    rprint(p, "Mean residual: %.16g", index_info->meanResidual);

    if (this->feature) {
      rprint(p, "Mean value: %.16g", index_info->meanValue);
    }

    fprint(p);
  }

  fprint(p);

  fprint(p);
}

/* Compute the offset into a binned features hypercube for a set of feature
 * values.
 */
static ulong offset_(BearHypercube *this, double *features) {
  ushort b;
  ulong f;
  ulong num_features;
  ulong o;
  ulong *strides;
  double feature;
  BearField **fields;

  /* Extract references.
   */
  num_features = this->numFeatures;
  fields = this->_fields;
  strides = this->_predictions->_strides;

  /* Walk through the features.
   */
  for (f = 0, o = 0; f < num_features; f++) {
    /* Extract the feature value.
     */
    feature = features[f];

    /* Look up the bin corresponding to this value for this field.
     */
    b = bear_field_lookup(fields[f], feature);

    /* Add to the offset.
     */
    o += b * strides[f];
  }

  /* Return the result.
   */
  return o;
}

/* Determine which iteration we are on, with 3 being the value for the final
 * iteration being completed.
 */
static byte iteration_(BearHypercube *this) {
  BearHypercubeWorking *w;

  if ((w = this->_w) == null) {
    return 3;
  }

  if (w->_data == null && this->_predictions == null) {
    return 0;
  }

  if (w->_data != null) {
    return 1;
  }

  return 2;
}
