/* Object class for a Bear hypercube for modeling a residual field with at
 * least one feature field and at least one example. The residual field is
 * always the last field.
 */
#ifndef BEAR_HYPERCUBE_H
#define BEAR_HYPERCUBE_H

#ifdef __cplusplus
extern "C" {
#endif

#include "BearField.h"
#include "Condition.h"
#include "Decider.h"
#include "LossFunction.h"
#include "Heap.h"
#include "Hypercube.h"
#include "KeyValueCollection.h"
#include "SparseKeyFrequency.h"
#include "gamma.h"

typedef struct {
  ulong totalFrequency;
  ulong totalFrequencyCheck;
  double meanResidual;
  double sumResidualFrequency;
  Decider *decider;
  KeyValuePair *_pair;
  KeyValueInfo *_info;
  KeyValueCollection *_data;
  UlongHypercube *_frequencies;
} BearHypercubeWorking;

typedef struct {
  ulong numFeatures;
  BearField **_fields;
  DoubleHypercube *_predictions;
  BearHypercubeWorking *_w;
} BearHypercube;

/* General constructor allowing streaming in an arbitrary amount of data, which
 * needs to be done in three iterations, with bear_hypercube_end_iteration()
 * being called at the end of each iteration. If bear_hypercube_is_trivial()
 * returns true after the second iteration, the object can no longer be used,
 * and a third iteration cannot be attempted.
 */
void bear_hypercube_new(
    BearHypercube **p_this,
    ulong num_features,
    Decider *decider);

/* Simplified constructor when the input data fits in memory. The 'values'
 * matrix must have the example index first and the field index second. The
 * residual field must be the last field, after the features.
 */
void bear_hypercube_new_memory(
    BearHypercube **p_this,
    ulong num_features,
    Decider *decider,
    ulong num_examples,
    double **values,
    ulong *frequencies);

/* Load.
 */
void bear_hypercube_new_load(BearHypercube **p_this, File *f);

/* Copy.
 */
void bear_hypercube_new_copy(BearHypercube **p_this, BearHypercube *other);

/* Destructor.
 */
void bear_hypercube_delete(BearHypercube **p_this);

/* Print.
 */
void bear_hypercube_print(BearHypercube *this, Printer *p);

/* Save.
 */
void bear_hypercube_save(BearHypercube *this, File *f);

/* Equals.
 */
bool bear_hypercube_equals(BearHypercube *this, BearHypercube *other);

/* Add an example, for any iteration.
 */
void bear_hypercube_add(BearHypercube *this, double *values, ulong frequency);

/* End an iteration. If the hypercube is found to be trivial, i.e., reduces to
 * or will reduce to a single cell, then the working object is destroyed, and
 * the hypercube should be destroyed by the calling code; it cannot be used
 * further.
 */
void bear_hypercube_end_iteration(BearHypercube *this);

/* Merge another hypercube into this one, deleting the other afterwards.
 */
void bear_hypercube_merge(BearHypercube *this, BearHypercube **p_other);

/* Whether the hypercube is trivial. If so, no predictions can be extracted
 * from it; the calling code should abort any attempt to use this hypercube.
 */
bool bear_hypercube_is_trivial(BearHypercube *this);

/* Get the residual prediction for the given set of feature values.
 */
double bear_hypercube_predict(BearHypercube *this, double *features);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _H */
