#include "BearModel.h"
#include <float.h>
#include <limits.h>
#include <math.h>

/* An upper bound on weight, for sanity's sake.
 */
#define MAX_WEIGHT_ 1e10

/* If defined, label predictions will be verified in the model-building
 * process against what is obtained from a full call to bear_model_predict().
#define VERIFY_LABEL_PREDICTIONS_
 */

typedef struct {
  ulong totalFrequency;
  ulong *frequencies;
  double maxLabel;
  double minLabel;
  double sum;
  double *labels;
  Lock lock;
} EmptyPredictionPass_;

typedef struct {
  double emptyLabelEstimate;
  double emptyLabelPrediction;
  double maxResidual;
  double minResidual;
  double *labelEstimates;
  double *labelPredictions;
  double *labels;
  double *parentResidualPredictions;
  double *residuals;
  Lock lock;
} EmptyResidualsPass_;

typedef struct {
  ulong *frequencies;
  double emptyLabelEstimate;
  double sum;
  double *labels;
  void *lossFunctionParams;
  Lock lock;
  LossFunction *lossFunction;
} EmptyLossPass_;

typedef struct {
  bool *examplesAreComplete;
  bool *featuresHaveMissingValues;
  ulong numCompleteExamples;
  ulong numFeatures;
  double **features;
  Lock lock;
} FindIncompletePass_;

typedef struct {
  bool *examplesAreComplete;
  double *parentResiduals;
  double **completenessValues;
  ulong completeFrequency;
  ulong incompleteFrequency;
  ulong *frequencies;
  double completeSum;
  double incompleteSum;
  Lock lock;
} CompletenessPredictionsPass_;

typedef struct {
  bool *examplesAreComplete;
  double completePrediction;
  double incompletePrediction;
  double *parentResiduals;
  double *parentResidualPredictions;
  double *predictions;
  double *residuals;
  Lock lock;
} CompletenessResidualsPass_;

typedef struct {
  ulong numFeatures;
  ulong *completeFrequencies;
  ulong *completeToAll;
  ulong *frequencies;
  double *completenessResiduals;
  double **completeValues;
  double **features;
} MixedCompletePass_;

typedef struct {
  bool *examplesAreComplete;
  ulong numFeatures;
  ulong *allToComplete;
  double maxResidual;
  double minResidual;
  double *completenessResiduals;
  double *parentResidualPredictions;
  double *residuals;
  double **completeValues;
  Lock lock;
  BearHypercube *completeHypercube;
  BearModel *this;
} MixedPredictionsPass_;

typedef struct {
  ulong numFeatures;
  double *parentResiduals;
  double **features;
  double **values;
} PureValuesPass_;

typedef struct {
  ulong numFeatures;
  double maxResidual;
  double minResidual;
  double *parentResiduals;
  double *parentResidualPredictions;
  double *residuals;
  double **values;
  Lock lock;
  BearHypercube *completeHypercube;
  BearModel *this;
} PurePredictionsPass_;

typedef struct {
  double *labelEstimates;
  double *labelPredictions;
  void *lossFunctionParams;
  LossFunction *lossFunction;
  BearModel *this;
#ifdef VERIFY_LABEL_PREDICTIONS_
  ulong numFullFeatures;
  double **fullFeatures;
#endif
} LabelPredictionsPass_;

typedef struct {
  ulong totalFrequency;
  ulong *frequencies;
  double sum;
  double *labels;
  double *labelEstimates;
  void *lossFunctionParams;
  LossFunction *lossFunction;
  Lock lock;
} LossPass_;

typedef struct {
  ulong nextPart;
  RandomWeightedNode *node;
} Attachment_;

static ulong id_ = 1;
static Lock lock_ = LOCK_INITIALIZER;

static void empty_prediction_(Parallel *par);

static void empty_residuals_(Parallel *par);

static void empty_loss_(Parallel *par);

static void find_incomplete_(Parallel *par);

static void completeness_predictions_(Parallel *par);

static void mixed_complete_(Parallel *par);

static void mixed_predictions_(Parallel *par);

static void pure_values_(Parallel *par);

static void pure_predictions_(Parallel *par);

static void label_predictions_(Parallel *par);

static void loss_(Parallel *par);

static void completeness_residuals_(Parallel *par);

static void add_(double *label, double prediction, BearModel *model);

static void attachment_new_(Attachment_ **p_this, RandomWeightedNode *node);

static void attachment_delete_(Attachment_ **p_this);

static sbyte attachment_compare_(Attachment_ *a, Attachment_ *b, void *x);

static void attachment_print_(Attachment_ *this, Printer *p);

static void veto_(Attachment_ *this, RandomWeighted *rw);


/* Constructor for an empty model, which is the root model of all models for a
 * given label.
 */
void bear_model_new_empty(
    BearModel **p_this,
    ulong num_examples,
    SwappableDoubles *labels,
    SwappableUlongs *frequencies,
    LossFunction *loss_function) {
  ulong total_frequency;
  ulong *_frequencies = null;
  double empty_label_estimate;
  double empty_label_prediction;
  double *label_estimates;
  double *label_predictions;
  double *parent_residual_predictions;
  double *residuals;
  double *_labels = null;
  EmptyPredictionPass_ prediction_pass;
  EmptyResidualsPass_ residuals_pass;
  EmptyLossPass_ loss_pass;
  BearModelWorking *w;
  BearModel *this;

  /* Sanity checks.
   */
  if (num_examples == 0) {
    throw("Must have at least one example");
  }

  if (loss_function == null) {
    throw("Loss function cannot be null");
  }

  /* Create the object.
   */
  this = newo(p_this);

  /* Set the model primary key ID.
   */
  lock(&lock_);
  this->id = id_++;
  unlock(&lock_);

  /* This is an empty model.
   */
  this->empty = true;

  /* Store reference to the loss function.
   */
  this->lossFunction = loss_function;

  /* Compute parameters of the loss function, which are owned by this empty
   * model.
   */
  loss_function->createParams(
      &this->_lossFunctionParams,
      num_examples,
      labels,
      frequencies);

  /* Create an empty assembly.
   */
  bear_assembly_new_empty(&this->_assembly);

  /* Create the working object.
   */
  bear_model_working_new_empty(
      &this->_w,
      num_examples,
      labels,
      frequencies);
  w = this->_w;

  /* Access the input data.
   */
  swappable_values_access(labels, &_labels);
  swappable_values_access(frequencies, &_frequencies);

  /* Compute the empty label prediction. We also track the minimum and maximum
   * label value.
   */
  lock_initialize(&prediction_pass.lock);
  prediction_pass.labels = _labels;
  prediction_pass.frequencies = _frequencies;
  prediction_pass.sum = 0;
  prediction_pass.totalFrequency = 0;
  prediction_pass.minLabel = DBL_MAX;
  prediction_pass.maxLabel = -DBL_MAX;
  parallel_for(num_examples, 13, empty_prediction_, &prediction_pass);

  /* Sanity check.
   */
  if ((total_frequency = prediction_pass.totalFrequency) == 0) {
    throw("Impossible");
  }

  /* Compute the empty label prediction.
   */
  empty_label_prediction = this->emptyLabelPrediction
    = prediction_pass.sum / prediction_pass.totalFrequency;

  /* Store the minimum and maximum input, i.e., label value.
   */
  this->minInput = prediction_pass.minLabel;
  this->maxInput = prediction_pass.maxLabel;

  /* Compute the empty label estimate, which may differ from the empty model
   * prediction for some loss functions, and is the quantity that is used to
   * compute the loss.
   */
  empty_label_estimate = this->emptyLabelEstimate
    = loss_function->computeLabelEstimate(
        empty_label_prediction,
        this->_lossFunctionParams);

  /* Get references.
   */
  parent_residual_predictions = w->_parentResidualPredictions;
  residuals = w->_residuals;
  label_predictions = w->_labelPredictions;
  label_estimates = w->_labelEstimates;

  /* In parallel, for each example, set the parent residual prediction and
   * label prediction, compute the residual and store it, keep track of the
   * minimum and maximum residual, and store the label estimate.
   */
  lock_initialize(&residuals_pass.lock);
  residuals_pass.emptyLabelPrediction = empty_label_prediction;
  residuals_pass.emptyLabelEstimate = empty_label_estimate;
  residuals_pass.parentResidualPredictions = parent_residual_predictions;
  residuals_pass.labelPredictions = label_predictions;
  residuals_pass.labelEstimates = label_estimates;
  residuals_pass.labels = _labels;
  residuals_pass.residuals = residuals;
  residuals_pass.minResidual = DBL_MAX;
  residuals_pass.maxResidual = -DBL_MAX;
  parallel_for(num_examples, 10, empty_residuals_, &residuals_pass);

  /* Store the minimum and maximum residual.
   */
  w->minResidual = residuals_pass.minResidual;
  w->maxResidual = residuals_pass.maxResidual;

  /* Compute the loss for this empty model, using the empty model label
   * estimate.
   */
  lock_initialize(&loss_pass.lock);
  loss_pass.lossFunction = loss_function;
  loss_pass.lossFunctionParams = this->_lossFunctionParams;
  loss_pass.sum = 0;
  loss_pass.frequencies = _frequencies;
  loss_pass.labels = _labels;
  loss_pass.emptyLabelEstimate = empty_label_estimate;
  parallel_for(num_examples, 6, empty_loss_, &loss_pass);

  /* Compute the loss.
   */
  this->loss = this->emptyLoss = loss_pass.sum / total_frequency;

  /* Deposit the working.
   */
  swappable_new_deposit(&this->_wS, bear_model_working, &this->_w);
  w = null;

  /* Release the local copy of the input data.
   */
  swappable_values_release(frequencies, &_frequencies);
  swappable_values_release(labels, &_labels);
}

/* Constructor for a general model. The assembly of the parent model must be
 * consistent with the assembly specified for this model. The base model is the
 * previously-existing model that this new construction is based on; for
 * gluing, it is just the parent, but for melting it is the "sibling" without
 * the extra attachment part melted in. The 'features' swappable should be that
 * for all features; this constructor will select out the ones it needs.
 */
void bear_model_new(
    BearModel **p_this,
    BearAssembly *assembly,
    BearModel *parent,
    BearModel *base,
    SwappableDoubles **features,
    bool *features_have_missing_values,
    Decider *decider) {
  bool have_incomplete;
  bool *examples_are_complete;
  bool *_features_have_missing_values = null;
  ulong c;
  ulong e;
  ulong f;
  ulong f_full;
  ulong num_complete_examples;
  ulong num_examples;
  ulong num_features;
  ulong num_models;
  ulong *_all_to_complete = null;
  ulong *_complete_frequencies = null;
  ulong *_complete_to_all = null;
  ulong *_frequencies = null;
  double complete_prediction;
  double incomplete_prediction;
  double loss;
  double max_residual;
  double min_residual;
  double *label_estimates;
  double *label_predictions;
  double *parent_residuals;
  double *parent_residual_predictions;
  double *residuals;
  double *_completeness_residuals = null;
  double *_labels = null;
  double **_complete_values = null;
  double **_completeness_values = null;
  double **_features = null;
  double **_values = null;
  void *loss_function_params;
  FindIncompletePass_ find_incomplete_pass;
  CompletenessPredictionsPass_ completeness_predictions_pass;
  CompletenessResidualsPass_ completeness_residuals_pass;
  MixedCompletePass_ mixed_complete_pass;
  MixedPredictionsPass_ mixed_predictions_pass;
  PureValuesPass_ pure_values_pass;
  PurePredictionsPass_ pure_predictions_pass;
  LabelPredictionsPass_ label_predictions_pass;
  LossPass_ loss_pass;
  LossFunction *loss_function;
  BearPart *part;
  BearHypercube *complete_hypercube;
  BearHypercube *_completeness_hypercube = null;
  BearModelWorking *parent_w;
  BearModelWorking *w;
  BearModel *model;
  BearModel *this;
#ifdef VERIFY_LABEL_PREDICTIONS_
  ulong max_f_full;
  ulong num_full_features;
  ulong p;
  double **_full_features = null;
#endif

  /* Our job is to model the residuals of the parent model using the features
   * specified in the last part of the assembly.
   *
   * Any example with any such feature value missing is called an "incomplete
   * example"; otherwise, it is a "complete example." If any of the features
   * for the last part have missing values, then there will be at least one
   * incomplete example. We cannot include incomplete examples when training
   * the model using a BearHypercube.
   *
   * However, in such cases, we first model the binary feature of whether an
   * example is complete or not, since it is possible that the distribution of
   * parent residuals for incomplete examples is different than that for
   * complete examples. This is called a "completeness model," but, as will
   * become clear, all we need from it is whether it is trivial (i.e., both
   * values of the binary completeness feature are merged, which necessarily
   * means that the labels are merged too) or not.
   *
   * Of course, this only makes sense if there is also at least one example
   * that is complete. If not, then this whole BearModel is trivial and
   * useless.
   *
   * Somewhat subtle is the fact that, regardless of whether the completeness
   * model is trivial or not, we still make a prediction for incomplete
   * examples (as the mean of their labels) and for complete examples (as the
   * mean of their labels). The reason for this is explained below. The
   * residuals of this completeness model are then used in the next step,
   * rather than the original residuals of the parent model.
   *
   * Regardless of whether we have incomplete examples or not, the next step is
   * to model the residuals (of the completeness model or of the parent model,
   * as the case may be), for all COMPLETE examples only, with a BearHypercube
   * of the features for this last part. This is called a "complete model,"
   * because it only models complete examples (with features).
   *
   * So in general, a BearModel conceptually consists of two internal models:
   * first a completeness model, and then a complete model.
   *
   * Each of these two internal models can be trivial or nontrivial. If both
   * are trivial, then this BearModel is trivial, and effectively useless.
   * Otherwise, this BearModel is nontrivial, and either or both of the
   * internal models are used to model and predict.
   *
   * The subtle reason mentioned above for why the incomplete and complete
   * predictions are always used, even if the completeness model is trivial, is
   * because the only scenario in which this matters is if the complete model
   * is nontrivial (otherwise both internal models are trivial, and this whole
   * BearModel would be trivial, and not used for anything further). Now, the
   * complete model only models the complete examples. On average it will make
   * a prediction that is the mean of the complete examples. If the incomplete
   * examples were not similarly modeled, then we would potentially introduce a
   * bias into the mean prediction across all examples. In such cases it makes
   * most sense to simply model the incomplete examples with their own mean,
   * i.e., the incomplete prediction. That removes the bias. But this is
   * ultimately no different than using both the incomplete and complete
   * predictions, and then modeling the complete examples with the residuals of
   * the complete prediction, since the overall predictions will just be the
   * sum of the two. So it is easier to just use the incomplete and complete
   * predictions in all cases. This actually also improves somewhat the
   * fidelity of the complete model, since the residuals will have mean zero,
   * which is better for paw quantization.
   *
   * So let's implement the above!
   *
   * Sanity checks:
   *
   * First, the empty model has its own special constructor above.
   */
  if (bear_assembly_is_empty(assembly)) {
    throw("Assembly cannot be empty");
  }

  /* Second, check that the parent's assembly is the correct parent assembly of
   * the specified assembly for this model.
   */
  if (!bear_assembly_is_parent(parent->_assembly, assembly)) {
    throw("Assembly of parent is not the parent of the specified assembly");
  }

  /* Third, the assembly of the base model must either be the parent (if
   * gluing) or a sibling (if melting) of the assembly specified for
   * construction.
   */
  if (!bear_assembly_is_parent(base->_assembly, assembly)
      && !bear_assembly_is_sibling(base->_assembly, assembly)) {
    throw("Assembly of base is neither the parent nor the sibling of the "
        "specified assembly");
  }

  /* Fourth, the parent model cannot be trivial.
   */
  if (bear_model_is_trivial(parent)) {
    throw("Parent model cannot be trivial");
  }

  /* Fifth, if the parent model is not an empty model, it cannot have zero
   * selection weight.
   */
  if (!parent->empty && parent->selectionWeight == 0) {
    throw("Non-empty parent model cannot have zero selection weight");
  }

  /* Create the object.
   */
  this = newo(p_this);

  /* Set the model primary key ID.
   */
  lock(&lock_);
  this->id = id_++;
  unlock(&lock_);

  /* Store a reference to the parent, and its ID.
   */
  this->parent = parent;
  this->parentId = parent->id;

  /* Copy from the parent.
   */
  this->emptyLoss = parent->emptyLoss;

  /* Reacquire the working in the parent model.
   */
  swappable_reacquire(parent->_wS);

  /* Extract from the parent.
   */
  parent_w = parent->_w;
  parent_residuals = parent_w->_residuals;
  this->minInput = parent_w->minResidual;
  this->maxInput = parent_w->maxResidual;

  /* Store a copy of the assembly specified to be built.
   */
  bear_assembly_new_copy(&this->_assembly, assembly);

  /* Create a DynamicSet for Attachment_ objects.
   */
  dynamic_set_new(
      &this->_attachments,
      attachment_delete_,
      attachment_compare_,
      attachment_print_);

  /* If the parent is the empty model, extract references for the loss
   * function. (If not, we will find the empty model below.)
   */
  if (parent->empty) {
    loss_function = parent->lossFunction;
    loss_function_params = parent->_lossFunctionParams;
  }

  /* Create the working object, which copies some information from the parent.
   */
  bear_model_working_new(&this->_w, parent);
  w = this->_w;

  /* Extract references.
   */
  parent_residual_predictions = w->_parentResidualPredictions;
  residuals = w->_residuals;

  /* Extract the number of examples.
   */
  num_examples = w->numExamples;

  /* Get a reference to the last part of the specified assembly, which is the
   * part that this model is based on (henceforth, "this part").
   */
  part = bear_assembly_last_part(assembly);

  /* Extract the number of features in this part.
   */
  num_features = part->numFeatures;

  /* We create local collections _features and _features_have_missing_values
   * that contain data for just the features of this part, in contrast to the
   * collection references 'features' and 'features_have_missing_values' passed
   * in to this constructor that contain data for all features.
   *
   * For _features_have_missing_values we create the array. For _features, we
   * create the top-level array.
   */
  new(&_features_have_missing_values, num_features);
  new(&_features, num_features);

  /* Walk through the features of this part. We also determine whether any of
   * the features of this part have missing values; if so, we will have at
   * least one incomplete example.
   */
  for (f = 0, have_incomplete = false; f < num_features; f++) {
    /* Extract the feature index in the full set.
     */
    f_full = part->_features[f];

    /* Access the data for this feature into our local matrix.
     */
    swappable_values_access(features[f_full], &_features[f]);

    /* Copy across the missing values flag, and, if it is true, set our overall
     * flag.
     */
    if((_features_have_missing_values[f]
          = features_have_missing_values[f_full])) {
      have_incomplete = true;
    }
  }

  /* Access the frequencies for all examples, a reference to the Swappable for
   * which is in the working object (copied from that of the parent model).
   */
  swappable_values_access(w->frequencies, &_frequencies);

  /* Create an array of flags signifying whether each given example is
   * complete or not. We store this in the working object as it will be of use
   * in later processing.
   */
  new(&w->_examplesAreComplete, num_examples);
  examples_are_complete = w->_examplesAreComplete;

  /* Populate that array, and count up how many complete examples there are.
   */
  lock_initialize(&find_incomplete_pass.lock);
  find_incomplete_pass.numFeatures = num_features;
  find_incomplete_pass.featuresHaveMissingValues
    = _features_have_missing_values;
  find_incomplete_pass.features = _features;
  find_incomplete_pass.examplesAreComplete = examples_are_complete;
  find_incomplete_pass.numCompleteExamples = 0;
  parallel_for(
      num_examples,
      5 + num_features * 4,
      find_incomplete_,
      &find_incomplete_pass);
  num_complete_examples = find_incomplete_pass.numCompleteExamples;

  /* Switch on whether we have any incomplete examples.
   */
  if (have_incomplete) {
    /* Sanity check.
     */
    if (num_complete_examples == num_examples) {
      throw("Impossible: found no incomplete examples");
    }

    /* Check if there are no complete examples.
     */
    if (num_complete_examples == 0) {
      /* There are no complete examples. This model is useless and trivial. Set
       * the flags that signify this.
       */
      this->completenessTrivial = true;
      this->completeTrivial = true;

      /* Clean up and return.
       */
      for (f = 0; f < num_features; f++) {
        swappable_values_release(features[part->_features[f]], &_features[f]);
      }

      swappable_release(parent->_wS);
      deletev(&_features);
      swappable_values_release(w->frequencies, &_frequencies);
      deletev(&_features_have_missing_values);

      return;
    }

    /* We have incomplete examples, but there is at least one complete example,
     * so this model might not be useless. We first compute the sums for the
     * complete and incomplete predictions, and populate the matrix of values
     * for a single binary feature equal in value to _examples_are_complete and
     * a single label being the residuals of the parent model.
     */
    newm(&_completeness_values, num_examples, 2);
    lock_initialize(&completeness_predictions_pass.lock);
    completeness_predictions_pass.examplesAreComplete = examples_are_complete;
    completeness_predictions_pass.completenessValues = _completeness_values;
    completeness_predictions_pass.parentResiduals = parent_residuals;
    completeness_predictions_pass.frequencies = _frequencies;
    completeness_predictions_pass.completeSum = 0;
    completeness_predictions_pass.incompleteSum = 0;
    completeness_predictions_pass.incompleteFrequency = 0;
    completeness_predictions_pass.completeFrequency = 0;
    parallel_for(
        num_examples,
        14,
        completeness_predictions_,
        &completeness_predictions_pass);

    /* Compute the incomplete and complete predictions.
     */
    incomplete_prediction = this->incompletePrediction
      = completeness_predictions_pass.incompleteSum
      / completeness_predictions_pass.incompleteFrequency;
    complete_prediction = this->completePrediction
      = completeness_predictions_pass.completeSum
      / completeness_predictions_pass.completeFrequency;

    /* Now create the BearHypercube for the completeness feature.
     */
    bear_hypercube_new_memory(
        &_completeness_hypercube,
        1,
        decider,
        num_examples,
        _completeness_values,
        _frequencies);

    /* Check if it is nontrivial.
     */
    this->completenessTrivial
      = bear_hypercube_is_trivial(_completeness_hypercube);

    /* Clean up.
     */
    deletemv(&_completeness_values, num_examples);

    /* We don't actually need the completeness hypercube any more; whether it
     * was nontrivial was all we cared about.
     */
    bear_hypercube_delete(&_completeness_hypercube);

    /* Now set the parent residual predictions to be the completeness
     * predictions, and compute the residuals.
     */
    new(&_completeness_residuals, num_examples);
    completeness_residuals_pass.incompletePrediction = incomplete_prediction;
    completeness_residuals_pass.completePrediction = complete_prediction;
    completeness_residuals_pass.parentResiduals = parent_residuals;
    completeness_residuals_pass.parentResidualPredictions
      = parent_residual_predictions;
    completeness_residuals_pass.examplesAreComplete = examples_are_complete;
    completeness_residuals_pass.residuals = _completeness_residuals;
    parallel_for(
        num_examples,
        5,
        completeness_residuals_,
        &completeness_residuals_pass);

    /* We now model these residuals of the completeness model.
     *
     * For incomplete examples, there is nothing more to do. We thus model the
     * complete examples only. Create the matrix of values and array of
     * frequencies for only these complete examples.
     */
    newm(&_complete_values, num_complete_examples, num_features + 1);
    new(&_complete_frequencies, num_complete_examples);

    /* Create the lookups of complete examples to all examples and vice versa.
     * This cannot be parallelized. Create the lookup arrays.
     */
    new(&_complete_to_all, num_complete_examples);
    new(&_all_to_complete, num_examples);

    /* Walk through the examples.
     */
    for (e = 0, c = 0; e < num_examples; e++) {
      /* Check if this is a complete example.
       */
      if (examples_are_complete[e]) {
        /* It is. Store the lookups.
         */
        _complete_to_all[c] = e;
        _all_to_complete[e] = c;

        /* Move to the next complete example.
         */
        c++;
      } else {
        /* Incomplete example. Set an invalid lookup.
         */
        _all_to_complete[e] = ULONG_MAX;
      }
    }

    /* Sanity check.
     */
    if (c != num_complete_examples) {
      throw("Impossible: %lu v %lu", c, num_complete_examples);
    }

    /* We can now populate the matrix of values and array of frequencies in
     * parallel.
     */
    mixed_complete_pass.completeToAll = _complete_to_all;
    mixed_complete_pass.numFeatures = num_features;
    mixed_complete_pass.features = _features;
    mixed_complete_pass.frequencies = _frequencies;
    mixed_complete_pass.completeValues = _complete_values;
    mixed_complete_pass.completeFrequencies = _complete_frequencies;
    mixed_complete_pass.completenessResiduals = _completeness_residuals;
    parallel_for(
        num_complete_examples,
        6 + num_features * 5,
        mixed_complete_,
        &mixed_complete_pass);

    /* Create the BearHypercube to model the complete examples.
     */
    bear_hypercube_new_memory(
        &this->_completeHypercube,
        num_features,
        decider,
        num_complete_examples,
        _complete_values,
        _complete_frequencies);
    complete_hypercube = this->_completeHypercube;

    /* Store whether it is trivial.
     */
    this->completeTrivial = bear_hypercube_is_trivial(complete_hypercube);

    /* If it is, delete it.
     */
    if (this->completeTrivial) {
      bear_hypercube_delete(&this->_completeHypercube);
    }

    /* Check if this overall model is trivial, which is only true if both the
     * completeness model and the complete model are trivial.
     */
    if (bear_model_is_trivial(this)) {
      /* This overall model is trivial. Clean up and return.
       */
      for (f = 0; f < num_features; f++) {
        swappable_values_release(features[part->_features[f]], &_features[f]);
      }

      deletemv(&_complete_values, num_complete_examples);
      deletev(&_complete_frequencies);
      deletev(&_complete_to_all);
      deletev(&_all_to_complete);
      deletev(&_completeness_residuals);
      swappable_release(parent->_wS);
      deletev(&_features);
      swappable_values_release(w->frequencies, &_frequencies);
      deletev(&_features_have_missing_values);

      return;
    }

    /* This model is not trivial. Compute predictions and residuals.
     */
    lock_initialize(&mixed_predictions_pass.lock);
    mixed_predictions_pass.this = this;
    mixed_predictions_pass.minResidual = DBL_MAX;
    mixed_predictions_pass.maxResidual = -DBL_MAX;
    mixed_predictions_pass.examplesAreComplete = examples_are_complete;
    mixed_predictions_pass.allToComplete = _all_to_complete;
    mixed_predictions_pass.completeHypercube = complete_hypercube;
    mixed_predictions_pass.completeValues = _complete_values;
    mixed_predictions_pass.parentResidualPredictions
      = parent_residual_predictions;
    mixed_predictions_pass.completenessResiduals = _completeness_residuals;
    mixed_predictions_pass.residuals = residuals;
    mixed_predictions_pass.numFeatures = num_features;
    parallel_for(
        num_examples,
        5 + num_features * 8,
        mixed_predictions_,
        &mixed_predictions_pass);
    min_residual = mixed_predictions_pass.minResidual;
    max_residual = mixed_predictions_pass.maxResidual;

    /* Clean up.
     */
    deletemv(&_complete_values, num_complete_examples);
    deletev(&_complete_frequencies);
    deletev(&_completeness_residuals);
    deletev(&_complete_to_all);
    deletev(&_all_to_complete);
  } else {
    /* We have no incomplete examples. The completeness model is by definition
     * trivial.
     */
    this->completenessTrivial = true;

    /* We simply need to model the residuals of the parent model. Create a
     * matrix for the values.
     */
    newm(&_values, num_examples, num_features + 1);

    /* Populate this matrix.
     */
    pure_values_pass.numFeatures = num_features;
    pure_values_pass.features = _features;
    pure_values_pass.values = _values;
    pure_values_pass.parentResiduals = parent_residuals;
    parallel_for(
        num_examples,
        6 + num_features * 7,
        pure_values_,
        &pure_values_pass);

    /* Create the BearHypercube.
     */
    bear_hypercube_new_memory(
        &this->_completeHypercube,
        num_features,
        decider,
        num_examples,
        _values,
        _frequencies);
    complete_hypercube = this->_completeHypercube;

    /* Check and store whether it is trivial.
     */
    if ((this->completeTrivial
          = bear_hypercube_is_trivial(complete_hypercube))) {
      /* This complete hypercube is trivial, and by definition the completeness
       * model is trivial, so this overall model is trivial. Clean up and
       * return.
       */
      for (f = 0; f < num_features; f++) {
        swappable_values_release(features[part->_features[f]], &_features[f]);
      }

      bear_hypercube_delete(&this->_completeHypercube);
      deletemv(&_values, num_examples);
      deletev(&_features_have_missing_values);
      swappable_release(parent->_wS);
      swappable_delete(&this->_wS);
      deletev(&_features);
      swappable_values_release(w->frequencies, &_frequencies);

      return;
    }

    /* Now compute predictions and residuals.
     */
    lock_initialize(&pure_predictions_pass.lock);
    pure_predictions_pass.minResidual = DBL_MAX;
    pure_predictions_pass.maxResidual = -DBL_MAX;
    pure_predictions_pass.this = this;
    pure_predictions_pass.completeHypercube = complete_hypercube;
    pure_predictions_pass.values = _values;
    pure_predictions_pass.parentResidualPredictions
      = parent_residual_predictions;
    pure_predictions_pass.parentResiduals = parent_residuals;
    pure_predictions_pass.residuals = residuals;
    pure_predictions_pass.numFeatures = num_features;
    parallel_for(
        num_examples,
        8 + num_features * 8,
        pure_predictions_,
        &pure_predictions_pass);
    min_residual = pure_predictions_pass.minResidual;
    max_residual = pure_predictions_pass.maxResidual;

    /* Delete the matrix of values.
     */
    deletemv(&_values, num_examples);
  }

  /* Store the minimum and maximum residual.
   */
  w->minResidual = min_residual;
  w->maxResidual = max_residual;

  /* Clean up.
   */
  deletev(&_features_have_missing_values);

  /* We now need to propagate all the predictions up to the empty model to
   * predict the label values. From those label predictions we ask the loss
   * function to compute the label estimates that minimize the expected loss
   * (which for MSE and Log are the same as the predictions, but for Balanced
   * Log are different, and for MAE convert centered ranks back to values).
   * From those estimates and the actual label values we compute the loss of
   * this model. Although with some loss models (like MSE) it might seem
   * mathematically equivalent to compute the loss from just the final
   * residuals, we actually bound each prediction for each model up the chain
   * to the range of the inputs to that model, so even in the case of MSE the
   * ultimate errors may differ from the residuals of this last model.
   *
   * Create arrays for the label predictions and label estimates.
   */
  label_predictions = new(&w->_labelPredictions, num_examples);
  label_estimates = new(&w->_labelEstimates, num_examples);

  /* Reacquire the working for all remaining models in this chain. We already
   * have this model and its parent. Walk through those models. The last will
   * be the empty model, from which we will get references to the loss function
   * and its params. This only fails if this is an elementary model, in which
   * the parent has already been extracted from above.
   */
  for (model = parent->parent; model != null; model = model->parent) {
    swappable_reacquire(model->_wS);

    if (model->empty) {
      loss_function = model->lossFunction;
      loss_function_params = model->_lossFunctionParams;
    }
  }

  /* Compute the number of models in our chain, just for ops per iteration.
   */
  for (model = this, num_models = 1;; model = model->parent, num_models++) {
    if (model->empty) {
      break;
    }
  }

#ifdef VERIFY_LABEL_PREDICTIONS_

  /* We are verifying all label predictions against those of explicit calls to
   * bear_model_predict() below. We need to set up for that.
   *
   * First, determine the maximum feature index used in the assembly.
   */
  for (p = 0, max_f_full = 0; p < assembly->numParts; p++) {
    for (f = 0; f < assembly->_parts[p]->numFeatures; f++) {
      max_f_full = maximum(max_f_full, assembly->_parts[p]->_features[f]);
    }
  }

  /* The number of features we need to care about is one more than this maximum
   * index.
   */
  num_full_features = max_f_full + 1;

  /* Access all the features we need.
   */
  swappable_values_accessm(features, num_full_features, &_full_features);

  /* Store in the pass object.
   */
  label_predictions_pass.numFullFeatures = num_full_features;
  label_predictions_pass.fullFeatures = _full_features;

  /* Now we need to reacquire all the hypercubes of all the models in this
   * model chain. Walk through them.
   */
  for (model = this; !model->empty; model = model->parent) {
    if (model->_completeHypercubeS != null) {
      swappable_reacquire(model->_completeHypercubeS);
    }
  }

#endif

  /* Compute the label predictions.
   */
  label_predictions_pass.this = this;
  label_predictions_pass.labelPredictions = label_predictions;
  label_predictions_pass.labelEstimates = label_estimates;
  label_predictions_pass.lossFunction = loss_function;
  label_predictions_pass.lossFunctionParams = loss_function_params;
  parallel_for(
      num_examples,
      4 + num_models * 5,
      label_predictions_,
      &label_predictions_pass);

#ifdef VERIFY_LABEL_PREDICTIONS_

  /* Clean up.
   */
  swappable_values_releasem(features, num_full_features, &_full_features);

  for (model = this; !model->empty; model = model->parent) {
    if (model->_completeHypercubeS != null) {
      swappable_release(model->_completeHypercubeS);
    }
  }

#endif

  /* Access the labels from the empty model, which is where we broke out.
   */
  swappable_values_access(model->_w->labels, &_labels);

  /* Compute the loss.
   */
  lock_initialize(&loss_pass.lock);
  loss_pass.sum = 0;
  loss_pass.totalFrequency = 0;
  loss_pass.frequencies = _frequencies;
  loss_pass.labels = _labels;
  loss_pass.labelEstimates = label_estimates;
  loss_pass.lossFunction = loss_function;
  loss_pass.lossFunctionParams = loss_function_params;
  parallel_for(num_examples, 4, loss_, &loss_pass);

  /* Normalize and store it.
   */
  this->loss = loss = loss_pass.sum / loss_pass.totalFrequency;

  /* Release the features, labels, and frequencies.
   */
  for (f = 0; f < num_features; f++) {
    swappable_values_release(features[part->_features[f]], &_features[f]);
  }

  deletev(&_features);
  swappable_values_release(model->_w->labels, &_labels);
  swappable_values_release(w->frequencies, &_frequencies);

  /* Compute the weight and selection weight.
   */
  this->weight = bear_model_weight(loss, this->emptyLoss);
  this->selectionWeight = bear_model_selection_weight(
      this->weight,
      base->weight);

  /* Release the working for all parent models.
   */
  for (model = parent; model != null; model = model->parent) {
    swappable_release(model->_wS);
  }

  /* Deposit the working, and the complete hypercube if it is not trivial.
   */
  swappable_new_deposit(&this->_wS, bear_model_working, &this->_w);

  if (!this->completeTrivial) {
    swappable_new_deposit(
        &this->_completeHypercubeS,
        bear_hypercube,
        &this->_completeHypercube);
  }
}

/* Load.
 */
void bear_model_new_load(BearModel **p_this, File *f) {
  char *_loss_function_name = null;
  BearModel *this;

  this = newo(p_this);
  load_label("[BearModel]", f);
  load_value(this->id, f);
  load_value(this->empty, f);

  if (!this->empty) {
    load_value(this->completenessTrivial, f);
    load_value(this->completeTrivial, f);
    load_value(this->incompletePrediction, f);
    load_value(this->completePrediction, f);
    load(this->_assembly, bear_assembly, f);
    load_value(this->weight, f);
    load_value(this->parentId, f);
  }

  load_value(this->minInput, f);
  load_value(this->maxInput, f);

  if (this->empty) {
    load(_loss_function_name, string, f);
    this->lossFunction = loss_function_get(_loss_function_name);
    string_delete(&_loss_function_name);
    this->lossFunction->loadParams(&this->_lossFunctionParams, f);
    load_value(this->emptyLabelPrediction, f);
    load_value(this->emptyLabelEstimate, f);
  } else if (!this->completeTrivial) {
    load(this->_completeHypercube, bear_hypercube, f);
  }

  load_label("[/BearModel]", f);
}

/* Destructor.
 */
void bear_model_delete(BearModel **p_this) {
  BearModel *this;

  if ((this = *p_this) == null) {
    return;
  }

  if (this->empty) {
    this->lossFunction->destroyParams(&this->_lossFunctionParams);
  }

  dynamic_set_delete(&this->_attachments);
  swappable_delete(&this->_completeHypercubeS);
  bear_hypercube_delete(&this->_completeHypercube);
  bear_assembly_delete(&this->_assembly);
  swappable_delete(&this->_wS);
  bear_model_working_delete(&this->_w);
  deleteo(p_this);
}

/* Print.
 */
void bear_model_print(BearModel *this, Printer *p) {
  sprint(p, "BearModel");

  if (this->id == 0) {
    throw("Model pointer bad: ID value is zero");
  }

  if (this->id >= id_) {
    throw("Model pointer bad: ID value of %lu greater than max ID %lu",
        this->id,
        id_);
  }

  rprint(p, "ID: %lu", this->id);

  if (this->parentId != 0) {
    rprint(p, "Parent model ID: %lu", this->parentId);
  }

  if (bear_model_is_trivial(this)) {
    rprint(p, "This model is trivial.");
  } else if (this->empty) {
    rprint(p, "This is an empty model.");
    rprint(p, "Minimum allowed label: %.16g", this->minInput);
    rprint(p, "Maximum allowed label: %.16g", this->maxInput);
    rprint(p, "Empty label prediction: %.16g", this->emptyLabelPrediction);

    if (this->emptyLabelEstimate != this->emptyLabelPrediction) {
      rprint(p, "Empty label estimate: %.16g", this->emptyLabelEstimate);
    }
  } else {
    bear_assembly_print(this->_assembly, p);
    rprint(p, "Weight: %.16g", this->weight);

    if (this->_wS != null) {
      rprint(p, "Loss: %.16g", this->loss);
      rprint(p, "Empty model loss: %.16g", this->emptyLoss);
      rprint(p, "Selection weight: %.16g", this->selectionWeight);
    }

    rprint(p, "Minimum input: %.16g", this->minInput);
    rprint(p, "Maximum input: %.16g", this->maxInput);

    rprint(p,
        "The completeness model is %strivial.",
        this->completenessTrivial ? "" : "non");
    rprint(p, "Incomplete prediction: %.16g", this->incompletePrediction);
    rprint(p, "Complete prediction: %.16g", this->completePrediction);

    rprint(p,
        "The complete model is %strivial.",
        this->completeTrivial ? "" : "non");

    if (!this->completeTrivial) {
      sprint(p, "Complete hypercube");

      if (this->_completeHypercubeS != null) {
        swappable_reacquire(this->_completeHypercubeS);
      }

      bear_hypercube_print(this->_completeHypercube, p);

      if (this->_completeHypercubeS != null) {
        swappable_release(this->_completeHypercubeS);
      }

      fprint(p);
    }

    if (this->_wS != null) {
      swappable_reacquire(this->_wS);
      bear_model_working_print(this->_w, p);
      swappable_release(this->_wS);
    }
  }

  if (this->parent != null) {
    sprint(p, "Parent");
    bear_model_print(this->parent, p);
    fprint(p);
  }

  fprint(p);
}

/* Save.
 */
void bear_model_save(BearModel *this, File *f) {
  save_label("[BearModel]", f);
  save_value(this->id, f);
  save_value(this->empty, f);

  if (!this->empty) {
    save_value(this->completenessTrivial, f);
    save_value(this->completeTrivial, f);
    save_value(this->incompletePrediction, f);
    save_value(this->completePrediction, f);
    save(this->_assembly, bear_assembly, f);
    save_value(this->weight, f);
    save_value(this->parentId, f);
  }

  save_value(this->minInput, f);
  save_value(this->maxInput, f);

  if (this->empty) {
    save(this->lossFunction->name, string, f);
    this->lossFunction->saveParams(this->_lossFunctionParams, f);
    save_value(this->emptyLabelPrediction, f);
    save_value(this->emptyLabelEstimate, f);
  } else if (!this->completeTrivial) {
    if (this->_completeHypercubeS != null) {
      swappable_reacquire(this->_completeHypercubeS);
    }

    save(this->_completeHypercube, bear_hypercube, f);

    if (this->_completeHypercubeS != null) {
      swappable_release(this->_completeHypercubeS);
    }
  }

  save_label("[/BearModel]", f);
}

/* Equals.
 */
bool bear_model_equals(BearModel *this, BearModel *other) {
  bool equals;

  if (other->id != this->id
      || other->empty != this->empty
      || other->completenessTrivial != this->completenessTrivial
      || other->completeTrivial != this->completeTrivial
      || other->incompletePrediction != this->incompletePrediction
      || other->completePrediction != this->completePrediction
      || other->weight != this->weight
      || other->parentId != this->parentId
      || (!this->empty
        &&!bear_assembly_equals(other->_assembly, this->_assembly))) {
    return false;
  }

  if (this->empty) {
    if (other->emptyLabelPrediction != this->emptyLabelPrediction
        || other->emptyLabelEstimate != this->emptyLabelEstimate) {
      return false;
    }
  } else if (!this->completeTrivial) {
    if (other->_completeHypercubeS != null) {
      swappable_reacquire(other->_completeHypercubeS);
    }

    if (this->_completeHypercubeS != null) {
      swappable_reacquire(this->_completeHypercubeS);
    }

    if (this->_completeHypercube == null) {
      throw("Impossible: this hypercube missing");
    }

    if (other->_completeHypercube == null) {
      throw("Impossible: other hypercube missing");
    }

    equals = bear_hypercube_equals(
        other->_completeHypercube,
        this->_completeHypercube);

    if (this->_completeHypercubeS != null) {
      swappable_release(this->_completeHypercubeS);
    }

    if (other->_completeHypercubeS != null) {
      swappable_release(other->_completeHypercubeS);
    }

    if (!equals) {
      return false;
    }
  }

  return true;
}

/* Make a prediction from a set of feature values. Note that this is the full
 * set of features, not just those relevant to this model.
 */
double bear_model_predict(BearModel *this, double *features) {
  bool complete;
  ulong f;
  ulong f_full;
  ulong num_features;
  double feature;
  double label;
  double *_features = null;
  BearPart *part;
  BearModel *model;

  /* Walk up through the models from this one to the empty model, starting with
   * the label prediction not yet set.
   */
  for (model = this, label = DBL_NAN;; model = model->parent) {
    /* Check if we are on the empty model.
     */
    if (model->empty) {
      /* We are up to the empty model. Add in its prediction.
       */
      add_(&label, model->emptyLabelPrediction, model);

      /* Return the label prediction.
       */
      return label;
    } else {
      /* We are not yet up to the empty model. Get a reference to the last part
       * of the assembly for this model.
       */
      part = bear_assembly_last_part(model->_assembly);

      /* Extract the number of features of this last part.
       */
      num_features = part->numFeatures;

      /* Create an array for the feature values that we will pass in for this
       * part.
       */
      new(&_features, num_features);

      /* Extract the feature values for this part, and determine whether the
       * example is complete for this last part, i.e., no feature values are
       * missing. Walk through the features of the last part.
       */
      for (f = 0, complete = true; f < num_features; f++) {
        /* Extract the feature index in the full set.
         */
        f_full = part->_features[f];

        /* Extract the value of that feature.
         */
        feature = features[f_full];

        /* Check if it is missing.
         */
        if (double_pawized_is_infinite_or_nan(feature)) {
          /* It is. Set flag and break out, since we don't need the remaining
           * feature values.
           */
          complete = false;

          break;
        }

        /* Store the feature value.
         */
        _features[f] = feature;
      }

      /* Switch on whether the example is complete for the last part of this
       * model or not.
       */
      if (complete) {
        /* Example is complete for the last part of this model. Add in the
         * prediction from the complete model, if it is nontrivial.
         */
        if (!model->completeTrivial) {
          add_(&label,
              bear_hypercube_predict(model->_completeHypercube, _features),
              null);
        }

        /* Add in the complete prediction.
         */
        add_(&label, model->completePrediction, model);
      } else {
        /* Example is incomplete for the last part of this model. Add in the
         * incomplete prediction.
         */
        add_(&label, model->incompletePrediction, model);
      }

      /* Delete our array of features.
       */
      deletev(&_features);
    }
  }

  /* There's no way that we can get here, but guard against it defensively
   * anyway.
   */
  throw("Impossible");
  return 0;
}

/* Constructor for the working object for the empty model.
 */
void bear_model_working_new_empty(
    BearModelWorking **p_this,
    ulong num_examples,
    SwappableDoubles *labels,
    SwappableUlongs *frequencies) {
  BearModelWorking *this;

  /* Create the object.
   */
  this = newo(p_this);

  /* Store information.
   */
  this->numExamples = num_examples;
  this->labels = labels;
  this->frequencies = frequencies;

  /* Create arrays that we will need.
   */
  new(&this->_parentResidualPredictions, num_examples);
  new(&this->_residuals, num_examples);
  new(&this->_labelPredictions, num_examples);
  new(&this->_labelEstimates, num_examples);
}

/* Constructor for the working object for a general model.
 */
void bear_model_working_new(BearModelWorking **p_this, BearModel *parent) {
  ulong num_examples;
  BearModelWorking *parent_w;
  BearModelWorking *this;

  /* Create the object.
   */
  this = newo(p_this);

  /* Store information required from the parent.
   */
  swappable_reacquire(parent->_wS);
  parent_w = parent->_w;
  num_examples = parent_w->numExamples;
  this->numExamples = parent_w->numExamples;
  this->frequencies = parent_w->frequencies;
  swappable_release(parent->_wS);

  /* Create arrays that we will need.
   */
  new(&this->_parentResidualPredictions, num_examples);
  new(&this->_residuals, num_examples);
}

/* Load a working object.
 */
void bear_model_working_new_load(BearModelWorking **p_this, File *f) {
  BearModelWorking *this;

  this = newo(p_this);
  load_label("[BearModelWorking]", f);
  load_value(this->numExamples, f);
  load_value(this->minResidual, f);
  load_value(this->maxResidual, f);
  load_new_values(this->_parentResidualPredictions, this->numExamples, f);
  load_new_values(this->_residuals, this->numExamples, f);
  load_new_values(this->_labelPredictions, this->numExamples, f);
  load_new_values(this->_labelEstimates, this->numExamples, f);
  load_value(this->labels, f);
  load_value(this->frequencies, f);
  load_label("[/BearModelWorking]", f);
}

/* Destructor for the working object.
 */
void bear_model_working_delete(BearModelWorking **p_this) {
  BearModelWorking *this;

  if ((this = *p_this) == null) {
    return;
  }

  deletev(&this->_examplesAreComplete);
  deletev(&this->_residuals);
  deletev(&this->_parentResidualPredictions);
  deletev(&this->_labelPredictions);
  deletev(&this->_labelEstimates);
  deleteo(p_this);
}

/* Print method for the working object.
 */
void bear_model_working_print(BearModelWorking *this, Printer *p) {
  ulong e;
  ulong *_frequencies = null;

  sprint(p, "BearModelWorking");
  rprint(p, "Number of examples: %lu", this->numExamples);
  swappable_values_access(this->frequencies, &_frequencies);
  sprint(p, "Examples");

  for (e = 0; e < this->numExamples; e++) {
    sprint(p, "Example %lu", e);
    rprint(p,
        "Parent residual prediction: %.16g",
        this->_parentResidualPredictions[e]);
    rprint(p, "Residual: %.16g", this->_residuals[e]);
    rprint(p, "Frequency: %lu", _frequencies[e]);
    fprint(p);
  }

  fprint(p);
  swappable_values_release(this->frequencies, &_frequencies);
  fprint(p);
}

/* Save the working object.
 */
void bear_model_working_save(BearModelWorking *this, File *f) {
  save_label("[BearModelWorking]", f);
  save_value(this->numExamples, f);
  save_value(this->minResidual, f);
  save_value(this->maxResidual, f);
  save_values(this->_parentResidualPredictions, this->numExamples, f);
  save_values(this->_residuals, this->numExamples, f);
  save_values(this->_labelPredictions, this->numExamples, f);
  save_values(this->_labelEstimates, this->numExamples, f);
  save_value(this->labels, f);
  save_value(this->frequencies, f);
  save_label("[/BearModelWorking]", f);
}

/* Whether the model is trivial, which is only true if both the completeness
 * model and the regular model are trivial.
 */
bool bear_model_is_trivial(BearModel *this) {
  return this->completenessTrivial && this->completeTrivial;
}

/* Compute the weight for a given loss and the empty-model loss.
 */
double bear_model_weight(double loss, double empty_loss) {
  return loss == 0
    ? MAX_WEIGHT_
    : bound(empty_loss / loss - 1, 0, MAX_WEIGHT_);
}

/* Compute the selection weight for a given weight and base-model weight.
 */
double bear_model_selection_weight(double weight, double base_weight) {
  return weight == MAX_WEIGHT_
    ? MAX_WEIGHT_
    : lower_bound(weight - base_weight, 0);
}

/* Register an ID value for a model loaded. This is only relevant for a model
 * loaded in a different program execution, and ensures that the serial primary
 * key is updated, if necessary, to not clash with any loaded models.
 */
void bear_model_register_id(ulong id) {
  lock(&lock_);
  id_ = maximum(id_, id + 1);
  unlock(&lock_);
}

/* Comparator based on assembly comparator.
 */
sbyte bear_model_compare_assembly(BearModel *a, BearModel *b, void *x) {
  return bear_assembly_compare(a->_assembly, b->_assembly, x);
}

/* Comparator based on decreasing weight, then simpler assembly.
 */
sbyte bear_model_compare_weight(BearModel *a, BearModel *b, void *x) {
  ulong num_features_a;
  ulong num_features_b;
  ulong p;
  double weight_a;
  double weight_b;
  BearAssembly *assembly_a;
  BearAssembly *assembly_b;

  weight_a = a->weight;
  weight_b = b->weight;

  if (weight_a > weight_b) {
    return -1;
  }

  if (weight_a < weight_b) {
    return +1;
  }

  assembly_a = a->_assembly;
  assembly_b = b->_assembly;

  if (assembly_a->numParts < assembly_b->numParts) {
    return -1;
  }

  if (assembly_a->numParts > assembly_b->numParts) {
    return +1;
  }

  for (p = 0, num_features_a = 0, num_features_b = 0;
      p < assembly_a->numParts;
      p++) {
    num_features_a += assembly_a->_parts[p]->numFeatures;
    num_features_b += assembly_b->_parts[p]->numFeatures;
  }

  if (num_features_a < num_features_b) {
    return -1;
  }

  if (num_features_a > num_features_b) {
    return +1;
  }

  for (p = 0; p < assembly_a->numParts; p++) {
    num_features_a = assembly_a->_parts[p]->numFeatures;
    num_features_b = assembly_b->_parts[p]->numFeatures;

    if (num_features_a > num_features_b) {
      return -1;
    }

    if (num_features_a < num_features_b) {
      return +1;
    }
  }

  return 0;
}

/* Pointer version of same.
 */
sbyte bear_model_pointer_compare_weight(
    BearModel **p_a,
    BearModel **p_b,
    void *x) {
  return bear_model_compare_weight(*p_a, *p_b, x);
}

/* Comparator based on ID.
 */
sbyte bear_model_compare_id(BearModel *a, BearModel *b, void *x) {
  ulong id_a;
  ulong id_b;

  id_a = a->id;
  id_b = b->id;

  if (id_a == 0 || id_b == 0) {
    throw("Impossible");
  }

  if (id_a < id_b) {
    return -1;
  }

  if (id_a > id_b) {
    return +1;
  }

  return 0;
}

/* Get the next part for an attachment to this model.
 */
ulong bear_model_get_next_attachment_part(
    BearModel *this,
    RandomWeightedNode *attachment_node) {
  ulong next_part;
  Attachment_ *found;
  Attachment_ *_attachment = null;
  BearModel *attachment_model;

  /* Create an Attachment_ for this model.
   */
  attachment_new_(&_attachment, attachment_node);

  /* See if it already exists.
   */
  if ((found = dynamic_set_find(this->_attachments, _attachment)) != null) {
    /* It does. Extract the next part, incrementing it in the set, and doing a
     * sanity check that it is not vetoed.
     */
    if ((next_part = found->nextPart++) == ULONG_MAX) {
      throw("Impossible");
    }

    /* Extract a reference to the attachment model.
     */
    attachment_model = attachment_node->object;

    /* If we're done with this attachment, flag that.
     */
    if (found->nextPart == attachment_model->_assembly->numParts) {
      found->nextPart = ULONG_MAX;
    }

    /* Delete the Attachment_ we created.
     */
    attachment_delete_(&_attachment);

    /* Return the next part.
     */
    return next_part;
  }

  /* It doesn't already exist. Insert the Attachment_ into the set.
   */
  dynamic_set_insert(this->_attachments, &_attachment);

  /* The next part is the first one.
   */
  return 0;
}

/* Apply any vetoes in this set of attachments for this model to a
 * RandomWeighted.
 */
void bear_model_apply_vetoes(BearModel *this, RandomWeighted *rw) {
  dynamic_set_foreach(this->_attachments, veto_, rw);
}

/* Parallel method for the empty model prediction.
 */
static void empty_prediction_(Parallel *par) {
  ulong e;
  ulong frequency;
  ulong total_frequency;
  ulong *frequencies;
  double label;
  double max_input;
  double min_input;
  double sum;
  double *labels;
  EmptyPredictionPass_ *pass;

  /* Initialize.
   */
  pass = par->x;
  labels = pass->labels;
  frequencies = pass->frequencies;

  /* Walk through the examples for this parallel thread.
   */
  for (e = par->start, sum = 0, total_frequency = 0, min_input = DBL_MAX,
      max_input = -DBL_MAX;
      e < par->end;
      e++) {
    /* Sanity checks.
     */
    if (double_pawized_is_infinite_or_nan(label = labels[e])) {
      throw("Invalid infinite or NaN label value %.16g for example %lu",
          label,
          e);
    }

    if ((frequency = frequencies[e]) == 0) {
      throw("Invalid zero frequency for example %lu", e);
    }

    /* Update the sum and total frequency.
     */
    sum += frequency * label;
    total_frequency += frequency;

    /* Update the min and max input values.
     */
    min_input = minimum(min_input, label);
    max_input = maximum(max_input, label);
  }

  /* Update the results in the pass object, under a lock.
   */
  lock(&pass->lock);
  pass->sum += sum;
  pass->totalFrequency += total_frequency;
  pass->minLabel = minimum(pass->minLabel, min_input);
  pass->maxLabel = maximum(pass->maxLabel, max_input);
  unlock(&pass->lock);
}

/* Parallel method for empty model residuals.
 */
static void empty_residuals_(Parallel *par) {
  ulong e;
  double empty_label_estimate;
  double empty_label_prediction;
  double max_residual;
  double min_residual;
  double residual;
  double *label_estimates;
  double *label_predictions;
  double *labels;
  double *parent_residual_predictions;
  double *residuals;
  EmptyResidualsPass_ *pass;

  /* Initialize.
   */
  pass = par->x;
  empty_label_prediction = pass->emptyLabelPrediction;
  empty_label_estimate = pass->emptyLabelEstimate;
  parent_residual_predictions = pass->parentResidualPredictions;
  label_predictions = pass->labelPredictions;
  label_estimates = pass->labelEstimates;
  residuals = pass->residuals;
  labels = pass->labels;

  /* Walk through the examples for this parallel thread.
   */
  for (e = par->start, min_residual = DBL_MAX, max_residual = -DBL_MAX;
      e < par->end;
      e++) {
    /* Store the "parent residual prediction" and the label prediction as just
     * the constant empty label prediction computed above.
     */
    parent_residual_predictions[e] = label_predictions[e]
      = empty_label_prediction;

    /* Store the label estimate, which may differ from the label prediction,
     * and is used in the computation of loss.
     */
    label_estimates[e] = empty_label_estimate;

    /* Compute the residual based on the parent residual prediction.
     */
    residuals[e] = residual = labels[e] - parent_residual_predictions[e];

    /* Update the minimum and maximum residual.
     */
    min_residual = minimum(min_residual, residual);
    max_residual = maximum(max_residual, residual);
  }

  /* Update the min and max residual in the pass object, under a lock.
   */
  lock(&pass->lock);
  pass->minResidual = minimum(pass->minResidual, min_residual);
  pass->maxResidual = maximum(pass->maxResidual, max_residual);
  unlock(&pass->lock);
}

/* Parallel method for the empty model loss.
 */
static void empty_loss_(Parallel *par) {
  ulong e;
  ulong *frequencies;
  double empty_label_estimate;
  double sum;
  double *labels;
  void *loss_function_params;
  LossFunction *loss_function;
  EmptyLossPass_ *pass;

  /* Initialize.
   */
  pass = par->x;
  labels = pass->labels;
  frequencies = pass->frequencies;
  loss_function = pass->lossFunction;
  loss_function_params = pass->lossFunctionParams;
  empty_label_estimate = pass->emptyLabelEstimate;

  /* Walk through the examples for this parallel thread.
   */
  for (e = par->start, sum = 0; e < par->end; e++) {
    /* Update the sum.
     */
    sum += frequencies[e] * loss_function->computeLoss(
        labels[e],
        empty_label_estimate,
        loss_function_params);
  }

  /* Update the min and max residual in the pass object, under a lock.
   */
  lock(&pass->lock);
  pass->sum += sum;
  unlock(&pass->lock);
}

/* Parallel method for finding incomplete examples.
 */
static void find_incomplete_(Parallel *par) {
  bool example_is_complete;
  bool *examples_are_complete;
  bool *features_have_missing_values;
  ulong e;
  ulong f;
  ulong num_complete_examples;
  ulong num_features;
  double feature;
  double **features;
  FindIncompletePass_ *pass;

  /* Initialize.
   */
  pass = par->x;
  num_features = pass->numFeatures;
  features_have_missing_values = pass->featuresHaveMissingValues;
  features = pass->features;
  examples_are_complete = pass->examplesAreComplete;

  /* Walk through the examples for this parallel thread.
   */
  for (e = par->start, num_complete_examples = 0; e < par->end; e++) {
    /* Walk through the features, checking if this example is complete.
     */
    for (f = 0, example_is_complete = true; f < num_features; f++) {
      /* Check if this feature has missing values.
       */
      if (features_have_missing_values[f]) {
        /* It does. Get the value of this feature for this example.
         */
        feature = features[f][e];

        /* Check if it is missing (infinite or NaN when pawized).
         */
        if (double_pawized_is_infinite_or_nan(feature)) {
          /* This value is missing. This example is incomplete.
           */
          example_is_complete = false;

          break;
        }
      }
    }

    /* Store the completeness flag for this example.
     */
    examples_are_complete[e] = example_is_complete;

    /* If the example is complete, increment the number of complete examples.
     */
    if (example_is_complete) {
      num_complete_examples++;
    }
  }

  /* Update the number of complete examples in the pass object, under a lock.
   */
  lock(&pass->lock);
  pass->numCompleteExamples += num_complete_examples;
  unlock(&pass->lock);
}

/* Parallel method for filling in the complete examples for mixed models.
 */
static void mixed_complete_(Parallel *par) {
  ulong c;
  ulong e;
  ulong f;
  ulong frequency;
  ulong num_features;
  ulong *complete_frequencies;
  ulong *complete_to_all;
  ulong *frequencies;
  double feature;
  double input;
  double *completeness_residuals;
  double **complete_values;
  double **features;
  MixedCompletePass_ *pass;

  /* Initialize.
   */
  pass = par->x;
  complete_to_all = pass->completeToAll;
  num_features = pass->numFeatures;
  features = pass->features;
  frequencies = pass->frequencies;
  completeness_residuals = pass->completenessResiduals;
  complete_values = pass->completeValues;
  complete_frequencies = pass->completeFrequencies;

  /* Walk through the complete examples for this parallel thread.
   */
  for (c = par->start; c < par->end; c++) {
    /* Look up the full example index.
     */
    e = complete_to_all[c];

    /* Walk through the features.
     */
    for (f = 0; f < num_features; f++) {
      /* Extract the feature value.
       */
      feature = features[f][e];

      /* Store it.
       */
      complete_values[c][f] = feature;
    }

    /* Extract the input value as the residual of the completeness model for
     * this example.
     */
    input = completeness_residuals[e];

    /* Store it in our matrix.
     */
    complete_values[c][num_features] = input;

    /* Extract the frequency for this example.
     */
    frequency = frequencies[e];

    /* Store it in our array for complete examples.
     */
    complete_frequencies[c] = frequency;
  }
}

/* Parallel method for computing the predictions and residuals for mixed
 * models.
 */
static void mixed_predictions_(Parallel *par) {
  bool *examples_are_complete;
  ulong c;
  ulong e;
  ulong num_features;
  ulong *all_to_complete;
  double input;
  double max_residual;
  double min_residual;
  double completeness_residual_prediction;
  double residual;
  double *completeness_residuals;
  double *parent_residual_predictions;
  double *residuals;
  double **complete_values;
  MixedPredictionsPass_ *pass;
  BearHypercube *complete_hypercube;
  BearModel *this;

  /* Initialize.
   */
  pass = par->x;
  this = pass->this;
  examples_are_complete = pass->examplesAreComplete;
  complete_values = pass->completeValues;
  all_to_complete = pass->allToComplete;
  complete_hypercube = pass->completeHypercube;
  parent_residual_predictions = pass->parentResidualPredictions;
  completeness_residuals = pass->completenessResiduals;
  num_features = pass->numFeatures;
  residuals = pass->residuals;

  /* Walk through the examples for this parallel thread. We keep track of the
   * minimum and maximum residual.
   */
  for (e = par->start, min_residual = DBL_MAX, max_residual = -DBL_MAX;
      e < par->end;
      e++) {
    /* Check if the complete model is nontrivial and this example is complete.
     */
    if (!this->completeTrivial && examples_are_complete[e]) {
      /* They are. Look up the complete example index.
       */
      c = all_to_complete[e];

      /* We use the predictor of the complete BearHypercube to predict the
       * residual of completeness model. Note that the first num_features
       * values in complete_values[c] are just the feature values we need to
       * specify.
       */
      completeness_residual_prediction
        = bear_hypercube_predict(complete_hypercube, complete_values[c]);

      /* Add this in to the parent residual prediction.
       */
      parent_residual_predictions[e] += completeness_residual_prediction;

      /* Extract the completeness model's residual as the input value.
       */
      input = completeness_residuals[e];

      /* Sanity check, since this must be the final value of this array.
       */
      if (input != complete_values[c][num_features]) {
        throw("Impossible");
      }

      /* Compute the residual.
       */
      residual = input - completeness_residual_prediction;
    } else {
      /* Either the complete model is trivial or this example is incomplete.
       * Copy across the residual from the completeness model.
       */
      residual = completeness_residuals[e];
    }

    /* Store the residual.
     */
    residuals[e] = residual;

    /* Update the min and max residual.
     */
    min_residual = minimum(min_residual, residual);
    max_residual = maximum(max_residual, residual);
  }

  /* Update the min and max residual, under the lock.
   */
  lock(&pass->lock);
  pass->minResidual = minimum(pass->minResidual, min_residual);
  pass->maxResidual = maximum(pass->maxResidual, max_residual);
  unlock(&pass->lock);
}

/* Parallel method for populating the values matrix for a pure model.
 */
static void pure_values_(Parallel *par) {
  ulong e;
  ulong f;
  ulong num_features;
  double feature;
  double input;
  double *parent_residuals;
  double **features;
  double **values;
  PureValuesPass_ *pass;

  /* Initialize.
   */
  pass = par->x;
  num_features = pass->numFeatures;
  features = pass->features;
  values = pass->values;
  parent_residuals = pass->parentResiduals;

  /* Walk through the examples for this parallel thread.
   */
  for (e = par->start; e < par->end; e++) {
    /* Walk through the features of this part.
     */
    for (f = 0; f < num_features; f++) {
      /* Extract the feature value.
       */
      feature = features[f][e];

      /* Sanity check.
       */
      if (double_pawized_is_infinite_or_nan(feature)) {
        throw("Impossible");
      }

      /* Store it in our matrix.
       */
      values[e][f] = feature;
    }

    /* Extract the input value from the parent model.
     */
    input = parent_residuals[e];

    /* Store it in our matrix.
     */
    values[e][num_features] = input;
  }
}

/* Parallel method for computing predictions and residuals for a pure model.
 */
static void pure_predictions_(Parallel *par) {
  ulong e;
  ulong num_features;
  double input;
  double min_residual;
  double max_residual;
  double parent_residual_prediction;
  double residual;
  double *parent_residual_predictions;
  double *parent_residuals;
  double *residuals;
  double **values;
  PurePredictionsPass_ *pass;
  BearHypercube *complete_hypercube;

  /* Initialize.
   */
  pass = par->x;
  complete_hypercube = pass->completeHypercube;
  values = pass->values;
  parent_residual_predictions = pass->parentResidualPredictions;
  parent_residuals = pass->parentResiduals;
  num_features = pass->numFeatures;
  residuals = pass->residuals;

  /* Walk through the examples for this thread. We keep track of the minimum
   * and maximum residual as we go.
   */
  for (e = par->start, min_residual = DBL_MAX, max_residual = -DBL_MAX;
      e < par->end;
      e++) {
    /* Use the predictor of the BearHypercube of this model to predict the
     * residual of the parent model. Note that the first num_features values in
     * values[e] are just the feature values we need to specify.
     */
    parent_residual_prediction
      = bear_hypercube_predict(complete_hypercube, values[e]);

    /* Store the prediction.
     */
    parent_residual_predictions[e] = parent_residual_prediction;

    /* Extract the parent model's residual as the input value.
     */
    input = parent_residuals[e];

    /* Sanity check, since this must be the final value of this array.
     */
    if (input != values[e][num_features]) {
      throw("Impossible");
    }

    /* Compute the residual.
     */
    residual = input - parent_residual_prediction;

    /* Store it.
     */
    residuals[e] = residual;

    /* Update min and max.
     */
    min_residual = minimum(min_residual, residual);
    max_residual = maximum(max_residual, residual);
  }

  /* Update the min and max residual, under the lock.
   */
  lock(&pass->lock);
  pass->minResidual = minimum(pass->minResidual, min_residual);
  pass->maxResidual = maximum(pass->maxResidual, max_residual);
  unlock(&pass->lock);
}

/* Parallel method for computing the label predictions.
 */
static void label_predictions_(Parallel *par) {
  ulong e;
  double label_prediction;
  double *label_estimates;
  double *label_predictions;
  void *loss_function_params;
  LabelPredictionsPass_ *pass;
  LossFunction *loss_function;
  BearModel *model;
  BearModel *this;
#ifdef VERIFY_LABEL_PREDICTIONS_
  char *_s = null;
  ulong f;
  ulong num_full_features;
  double label_prediction_alt;
  double *_features = null;
  double **full_features;
#endif

  /* Initialize.
   */
  pass = par->x;
  this = pass->this;
  label_predictions = pass->labelPredictions;
  label_estimates = pass->labelEstimates;
  loss_function = pass->lossFunction;
  loss_function_params = pass->lossFunctionParams;

#ifdef VERIFY_LABEL_PREDICTIONS_

  /* We will verify the label prediction for each example against what we get
   * with a full call to bear_model_predict(). This is of course a lot slower,
   * and unnecessary, so is only done for debugging.
   *
   * Extract from the pass object.
   */
  num_full_features = pass->numFullFeatures;
  full_features = pass->fullFeatures;

  /* Create an array for the features for each example.
   */
  new(&_features, num_full_features);

#endif

  /* Walk through the examples for this parallel thread.
   */
  for (e = par->start; e < par->end; e++) {
    /* Walk up from this model. We initialize the label prediction as not set.
     */
    for (model = this, label_prediction = DBL_NAN;; model = model->parent) {
      /* Sanity check.
       */
      if (model == null) {
        throw("Impossible");
      }

      /* Add the prediction for this model to the label. This includes
       * contributions from both the completeness model and the complete model,
       * where each exists.
       */
      add_(&label_prediction, model->_w->_parentResidualPredictions[e], model);

      /* If this model is the empty model, break out.
       */
      if (model->empty) {
        break;
      }
    }

    /* Store the label prediction.
     */
    label_predictions[e] = label_prediction;

    /* Ask the loss function to compute the label estimate from the label
     * prediction.
     */
    label_estimates[e] = loss_function->computeLabelEstimate(
        label_prediction,
        loss_function_params);

#ifdef VERIFY_LABEL_PREDICTIONS_

    /* Populate the array of features for this example.
     */
    for (f = 0; f < num_full_features; f++) {
      _features[f] = full_features[f][e];
    }

    /* Call the full prediction method.
     */
    label_prediction_alt = bear_model_predict(this, _features);

    /* Compare to our direct computation.
     */
    if (fabs(label_prediction_alt - label_prediction) > 1e-14) {
      aprint(&_s, 1, bear_assembly, this->_assembly);

      throw("Impossible: %.16g v %.16g in:\n%s",
          label_prediction_alt,
          label_prediction,
          _s);
    }

#endif

  }

#ifdef VERIFY_LABEL_PREDICTIONS_

  /* Clean up.
   */
  deletev(&_features);

#endif

}

/* Parallel method for computing the loss.
 */
static void loss_(Parallel *par) {
  ulong e;
  ulong frequency;
  ulong total_frequency;
  ulong *frequencies;
  double sum;
  double *labels;
  double *label_estimates;
  void *loss_function_params;
  LossPass_ *pass;
  LossFunction *loss_function;

  /* Initialize.
   */
  pass = par->x;
  frequencies = pass->frequencies;
  labels = pass->labels;
  label_estimates = pass->labelEstimates;
  loss_function = pass->lossFunction;
  loss_function_params = pass->lossFunctionParams;

  /* Walk through the examples for this parallel thread.
   */
  for (e = par->start, sum = 0, total_frequency = 0; e < par->end; e++) {
    /* Extract the frequency and add it to the total.
     */
    total_frequency += (frequency = frequencies[e]);

    /* Compute the loss and add it to the total.
     */
    sum += frequency * loss_function->computeLoss(
        labels[e],
        label_estimates[e],
        loss_function_params);
  }

  /* Update the totals, under the lock.
   */
  lock(&pass->lock);
  pass->sum += sum;
  pass->totalFrequency += total_frequency;
  unlock(&pass->lock);
}

/* Parallel method for computing the completeness predictions and populating
 * the completeness feature.
 */
static void completeness_predictions_(Parallel *par) {
  bool *examples_are_complete;
  ulong e;
  ulong complete_frequency;
  ulong frequency;
  ulong incomplete_frequency;
  ulong *frequencies;
  double complete_sum;
  double incomplete_sum;
  double parent_residual;
  double *parent_residuals;
  double **completeness_values;
  CompletenessPredictionsPass_ *pass;

  /* Initialize.
   */
  pass = par->x;
  examples_are_complete = pass->examplesAreComplete;
  completeness_values = pass->completenessValues;
  parent_residuals = pass->parentResiduals;
  frequencies = pass->frequencies;

  /* Walk through the examples for this parallel thread.
   */
  for (e = par->start,
      incomplete_sum = 0,
      complete_sum = 0,
      incomplete_frequency = 0,
      complete_frequency = 0;
      e < par->end;
      e++) {
    parent_residual = parent_residuals[e];
    frequency = frequencies[e];

    if (examples_are_complete[e]) {
      complete_sum += frequency * parent_residual;
      complete_frequency += frequency;
      completeness_values[e][0] = 1;
    } else {
      incomplete_sum += frequency * parent_residual;
      incomplete_frequency += frequency;
      completeness_values[e][0] = 0;
    }

    completeness_values[e][1] = parent_residual;
  }

  /* Update the totals, under the lock.
   */
  lock(&pass->lock);
  pass->incompleteSum += incomplete_sum;
  pass->completeSum += complete_sum;
  pass->incompleteFrequency += incomplete_frequency;
  pass->completeFrequency += complete_frequency;
  unlock(&pass->lock);

}

/* Parallel method for computing the completeness model residuals.
 */
static void completeness_residuals_(Parallel *par) {
  bool *examples_are_complete;
  ulong e;
  double complete_prediction;
  double incomplete_prediction;
  double input;
  double prediction;
  double residual;
  double *parent_residuals;
  double *parent_residual_predictions;
  double *residuals;
  CompletenessResidualsPass_ *pass;

  /* Initialize.
   */
  pass = par->x;
  incomplete_prediction = pass->incompletePrediction;
  complete_prediction = pass->completePrediction;
  examples_are_complete = pass->examplesAreComplete;
  parent_residuals = pass->parentResiduals;
  parent_residual_predictions = pass->parentResidualPredictions;
  residuals = pass->residuals;

  /* Walk through the examples for this parallel thread.
   */
  for (e = par->start; e < par->end; e++) {
    /* Select and store the prediction.
     */
    prediction = parent_residual_predictions[e] = examples_are_complete[e]
      ? complete_prediction
      : incomplete_prediction;

    /* Extract the parent model's residual as the input value.
     */
    input = parent_residuals[e];

    /* Compute the residual.
     */
    residual = input - prediction;

    /* Store it.
     */
    residuals[e] = residual;
  }
}

/* Add a new prediction to the current label prediction from child models.
 */
static void add_(double *label, double prediction, BearModel *model) {
  /* Check if the prediction is zero.
   */
  if (prediction == 0) {
    /* Prediction is zero. Check if the label has been set yet.
     */
    if (double_is_nan(*label)) {
      /* Label is not set. Set it to zero.
       */
      *label = 0;
    } else {
      /* Label is set. Nothing more to do, since adding zero does nothing, and
       * we previously checked its bounds, if applicable.
       */
      return;
    }
  } else {
    /* Prediction is nonzero. Check if the label has been set yet.
     */
    if (double_is_nan(*label)) {
      /* Label is not set. Set it to the prediction.
       */
      *label = prediction;
    } else {
      /* Label is set. Add in the prediction.
       */
      *label += prediction;
    }
  }

  /* Check if the model has been supplied.
   */
  if (model != null) {
    /* It has. Bound this new label prediction to the bounds of the data input
     * to this model.
     */
    *label = bound(*label, model->minInput, model->maxInput);
  }
}

/* Constructor for an Attachment_;
 */
static void attachment_new_(Attachment_ **p_this, RandomWeightedNode *node) {
  Attachment_ *this;

  this = newo(p_this);
  this->node = node;
}

/* Destructor for an Attachment_;
 */
static void attachment_delete_(Attachment_ **p_this) {
  deleteo(p_this);
}

/* Comparator for an Attachment_;
 */
static sbyte attachment_compare_(Attachment_ *a, Attachment_ *b, void *x) {
  return pointer_compare(a->node, b->node, null);
}

/* Print for an Attachment_;
 */
static void attachment_print_(Attachment_ *this, Printer *p) {
  BearModel *model;

  sprint(p, "Attachment_");
  model = this->node->object;
  bear_model_print(model, p);
  rprint(p, "Next part: %lu", this->nextPart);
  fprint(p);
}

/* Check an Attachment_ for veto status, and, if so, add a veto node to the
 * RandomWeighted.
 */
static void veto_(Attachment_ *this, RandomWeighted *rw) {
  if (this->nextPart == ULONG_MAX) {
    random_weighted_veto(rw, this->node);
  }
}
