/* Object class for a Bear model. Internally, each BearModel object only owns
 * data for the last part of the model, but also has a reference to its parent
 * model, which is sufficient to rebuild the entire model.
 *
 * The empty model is the root model of every model for that label.
 *
 * Each model keeps track of its loss, as determined by the supplied function,
 * and from this computes a (prediction) weight, which is the empty model loss
 * divided by the model's loss, minus one, but lower-bounded to zero. (The
 * selection weight of a model is the excess of the prediction weight over that
 * of its based mode.)
 *
 * The model also maintains an array of RandomWeightedNode objects to veto,
 * used by the owning Bear class.
 *
 * The storage model for the class is divided into three parts:
 *
 * (1) The BearModel class always keeps in memory simple metadata about the
 *     model, including its assembly. This metadata is used by Bear in
 *     determining which new models to build. The Bear class never swaps out
 *     any BearModel while it is training in its constructor.
 *
 * (2) Within the BearModel, the constructed BearHypercube is kept as a
 *     Swappable object. Since it is needed for predictions, it is extracted
 *     back out of the Swappable before the BearModel is saved.
 *
 * (3) Also within the BearModel is a BearModelWorking object which is also
 *     kept as a Swappable object. This object keeps all of the data required
 *     during Bear's training, such as the per-example data. Since this data is
 *     not needed for predictions, this object is not extracted back out prior
 *     to saving the BearModel. Consequently, since this data is never saved
 *     and reloaded in a separate program execution, pointer values can be
 *     saved directly when it is swapped out.
 */
#ifndef BEAR_MODEL_H
#define BEAR_MODEL_H

#ifdef __cplusplus
extern "C" {
#endif

#include "BearAssembly.h"
#include "BearHypercube.h"
#include "RandomWeighted.h"

struct BearModel;

typedef struct {
  bool *_examplesAreComplete;
  ulong numExamples;
  SwappableUlongs *frequencies;
  double minResidual;
  double maxResidual;
  double *_labelEstimates;
  double *_labelPredictions;
  double *_parentResidualPredictions;
  double *_residuals;
  SwappableDoubles *labels;
} BearModelWorking;

typedef struct BearModel {
  bool completeTrivial;
  bool completenessTrivial;
  bool empty;
  ulong id;
  ulong parentId;
  double completePrediction;
  double incompletePrediction;
  double emptyLabelEstimate;
  double emptyLabelPrediction;
  double emptyLoss;
  double loss;
  double maxInput;
  double minInput;
  double selectionWeight;
  double weight;
  LossFunction *lossFunction;
  BearAssembly *_assembly;
  DynamicSet *_attachments;
  BearHypercube *_completeHypercube;
  Swappable *_completeHypercubeS;
  BearModelWorking *_w;
  Swappable *_wS;
  struct BearModel *parent;
  void *_lossFunctionParams;
} BearModel;

/* Constructor for an empty model, which is the root model of all models for a
 * given label.
 */
void bear_model_new_empty(
    BearModel **p_this,
    ulong num_examples,
    SwappableDoubles *labels,
    SwappableUlongs *frequencies,
    LossFunction *loss_function);

/* Constructor for a general model. The assembly of the parent model must be
 * consistent with the assembly specified for this model. The base model is the
 * previously-existing model that this new construction is based on; for
 * gluing, it is just the parent, but for melting it is the "sibling" without
 * the extra attachment part melted in. The 'features' swappable should be that
 * for all features; this constructor will select out the ones it needs.
 */
void bear_model_new(
    BearModel **p_this,
    BearAssembly *assembly,
    BearModel *parent,
    BearModel *base,
    SwappableDoubles **features,
    bool *features_have_missing_values,
    Decider *decider);

/* Load.
 */
void bear_model_new_load(BearModel **p_this, File *f);

/* Destructor.
 */
void bear_model_delete(BearModel **p_this);

/* Print.
 */
void bear_model_print(BearModel *this, Printer *p);

/* Save.
 */
void bear_model_save(BearModel *this, File *f);

/* Equals.
 */
bool bear_model_equals(BearModel *this, BearModel *other);

/* Make a prediction from a set of feature values. Note that this is the full
 * set of features, not just those relevant to this model.
 */
double bear_model_predict(BearModel *this, double *features);

/* Constructor for the working object for the empty model.
 */
void bear_model_working_new_empty(
    BearModelWorking **p_this,
    ulong num_examples,
    SwappableDoubles *labels,
    SwappableUlongs *frequencies);

/* Constructor for the working object for a model.
 */
void bear_model_working_new(BearModelWorking **p_this, BearModel *parent);

/* Load a working object.
 */
void bear_model_working_new_load(BearModelWorking **p_this, File *f);

/* Destructor for the working object.
 */
void bear_model_working_delete(BearModelWorking **p_this);

/* Print method for the working object.
 */
void bear_model_working_print(BearModelWorking *this, Printer *p);

/* Save the working object.
 */
void bear_model_working_save(BearModelWorking *this, File *f);

/* Whether the model is trivial, which is only true if both the completeness
 * model and the regular model are trivial.
 */
bool bear_model_is_trivial(BearModel *this);

/* Compute the weight for a given loss and the empty-model loss.
 */
double bear_model_weight(double loss, double empty_loss);

/* Compute the selection weight for a given weight and base-model weight.
 */
double bear_model_selection_weight(double weight, double base_weight);

/* Get the next part for an attachment to this model.
 */
ulong bear_model_get_next_attachment_part(
    BearModel *this,
    RandomWeightedNode *attachment_node);

/* Apply any vetoes in this set of attachments for this model to a
 * RandomWeighted.
 */
void bear_model_apply_vetoes(BearModel *this, RandomWeighted *rw);

/* Add a RandomWeightedNode to the list of vetoes for this model.
 */
void bear_model_add_veto(BearModel *this, RandomWeightedNode *node);

/* Apply the set of vetoes in this model to a RandomWeighted.
 */
void bear_model_apply_vetoes(BearModel *this, RandomWeighted *rw);

/* Register an ID value for a model loaded. This is only relevant for a model
 * loaded in a different program execution, and ensures that the serial primary
 * key is updated, if necessary, to not clash with any loaded models.
 */
void bear_model_register_id(ulong id);

/* Comparator based on assembly comparator.
 */
sbyte bear_model_compare_assembly(BearModel *a, BearModel *b, void *x);

/* Comparator based on decreasing weight, then simpler assembly.
 */
sbyte bear_model_compare_weight(BearModel *a, BearModel *b, void *x);

/* Pointer version of same.
 */
sbyte bear_model_pointer_compare_weight(
    BearModel **p_a,
    BearModel **p_b,
    void *x);

/* Comparator based on ID.
 */
sbyte bear_model_compare_id(BearModel *a, BearModel *b, void *x);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _H */
