/* Simple object class for handling the case of multiple decisions, where we
 * want a single p-value of confidence for the entire process, but we don't
 * know how ahead of time how many decisions we are going to need to make.
 *
 * This class solves that problem by forcing you to execute your program twice,
 * with approximately the same amount of time allocated to each run.
 *
 * - On the first run, the number of decisions SO FAR will be used to determine
 *   the p-value to be used for each decision. This provides too much p-value
 *   for all but the last call, and in general especially so for the earliest
 *   calls, but at least provides some level of multiple-decision sanity, for
 *   this test run of the program.
 *
 * - On the second run, the number of decisions for the entire first run will
 *   be used to determine the p-value for EVERY decision. This is now a
 *   conservative estimate, as we want. It is possible that fewer decisions
 *   will be actually be made on this second run, since each now has a higher
 *   threshold for significance (which could prevent downstream decisions from
 *   even being considered), but we accept that conservative bias, and in any
 *   case the total number of decisions should still be of the right order of
 *   magnitude, which is what really matters.
 *
 * The destructor throws if two iterations have not been ended.
 */
#ifndef DECIDER_H
#define DECIDER_H

#ifdef __cplusplus
extern "C" {
#endif

#include "jpc.h"

typedef struct {
  byte iteration;
  ulong numDecisions;
  double pValue;
  Lock lock;
} Decider;

/* Constructor. The default p-value of 0.05 is used. If another p-value is ever
 * needed, a new alternative constructor could easily be made.
 */
void decider_new(Decider **p_this);

/* Destructor.
 */
void decider_delete(Decider **p_this);

/* Get the p-value to be used for each of the new decisions being added, on
 * either iteration. (The number of new decisions is not used on the second
 * iteration, but it is still sanity-checked.)
 */
double decider_get_p_value(Decider *this, ulong num_new_decisions);

/* End an iteration.
 */
void decider_end_iteration(Decider *this);

/* End all iterations that haven't been completed. Used for aborting, so that
 * the destructor does not throw.
 */
void decider_end_iterations(Decider *this);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _H */
