#include "Dictionary.h"
#include <limits.h>

static void destroy_(KeyValuePair *pair, KeyValueInfo *info);

static void print_(KeyValuePairPrintInfo *kvppi, Printer *p);

static sbyte compare_(KeyValuePair *a, KeyValuePair *b, KeyValueInfo *info);

static void merge_(KeyValuePair *acc, KeyValuePair *add, KeyValueInfo *info);

static void check_not_finalized_(Dictionary *this, char *action);

static void check_finalized_(Dictionary *this, char *action);

static void pair_new_(DictionaryPair **p_this,
    char *key,
    void *object,
    ulong serial);

static void pair_delete_(DictionaryPair **p_this);

static sbyte chrono_pointer_compare_(
    DictionaryPair **p_a,
    DictionaryPair **p_b,
    void *x);


void dictionary_newI(
    Dictionary **p_this,
    void destroy(void **p_this),
    void print(void *this, Printer *)) {
  ulong maxes[] = {ULONG_MAX, ULONG_MAX};
  KeyValueInfo *info;
  Dictionary *this;

  if (destroy == null) {
    throw("Destructor cannot be null");
  }

  this = newo(p_this);

  this->destroy = destroy;
  this->print = print;

  key_value_info_new(&this->_info, 1, maxes, 2, maxes);
  info = this->_info;
  info->x = this;
  info->destroy = destroy_;
  info->print = print_;
  info->merge = merge_;
  info->compare = compare_;

  key_value_collection_new(&this->_collection, this->_info);
}

void dictionary_delete(Dictionary **p_this) {
  Dictionary *this;

  if ((this = *p_this) == null) {
    return;
  }

  key_value_array_delete(&this->_array);
  key_value_collection_delete(&this->_collection);

  key_value_info_delete(&this->_info);
  deleteo(p_this);
}

void dictionary_print(Dictionary *this, Printer *p) {
  DictionaryPairs *_pairs = null;

  sprint(p, "Dictionary");

  if (dictionary_is_finalized(this)) {
    dictionary_pairs_new(&_pairs, this, false);
    dictionary_pairs_print(_pairs, p);
    dictionary_pairs_delete(&_pairs);
  } else {
    key_value_collection_print(this->_collection, p);
  }

  fprint(p);
}

void dictionary_donate(Dictionary *this, char **p_key, void *p_object) {
  void **p_o;
  KeyValuePair *_pair = null;

  check_not_finalized_(this, "donate to");

  p_o = p_object;

  key_value_pair_new(&_pair, this->_info);
  key_value_pair_donate_to_key(_pair, this->_info, 0, p_key);
  key_value_pair_donate_to_value(_pair, this->_info, 0, p_o);
  key_value_pair_set_value(_pair, this->_info, 1, this->serial++);
  key_value_collection_donate(this->_collection, _pair);
  key_value_pair_delete(&_pair, this->_info);
}

void dictionary_finalize(Dictionary *this) {
  check_not_finalized_(this, "re-finalize");

  key_value_array_new_from_collection(
      &this->_array,
      &this->_collection);
}

bool dictionary_is_finalized(Dictionary *this) {
  return this->_array != null;
}

void dictionary_reopen(Dictionary *this) {
  check_finalized_(this, "reopen");

  key_value_collection_new_from_array(
      &this->_collection,
      &this->_array);
}

void *dictionary_find(Dictionary *this, char *key) {
  KeyValuePair *found;
  KeyValuePair *_pair = null;

  check_finalized_(this, "find in");

  key_value_pair_new(&_pair, this->_info);
  key_value_pair_reference_to_key(_pair, this->_info, 0, key);
  found = key_value_array_find(this->_array, _pair);
  key_value_pair_reference_to_key(_pair, this->_info, 0, null);
  key_value_pair_delete(&_pair, this->_info);

  if (found == null) {
    return null;
  }

  return key_value_pair_reference_from_value(found, this->_info, 0);
}

ulong dictionary_cardinality(Dictionary *this) {
  check_finalized_(this, "get cardinality of");

  return key_value_array_cardinality(this->_array);
}

void dictionary_pairs_new(
    DictionaryPairs **p_this,
    Dictionary *dict,
    bool chronological) {
  char *key;
  ulong u;
  ulong serial;
  KeyValuePair *pair;
  void *object;
  DictionaryPairs *this;

  check_finalized_(dict, "get pairs from");

  this = newo(p_this);

  this->cardinality = dictionary_cardinality(dict);
  new(&this->_pairs, this->cardinality);
  this->print = dict->print;

  for (u = 0; u < this->cardinality; u++) {
    pair = key_value_array_get(dict->_array, u);
    key = key_value_pair_reference_from_key(pair, dict->_info, 0);
    object = key_value_pair_reference_from_value(pair, dict->_info, 0);
    serial = key_value_pair_get_value(pair, dict->_info, 1);
    pair_new_(&this->_pairs[u], key, object, serial);
  }

  if (chronological) {
    qsortx(this->_pairs,
        this->cardinality,
        chrono_pointer_compare_,
        false,
        null);
  }
}

void dictionary_pairs_delete(DictionaryPairs **p_this) {
  DictionaryPairs *this;

  if ((this = *p_this) == null) {
    return;
  }

  delete_(&this->_pairs, this->cardinality, pair);

  deleteo(p_this);
}

void dictionary_pairs_print(DictionaryPairs *this, Printer *p) {
  ulong u;
  DictionaryPair *pair;

  sprint(p, "DictionaryPairs");
  rprint(p, "Cardinality: %lu", this->cardinality);

  if (this->cardinality > 0) {
    sprint(p, "Pairs");

    for (u = 0; u < this->cardinality; u++) {
      pair = this->_pairs[u];

      sprint(p, "Key '%s'", pair->key);

      if (this->print == null) {
        rprint(p, "Value: object at %p", pair->object);
      } else {
        sprint(p, "Value");
        this->print(pair->object, p);
        fprint(p);
      }

      rprint(p, "Serial: %lu", pair->serial);
      fprint(p);
    }

    fprint(p);
  }

  fprint(p);
}


static void destroy_(KeyValuePair *pair, KeyValueInfo *info) {
  char *_key = null;
  void *_object = null;
  Dictionary *this;

  if (!key_value_pair_try_donate_from_key(pair, info, 0, &_key)) {
    return;
  }

  string_delete(&_key);

  key_value_pair_donate_from_value(pair, info, 0, &_object);
  this = info->x;
  this->destroy(&_object);
}

static void print_(KeyValuePairPrintInfo *kvppi, Printer *p) {
  char *key;
  ulong serial;
  void *object;
  Dictionary *this;

  sprint(p, "DictionaryEntry");
  key = key_value_pair_reference_from_key(kvppi->this, kvppi->info, 0);
  rprint(p, "Key: '%s'", key);

  this = kvppi->info->x;
  object = key_value_pair_reference_from_value(kvppi->this, kvppi->info, 0);

  if (this->print == null) {
    rprint(p, "Value: object at %p", object);
  } else {
    sprint(p, "Value");
    this->print(object, p);
    fprint(p);
  }

  serial = key_value_pair_get_value(kvppi->this, kvppi->info, 1);
  rprint(p, "Serial: %lu", serial);
  fprint(p);
}

static sbyte compare_(KeyValuePair *a, KeyValuePair *b, KeyValueInfo *info) {
  char *key_a;
  char *key_b;

  key_a = key_value_pair_reference_from_key(a, info, 0);
  key_b = key_value_pair_reference_from_key(b, info, 0);

  return string_compare(key_a, key_b, null);
}

static void merge_(KeyValuePair *acc, KeyValuePair *add, KeyValueInfo *info) {
  char *_key_add = null;
  void *_object = null;
  ulong serial_acc;
  ulong serial_add;
  Dictionary *this;

  key_value_pair_donate_from_key(add, info, 0, &_key_add);
  string_delete(&_key_add);

  serial_acc = key_value_pair_get_value(acc, info, 1);
  serial_add = key_value_pair_get_value(add, info, 1);

  this = info->x;

  if (serial_acc > serial_add) {
    key_value_pair_donate_from_value(add, info, 0, &_object);
    this->destroy(&_object);
  } else if (serial_acc < serial_add) {
    key_value_pair_donate_from_value(acc, info, 0, &_object);
    this->destroy(&_object);
    key_value_pair_donate_from_value(add, info, 0, &_object);

    key_value_pair_donate_to_value(acc, info, 0, &_object);
    key_value_pair_set_value(acc, info, 1, serial_add);
  } else {
    throw("Impossible");
  }
}

static void pair_new_(DictionaryPair **p_this,
    char *key,
    void *object,
    ulong serial) {
  DictionaryPair *this;

  this = newo(p_this);
  this = *p_this;

  this->key = key;
  this->object = object;
  this->serial = serial;
}

static void pair_delete_(DictionaryPair **p_this) {
  DictionaryPair *this;

  if ((this = *p_this) == null) {
    return;
  }

  deleteo(p_this);
}

static sbyte chrono_pointer_compare_(
    DictionaryPair **p_a,
    DictionaryPair **p_b,
    void *x) {
  ulong serial_a;
  ulong serial_b;
  DictionaryPair *a;
  DictionaryPair *b;

  a = *p_a;
  b = *p_b;

  serial_a = a->serial;
  serial_b = b->serial;

  if (serial_a < serial_b) {
    return -1;
  }

  if (serial_a == serial_b) {
    throw("Impossible");
  }

  return +1;
}

static void check_not_finalized_(Dictionary *this, char *action) {
  if (dictionary_is_finalized(this)) {
    throw("Cannot %s a finalized dictionary", action);
  }
}

static void check_finalized_(Dictionary *this, char *action) {
  if (!dictionary_is_finalized(this)) {
    throw("Cannot %s a dictionary until it is finalized", action);
  }
}

