/* Implementation of dictionaries of strings to objects of a single type (i.e.
 * only a single version of each method can be supplied), using KeyValueArrays.
 *
 * If more than one pair is donated for a given key, then each extra donation
 * replaces the previous ones.
 *
 * When all key-value pairs have been added, dictionary_finalize() can be
 * called, which extracts the KeyValueCollection into the final KeyValueArray.
 *
 * Internally, the key-value pair is constructed as follows: The single key
 * takes ownership of the supplied key string. The first value takes ownership
 * of the object pointer. A second value specifies how many pairs were donated
 * before this one, which allows the latest one to take precedence in the case
 * that equal keys are found.
 */
#ifndef DICTIONARY_H
#define DICTIONARY_H

#ifdef __cplusplus
extern "C" {
#endif

#include "KeyValueCollection.h"

typedef struct {
  ulong serial;
  KeyValueInfo *_info;
  KeyValueArray *_array;
  KeyValueCollection *_collection;
  void (*destroy)(void **p_this);
  void (*print)(void *this, Printer *p);
} Dictionary;

typedef struct {
  char *key;
  void *object;
  ulong serial;
} DictionaryPair;

typedef struct {
  ulong cardinality;
  DictionaryPair **_pairs;
  void (*print)(void *this, Printer *p);
} DictionaryPairs;

/* Constructor.
 */
#define dictionary_new(p_this, destroy, print) \
  dictionary_newI( \
      (p_this), \
      (void (*)(void **))(destroy), \
      (void (*)(void *, Printer *))(print))

void dictionary_newI(
    Dictionary **p_this,
    void destroy(void **p_this),
    void print(void *this, Printer *p));

/* Destructor.
 */
void dictionary_delete(Dictionary **p_this);

/* Print.
 */
void dictionary_print(Dictionary *this, Printer *p);

/* Donate a key-value pair. This donates both the key string and the object.
 */
void dictionary_donate(Dictionary *this, char **p_key, void *p_object);

/* Finalize the Dictionary, which extracts into the final _array.
 */
void dictionary_finalize(Dictionary *this);

/* Whether the dictionary is finalized.
 */
bool dictionary_is_finalized(Dictionary *this);

/* Reopen the Dictionary, which extracts the array back into the collection,
 * allowing more pairs to be donated.
 */
void dictionary_reopen(Dictionary *this);

/* Find the given key. Returns a reference to the value object, or null if the
 * key was not found.
 */
void *dictionary_find(Dictionary *this, char *key);

/* The cardinality of the Dictionary, i.e. the number of pairs in it.
 */
ulong dictionary_cardinality(Dictionary *this);

/* Construct a DictionaryPairs object from a Dictionary, which is easier to
 * iterate through. Only references to the objects are returned in the pairs.
 * If 'chronological' is true, the pairs are sorted in chronological order of
 * being added; otherwise, they are sorted by key.
 */
void dictionary_pairs_new(
    DictionaryPairs **p_this,
    Dictionary *dict,
    bool chronological);

/* Destructor for the DictionaryPairs object.
 */
void dictionary_pairs_delete(DictionaryPairs **p_this);

/* Print for the DictionaryPairs object.
 */
void dictionary_pairs_print(DictionaryPairs *this, Printer *p);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _H */
