#include "File.h"

static void new_vf_(File **p_this,
    char *mode,
    char *fmt,
    va_list ap1,
    va_list ap2);

static bool try_new_vf_(
    File **p_this,
    char *mode,
    char *fmt,
    va_list ap1,
    va_list ap2);


void file_new(File **p_this, char *filename, char *mode) {
  if (!file_try_new(p_this, filename, mode)) {
    throw("Could not open file '%s' in mode '%s'", filename, mode);
  }
}

bool file_try_new(
    File **p_this,
    char *filename,
    char *mode) {
  if (mode == null) {
    file_new_memory(p_this);

    return true;
  }

  return file_try_new_storage(p_this, filename, mode);
}

void file_new_f(File **p_this, char *mode, char *fmt, ...) {
  va_list ap1;
  va_list ap2;

  va_start(ap1, fmt);
  va_start(ap2, fmt);
  new_vf_(p_this, mode, fmt, ap1, ap2);
  va_end(ap1);
  va_end(ap2);
}

bool file_try_new_f(File **p_this, char *mode, char *fmt, ...) {
  bool success;
  va_list ap1;
  va_list ap2;

  va_start(ap1, fmt);
  va_start(ap2, fmt);
  success = try_new_vf_(p_this, mode, fmt, ap1, ap2);
  va_end(ap1);
  va_end(ap2);

  return success;
}

void file_new_memory(File **p_this) {
  File *this;

  this = newo(p_this);

  memory_file_new(&this->_mf);
}

void file_new_storage(File **p_this, char *filename, char *mode) {
  if (!file_try_new_storage(p_this, filename, mode)) {
    throw("Could not open storage file '%s' in mode '%s'", filename, mode);
  }
}

bool file_try_new_storage(
    File **p_this,
    char *filename,
    char *mode) {
  File *this;

  this = newo(p_this);

  if (!storage_file_try_new((StorageFile **)&this->_sf, filename, mode)) {
    file_delete(p_this);

    return false;
  }

  return true;
}

void file_new_wrap_ansifile(File **p_this, FILE *ansifile, char *mode) {
  File *this;

  this = newo(p_this);

  storage_file_new_wrap_ansifile((StorageFile **)&this->_sf, ansifile, mode);
}

void file_new_memory_copy_from_filename(File **p_this, char *filename) {
  file_new_copy_from_filename(p_this, null, null, filename);
}

void file_new_memory_copy_storage(File **p_this, struct StorageFile *ef) {
  file_new_copy_storage(p_this, null, null, ef);

}

void file_new_copy_from_filename(
    File **p_this,
    char *filename,
    char *mode,
    char *other_filename) {
  File *_other = null;

  file_new(&_other, other_filename, "r");
  file_new_copy(p_this, filename, mode, _other);
  file_delete(&_other);
}

void file_new_copy(File **p_this, char *filename, char *mode, File *other) {
  File *this;

  file_new(p_this, filename, mode);
  this = *p_this;

  file_copy_other(this, other);
}

void file_new_copy_storage(
    File **p_this,
    char *filename,
    char *mode,
    struct StorageFile *ef) {
  File *this;

  file_new(p_this, filename, mode);
  this = *p_this;

  file_copy_storage(this, ef);
}

void file_new_memory_copy_string(File **p_this, char *s) {
  char c;
  char *p;
  File *this;

  file_new(p_this, null, null);
  this = *p_this;

  for (p = s; (c = *p) != '\0'; p++) {
    file_write_char(this, c);
  }
}

void file_delete(File **p_this) {
  File *this;

  if ((this = *p_this) == null) {
    return;
  }

  memory_file_delete(&this->_mf);
  storage_file_delete((StorageFile **)&this->_sf);

  deleteo(p_this);
}

void file_print(File *this, Printer *p) {
  sprint(p, "File");

  if (file_is_memory(this)) {
    rprint(p, "Memory file: true");
    rprint(p, "External file: false");
    memory_file_print(this->_mf, p);
  } else {
    rprint(p, "Memory file: false");
    rprint(p, "External file: true");
    storage_file_print((StorageFile *)this->_sf, p);
  }

  fprint(p);
}

void file_release(File *this) {
  if (this->_mf == null) {
    throw("Can only release memory files");
  }

  memory_file_release(this->_mf);
}

void file_reacquire(File *this) {
  if (this->_mf == null) {
    throw("Can only reacquire memory files");
  }

  memory_file_reacquire(this->_mf);
}

bool file_is_gzipped(File *this) {
  return file_is_storage(this)
    && storage_file_is_gzipped((StorageFile *)this->_sf);
}

bool file_is_memory(File *this) {
  return this->_mf != null;
}

bool file_is_storage(File *this) {
  return this->_sf != null;
}

char *file_get_filename(File *this) {
  return file_is_storage(this) ? ((StorageFile *)this->_sf)->_filename : null;
}

char *file_get_mode(File *this) {
  return file_is_storage(this) ? ((StorageFile *)this->_sf)->_mode : null;
}

void file_write_byte(File *this, byte b) {
  if (file_is_memory(this)) {
    memory_file_write_byte(this->_mf, b);
  } else {
    storage_file_write_byte((StorageFile *)this->_sf, b);
  }
}

void file_write_bytes(File *this, byte *bs, ulong num_bytes) {
  if (file_is_memory(this)) {
    memory_file_write_bytes(this->_mf, bs, num_bytes);
  } else {
    storage_file_write_bytes((StorageFile *)this->_sf, bs, num_bytes);
  }
}

void file_write_bytesm(File *this, byte **bm, ulong n0, ulong num_bytes) {
  ulong u;

  if (file_is_memory(this)) {
    memory_file_write_bytesm(this->_mf, bm, n0, num_bytes);
  } else {
    for (u = 0; u < n0; u++) {
      storage_file_write_bytes((StorageFile *)this->_sf, bm[u], num_bytes);
    }
  }
}

void file_write_bytesim(
    File *this,
    byte **bm,
    ulong n0,
    ulong *n1s,
    ulong size) {
  ulong u;

  if (file_is_memory(this)) {
    memory_file_write_bytesim(this->_mf, bm, n0, n1s, size);
  } else {
    for (u = 0; u < n0; u++) {
      storage_file_write_bytes(
          (StorageFile *)this->_sf,
          bm[u],
          n1s[u] * size);
    }
  }
}

void file_write_char(File *this, char c) {
  file_write_byte(this, (byte)c);
}

void file_write_string(File *this, char *s) {
  char c;
  char *p;

  for (p = s; (c = *p) != '\0'; p++) {
    file_write_byte(this, (byte)c);
  }
}

byte file_read_byte(File *this) {
  if (file_is_memory(this)) {
    return memory_file_read_byte(this->_mf);
  } else {
    return storage_file_read_byte((StorageFile *)this->_sf);
  }
}

bool file_try_read_byte(File *this, byte *b) {
  if (file_is_memory(this)) {
    return memory_file_try_read_byte(this->_mf, b);
  } else {
    return storage_file_try_read_byte((StorageFile *)this->_sf, b);
  }
}

void file_read_bytes(File *this, byte *bs, ulong num_bytes) {
  if (file_is_memory(this)) {
    if (bs == null) {
      memory_file_set_position(
          this->_mf,
          memory_file_get_position(this->_mf) + num_bytes);
    } else {
      memory_file_read_bytes(this->_mf, bs, num_bytes);
    }
  } else {
    storage_file_read_bytes((StorageFile *)this->_sf, bs, num_bytes);
  }
}

void file_read_bytesm(File *this, byte **bm, ulong n0, ulong num_bytes) {
  ulong u;

  if (file_is_memory(this)) {
    memory_file_read_bytesm(this->_mf, bm, n0, num_bytes);
  } else {
    for (u = 0; u < n0; u++) {
      storage_file_read_bytes((StorageFile *)this->_sf, bm[u], num_bytes);
    }
  }
}

void file_read_bytesim(
    File *this,
    byte **bm,
    ulong n0,
    ulong *n1s,
    ulong size) {
  ulong u;

  if (file_is_memory(this)) {
    memory_file_read_bytesim(this->_mf, bm, n0, n1s, size);
  } else {
    for (u = 0; u < n0; u++) {
      storage_file_read_bytes((StorageFile *)this->_sf, bm[u], n1s[u] * size);
    }
  }

}

ulong file_get_position(File *this) {
  if (file_is_memory(this)) {
    return memory_file_get_position(this->_mf);
  } else {
    return storage_file_get_position((StorageFile *)this->_sf);
  }
}

void file_set_position(File *this, ulong position) {
  if (file_is_memory(this)) {
    memory_file_set_position(this->_mf, position);
  } else {
    storage_file_set_position((StorageFile *)this->_sf, position);
  }
}

void file_rewind(File *this) {
  if (file_is_memory(this)) {
    memory_file_rewind(this->_mf);
  } else {
    storage_file_rewind((StorageFile *)this->_sf);
  }
}

void file_fast_forward(File *this) {
  if (file_is_memory(this)) {
    memory_file_fast_forward(this->_mf);
  } else {
    storage_file_fast_forward((StorageFile *)this->_sf);
  }
}

void file_backup_one_byte(File *this) {
  ulong pos;

  if ((pos = file_get_position(this)) == 0) {
    throw("Cannot backup one byte when file is at position 0");
  }

  file_set_position(this, pos - 1);
}

ulong file_get_size(File *this) {
  if (file_is_memory(this)) {
    return memory_file_get_size(this->_mf);
  } else {
    return storage_file_get_size((StorageFile *)this->_sf);
  }
}

void file_flush(File *this) {
  if (file_is_storage(this)) {
    storage_file_flush((StorageFile *)this->_sf);
  }
}

FILE *file_get_ansifile(File *this) {
  if (!file_is_storage(this)) {
    throw("Cannot get ansifile for a memory file");
  }

  return storage_file_get_ansifile((StorageFile *)this->_sf);
}

bool file_contents_equals_blob(File *this, ulong len, byte *blob) {
  if (file_is_memory(this)) {
    return memory_file_contents_equals_blob(this->_mf, len, blob);
  } else {
    return storage_file_contents_equals_blob(
        (StorageFile *)this->_sf,
        len,
        blob);
  }
}

void file_assert_contents_equals_blob(File *this, ulong len, byte *blob) {
  if (file_is_memory(this)) {
    memory_file_assert_contents_equals_blob(this->_mf, len, blob);
  } else {
    storage_file_assert_contents_equals_blob(
        (StorageFile *)this->_sf,
        len,
        blob);
  }
}

bool file_contents_equals_string(File *this, char *s) {
  return file_contents_equals_blob(this, string_length(s), (byte *)s);
}

void file_assert_contents_equals_string(File *this, char *s) {
  file_assert_contents_equals_blob(this, string_length(s), (byte *)s);
}

void file_copy_other(File *this, File *other) {
  byte b;

  while (file_try_read_byte(other, &b)) {
    file_write_byte(this, b);
  }
}

void file_copy_storage(File *this, struct StorageFile *ef) {
  byte b;

  while (storage_file_try_read_byte((StorageFile *)ef, &b)) {
    file_write_byte(this, b);
  }
}

bool file_exists(char *filename) {
  return storage_file_exists(filename);
}

bool file_exists_wait(char *filename, byte max_seconds) {
  return storage_file_exists_wait(filename, max_seconds);
}

bool file_partial_exists(char *filename) {
  return storage_file_partial_exists(filename);
}

bool file_either_exists(char *filename) {
  return storage_file_either_exists(filename);
}

bool file_is_empty(char *filename) {
  return storage_file_is_empty(filename);
}

bool file_is_writable(char *filename) {
  return storage_file_is_writable(filename);
}

void file_create(char *filename) {
  storage_file_create(filename);
}

void file_remove(char *filename) {
  storage_file_remove(filename);
}

bool file_try_remove(char *filename) {
  return storage_file_try_remove(filename);
}

void file_remove_partial(char *filename) {
  storage_file_remove_partial(filename);
}

void file_remove_both(char *filename) {
  storage_file_remove_both(filename);
}

void file_rename(char *orig_filename, char *new_filename) {
  storage_file_rename(orig_filename, new_filename);
}

ulong blob_new_file_contents(byte **p_this, File *f) {
  if (file_is_memory(f)) {
    return blob_new_memory_file_contents(p_this, f->_mf);
  }

  return blob_new_storage_file_contents(p_this, (StorageFile *)f->_sf);
}

ulong blob_new_file_contents_from_filename(byte **p_this, char *filename) {
  return blob_new_storage_file_contents_from_filename(p_this, filename);
}

void string_new_file_contents(char **p_this, File *f) {
  if (file_is_memory(f)) {
    string_new_memory_file_contents(p_this, f->_mf);
  }

  string_new_storage_file_contents(p_this, (StorageFile *)f->_sf);
}

void string_new_file_contents_from_filename(char **p_this, char *filename) {
  string_new_storage_file_contents_from_filename(p_this, filename);
}

sbyte file_progress_status(char *filename) {
  return storage_file_progress_status(filename);
}

sbyte files_progress_status(ulong num_filenames, char **filenames) {
  return storage_files_progress_status(num_filenames, filenames);
}


void load_label(char *s, File *f) { \
  char *_s = null;
  ulong len;

  load_value(len, f);

  if (len != string_length(s)) {
    throw( \
        "File corrupted: expected label '%s' but loaded a label length of %lu "
        "rather than %lu",
        s,
        len,
        string_length(s));
  }

  string_new_load_from_length(&_s, len, f);

  if (!string_equals(_s, s)) {
    throw("File corrupted: read label '%s' instead of expected '%s'", _s, s);
  }

  string_delete(&_s);
}

static void new_vf_(File **p_this,
    char *mode,
    char *fmt,
    va_list ap1,
    va_list ap2) {
  char *_filename = null;

  string_new_vf(&_filename, fmt, ap1, ap2);
  file_new(p_this, _filename, mode);
  string_delete(&_filename);
}

static bool try_new_vf_(
    File **p_this,
    char *mode,
    char *fmt,
    va_list ap1,
    va_list ap2) {
  bool success;
  char *_filename = null;

  string_new_vf(&_filename, fmt, ap1, ap2);
  success = file_try_new(p_this, _filename, mode);
  string_delete(&_filename);

  return success;
}
