#include "Heap.h"

#define parent_(l) (((l) - 1) / 2)

#define left_(u) (((u) << 1) + 1)

#define right_(u) (((u) << 1) + 2)

#define elem_(this, u) \
  ((void *)((byte *)(this)->_data + (u) * (this)->elementSize))

static void copy_element_(Heap *this, void *src, void *dst);

static void move_element_(Heap *this, void *src, void *dst);

static void check_full_(Heap *this);

static void check_empty_(Heap *this);

static void exchange_(Heap *this, ulong u, ulong v);

static void check_heap_(Heap *this);

static void check_node_(Heap *this, ulong u);

static void print_subtree_(Heap *this, Printer *p, ulong u);

static void heapify_(Heap *this, ulong u);

static void decrease_key_(Heap *this, ulong u, void *key);


void heap_newI(Heap **p_this,
    void *maxPossible,
    ulong element_size,
    sbyte compare(void *a, void *b, void *x),
    void print(HeapObjectPrintInfo *hopi, Printer *p),
    void *x) {
  Heap *this;

  this = newo(p_this);

  this->heapified = true;
  this->elementSize = element_size;
  this->compare = compare;
  this->print = print;
  this->x = x;

  new(&this->_maxPossible, element_size);
  copy_element_(this, maxPossible, this->_maxPossible);

  new(&this->_data, element_size);
  new(&this->_temp, element_size);

  this->maxNum = 1;
  this->num = 0;
}

void heap_delete(Heap **p_this) {
  Heap *this;

  if ((this = *p_this) == null) {
    return;
  }

  deletev(&this->_temp);
  deletev(&this->_data);
  deletev(&this->_maxPossible);

  deleteo(p_this);
}

void heap_print(Heap *this, Printer *p) {
  sprint(p, "Heap");
  rprint(p, "Element size: %lu", this->elementSize);
  rprint(p, "Maximum number of elements: %lu", this->maxNum);
  rprint(p, "Current number of elements: %lu", this->num);
  rprint(p, "Heapified: %s", bool_string(this->heapified));

  if (this->num != 0) {
    sprint(p, "Elements");
    print_subtree_(this, p, 0);
    fprint(p);
  }

  if (this->heapified) {
    check_heap_(this);
  }

  fprint(p);
}

void heap_add(Heap *this, void *a) {
  check_full_(this);

  copy_element_(this, a, elem_(this, this->num));

  if (++this->num > 1) {
    this->heapified = false;
  }
}

void heap_build(Heap *this) {
  long l;

  if (this->heapified) {
    return;
  }

  for (l = parent_(this->num - 1); l >= 0; l--) {
    heapify_(this, l);
  }

  this->heapified = true;
}

ulong heap_num_elements(Heap *this) {
  return this->num;
}

bool heap_is_empty(Heap *this) {
  return heap_num_elements(this) == 0;
}

void *heap_get_min(Heap *this) {
  if (this->num == 0) {
    throw("Cannot get minimum from an empty heap");
  }

  if (!this->heapified) {
    throw("Cannot get minimum before heap is heapified");
  }

  return this->_data;
}

void heap_extract_min(Heap *this, void *min) {
  if (this->num == 0) {
    throw("Cannot get minimum from an empty heap");
  }

  if (!this->heapified) {
    throw("Cannot extract minimum before heap is heapified");
  }

  copy_element_(this, this->_data, min);

  if (--this->num > 0) {
    move_element_(this, elem_(this, this->num), this->_data);
  }

  heapify_(this, 0);

  check_empty_(this);
}

void heap_insert(Heap *this, void *a) {
  check_full_(this);

  copy_element_(this, this->_maxPossible, elem_(this, this->num++));
  decrease_key_(this, this->num - 1, a);
}

void *heap_get(Heap *this, ulong u) {
  return this->_data + u * this->elementSize;
}

static void check_full_(Heap *this) {
  ulong size;

  if (this->num < this->maxNum) {
    return;
  }

  size = this->elementSize;
  renewup(&this->_data, this->maxNum * size, this->maxNum * 2 * size);
  this->maxNum *= 2;
}

static void check_empty_(Heap *this) {
  ulong size;

  if (this->num >= this->maxNum / 4) {
    return;
  }

  size = this->elementSize;
  renewv(&this->_data, this->maxNum * size, this->maxNum * size / 2);
  this->maxNum /= 2;
}

static void decrease_key_(Heap *this, ulong u, void *key) {
  char *_s1 = null;
  char *_s2 = null;
  void *x;

  x = this->x;

  if (this->compare(key, elem_(this, u), x) >= 0) {
    mprint(&_s1, 1, this->print, key);
    mprint(&_s2, 1, this->print, elem_(this, u));
    throw("New key doesn't decrease current key for element %lu\n"
        "New key:\n%sCurrent key:\n%s",
        u,
        _s1,
        _s2);
  }

  copy_element_(this, key, elem_(this, u));

  while (u && this->compare(elem_(this, parent_(u)), elem_(this, u), x) > 0) {
    exchange_(this, u, parent_(u));
    u = parent_(u);
  }
}

static void exchange_(Heap *this, ulong u, ulong v) {
  void *temp;
  void *p_u;
  void *p_v;

  temp = this->_temp;
  p_u = elem_(this, u);
  p_v = elem_(this, v);

  copy_element_(this, p_u, temp);
  copy_element_(this, p_v, p_u);
  copy_element_(this, temp, p_v);
}

static void check_heap_(Heap *this) {
  check_node_(this, 0);
}

static void check_node_(Heap *this, ulong u) {
  ulong l;
  ulong r;
  void *x;

  if (u >= this->num) {
    return;
  }

  l = left_(u);
  r = right_(u);
  x = this->x;

  if (l < this->num && this->compare(elem_(this, l), elem_(this, u), x) < 0) {
    throw("Heap property violated for elements %lu and %lu", u, l);
  }

  if (r < this->num && this->compare(elem_(this, r), elem_(this, u), x) < 0) {
    throw("Heap property violated for elements %lu and %lu", u, r);
  }

  check_node_(this, l);
  check_node_(this, r);
}

static void print_subtree_(Heap *this, Printer *p, ulong u) {
  HeapObjectPrintInfo hopi;

  if (u >= this->num) {
    return;
  }

  sprint(p, "Index %lu", u);
  hopi.this = elem_(this, u);
  hopi.x = this->x;
  this->print(&hopi, p);
  print_subtree_(this, p, left_(u));
  print_subtree_(this, p, right_(u));
  fprint(p);
}

static void copy_element_(Heap *this, void *src, void *dst) {
  copy_memory(src, dst, this->elementSize);
}

static void move_element_(Heap *this, void *src, void *dst) {
  move_memory(src, dst, this->elementSize);
}

static void heapify_(Heap *this, ulong u) {
  ulong l;
  ulong r;
  ulong num;
  ulong smallest;
  void *x;
  sbyte (*compare)(void *a, void *b, void *x);

  num = this->num;

  if (u >= num) {
    return;
  }

  l = left_(u);
  r = right_(u);

  compare = this->compare;
  x = this->x;

  if (l < num && compare(elem_(this, l), elem_(this, u), x) < 0) {
    smallest = l;
  } else {
    smallest = u;
  }

  if (r < num && compare(elem_(this, r), elem_(this, smallest), x) < 0) {
    smallest = r;
  }

  if (smallest != u) {
    exchange_(this, u, smallest);
    heapify_(this, smallest);
  }
}
