/* Heaps of value or reference objects. Implemented as a min-heap, but this can
 * be reversed by adjusting the sense of the comparator.
 */
#ifndef HEAP_H
#define HEAP_H

#ifdef __cplusplus
extern "C" {
#endif

#include "jpc.h"

typedef struct {
  void *this;
  void *x;
} HeapObjectPrintInfo;

typedef struct {
  bool heapified;
  byte *_maxPossible;
  byte *_data;
  byte *_temp;
  ulong maxNum;
  ulong num;
  ulong elementSize;
  sbyte (*compare)(void *a, void *b, void *x);
  void (*print)(HeapObjectPrintInfo *hopi, Printer *p);
  void *x;
} Heap;

/* Constructor. The size of the type is taken from *maxPossible, which points
 * to an element that contains the maximum possible value for the type (as
 * determined by the compare() method).
 */
#define heap_new(p_this, max, compare, print, x) \
  heap_newI( \
      (p_this), \
      (max), \
      sizeof(*(max)), \
      (sbyte (*)(void *, void *, void *))(compare), \
      (print), \
      (void *)(x))

void heap_newI(Heap **p_this,
    void *maxPossible,
    ulong element_size,
    sbyte compare(void *a, void *b, void *x),
    void print(HeapObjectPrintInfo *hopi, Printer *p),
    void *x);

/* Destructor.
 */
void heap_delete(Heap **p_this);

/* Print.
 */
void heap_print(Heap *this, Printer *p);

/* Add an item to the heap.
 */
void heap_add(Heap *this, void *a);

/* Build the heap.
 */
void heap_build(Heap *this);

/* The number of elements in the heap.
 */
ulong heap_num_elements(Heap *this);

/* Whether the heap is empty.
 */
bool heap_is_empty(Heap *this);

/* Get a pointer to the minimum item in the heap, without extracting it.
 */
void *heap_get_min(Heap *this);

/* Extract the minimum item from the heap.
 */
void heap_extract_min(Heap *this, void *min);

/* Insert an item into the heap.
 */
void heap_insert(Heap *this, void *a);

/* Get a pointer to the u-th item in the heap storage. Only used to iterate
 * through all items.
 */
void *heap_get(Heap *this, ulong u);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _H */
