#include "Hypercube.h"

typedef struct {
  bool *walkDimensions;
  ulong d;
  ulong o;
  ulong ops;
  ulong *specifiedIndexes;
  ulong *_indexes;
  void (*callback)(
      Hypercube *this,
      bool *walk_dimensions,
      ulong *indexes,
      ulong o,
      void *x);
  void *x;
  Hypercube *this;
} Pass_;

static void check_d_(Hypercube *this, ulong d);

static void check_o_(Hypercube *this, ulong o);

static void foreach_(Parallel *par);

static void print_(Hypercube *this, Printer *p, ulong d, ulong s);


void hypercube_newI(
    Hypercube **p_this,
    char *type,
    ulong size,
    ulong num_dimensions,
    ulong *dimensions,
    void print(void *this, Printer *p)) {
  ulong d;
  ulong d_rev;
  ulong dimension;
  ulong num_elements;
  ulong stride;
  ulong *strides;
  Hypercube *this;

  this = newo(p_this);
  string_new_copy(&this->_type, type);
  this->size = size;
  this->numDimensions = num_dimensions;
  new(&this->_dimensions, num_dimensions);
  this->print = print;

  for (d = 0, num_elements = 1; d < num_dimensions; d++) {
    dimension = this->_dimensions[d] = dimensions[d];

    if (dimension == 0) {
      throw("Dimension %lu is zero, which is not allowed", d);
    }

    num_elements *= dimension;
  }

  this->numElements = num_elements;
  new(&this->_data, num_elements * size);
  new(&this->_strides, num_dimensions);
  strides = this->_strides;

  for (d_rev = 0, stride = 1; d_rev < num_dimensions; d_rev++) {
    d = num_dimensions - 1 - d_rev;
    strides[d] = stride;
    stride *= dimensions[d];
  }
}

void hypercube_new_copy(Hypercube **p_this, Hypercube *other) {
  Hypercube *this;

  hypercube_newI(
      p_this,
      other->_type,
      other->size,
      other->numDimensions,
      other->_dimensions,
      other->print);
  this = *p_this;
  copy_values(other->_data, this->_data, this->numElements * this->size);
}

void hypercube_new_load(Hypercube **p_this, File *f) {
  char *_type = null;
  ulong num_dimensions;
  ulong size;
  ulong *_dimensions = null;
  Hypercube *this;
  void (*print)(void *this, Printer *p);

  load_label("[Hypercube]", f);
  load(_type, string, f);

  if (string_equals(_type, "ulong")) {
    print = (void (*)(void *, Printer *))ulong_print;
  } else if (string_equals(_type, "paw")) {
    print = (void (*)(void *, Printer *))paw_print;
  } else if (string_equals(_type, "double")) {
    print = (void (*)(void *, Printer *))double_print;
  } else {
    throw("Type '%s' not yet implemented", _type);
    print = null;
  }

  load_value(size, f);
  load_value(num_dimensions, f);
  load_new_values(_dimensions, num_dimensions, f);
  hypercube_newI(p_this, _type, size, num_dimensions, _dimensions, print);
  this = *p_this;
  string_delete(&_type);
  deletev(&_dimensions);
  load_values(this->_data, this->numElements * size, f);
  load_label("[/Hypercube]", f);
}

void ulong_hypercube_new_from_sparse_key_frequency_array(
    Hypercube **p_this,
    KeyValueArray **p_array,
    ulong *dimensions) {
  ulong d;
  ulong frequency;
  ulong num_dimensions;
  ulong num_pairs;
  ulong o;
  ulong u;
  ulong *strides;
  KeyValuePair *pair;
  KeyValueInfo *info;
  KeyValueArray *array;
  Hypercube *this;

  array = *p_array;
  info = array->_info;

  if (info->numValues != 1) {
    throw("Number of values (%lu) must be 1", info->numValues);
  }

  num_dimensions = info->numKeys;
  ulong_hypercube_new(p_this, num_dimensions, dimensions);
  this = *p_this;

  strides = this->_strides;
  num_pairs = array->numPairs;

  for (u = 0; u < num_pairs; u++) {
    pair = key_value_array_get(array, u);

    for (d = 0, o = 0; d < num_dimensions; d++) {
      o += key_value_pair_get_key(pair, info, d) * strides[d];
    }

    frequency = sparse_key_frequency_pair_get_frequency(pair, info);
    hypercube_set(this, o, &frequency);
  }

  key_value_array_delete(p_array);
}

void hypercube_delete(Hypercube **p_this) {
  Hypercube *this;

  if ((this = *p_this) == null) {
    return;
  }

  deletev(&this->_strides);
  deletev(&this->_data);
  deletev(&this->_dimensions);
  string_delete(&this->_type);
  deleteo(p_this);
}

void hypercube_print(Hypercube *this, Printer *p) {
  ulong d;

  sprint(p, "Hypercube");
  rprint(p, "Bytes per element: %lu", this->size);
  rprint(p, "Number of dimensions: %lu", this->numDimensions);
  rprint(p, "Number of elements: %lu", this->numElements);

  sprint(p, "Dimensions");

  for (d = 0; d < this->numDimensions; d++) {
    rprint(p, "Dimension %lu: %lu", d, this->_dimensions[d]);
  }

  fprint(p);
  sprint(p, "Strides");

  for (d = 0; d < this->numDimensions; d++) {
    rprint(p, "Stride %lu: %lu", d, this->_strides[d]);
  }

  fprint(p);
  sprint(p, "Contents");
  print_(this, p, 0, 0);
  fprint(p);

  fprint(p);
}

void hypercube_save(Hypercube *this, File *f) {
  save_label("[Hypercube]", f);
  save(this->_type, string, f);
  save_value(this->size, f);
  save_value(this->numDimensions, f);
  save_values(this->_dimensions, this->numDimensions, f);
  save_values(this->_data, this->numElements * this->size, f);
  save_label("[/Hypercube]", f);
}

bool hypercube_equals(Hypercube *this, Hypercube *other) {
  return other->size == this->size
    && other->numDimensions == this->numDimensions
    && values_equal(other->_dimensions, this->_dimensions, this->numDimensions)
    && values_equal(other->_data, this->_data, this->numElements * this->size);
}

void hypercube_set(Hypercube *this, ulong o, void *value) {
  check_o_(this, o);
  copy_memory(value, &this->_data[o * this->size], this->size);
}

void hypercube_get(Hypercube *this, ulong o, void *value) {
  check_o_(this, o);
  copy_memory(&this->_data[o * this->size], value, this->size);
}

void hypercube_get_from_indexes(Hypercube *this, ulong *indexes, void *value) {
  ulong d;
  ulong index;
  ulong o;

  for (o = 0, d = 0; d < this->numDimensions; d++) {
    if ((index = indexes[d]) >= this->_dimensions[d]) {
      throw("Index %lu (%lu) is invalid (max %lu)",
          d,
          indexes[d],
          this->_dimensions[d] - 1);
    }

    o += this->_strides[d] * index;
  }

  hypercube_get(this, o, value);
}

ulong hypercube_i_from_o(Hypercube *this, ulong d, ulong o) {
  check_d_(this, d);
  check_o_(this, o);
  return o / this->_strides[d] % this->_dimensions[d];
}

void hypercube_foreach(
    Hypercube *this,
    bool *walk_dimensions,
    ulong *specified_indexes,
    void callback(Hypercube *this,
      bool *walk_dimensions,
      ulong *indexes,
      ulong o,
      void *x),
    void *x,
    ulong ops_per_callback) {
  bool *_walk_dimensions = null;
  ulong d;
  ulong *_specified_indexes = null;
  Pass_ pass;

  if (this->numDimensions == 0) {
    callback(this, walk_dimensions, null, 0, x);

    return;
  }

  new(&_walk_dimensions, this->numDimensions);
  new(&_specified_indexes, this->numDimensions);

  for (d = 0, pass.ops = ops_per_callback; d < this->numDimensions; d++) {
    _walk_dimensions[d] = walk_dimensions == null ? true : walk_dimensions[d];

    if (specified_indexes != null) {
      _specified_indexes[d] = specified_indexes[d];
    }

    if (_walk_dimensions[d]) {
      if (_specified_indexes[d] != 0) {
        throw("specified_indexes[%lu] set to %lu but this is a "
            "walk dimension, so must be 0",
            d,
            _specified_indexes[d]);
      }

      pass.ops *= this->_dimensions[d];
    } else {
      if (_specified_indexes[d] >= this->_dimensions[d]) {
        throw("specified_indexes[%lu] set to %lu but dimensions[%lu] = %lu",
            d,
            specified_indexes[d],
            d,
            this->_dimensions[d]);
      }
    }
  }

  pass.walkDimensions = _walk_dimensions;
  pass.specifiedIndexes = _specified_indexes;
  pass._indexes = null;
  pass.d = 0;
  pass.o = 0;
  pass.callback = callback;
  pass.x = x;
  pass.this = this;
  parallel_single(foreach_, &pass);
  deletev(&_specified_indexes);
  deletev(&_walk_dimensions);
}

static void foreach_(Parallel *par) {
  ulong d;
  ulong i;
  Pass_ *in;
  Pass_ out;
  Hypercube *this;

  in = par->x;
  this = in->this;
  d = in->d;

  out.walkDimensions = in->walkDimensions;
  out.specifiedIndexes = in->specifiedIndexes;
  out._indexes = null;
  out.d = d + 1;
  out.o = in->o;
  out.callback = in->callback;
  out.x = in->x;
  out.this = this;
  out.ops = in->ops;

  if (d > 0) {
    new(&out._indexes, d);
    copy_values(in->_indexes, out._indexes, d - 1);
  }

  for (i = par->start; i < par->end; i++) {
    if (d > 0) {
      out._indexes[d - 1] = in->walkDimensions[d - 1]
        ? i
        : in->specifiedIndexes[d - 1];
      out.o = in->o + out._indexes[d - 1] * this->_strides[d - 1];
    }

    if (d == this->numDimensions) {
      out.callback(this, out.walkDimensions, out._indexes, out.o, out.x);
    } else if (in->walkDimensions == null || in->walkDimensions[d]) {
      out.ops /= this->_dimensions[d];
      parallel_for(this->_dimensions[d], out.ops, foreach_, &out);
    } else {
      parallel_single(foreach_, &out);
    }
  }

  if (d > 0) {
    deletev(&out._indexes);
  }
}

static void check_o_(Hypercube *this, ulong o) {
  if (o >= this->numElements) {
    throw("Overall element index %lu is invalid (max %lu)",
        o,
        this->numElements - 1);
  }
}

static void check_d_(Hypercube *this, ulong d) {
  if (d >= this->numDimensions) {
    throw("Dimension index %lu is invalid (max %lu)",
        d,
        this->numDimensions - 1);
  }
}

static void print_(Hypercube *this, Printer *p, ulong d, ulong s) {
  byte *_value = null;
  ulong dimension;
  ulong i;
  ulong num_dimensions;
  ulong o;
  ulong *strides;

  num_dimensions = this->numDimensions;

  if (num_dimensions == 0) {
    new(&_value, this->size);
    hypercube_get(this, 0, _value);
    this->print(_value, p);
    deletev(&_value);

    return;
  }

  strides = this->_strides;
  dimension = this->_dimensions[d];
  new(&_value, this->size);

  for (i = 0; i < dimension; i++) {
    o = s + strides[d] * i;
    sprint(p, "i%lu = %lu", d, i);

    if (d == num_dimensions - 1) {
      hypercube_get(this, o, _value);
      this->print(_value, p);
    } else {
      print_(this, p, d + 1, o);
    }

    fprint(p);
  }

  deletev(&_value);
}
