/* Object class for arbitrary d-dimensional hypercubes of values. The data is
 * stored as a single large array, indexed by the overall index 'o'. The index
 * for a given dimension 'd' is denoted 'i'.
 */
#ifndef HYPERCUBE_H
#define HYPERCUBE_H

#ifdef __cplusplus
extern "C" {
#endif

#include "KeyValueArray.h"
#include "SparseKeyFrequency.h"

typedef struct {
  byte *_data;
  char *_type;
  ulong numDimensions;
  ulong numElements;
  ulong size;
  ulong *_dimensions;
  ulong *_strides;
  void (*print)(void *this, Printer *p);
} Hypercube;

/* Constructor.
 */
#define hypercube_new(p_this, c, nd, ds) \
  hypercube_newI( \
      (p_this), \
      #c, \
      sizeof(c), \
      (nd), \
      (ds), \
      (void (*)(void *, Printer *))c##_print)

void hypercube_newI(
    Hypercube **p_this,
    char *type,
    ulong size,
    ulong num_dimensions,
    ulong *dimensions,
    void print(void *this, Printer *p));

/* Syntactic sugar for specific types.
 */
#define UlongHypercube Hypercube

#define ulong_hypercube_new(p_this, nd, ds) \
  hypercube_new((p_this), ulong, (nd), (ds))

#define PawHypercube Hypercube

#define paw_hypercube_new(p_this, nd, ds) \
  hypercube_new((p_this), paw, (nd), (ds))

#define DoubleHypercube Hypercube

#define double_hypercube_new(p_this, nd, ds) \
  hypercube_new((p_this), double, (nd), (ds))

/* Copy constructor.
 */
void hypercube_new_copy(Hypercube **p_this, Hypercube *other);

/* Ulong constructor, copying from a SparseKeyFrequencyArray, and then
 * destroying the latter.
 */
void ulong_hypercube_new_from_sparse_key_frequency_array(
    Hypercube **p_this,
    KeyValueArray **p_array,
    ulong *dimensions);

/* Load.
 */
void hypercube_new_load(Hypercube **p_this, File *f);

/* Destructor.
 */
void hypercube_delete(Hypercube **p_this);

/* Print.
 */
void hypercube_print(Hypercube *this, Printer *p);

/* Save.
 */
void hypercube_save(Hypercube *this, File *f);

/* Equals.
 */
bool hypercube_equals(Hypercube *this, Hypercube *other);

/* Set the value at the given overall index.
 */
void hypercube_set(Hypercube *this, ulong o, void *value);

/* Get the value at the given overall index.
 */
void hypercube_get(Hypercube *this, ulong o, void *value);

/* Get the value at the given set of indexes.
 */
void hypercube_get_from_indexes(Hypercube *this, ulong *indexes, void *value);

/* Get the index i for the d-th dimension from the overall index o.
 */
ulong hypercube_i_from_o(Hypercube *this, ulong d, ulong o);

/* Walk through each of a number of dimensions. If 'walk_dimensions' is null,
 * then all dimensions are walked through; if not, it specifies the dimensions
 * to be walked through. If 'specified_indexes' is non-null, then it specifies
 * the 'i' indexes of all dimensions not to be walked through (those walked
 * through must be zero); if it is null, each such 'i' index is set to zero.
 * The callback function is passed the object, the walk dimensions, the array
 * of 'i' indexes for the element being called, the overall index 'o' for that
 * element, and the passthrough x.
 */
void hypercube_foreach(
    Hypercube *this,
    bool *walk_dimensions,
    ulong *specified_indexes,
    void callback(Hypercube *this,
      bool *walk_dimensions,
      ulong *indexes,
      ulong o,
      void *x),
    void *x,
    ulong ops_per_callback);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _H */
