/* Simple implementation of a JSON library. The following object classes are
 * defined:
 *
 *   Json:  General JSON object (in the programming sense of the word "object")
 *   JsonString:  String
 *   JsonNumber:  Number
 *   JsonArray:  Array
 *   JsonObject:  JSON "Object", i.e. dictionary
 *   JsonObjectPair:  "Object" (dictionary) pair
 *   JsonObjectPairs:  Set of "Object" (dictionary) pairs
 */
#ifndef JSON_H
#define JSON_H

#ifdef __cplusplus
extern "C" {
#endif

#include "Dictionary.h"
#include "File.h"

typedef enum {
  JSON_EMPTY,
  JSON_UNKNOWN,
  JSON_NULL,
  JSON_TRUE,
  JSON_FALSE,
  JSON_STRING,
  JSON_NUMBER,
  JSON_ARRAY,
  JSON_OBJECT
} JsonType;

typedef struct {
  JsonType type;
  char *_e;
  void *_o;
} Json;

typedef struct {
  char *_s;
} JsonString;

typedef struct {
  boolbit negative;
  boolbit integer;
  char *_s;
} JsonNumber;

typedef struct {
  ulong length;
  Json **_js;
} JsonArray;

typedef struct {
  Dictionary *_d;
} JsonObject;

typedef struct {
  char *key;
  Json *j;
} JsonObjectPair;

typedef struct {
  ulong cardinality;
  JsonObjectPair **_pairs;
} JsonObjectPairs;

/* Constructor for a general Json object, parsing enough of the string pointed
 * to by 'pp' to create one complete Json object. The pointer to string pointer
 * 'p_p' is updated to point to the first character after the characters
 * ingested to create the object.
 */
void json_new(Json **p_this, char **p_p);

/* Constructor for a general Json object that is empty.
 */
void json_new_empty(Json **p_this);

/* Constructor for a general Json object that is an empty array.
 */
void json_new_empty_array(Json **p_this);

/* Constructor for a general Json object that is expected to consume the entire
 * string. Throws an exception if the entire string was not consumed.
 */
void json_new_entire(Json **p_this, char *s);

/* Copy constructor for a general Json object.
 */
void json_new_copy(Json **p_this, Json *other);

/* Constructor for a general Json object that is extracted from another by
 * means of the normal path notation.
 */
void json_new_extract(Json **p_this, Json *other, char *path);

/* Destructor for a general Json object.
 */
void json_delete(Json **p_this);

/* Print for a general Json object.
 */
void json_print(Json *this, Printer *p);

/* Get the type of a general Json object, as a string. These are upper camel
 * case versions of the JsonType enum above. This means that those that
 * correspond to any of the object classes defined above are equal to the class
 * names.
 */
char *json_get_type_string(Json *this);

/* Whether a general Json object is empty.
 */
bool json_is_empty(Json *this);

/* Whether a general Json object is of unknown type.
 */
bool json_is_unknown(Json *this);

/* Whether a general Json object is null.
 */
bool json_is_null(Json *this);

/* Whether a general Json object is a scalar (i.e. not a collection).
 */
bool json_is_scalar(Json *this);

/* Whether a general Json object is a value object, i.e. is neither empty nor
 * of unknown type.
 */
bool json_is_value(Json *this);

/* Whether a general Json object is an array.
 */
bool json_is_array(Json *this);

/* Whether a general Json object is an "Object", i.e. dictionary.
 */
bool json_is_object(Json *this);

/* Get a pointer to the JsonString in a general Json object. Throws if it is
 * not a JsonString.
 */
JsonString *json_get_json_string(Json *this);

/* Get a pointer to the JsonNumber in a general Json object. Throws if it is
 * not a JsonNumber.
 */
JsonNumber *json_get_json_number(Json *this);

/* Get a pointer to the JsonArray in a general Json object. Throws if it is
 * not a JsonArray.
 */
JsonArray *json_get_json_array(Json *this);

/* Get a pointer to the JsonObject in a general Json object. Throws if it is
 * not a JsonObject.
 */
JsonObject *json_get_json_object(Json *this);

/* Get the value of a general Json object, interpreted as a boolean. Throws if
 * it can't be so interpreted.
 */
bool json_get_bool(Json *this);

/* Get the value of a general Json object, interpreted as a ulong. Throws if it
 * can't be so interpreted.
 */
ulong json_get_ulong(Json *this);

/* Get the value of a JsonNumber, interpreted as a ulong. Throws if it can't be
 * so interpreted.
 */
ulong json_number_get_ulong(JsonNumber *this);

/* Get the value of a general Json object, interpreted as a long. Throws if it
 * can't be so interpreted.
 */
long json_get_long(Json *this);

/* Get the value of a JsonNumber, interpreted as a long. Throws if it can't be
 * so interpreted.
 */
long json_number_get_long(JsonNumber *this);

/* Get the value of a general Json object, interpreted as a double. Throws if
 * it can't be so interpreted.
 */
double json_get_double(Json *this);

/* Get the value of a JsonNumber, interpreted as a double. Throws if it can't
 * be so interpreted.
 */
double json_number_get_double(JsonNumber *this);

/* Get a pointer to the value of a general Json that is a JsonString. Throws if
 * it isn't. Note that string_new_json() can convert any Json to a string.
 */
char *json_get_string(Json *this);

/* Try to get the value of a general Json object, interpreted as a boolean.
 */
bool json_try_get_bool(Json *this, bool *p_b);

/* Try to get the value of a general Json object, interpreted as a ulong.
 */
bool json_try_get_ulong(Json *this, ulong *p_u);

/* Try to get the value of a JsonNumber, interpreted as a ulong.
 */
bool json_number_try_get_ulong(JsonNumber *this, ulong *p_u);

/* Try to get the value of a general Json object, interpreted as a long.
 */
bool json_try_get_long(Json *this, long *p_l);

/* Try to get the value of a JsonNumber, interpreted as a long.
 */
bool json_number_try_get_long(JsonNumber *this, long *p_l);

/* Try to get the value of a general Json object, interpreted as a double.
 */
bool json_try_get_double(Json *this, double *p_d);

/* Try to get the value of a JsonNumber, interpreted as a double.
 */
bool json_number_try_get_double(JsonNumber *this, double *p_d);

/* Try to get the value of a general Json object as a string; it must be a
 * JsonString to succeed. Note that string_new_json() can render any Json as a
 * string.
 */
bool json_try_get_string(Json *this, char **p_s);

/* Write a general Json object to a File.
 *
 * Here and below, if 'pretty' is true, then any Json arrays or "Objects"
 * (dictionaries) are pretty-printed.
 */
void json_write_file(Json *this, File *file, bool pretty, byte i);

/* Append a general Json object to the end of a general Json object (which must
 * be an array). Takes ownership of the appended object.
 */
void json_append_array(Json *this, Json **p_other);

/* Append the elements of a general Json object (which must be an array) to the
 * end of a general Json object (which must also be an array). Takes ownership
 * of the appended elements.
 */
void json_append_array_with_array(Json *this, Json **p_other);

/* Donate a key-value pair with the specified key and the given Json object to
 * a general Json object (which must be an "Object," i.e. dictionary). They key
 * and the Json object are donated.
 */
void json_donate_to_object(Json *this, char **p_key, Json **p_other);

/* Create a string that represents a general Json object in standard
 * representation, prettyified if 'pretty' is true, with an optional indent.
 */
void string_new_json(char **p_this, Json *j, bool pretty, byte i);

/* Constructor for a JsonString object, consuming as much of the string pointed
 * to by 'pp' as required. Such a JsonString must be enclosed in double quotes.
 *
 * Here and below, any error(s) encountered have their exception string(s)
 * copied into a new string pointed to by perr, rather than throwing.
 */
void json_string_new(JsonString **p_this, char **p_p, char **p_err);

/* Copy constructor for a JsonString.
 */
void json_string_new_copy(JsonString **p_this, JsonString *other);

/* Destructor for a JsonString.
 */
void json_string_delete(JsonString **p_this);

/* Print for a JsonString.
 */
void json_string_print(JsonString *this, Printer *p);

/* Write a JsonString to a File.
 */
void json_string_write_file(JsonString *this, File *file);

/* Consuming constructor for a JsonNumber object.
 */
void json_number_new(JsonNumber **p_this, char **p_p, char **p_err);

/* Copy constructor for a JsonNumber object.
 */
void json_number_new_copy(JsonNumber **p_this, JsonNumber *other);

/* Destructor for a JsonNumber object.
 */
void json_number_delete(JsonNumber **p_this);

/* Print for a JsonNumber object.
 */
void json_number_print(JsonNumber *this, Printer *p);

/* Print a JsonNumber to a File.
 */
void json_number_write_file(JsonNumber *this, File *file);

/* Consuming constructor for a JsonArray object.
 */
void json_array_new(JsonArray **p_this, char **p_p, char **p_err);

/* Constructor for an empty JsonArray object.
 */
void json_array_new_empty(JsonArray **p_this);

/* Entire consuming constructor for a JsonArray object.
 */
void json_array_new_entire(JsonArray **p_this, char *s);

/* Copy constructor for a JsonArray object.
 */
void json_array_new_copy(JsonArray **p_this, JsonArray *other);

/* Destructor for a JsonArray object.
 */
void json_array_delete(JsonArray **p_this);

/* Print for a JsonArray object.
 */
void json_array_print(JsonArray *this, Printer *p);

/* Append a general Json object to the end of a JsonArray. Takes over ownership
 * of the general Json object.
 */
void json_array_append(JsonArray *this, Json **p_j);

/* Append the elements of a general Json object, which must be an array, to the
 * end of a JsonArray. Takes over ownership of the general Json object and its
 * elements.
 */
void json_array_append_array(JsonArray *this, Json **p_j);

/* Get the number of elements in a JsonArray object.
 */
ulong json_array_length(JsonArray *this);

/* Get a pointer to the specified element of a JsonArray object. Negative
 * values of 'index' are counted from the end. If the index is out of range,
 * null is returned.
 */
Json *json_array_get(JsonArray *this, long index);

/* Write a JsonArray to a File.
 */
void json_array_write_file(JsonArray *this, File *file, bool pretty, byte i);

/* Create an array of long values by consuming an entire JSON string.
 */
void long_array_new_from_json_entire(long **p_this, ulong *p_n, char *s);

/* Create an array of ulong values by consuming an entire JSON string.
 */
ulong *ulong_array_new_from_json_entire(ulong **p_this, ulong *p_n, char *s);

/* Create an array of string values by consuming an entire JSON string.
 */
char **string_array_new_from_json_entire(char ***p_this, ulong *p_n, char *s);

/* Consuming constructor for a Json "Object" (i.e. dictionary).
 */
void json_object_new(JsonObject **p_this, char **p_p, char **p_err);

/* Constructor for an empty Json "Object" (i.e. dictionary).
 */
void json_object_new_empty(JsonObject **p_this);

/* Copy constructor for a Json "Object" (i.e. dictionary).
 */
void json_object_new_copy(JsonObject **p_this, JsonObject *other);

/* Destructor for a Json "Object" (i.e. dictionary).
 */
void json_object_delete(JsonObject **p_this);

/* Print for a Json "Object" (i.e. dictionary).
 */
void json_object_print(JsonObject *this, Printer *p);

/* Donate a key-value pair to a Json "Object" (i.e. dictionary). The key and
 * the general Json object value are taken over by the Json "Object."
 */
void json_object_donate(JsonObject *this, char **p_key, Json **p_j);

/* Get a pointer to the object in the Json "Object" (i.e. dictionary) with
 * specified key. Returns null if the key does not appear in the "Object."
 */
void *json_object_get(JsonObject *this, char *key);

/* Get the cardinality of a Json "Object" (i.e. dictionary).
 */
ulong json_object_cardinality(JsonObject *this);

/* Write a Json "Object" (i.e. dictionary) to a File.
 */
void json_object_write_file(JsonObject *this, File *file, bool pretty, byte i);

/* Constructor for a JsonObjectPairs collection object of all of the key-value
 * pairs in a Json "Object" (i.e. dictionary), ordered in chronological order
 * of being donated to the "Object." Note that this does not take over
 * ownership of the objects in the "Object."
 */
void json_object_pairs_new(JsonObjectPairs **p_this, JsonObject *j);

/* Destructor for a JsonObjectPairs collection object.
 */
void json_object_pairs_delete(JsonObjectPairs **p_this);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _H */
