#include "KeyValueArray.h"

static KeyValueArray *new_(KeyValueArray **p_this,
    KeyValueInfo *info,
    ulong max_num_pairs,
    void *data);

static void trim_(KeyValueArray *this);

static void check_finalized_(KeyValueArray *this, char *action);


void key_value_array_new(
    KeyValueArray **p_this,
    KeyValueInfo *info,
    ulong max_num_pairs) {
  new_(p_this, info, max_num_pairs, null);
}

void key_value_array_new_fixed(
    KeyValueArray **p_this,
    KeyValueInfo *info,
    ulong num_pairs) {
  key_value_array_new(p_this, info, num_pairs);
  key_value_array_set_full(*p_this);
}

void key_value_array_new_copy(KeyValueArray **p_this, KeyValueArray *other) {
  KeyValueArray *this;

  if (!other->finalized) {
    throw("Cannot copy an unfinalized KeyValueArray");
  }

  this = new_(p_this, other->_info, other->numPairs, null);

  this->numPairs = other->numPairs;
  this->finalized = true;
  copy_memory(other->data, this->data, this->numPairs * this->pairWidth);
}

void key_value_array_new_load(KeyValueArray **p_this, File *f) {
  ulong max_num_pairs;
  KeyValueInfo *_info = null;
  KeyValueArray *this;

  load_label("[KeyValueArray]", f);
  load_value(max_num_pairs, f);
  load(_info, key_value_info, f);
  key_value_array_new(p_this, _info, max_num_pairs);
  this = *p_this;
  key_value_info_delete(&_info);
  load_values(this->_data, this->maxNumPairs * this->pairWidth, f);
  load_label("[/KeyValueArray]", f);

  key_value_array_set_full(this);
}

void key_value_array_new_hyperslice(
    KeyValueArray **p_this,
    KeyValueArray *other,
    ulong pair_offset,
    ulong num_pairs) {
  if (!other->finalized) {
    throw("Cannot create hyperslice of unfinalized KeyValueArray");
  }

  if (pair_offset + num_pairs > other->numPairs) {
    throw("Hyperslice pair offset %lu with number of pairs %lu "
        "exceeds number of pairs in KeyValueArray (%lu)",
        pair_offset,
        num_pairs,
        other->numPairs);
  }

  new_(p_this,
      other->_info,
      num_pairs,
      other->data + other->pairWidth * pair_offset);
}

void key_value_array_new_transform(
    KeyValueArray **p_this,
    KeyValueArray *other,
    KeyValueInfo *info_new,
    void (*transform)(
      KeyValuePair *pair_old,
      KeyValueInfo *info_old,
      KeyValuePair *pair_new,
      KeyValueInfo *info_new,
      void *x),
    void *x) {
  bool finalized;
  ulong num_pairs;
  ulong u;
  KeyValuePair *pair_new;
  KeyValuePair *pair_old;
  KeyValueInfo *info_old;
  KeyValueArray *this;

  finalized = other->finalized;
  num_pairs = other->numPairs;

  new_(p_this, info_new, num_pairs, null);
  this = *p_this;

  this->numPairs = num_pairs;
  info_old = other->_info;

  for (u = 0; u < num_pairs; u++) {
    pair_old = key_value_array_get(other, u);
    pair_new = key_value_array_get(this, u);
    transform(pair_old, info_old, pair_new, info_new, x);
  }

  if (finalized) {
    key_value_array_finalize(this);
  }
}

void key_value_array_delete(KeyValueArray **p_this) {
  ulong u;
  KeyValueInfo *info;
  KeyValuePair *pair;
  KeyValueArray *this;

  if ((this = *p_this) == null) {
    return;
  }

  if (!this->hyperslice) {
    info = this->_info;

    if (info->destroy != null) {
      for (u = 0; u < this->numPairs; u++) {
        pair = key_value_array_get(this, u);
        info->destroy(pair, info);
      }
    }

    deletev(&this->_data);
  }

  key_value_info_delete(&this->_info);

  deleteo(p_this);
}

void key_value_array_print(KeyValueArray *this, Printer *p) {
  ulong u;
  KeyValuePairPrintInfo kvppi;

  sprint(p, "KeyValueArray");
  rprint(p,
      "Maximum number of key-value pairs: %lu",
      this->maxNumPairs);
  rprint(p,
      "Bytes per key-value pair: %lu",
      this->pairWidth);
  rprint(p,
      "Current number of key-value pairs: %lu",
      this->numPairs);
  rprint(p, "Finalized: %s", bool_string(this->finalized));
  rprint(p, "Hyperslice: %s", bool_string(this->hyperslice));

  if (this->numPairs > 0) {
    sprint(p, "Contents");
    kvppi.info = this->_info;

    for (u = 0; u < this->numPairs; u++) {
      sprint(p, "Pair %lu", u);
      kvppi.this = key_value_array_get(this, u);
      key_value_pair_print(&kvppi, p);
      fprint(p);
    }

    fprint(p);
  }

  fprint(p);
}

void key_value_array_save(KeyValueArray *this, File *f) {
  if (this->hyperslice) {
    throw("Cannot save a hyperslice array");
  }

  if (!this->finalized) {
    throw("Only finalized arrays can be saved");
  }

  save_label("[KeyValueArray]", f);
  save_value(this->maxNumPairs, f);
  save(this->_info, key_value_info, f);
  save_values(this->_data, this->maxNumPairs * this->pairWidth, f);
  save_label("[/KeyValueArray]", f);
}

bool key_value_array_equals(KeyValueArray *this, KeyValueArray *other) {
  if (this->hyperslice || other->hyperslice) {
    throw("Cannot compare hyperslice arrays");
  }

  if (!this->finalized || !other->finalized) {
    throw("Only finalized arrays can be compared");
  }
  if (other->maxNumPairs != this->maxNumPairs) {
    return false;
  }

  if (!key_value_info_equals(other->_info, this->_info)) {
    return false;
  }

  return memory_equals(other->_data, this->_data, this->maxNumPairs);
}

void key_value_array_donate(KeyValueArray *this, KeyValuePair *pair) {
  ulong num_pairs;
  ulong pair_width;

  num_pairs = this->numPairs;
  pair_width = this->pairWidth;

  if (num_pairs == this->maxNumPairs) {
    throw("Attempt to donate to a KeyValueArray that is full");
  }

  key_value_pair_move(
      pair,
      this->data + num_pairs * pair_width,
      this->_info);

  this->numPairs++;
}

void key_value_array_set_full(KeyValueArray *this) {
  this->numPairs = this->maxNumPairs;
  this->finalized = true;
}

bool key_value_array_is_full(KeyValueArray *this) {
  return this->numPairs == this->maxNumPairs;
}

KeyValuePair *key_value_array_get(KeyValueArray *this, ulong index) {
  if (index >= this->numPairs) {
    if (this->numPairs == 0) {
      throw("Attempt to get element %lu from empty KeyValueArray", index);
    } else {
      throw("Attempt to get element %lu from KeyValueArray (max %lu)",
          index,
          this->numPairs - 1);
    }
  }

  return this->data + this->pairWidth * index;
}

ulong key_value_array_get_key(KeyValueArray *this, ulong index, ulong k) {
  return key_value_pair_get_key(
      key_value_array_get(this, index),
      this->_info,
      k);
}

void key_value_array_finalize(KeyValueArray *this) {
  bool hyperslice;
  byte *out;
  byte *in;
  byte *end;
  sbyte c;
  ulong num_pairs;
  ulong pair_width;
  KeyValueInfo *info;

  if (this->finalized) {
    return;
  }

  pair_width = this->pairWidth;

  if (pair_width == 0) {
    this->finalized = true;

    return;
  }

  hyperslice = this->hyperslice;
  num_pairs = this->numPairs;
  info = this->_info;

  qsortxn(this->data,
      num_pairs,
      pair_width,
      (sbyte (*)(void *, void *, void *))info->compare,
      false,
      info);

  for (out = this->data,
      in = out + pair_width,
      end = out + num_pairs * pair_width;
      in < end;
      in += pair_width) {
    if (info->getInvalid != null && info->getInvalid(in, info)) {
      if (info->destroy != null) {
        info->destroy(in, info);
      }

      clear_memory(in, pair_width);
      num_pairs--;

      continue;
    }

    if (info->getInvalid != null && info->getInvalid(out, info)) {
      if (info->destroy != null) {
        info->destroy(out, info);
      }

      num_pairs--;
      move_memory(in, out, pair_width);
    } else if ((c = info->compare(in, out, info)) > 0) {
      out += pair_width;

      if (out != in) {
        move_memory(in, out, pair_width);
      }
    } else if (c == 0) {
      num_pairs--;

      if (info->merge != null) {
        info->merge(out, in, info);
        clear_memory(in, pair_width);
      }
    } else {
      throw("Impossible: c = %d", c);
    }
  }

  if (info->getInvalid != null && info->getInvalid(this->data, info)) {
    if (num_pairs != 1) {
      throw("Impossible: Expected algorithm to give num_pairs == 1 "
          "for no valid records");
    }

    if (info->destroy != null) {
      info->destroy(this->data, info);
    }

    clear_memory(this->data, pair_width);

    num_pairs = 0;
  }

  this->numPairs = num_pairs;

  if (!hyperslice) {
    trim_(this);
  }

  this->finalized = true;
}

void key_value_array_refinalize(KeyValueArray *this) {
  this->finalized = false;
  key_value_array_finalize(this);
}

ulong key_value_array_cardinality(KeyValueArray *this) {
  check_finalized_(this, "get the cardinality of");

  return this->numPairs;
}

bool key_value_array_is_empty(KeyValueArray *this) {
  return key_value_array_cardinality(this) == 0;
}

KeyValuePair *key_value_array_minimum(KeyValueArray *this) {
  check_finalized_(this, "get the minimum of");

  if (this->numPairs == 0) {
    return null;
  }

  return key_value_array_get(this, 0);
}

KeyValuePair *key_value_array_maximum(KeyValueArray *this) {
  check_finalized_(this, "get the maximum of");

  if (this->numPairs == 0) {
    return null;
  }

  return key_value_array_get(this, this->numPairs - 1);
}

KeyValuePair *key_value_array_find(KeyValueArray *this, KeyValuePair *pair) {
  check_finalized_(this, "find in");

  if (this->numPairs == 0) {
    return null;
  }

  return bsearchxn(pair,
      this->data,
      this->numPairs,
      this->pairWidth,
      (sbyte (*)(void *, void *, void *))this->_info->compare,
      this->_info);
}

ulong key_value_array_find_index(KeyValueArray *this, KeyValuePair *pair) {
  char *_e = null;
  ulong index;
  KeyValuePairPrintInfo kvppi;

  if (!key_value_array_try_find_index(this, pair, &index)) {
    kvppi.this = pair;
    kvppi.info = this->_info;
    aprint(&_e, 1, key_value_pair, &kvppi);
    string_renew_prepend(&_e, "Could not find the pair in the array:\n");

    throw(_e);
  }

  return index;
}

bool key_value_array_try_find_index(
    KeyValueArray *this,
    KeyValuePair *pair,
    ulong *index) {
  KeyValuePair *pair_found;

  pair_found = key_value_array_find(this, pair);

  if (pair_found == null) {
    return false;
  }

  *index = (pair_found - this->data) / this->pairWidth;

  return true;
}

bool key_value_array_try_merge_pair(KeyValueArray *this, KeyValuePair *pair) {
  KeyValuePair *field;
  KeyValueInfo *info;

  check_finalized_(this, "merge into");

  if ((field = key_value_array_find(this, pair)) == null) {
    return false;
  }

  info = this->_info;

  if (info->merge != null) {
    info->merge(field, pair, info);
  }

  if (info->destroy != null) {
    clear_memory(pair, this->pairWidth);
  }

  return true;
}

void key_value_array_merge(KeyValueArray *this, KeyValueArray *other) {
  bool ia;
  bool ra;
  bool rb;
  byte *a;
  byte *ae;
  byte *b;
  byte *be;
  byte *i;
  byte *lo;
  byte *new_data;
  byte *o;
  byte *other_data;
  byte *_old_data = null;
  sbyte c;
  ulong new_max_num_pairs;
  ulong old_num_pairs;
  ulong other_num_pairs;
  ulong pair_width;
  KeyValueInfo *info;

  if (this->hyperslice) {
    throw("Cannot merge into a hyperslice");
  }

  if (!this->finalized || !other->finalized) {
    throw("Can only merge finalized KeyValueArrays");
  }

  info = this->_info;

  if (!key_value_info_equals(info, other->_info)) {
    throw("Cannot merge KeyValueArrays with different KeyValueInfos");
  }

  pair_width = this->pairWidth;

  old_num_pairs = this->numPairs;

  if (this->maxNumPairs != old_num_pairs) {
    throw("Impossible");
  }

  other_num_pairs = other->numPairs;

  if (other->maxNumPairs != other_num_pairs) {
    throw("Impossible");
  }

  /* Extract the data array from this object.
   */
  donate(&this->_data, &_old_data);
  this->data = null;

  /* Compute and set the new maximum number of pairs as the total number of
   * pairs in the two inputs. This will be trimmed later if there are any
   * merges.
   */
  this->maxNumPairs = new_max_num_pairs = old_num_pairs + other_num_pairs;

  /* Create the new data array for this object.
   */
  new(&this->_data, new_max_num_pairs * pair_width);
  this->data = this->_data;

  /* Extract pointers to the other two data arrays.
   */
  new_data = this->_data;
  other_data = other->_data;

  /* The pointer 'a' walks through the old data, and ends when it gets to 'ae'.
   */
  a = _old_data;
  ae = a + old_num_pairs * pair_width;

  /* Likewise, the pointer 'b' walks through the other data, and ends when it
   * gets to 'be'.
   */
  b = other_data;
  be = b + other_num_pairs * pair_width;

  /* The pointer 'o' points to the current position in the output (new) array.
   */
  o = new_data;

  /* The pointer 'lo' points to the last position in the new data array that we
   * wrote to. Initially it is null.
   */
  lo = null;

  /* The Boolean 'ra' keeps track of whether 'a' is still 'running,' i.e. it
   * has not hit the end of the old data, and 'ra' likewise for 'b' in the
   * other data. We continue while either is still valid.
   */
  for (ra = a < ae, rb = b < be; ra || rb;) {
    /* If both 'a' and 'b' are still running, then we need to compare them to
     * see which one compares less. If that's 'a', then set 'ia' to true, which
     * means that the input is 'a', and we set 'i', the input pointer, to be
     * 'a'; otherwise, we set 'ia' to false, and set 'i' to be 'b'. If only one
     * of 'ra' or 'rb' is true, then we just use whichever is still running.
     */
    if (ra && rb) {
      if (info->compare(a, b, info) <= 0) {
        ia = true;
        i = a;
      } else {
        ia = false;
        i = b;
      }
    } else if (ra) {
      ia = true;
      i = a;
    } else {
      ia = false;
      i = b;
    }

    /* We now compare the selected input, 'i', to the last output we wrote,
     * 'lo'. If there was no previous 'lo', then we set the result to be -1,
     * which ensures that we donate 'i' rather than trying to merge it with
     * 'lo'.
     */
    c = lo == null ? -1 : info->compare(lo, i, info);

    /* If the input arrays were finalized, then we should never encounter a new
     * input that is less than the last written output.
     */
    if (c > 0) {
      throw("Impossible: Not finalized");
    }

    /* If the last output 'lo' compares less than the input 'i', then we simply
     * donate from 'i' to 'o'. We set 'lo' to 'o' and increment 'o' to the next
     * pair.
     *
     * Otherwise, if 'lo' and 'i' compare equal, then we need to merge 'i' into
     * 'lo'. If the merge() method is not null, we call it. Otherwise, if the
     * merge() method is null, then there is nothing to do for 'o'. If the
     * destroy() method is not null, we need to call it on 'i'.
     */
    if (c < 0) {
      key_value_pair_move(i, o, info);
      lo = o;
      o += pair_width;
    } else {
      if (info->merge != null) {
        info->merge(lo, i, info);
      } else {
        if (info->destroy != null) {
          info->destroy(i, info);
        }
      }
    }

    /* Clear the input pair, if there is a destructor.
     */
    if (info->destroy != null) {
      clear_memory(i, pair_width);
    }

    /* If the input was 'a', i.e. the old data from 'this', then we increment
     * 'a' to the next pair. Otherwise, the input was 'b', i.e. from 'other'.
     */
    if (ia) {
      a += pair_width;
      ra = a < ae;
    } else {
      b += pair_width;
      rb = b < be;
    }
  }

  /* Delete the old data array.
   */
  deletev(&_old_data);

  /* Compute the new number of pairs in the output, explicitly.
   */
  this->numPairs = (o - new_data) / pair_width;

  /* If we did any merges, then the array is not full, so trim it.
   */
  trim_(this);

  /* We have now ensured that the output is finalized, so we can just set this
   * flag to true.
   */
  this->finalized = true;

  /* We dealt with all pairs in 'other', either by donating them to the new
   * array in 'this' or by destroying them in the merge process. We cleared all
   * of them, just to ensure that no dangling pointers would be left. Set the
   * number of pairs in 'other' to be zero.
   */
  other->numPairs = 0;
}

void key_value_array_transform(
    KeyValueArray *this,
    void (*transform)(
      KeyValuePair *pair_old,
      KeyValuePair *pair_new,
      KeyValueInfo *info,
      void *x),
    void *x) {
  bool finalized;
  ulong num_pairs;
  ulong u;
  KeyValuePair *pair_new;
  KeyValuePair *_pair_old = null;
  KeyValueInfo *info;

  finalized = this->finalized;
  this->finalized = false;
  num_pairs = this->numPairs;
  info = this->_info;

  key_value_pair_new(&_pair_old, info);

  for (u = 0; u < num_pairs; u++) {
    pair_new = key_value_array_get(this, u);
    key_value_pair_move(pair_new, _pair_old, info);
    transform(_pair_old, pair_new, info, x);

    if (info->destroy != null) {
      info->destroy(_pair_old, info);
    }
  }

  if (info->destroy != null) {
    clear_memory(_pair_old, this->pairWidth);
  }

  key_value_pair_delete(&_pair_old, info);

  if (finalized) {
    key_value_array_finalize(this);
  }
}

void key_value_array_verify_finalized(KeyValueArray *this) {
  sbyte c;
  ulong pair_width;
  KeyValueInfo *info;
  KeyValuePair *pair;
  KeyValuePair *pairn;
  KeyValuePair *paire;

  if (!this->finalized) {
    throw("KeyValueArray is not flagged as finalized");
  }

  pair_width = this->pairWidth;
  info = this->_info;

  for (pair = this->data, pairn = pair + pair_width,
      paire = pair + pair_width * this->numPairs;
      pairn < paire;
      pair = pairn, pairn += pair_width) {
    if ((c = info->compare(pair, pairn, info)) >= 0) {
      throw(c > 0 ? "Keys out of order" : "Duplicated key");
    }
  }
}


static KeyValueArray *new_(KeyValueArray **p_this,
    KeyValueInfo *info,
    ulong max_num_pairs,
    void *data) {
  KeyValueArray *this;

  this = newo(p_this);

  this->maxNumPairs = max_num_pairs;
  key_value_info_new_copy(&this->_info, info);

  this->pairWidth = info->numBytes;

  if (data != null) {
    this->data = data;
    this->hyperslice = true;
    this->numPairs = max_num_pairs;
  } else {
    new(&this->_data, max_num_pairs * this->pairWidth);
    this->data = this->_data;
  }

  return this;
}

static void trim_(KeyValueArray *this) {
  if (this->hyperslice) {
    throw("Cannot trim a hyperslice");
  }

  if (this->numPairs == this->maxNumPairs) {
    return;
  }

  renewv(&this->_data,
      this->maxNumPairs * this->pairWidth,
      this->numPairs * this->pairWidth);
  this->data = this->_data;
  this->maxNumPairs = this->numPairs;
}

static void check_finalized_(KeyValueArray *this, char *action) {
  if (!this->finalized) {
    throw("Cannot %s an unfinalized KeyValueArray", action);
  }
}
