/* Object class for arrays of KeyValuePairs. Each array is created with a fixed
 * maximum number of pairs, with storage for those allocated, and the current
 * position pointing to the start of that memory. Pairs donated to the array
 * are appended within that allocated memory. When all desired pairs have been
 * donated, the array can be "finalized", i.e. sorted, merged, and trimmed: all
 * the pairs will be sorted; equal keys will be merged according to the merge()
 * function within the KeyValueInfo object; and the array will be trimmed to
 * the actual size required. Keys can then be found within the array in log
 * time by binary search.
 *
 * The minimum number of pairs per thread for parallel execution is set in this
 * class; some other classes also read this value.
 *
 * An array can also be constructed that is a reference to a contiguous
 * hyperslice of another existing, finalized array.
 */
#ifndef KEY_VALUE_ARRAY_H
#define KEY_VALUE_ARRAY_H

#ifdef __cplusplus
extern "C" {
#endif

#include "KeyValuePair.h"
#include "TextFile.h"
#include "numbers.h"
#include "qsortx.h"

typedef struct {
  bool finalized;
  bool hyperslice;
  byte *data;
  byte *_data;
  ulong maxNumPairs;
  ulong numPairs;
  ulong pairWidth;
  KeyValueInfo *_info;
} KeyValueArray;

/* Constructor.
 */
void key_value_array_new(
    KeyValueArray **p_this,
    KeyValueInfo *info,
    ulong max_num_pairs);

/* Constructor specifying a fixed-sized array.
 */
void key_value_array_new_fixed(
    KeyValueArray **p_this,
    KeyValueInfo *info,
    ulong num_pairs);

/* Copy constructor.
 */
void key_value_array_new_copy(KeyValueArray **p_this, KeyValueArray *other);

/* Load.
 */
void key_value_array_new_load(KeyValueArray **p_this, File *f);

/* Constructor that actually refers to a contiguous hyperslice of another
 * array.
 */
void key_value_array_new_hyperslice(
    KeyValueArray **p_this,
    KeyValueArray *other,
    ulong pair_offset,
    ulong num_pairs);

/* Constructor that transforms an existing array. Generally used when the
 * info object changes; otherwise, key_value_array_transform() can just be
 * used to transform an array in-place.
 *
 * The pass-through 'x' is distinct from that in either the old or new info
 * objects, and can be used to pass through information related specifically
 * to the transformation.
 *
 * If the array was finalized, it is again finalized; otherwise, it is not.
 */
void key_value_array_new_transform(
    KeyValueArray **p_this,
    KeyValueArray *other,
    KeyValueInfo *info_new,
    void (*transform)(
      KeyValuePair *pair_old,
      KeyValueInfo *info_old,
      KeyValuePair *pair_new,
      KeyValueInfo *info_new,
      void *x),
    void *x);

/* Destructor.
 */
void key_value_array_delete(KeyValueArray **p_this);

/* Print.
 */
void key_value_array_print(KeyValueArray *this, Printer *p);

/* Save.
 */
void key_value_array_save(KeyValueArray *this, File *f);

/* Equals.
 */
bool key_value_array_equals(KeyValueArray *this, KeyValueArray *other);

/* Donate a pair to the array. The donated pair is appended to the end of the
 * pairs currently in the array. If objects are stored, all fields of 'pair'
 * are zeroed, since they have been donated to the array. For pairs not
 * containing objects, the more intuitive method name key_value_array_add() can
 * be used, but it is the same method.
 */
void key_value_array_donate(KeyValueArray *this, KeyValuePair *pair);
#define key_value_array_add key_value_array_donate

/* Specify that the array is full. Used when a fixed-size array is being
 * preallocated like a regular array.
 */
void key_value_array_set_full(KeyValueArray *this);

/* Whether the array is full.
 */
bool key_value_array_is_full(KeyValueArray *this);

/* Get the pointer to a pair from its index, which is a simple arithmetic
 * calculation. Throws if the index is not within bound for the number of pairs
 * actually in the array.
 */
KeyValuePair *key_value_array_get(KeyValueArray *this, ulong index);

/* Get the value of the k-th key of the index-th pair.
 */
ulong key_value_array_get_key(KeyValueArray *this, ulong index, ulong k);

/* Finalize the array, by sorting all pairs, merging any pairs that compare
 * equal, and trimming the result. Returns immediately if already finalized.
 */
void key_value_array_finalize(KeyValueArray *this);

/* Refinalize the array, by marking it as not finalized and then finalizing it
 * again. This operation is appropriate if the compare() method depends on a
 * passthrough parameter that has been changed.
 */
void key_value_array_refinalize(KeyValueArray *this);

/* Get the cardinality of the array. Throws if not finalized.
 */
ulong key_value_array_cardinality(KeyValueArray *this);

/* Whether the array is empty. Throws if not finalized.
 */
bool key_value_array_is_empty(KeyValueArray *this);

/* Get the pointer to the first (minimum) pair in the sorted array, or null if
 * the array is empty. Throws if the array is not finalized.
 */
KeyValuePair *key_value_array_minimum(KeyValueArray *this);

/* Get the pointer to the last (maximum) pair in the sorted array, or null if
 * the array is empty. Throws if the array is not finalized.
 */
KeyValuePair *key_value_array_maximum(KeyValueArray *this);

/* Find the pair with the key of the provided pair by binary search, returning
 * the pointer to it, or null if the key is not present. Throws if the array is
 * not finalized. The value of 'pair' is ignored.
 */
KeyValuePair *key_value_array_find(KeyValueArray *this, KeyValuePair *pair);

/* Find the index of the pair with the given key. Throws if the array is not
 * finalized.
 */
ulong key_value_array_find_index(KeyValueArray *this, KeyValuePair *pair);

/* Try to find the index of the pair with the given key. Throws if the array is
 * not finalized.
 */
bool key_value_array_try_find_index(
    KeyValueArray *this,
    KeyValuePair *pair,
    ulong *index);

/* Try to merge a pair into a finalized array. If the key of 'pair' is found in
 * the array, then 'pair' is donated to the array, and merged in, and the
 * function returns true; otherwise, the function returns false, and 'pair' is
 * not donated. Throws if the array is not finalized.
 */
bool key_value_array_try_merge_pair(KeyValueArray *this, KeyValuePair *pair);

/* Merge another array into this array. Both must be finalized. This operates
 * by creating a data array whose length is the sum of the two arrays,
 * merge-sorts the two sets of data into this new array, trims it, destroys the
 * original data array for this array, and replaces it by the merged data
 * array.
 */
void key_value_array_merge(KeyValueArray *this, KeyValueArray *other);

/* Transform an array in-place, which is only possible if the info object is
 * not changing.
 *
 * The 'transform()' method does not need to worry that it is transforming
 * in-place: the 'pair_old' passed to it is actually a copy, so it can set
 * values in 'pair_new' freely. Any objects owned by the old pair are
 * automatically destroyed.
 *
 * The pass-through 'x' is distinct from that in either the old or new info
 * objects, and can be used to pass through information related specifically
 * to the transformation.
 *
 * If the array was finalized, it is again finalized; otherwise, it is not.
 *
 * This method also works on hyperslices.
 */
void key_value_array_transform(
    KeyValueArray *this,
    void (*transform)(
      KeyValuePair *pair_old,
      KeyValuePair *pair_new,
      KeyValueInfo *info,
      void *x),
    void *x);

/* Verify that the array is properly finalized, i.e. sorted, deduplicated, and
 * trimmed.
 */
void key_value_array_verify_finalized(KeyValueArray *this);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _H */
