#include "KeyValueCollection.h"

#define DEFAULT_CHUNK_SIZE_ 64

static ulong chunk_size_ = DEFAULT_CHUNK_SIZE_;

static bool try_merge_sorted_(KeyValueCollection *this, KeyValuePair *pair);

static void add_to_unsorted_(KeyValueCollection *this, KeyValuePair *pair);

static void add_to_tree_(KeyValueCollection *this, KeyValueArray **p_chunk);


void key_value_collection_new(
    KeyValueCollection **p_this,
    KeyValueInfo *info) {
  KeyValueCollection *this;

  this = newo(p_this);

  key_value_info_new_copy(&this->_info, info);
  this->chunkSize = chunk_size_;
}

void key_value_collection_new_from_array(
    KeyValueCollection **p_this,
    KeyValueArray **p_array) {
  KeyValueCollection *this;

  key_value_collection_new(p_this, (*p_array)->_info);
  this = *p_this;

  add_to_tree_(this, p_array);
}

void key_value_collection_delete(KeyValueCollection **p_this) {
  KeyValueCollection *this;

  if ((this = *p_this) == null) {
    return;
  }

  key_value_node_delete(&this->_root);
  key_value_array_delete(&this->_chunk);
  key_value_info_delete(&this->_info);

  deleteo(p_this);
}

void key_value_collection_print(KeyValueCollection *this, Printer *p) {
  sprint(p, "KeyValueCollection");
  rprint(p, "Chunk size: %lu", this->chunkSize);
  rprint(p, "Level: %u", this->level);
  rprint(p, "Next index: %lu", this->next);

  if (this->_chunk != null) {
    sprint(p, "Chunk");
    key_value_array_print(this->_chunk, p);
    fprint(p);
  }

  if (this->_root != null) {
    sprint(p, "Tree");
    key_value_node_print(this->_root, p);
    fprint(p);
  }

  fprint(p);
}

ulong key_value_collection_set_chunk_size(ulong chunk_size) {
  ulong current;

  if (chunk_size == 0) {
    throw("Chunk size cannot be zero");
  }

  current = chunk_size_;
  chunk_size_ = chunk_size;
  return current;
}

ulong key_value_collection_set_default_chunk_size(void) {
  return key_value_collection_set_chunk_size(DEFAULT_CHUNK_SIZE_);
}

ulong key_value_collection_get_chunk_size(void) {
  return chunk_size_;
}

bool key_value_collection_donate(
    KeyValueCollection *this,
    KeyValuePair *pair) {
  if (try_merge_sorted_(this, pair)) {
    return true;
  }

  add_to_unsorted_(this, pair);

  return false;
}

void key_value_array_new_from_collection(
    KeyValueArray **p_this,
    KeyValueCollection **p_collection) {
  KeyValueCollection *collection;
  KeyValueArray *this;

  collection = *p_collection;

  if (collection->_chunk != null) {
    add_to_tree_(collection, &collection->_chunk);
  }

  if (collection->_root == null) {
    key_value_array_new(p_this, collection->_info, 0);
    this = *p_this;
    this->finalized = true;
    key_value_collection_delete(p_collection);

    return;
  }

  key_value_node_force_merge_children(collection->_root);
  donate(&collection->_root->_array, p_this);
  key_value_collection_delete(p_collection);

  if (p_this == null) {
    throw("Impossible: Force merge children didn't create array");
  }

  this = *p_this;

  if (this->numPairs != this->maxNumPairs) {
    throw("Impossible: %lu v %lu", this->numPairs, this->maxNumPairs);
  }
}


static bool try_merge_sorted_(KeyValueCollection *this, KeyValuePair *pair) {
  if (this->_root == null) {
    return false;
  }

  return key_value_node_try_merge_pair(this->_root, pair);
}

static void add_to_unsorted_(
    KeyValueCollection *this,
    KeyValuePair *pair) {
  if (this->_chunk == null) {
    key_value_array_new(
        &this->_chunk,
        this->_info,
        this->chunkSize);
  }

  key_value_array_donate(this->_chunk, pair);

  if (key_value_array_is_full(this->_chunk)) {
    add_to_tree_(this, &this->_chunk);
  }
}

static void add_to_tree_(KeyValueCollection *this, KeyValueArray **p_chunk) {
  KeyValueNode *_new_root = null;

  key_value_array_finalize(*p_chunk);

  if (this->_root == null) {
    key_value_node_new(&this->_root, null);
  }

  key_value_node_add_chunk(
      this->_root,
      p_chunk,
      this->level,
      this->next);

  if ((this->next & (this->next + 1)) == 0) {
    this->level++;
    key_value_node_new(&_new_root, null);
    _new_root->_left = this->_root;
    this->_root->parent = _new_root;
    this->_root = _new_root;
  }

  this->next++;
}
