/* Object class for building a collection of key-value pairs. This class has
 * been optimized primarily for memory efficiency (to be able to store as many
 * pairs in available memory as possible), then insertion time, and finally
 * retrieval time. The designed use case is for the collection to be initially
 * populated in a batch run, where keys are expected to be updated repeatedly
 * during this process (e.g. creating a set or frequency table of keys). After
 * the collection is fully populated, it is extracted into a KeyValueArray.
 *
 * The efficiency is guaranteed to be no worse than O(n log^2 n) for high
 * cardinality, and more efficient for low cardinality. Donated pairs are
 * donated into a "chunk" KeyValueArray, which when full is sorted and merged,
 * and added to the successive levels of a binary tree, left to right along
 * each level, where the left children are always merged. It can be shown that
 * this behavior ensures the O(n log^2 n) worst-case efficiency.
 */
#ifndef KEY_VALUE_COLLECTION_H
#define KEY_VALUE_COLLECTION_H

#ifdef __cplusplus
extern "C" {
#endif

#include "KeyValueArray.h"
#include "KeyValueNode.h"

typedef struct {
  byte level;
  ulong chunkSize;
  ulong next;
  KeyValueInfo *_info;
  KeyValueArray *_chunk;
  KeyValueNode *_root;
} KeyValueCollection;

/* Constructor.
 */
void key_value_collection_new(
    KeyValueCollection **p_this,
    KeyValueInfo *info);

/* Constructor, initializing from an existing array, which is taken over.
 */
void key_value_collection_new_from_array(
    KeyValueCollection **p_this,
    KeyValueArray **p_array);

/* Destructor.
 */
void key_value_collection_delete(KeyValueCollection **p_this);

/* Print.
 */
void key_value_collection_print(KeyValueCollection *this, Printer *p);

/* Set the chunk size to use. Returns the previous value.
 */
ulong key_value_collection_set_chunk_size(ulong chunk_size);

/* Set the default chunk size. Returns the previous value.
 */
ulong key_value_collection_set_default_chunk_size(void);

/* Get the current chunk size.
 */
ulong key_value_collection_get_chunk_size(void);

/* Donate a pair to the collection. Returns true if it was able to be merged in
 * with an existing sorted pair, or false otherwise (i.e. was donated to the
 * unsorted chunk). For pairs that do not contain objects, the more intuitive
 * method name key_value_collection_add() can be used, but it is the same
 * method.
 */
bool key_value_collection_donate(KeyValueCollection *this, KeyValuePair *pair);
#define key_value_collection_add key_value_collection_donate

/* Extract the final data of the collection into a new array, and destroy the
 * collection.
 */
void key_value_array_new_from_collection(
    KeyValueArray **p_this,
    KeyValueCollection **p_collection);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _H */
