#include "KeyValueInfo.h"
#include <limits.h>

static sbyte compare_(KeyValuePair *a, KeyValuePair *b, KeyValueInfo *info);


void key_value_info_new(
    KeyValueInfo **p_this,
    ulong num_keys,
    ulong *max_keys,
    ulong num_values,
    ulong *max_values) {
  static const byte m_[] = {
    0x80, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc, 0xfe, 0xff};

  ulong bit_offset;
  ulong f;
  ulong k;
  ulong last_key_bit_offset;
  ulong last_key_bit_offset_plus_1;
  ulong num_fields;
  ulong v;
  ulong *max_fields;
  KeyValueInfo *this;
  KeyValueField *last_key_field;

  if ((num_fields = num_keys + num_values) == 0) {
    throw("Must have at least one key or one value");
  }

  this = newo(p_this);

  this->numFields = num_fields;
  this->numKeys = num_keys;
  this->numValues = num_values;

  new(&this->_maxFields, num_fields);
  max_fields = this->_maxFields;
  this->maxKeys = this->_maxFields;
  this->maxValues = this->_maxFields + num_keys;

  for (k = 0; k < num_keys; k++) {
    this->maxKeys[k] = max_keys[k];
  }

  for (v = 0; v < num_values; v++) {
    this->maxValues[v] = max_values[v];
  }

  new(&this->_fields, this->numFields);
  this->keyKvfs = this->_fields;
  this->valueKvfs = this->_fields + num_keys;

  bit_offset = 0;

  for (f = 0; f < num_fields; f++) {
    key_value_field_new(&this->_fields[f], max_fields[f], &bit_offset);
  }

  this->numBits = bit_offset;
  this->numBytes = (bit_offset + 7) / 8;

  if (num_keys > 0) {
    last_key_field = this->keyKvfs[num_keys - 1];

    if ((last_key_bit_offset_plus_1
          = last_key_field->bitOffset + last_key_field->bitWidth) > 0) {
      last_key_bit_offset = last_key_bit_offset_plus_1 - 1;
      this->lastKeyByte = last_key_bit_offset / 8;
      this->lastKeyByteMask = m_[last_key_bit_offset % 8];
    }
  }

  this->compare = compare_;
}

void key_value_info_new_copy(
    KeyValueInfo **p_this,
    KeyValueInfo *other) {
  KeyValueInfo *this;

  key_value_info_new(
      p_this,
      other->numKeys,
      other->maxKeys,
      other->numValues,
      other->maxValues);
  this = *p_this;
  this->x = other->x;
  this->destroy = other->destroy;
  this->print = other->print;
  this->merge = other->merge;
  this->compare = other->compare;
  this->getInvalid = other->getInvalid;
  this->setInvalid = other->setInvalid;
}

void key_value_info_new_load(KeyValueInfo **p_this, File *f) {
  ulong num_keys;
  ulong num_values;
  ulong *_max_keys = null;
  ulong *_max_values = null;

  load_label("[KeyValueInfo]", f);
  load_value(num_keys, f);
  load_new_values(_max_keys, num_keys, f);
  load_value(num_values, f);
  load_new_values(_max_values, num_values, f);
  load_label("[/KeyValueInfo]", f);

  key_value_info_new(
      p_this,
      num_keys,
      _max_keys,
      num_values,
      _max_values);

  deletev(&_max_values);
  deletev(&_max_keys);
}

void key_value_info_delete(KeyValueInfo **p_this) {
  KeyValueInfo *this;

  if ((this = *p_this) == null) {
    return;
  }

  delete(&this->_fields, this->numFields, key_value_field);
  deletev(&this->_maxFields);
  deleteo(p_this);
}

void key_value_info_print(KeyValueInfo *this, Printer *p) {
  ulong k;
  ulong v;

  sprint(p, "KeyValueInfo");
  rprint(p, "Number of keys: %lu", this->numKeys);
  rprint(p, "Number of values: %lu", this->numValues);
  rprint(p, "Number of fields: %lu", this->numFields);
  rprint(p, "Number of bits: %lu", this->numBits);
  rprint(p, "Number of bytes: %lu", this->numBytes);

  for (k = 0; k < this->numKeys; k++) {
    sprint(p, "Key %lu", k);
    key_value_field_print(this->keyKvfs[k], p);
    fprint(p);
  }

  for (v = 0; v < this->numValues; v++) {
    sprint(p, "Value %lu", v);
    key_value_field_print(this->valueKvfs[v], p);
    fprint(p);
  }

  rprint(p, "Last key byte = %lu", this->lastKeyByte);
  rprint(p, "Last key byte mask = 0x%02x", this->lastKeyByteMask);
  fprint(p);
}

void key_value_info_save(KeyValueInfo *this, File *f) {
  save_label("[KeyValueInfo]", f);
  save_value(this->numKeys ,f);
  save_values(this->maxKeys, this->numKeys, f);
  save_value(this->numValues, f);
  save_values(this->maxValues, this->numValues, f);
  save_label("[/KeyValueInfo]", f);
}

bool key_value_info_equals(KeyValueInfo *a, KeyValueInfo *b) {
  ulong f;

  if (a == b) {
    return true;
  }

  if (a->numKeys != b->numKeys) {
    return false;
  }

  if (a->numValues != b->numValues) {
    return false;
  }

  for (f = 0; f < a->numFields; f++) {
    if (a->_fields[f]->maxValue != b->_fields[f]->maxValue) {
      return false;
    }
  }

  return true;
}


static sbyte compare_(KeyValuePair *a, KeyValuePair *b, KeyValueInfo *info) {
  byte ba;
  byte bb;
  byte mask;
  byte *e;

  for (e = a + info->lastKeyByte; a < e; a++, b++) {
    ba = *a;
    bb = *b;

    if (ba > bb) {
      return +1;
    }

    if (ba < bb) {
      return -1;
    }
  }

  mask = info->lastKeyByteMask;
  ba = *a & mask;
  bb = *b & mask;

  if (ba > bb) {
    return +1;
  }

  if (ba < bb) {
    return -1;
  }

  return 0;
}
