/* Object class that specifies completely the transformation between a
 * KeyValuePair and its representation as an array of bytes, as well as all
 * methods required for such pairs. Immutable class. This class creates a
 * KeyValueField for each field, which in turn creates a KeyValueByte for each
 * byte in the field.
 *
 * Pairs ostensibly only contain ulong fields, but these can (and are) used in
 * some cases to store pointers to objects, for both key fields as well as
 * value fields. Thus collections of pairs may, in general, be collections of
 * objects, not just values, and must be in possession of the functions
 * necessary for handling such objects.
 */
#ifndef KEY_VALUE_INFO_H
#define KEY_VALUE_INFO_H

#ifdef __cplusplus
extern "C" {
#endif

#include "KeyValueField.h"
#include "TextFile.h"
#include "numbers.h"

typedef byte KeyValuePair;
struct KeyValuePairPrintInfo;

/* The class definition itself. The constructor below only requires the
 * primary information. Other methods are more conveniently and transparently
 * set after calling the constructor. They are:
 *   x:
 *     Passback pointer. This does not need to be passed as an explicit
 *     parameter in the following methods, because the KeyValueInfo itself is
 *     always passed back, which includes x.
 *   destroy():
 *     Destructor for any objects contained in the pair. May be null. As with
 *     all destructors, this method must handle pairs that have been zeroed,
 *     i.e. all pointers null.
 *   print():
 *     Print method for the pair. If null, all keys and values are simply
 *     printed.
 *   merge():
 *     Specifies how to merge pairs that compare equal. May be null, since some
 *     classes do not require merges. Note that this method is called for pairs
 *     for which compare() returns zero, and 'acc' must compare the same after
 *     this method is called than before; this usually means that its keys
 *     cannot be changed.
 *   compare():
 *     Comparator. If null, then a default comparator is used, which compares
 *     all of the key fields, in order. Usually uses all key fields in its
 *     comparison (as that defines the difference between 'key' and 'value'
 *     fields), but in general this is not enforced; beware if you do anything
 *     different.
 *   get_invalid():
 *     Optional method for setting an "invalid" status for a pair.
 *   set_invalid():
 *     Optional method for getting the "invalid" status for a pair.
 */
typedef struct KeyValueInfo {
  byte lastKeyByteMask;
  ulong numKeys;
  ulong numValues;
  ulong numFields;
  ulong numBits;
  ulong numBytes;
  ulong lastKeyByte;
  ulong *_maxFields;
  ulong *maxKeys;
  ulong *maxValues;
  void *x;
  KeyValueField **_fields;
  KeyValueField **keyKvfs;
  KeyValueField **valueKvfs;
  void (*destroy)(KeyValuePair *pair, struct KeyValueInfo *info);
  void (*print)(struct KeyValuePairPrintInfo *kvppi, Printer *p);
  void (*merge)(KeyValuePair *acc,
      KeyValuePair *add,
      struct KeyValueInfo *info);
  sbyte (*compare)(
      KeyValuePair *a,
      KeyValuePair *b,
      struct KeyValueInfo *info);
  bool (*getInvalid)(KeyValuePair *pair, struct KeyValueInfo *info);
  void (*setInvalid)(KeyValuePair *pair, struct KeyValueInfo *info);
} KeyValueInfo;

/* Constructor, with just the primary information, namely, how many key and
 * value fields there are to be, and the maximum possible values of each.
 */
void key_value_info_new(
    KeyValueInfo **p_this,
    ulong num_keys,
    ulong *max_keys,
    ulong num_values,
    ulong *max_values);

/* Copy.
 */
void key_value_info_new_copy(KeyValueInfo **p_this, KeyValueInfo *other);

/* Load.
 */
void key_value_info_new_load(KeyValueInfo **p_this, File *f);

/* Destructor.
 */
void key_value_info_delete(KeyValueInfo **p_this);

/* Print.
 */
void key_value_info_print(KeyValueInfo *this, Printer *p);

/* Save.
 */
void key_value_info_save(KeyValueInfo *this, File *f);

/* Equals.
 */
bool key_value_info_equals(KeyValueInfo *a, KeyValueInfo *b);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _H */
