/* Object class that is a helper class for the KeyValueCollection class. A
 * binary tree made up of this node class is populated to hold the finalized
 * arrays. Nodes are populated left-to-right along each level, where every time
 * a level is filled, a new root node is created and the existing tree moved to
 * be its left child. The guarantee of O(n log^2 n) efficiency is obtained by
 * the following extra step: every time a right child is added, it is merged
 * with its left sibling and the result placed in the parent node, with the
 * children deleted; and so on, recursively, if this parent node is itself a
 * right child. This yields a binary tree where only the left children contain
 * KeyValueArrays. It can be shown that this achieves the promised O(n log^2 n)
 * behavior even in the worst case when all values added are distinct.
 */
#ifndef KEY_VALUE_NODE_H
#define KEY_VALUE_NODE_H

#ifdef __cplusplus
extern "C" {
#endif

#include "KeyValueArray.h"

typedef struct KeyValueNode {
  KeyValueArray *_array;
  struct KeyValueNode *_left;
  struct KeyValueNode *_right;
  struct KeyValueNode *parent;
} KeyValueNode;

/* Constructor. We just need to be supplied the pointer to the parent, or null
 * if this is the collection's root node.
 */
void key_value_node_new(KeyValueNode **p_this, KeyValueNode *parent);

/* Destructor.
 */
void key_value_node_delete(KeyValueNode **p_this);

/* Print.
 */
void key_value_node_print(KeyValueNode *this, Printer *p);

/* Try to merge a pair into the tree rooted on this node. If the key of 'pair'
 * is found in the tree, then 'pair' is donated to the tree, and merged in, and
 * the function returns true; otherwise, the function returns false, and 'pair'
 * is not donated.
 */
bool key_value_node_try_merge_pair(KeyValueNode *this, KeyValuePair *pair);

/* Add a chunk to the tree rooted on this node.
 */
void key_value_node_add_chunk(
    KeyValueNode *this,
    KeyValueArray **_chunk,
    byte level,
    ulong index);

/* Force merging of all of the children of this node. Called when the
 * collection is being finalized and extracted into a single array.
 */
void key_value_node_force_merge_children(KeyValueNode *this);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _H */
