#include "KeyValuePair.h"
#include <limits.h>

static void set_(KeyValuePair *this, KeyValueInfo *info, ulong f, ulong v);

static ulong get_(KeyValuePair *this, KeyValueInfo *info, ulong f);

static void donate_(KeyValuePair *this,
    KeyValueInfo *info,
    ulong f,
    void *p_p_o,
    bool to);

static bool try_donate_(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong f,
    void *p_p_o,
    bool to);

static void check_key_(ulong k, KeyValueInfo *info);

static void check_value_(ulong v, KeyValueInfo *info);

static void check_field_(ulong f, KeyValueInfo *info);


void key_value_pair_new(KeyValuePair **p_this, KeyValueInfo *info) {
  new(p_this, info->numBytes);
}

void key_value_pair_delete(KeyValuePair **p_this, KeyValueInfo *info) {
  KeyValuePair *this;

  if ((this = *p_this) == null) {
    return;
  }

  if (info->destroy != null) {
    info->destroy(this, info);
  }

  deletev(p_this);
}

void key_value_pair_print(KeyValuePairPrintInfo *kvppi, Printer *p) {
  char keystr[32];
  ulong k;
  ulong v;
  KeyValuePair *this;
  KeyValueInfo *info;

  this = kvppi->this;
  info = kvppi->info;

  sprint(p, "KeyValuePair");

  if (info->print != null) {
    info->print(kvppi, p);
  } else {
    if (info->getInvalid != null && info->getInvalid(this, info)) {
      rprint(p, "(Invalid)");
    } else {
      keystr[0] = '\0';

      for (k = 0; k < info->numKeys; k++) {
        if (info->numKeys > 1) {
          sprintf(keystr, " %lu", k);
        }

        rprint(p,
            "Key%s: %lu (max %lu)",
            keystr,
            key_value_pair_get_key(this, info, k),
            info->keyKvfs[k]->maxValue);
      }

      keystr[0] = '\0';

      for (v = 0; v < info->numValues; v++) {
        if (info->numValues > 1) {
          sprintf(keystr, " %lu", v);
        }

        rprint(p,
            "Value%s: %lu (max %lu)",
            keystr,
            key_value_pair_get_value(this, info, v),
            info->valueKvfs[v]->maxValue);
      }
    }
  }

  fprint(p);
}

void key_value_pair_set_key(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong k,
    ulong key) {
  if (k >= info->numKeys) {
    throw("Key index %lu is invalid (max %lu)", k, info->numKeys - 1);
  }

  set_(this, info, k, key);
}

void key_value_pair_donate_to_key(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong k,
    void *p_p_o) {
  check_key_(k, info);
  key_value_pair_donate_to_field(this, info, k, p_p_o);
}

void key_value_pair_reference_to_key(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong k,
    void *p) {
  check_key_(k, info);
  key_value_pair_reference_to_field(this, info, k, p);
}

ulong key_value_pair_get_key(KeyValuePair *this, KeyValueInfo *info, ulong k) {
  if (k >= info->numKeys) {
    throw("Key index %lu is invalid (max %lu)", k, info->numKeys - 1);
  }

  return get_(this, info, k);
}

void key_value_pair_donate_from_key(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong k,
    void *p_p_o) {
  check_key_(k, info);
  key_value_pair_donate_from_field(this, info, k, p_p_o);
}

bool key_value_pair_try_donate_from_key(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong k,
    void *p_p_o) {
  check_key_(k, info);
  return key_value_pair_try_donate_from_field(this, info, k, p_p_o);
}

void *key_value_pair_reference_from_key(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong k) {
  check_key_(k, info);
  return key_value_pair_reference_from_field(this, info, k);
}

void key_value_pair_set_value(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong v,
    ulong value) {
  if (v >= info->numValues) {
    throw("Value index %lu is invalid (max %lu)", v, info->numValues - 1);
  }

  set_(this, info, info->numKeys + v, value);
}

void key_value_pair_donate_to_value(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong v,
    void *p_p_o) {
  check_value_(v, info);
  key_value_pair_donate_to_field(this, info, info->numKeys + v, p_p_o);
}

ulong key_value_pair_get_value(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong v) {
  if (v >= info->numValues) {
    throw("Value index %lu is invalid (max %lu)", v, info->numValues - 1);
  }

  return get_(this, info, info->numKeys + v);
}

void key_value_pair_donate_from_value(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong v,
    void *p_p_o) {
  check_value_(v, info);
  key_value_pair_donate_from_field(this, info, info->numKeys + v, p_p_o);
}

bool key_value_pair_try_donate_from_value(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong v,
    void *p_p_o) {
  check_value_(v, info);
  return key_value_pair_try_donate_from_field(
      this,
      info,
      info->numKeys + v,
      p_p_o);
}

void *key_value_pair_reference_from_value(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong v) {
  check_value_(v, info);
  return key_value_pair_reference_from_field(this, info, info->numKeys + v);
}

void key_value_pair_set_field(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong f,
    ulong value) {
  if (f >= info->numFields) {
    throw("Field index %lu is invalid (max %lu)", f, info->numFields - 1);
  }

  set_(this, info, f, value);
}

void key_value_pair_donate_to_field(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong f,
    void *p_p_o) {
  donate_(this, info, f, p_p_o, true);
}

ulong key_value_pair_get_field(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong f) {
  if (f >= info->numFields) {
    throw("Field index %lu is invalid (max %lu)", f, info->numFields - 1);
  }

  return get_(this, info, f);
}

void key_value_pair_donate_from_field(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong f,
    void *p_p_o) {
  donate_(this, info, f, p_p_o, false);
}

bool key_value_pair_try_donate_from_field(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong f,
    void *p_p_o) {
  return try_donate_(this, info, f, p_p_o, false);
}

void *key_value_pair_reference_from_field(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong f) {
  check_field_(f, info);

  return (void *)key_value_pair_get_field(this, info, f);
}

void key_value_pair_reference_to_field(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong f,
    void *p) {
  check_field_(f, info);

  key_value_pair_set_field(this, info, f, (ulong)p);
}

void key_value_pair_move(
    KeyValuePair *src,
    KeyValuePair *dst,
    KeyValueInfo *info) {
  copy_memory(src, dst, info->numBytes);

  if (info->destroy != null) {
    clear_memory(src, info->numBytes);
  }
}

void key_value_pair_set_invalid(KeyValuePair *this, KeyValueInfo *info) {
  info->setInvalid(this, info);
}

bool key_value_pair_get_invalid(KeyValuePair *this, KeyValueInfo *info) {
  return info->getInvalid(this, info);
}

void key_value_pair_convert(
    KeyValuePair *old_pair,
    KeyValueInfo *old_info,
    KeyValuePair *new_pair,
    KeyValueInfo *new_info) {
  ulong f;
  ulong field;
  ulong num_fields;

  num_fields = old_info->numFields;

  if (new_info->numFields != num_fields) {
    throw("Number of fields wrong (%lu v %lu)",
        num_fields,
        new_info->numFields);
  }

  for (f = 0; f < num_fields; f++) {
    field = key_value_pair_get_field(old_pair, old_info, f);
    key_value_pair_set_field(new_pair, new_info, f, field);
  }
}


static void set_(KeyValuePair *this, KeyValueInfo *info, ulong f, ulong v) {
  byte *b;
  byte *be;
  sbyte shift;
  ulong extracted;
  KeyValueByte *byt;
  KeyValueByte **p_byte;
  KeyValueField *field;

  if (f >= info->numFields) {
    throw("Impossible: Attempt to set field %lu (max %lu)",
        f,
        info->numFields - 1);
  }

  field = info->_fields[f];

  if (v > field->maxValue) {
    throw("Value (%lu) of field %lu greater than maximum allowed (%lu)",
        v,
        f,
        field->maxValue);
  }

  for (b = this + field->firstOffset,
      be = b + field->numBytes,
      p_byte = field->_bytes;
      b < be;
      b++, p_byte++) {
    byt = *p_byte;
    extracted = v & byt->valueMask;

    shift = byt->shift;

    if (shift > 0) {
      extracted >>= shift;
    } else if (shift < 0) {
      extracted <<= -shift;
    }

    *b &= byt->writeByteMask;
    *b |= (byte)extracted;
  }
}

static ulong get_(KeyValuePair *this, KeyValueInfo *info, ulong f) {
  byte *b;
  byte *be;
  sbyte shift;
  ulong value;
  ulong extracted;
  KeyValueByte *byt;
  KeyValueByte **p_byte;
  KeyValueField *field;

  if (f >= info->numFields) {
    throw("Impossible: Attempt to get field %lu (max %lu)",
        f,
        info->numFields - 1);
  }

  field = info->_fields[f];

  value = 0;

  for (b = this + field->firstOffset,
      be = b + field->numBytes,
      p_byte = field->_bytes;
      b < be;
      b++, p_byte++) {
    byt = *p_byte;
    extracted = (ulong)(*b & byt->readByteMask);

    shift = byt->shift;

    if (shift > 0) {
      extracted <<= shift;
    } else if (shift < 0) {
      extracted >>= -shift;
    }

    value |= extracted;
  }

  return value;
}

static void donate_(KeyValuePair *this,
    KeyValueInfo *info,
    ulong f,
    void *p_p_o,
    bool to) {
  if (!try_donate_(this, info, f, p_p_o, to)) {
    throw("Source pointer empty");
  }
}

static bool try_donate_(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong f,
    void *p_p_o,
    bool to) {
  void *p;
  void *q;
  void *src;
  void *dst;

  check_field_(f, info);

  if (p_p_o == null) {
    throw("Pointer to %s pointer null", to ? "source" : "destination");
  }

  p = *(void **)p_p_o;
  q = (void *)key_value_pair_get_field(this, info, f);

  if (to) {
    src = p;
    dst = q;
  } else {
    src = q;
    dst = p;
  }

  if (src == null) {
    return false;
  }

  if (dst != null) {
    throw("Destination pointer not empty");
  }

  if (to) {
    key_value_pair_set_field(this, info, f, (ulong)src);
    *(void **)p_p_o = null;
  } else {
    *(void **)p_p_o = src;
    key_value_pair_set_field(this, info, f, 0);
  }

  return true;
}

static void check_key_(ulong k, KeyValueInfo *info) {
  if (k >= info->numKeys) {
    if (info->numKeys == 0) {
      throw("Invalid: No keys in KeyValueInfo");
    }

    throw("Key index %lu is invalid (max %lu)", k, info->numKeys - 1);
  }
}

static void check_value_(ulong v, KeyValueInfo *info) {
  if (v >= info->numValues) {
    if (info->numValues == 0) {
      throw("Invalid: No values in KeyValueInfo");
    }

    throw("Value index %lu is invalid (max %lu)", v, info->numValues - 1);
  }
}

static void check_field_(ulong f, KeyValueInfo *info) {
  ulong max;

  if (f >= info->numFields) {
    throw("Field index %lu is invalid (max %lu)", f, info->numFields - 1);
  }

  max = info->_maxFields[f];

  if (max != ULONG_MAX) {
    throw("Cannot store a pointer in field %lu (%s %lu) as its maximum value "
        "is %lu: must be ULONG_MAX = %lu",
        f,
        f < info->numKeys ? "key" : "value",
        f < info->numKeys ? f : f - info->numKeys,
        max,
        ULONG_MAX);
  }
}
