/* Pseudo-object class that provides simple but efficient key-value pairs. Each
 * pair contains a fixed, nonzero number of ulong fields, with some number of
 * them together being considered the 'key,' and the remainder the 'value.'
 * Each such field has a fixed bit-width, specified at construction time. Each
 * pair is stored in an integral number of bytes, making random access to any
 * such pair efficient, but the fields within the pair are all bit-packed
 * together.
 *
 * The KeyValuePair class is itself actually just a typedef of byte, with the
 * KeyValueInfo class performing all the magic of mapping these bytes to the
 * fields of each pair. This typedef therefore appears in KeyValueInfo.h.
 * Thus, KeyValuePair is only a pseudo-object class, not actually containing a
 * typedef struct for objects, but rather being closer in nature to arrays (as,
 * indeed, it is just an array of bytes).
 *
 * Although each field can only store a ulong, if its maximum value is
 * specicied as ULONG_MAX, then any 64-bit quantity can be stored in it: pairs
 * can own objects by storing owning pointers (the KeyValueInfo class contains
 * methods for dealing with those owned objects), and other classes store longs
 * and doubles in it, through union converters. (These are all confirmed in
 * jpc_test.c to be 64-bit quantities.)
 */
#ifndef KEY_VALUE_PAIR_H
#define KEY_VALUE_PAIR_H

#ifdef __cplusplus
extern "C" {
#endif

#include "KeyValueInfo.h"

typedef struct KeyValuePairPrintInfo {
  KeyValuePair *this;
  KeyValueInfo *info;
} KeyValuePairPrintInfo;

/* Constructor, based on a KeyValueInfo. Note that, per the above, this
 * actually just creates an array of bytes, rather than a pointer to a struct.
 */
void key_value_pair_new(KeyValuePair **p_this, KeyValueInfo *info);

/* Destructor. Since this is a pseudo-object class like an array, the
 * KeyValueInfo must again be supplies to provide all information required to
 * destroy the array of bytes, and any objects stored in the pair.
 */
void key_value_pair_delete(KeyValuePair **p_this, KeyValueInfo *info);

/* Print.
 */
void key_value_pair_print(KeyValuePairPrintInfo *kvppi, Printer *p);

/* Set the k-th key field to 'key'.
 */
void key_value_pair_set_key(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong k,
    ulong key);

/* Donate a pointer to the k-th key field.
 */
void key_value_pair_donate_to_key(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong k,
    void *p_p_o);

/* Put a reference to the pointer in the k-th key field.
 */
void key_value_pair_reference_to_key(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong k,
    void *p);

/* Donate a pointer from the k-th key field.
 */
void key_value_pair_donate_from_key(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong k,
    void *p_p_o);

/* Tries to donate a pointer from the k-th key field. Returns whether a pointer
 * was found.
 */
bool key_value_pair_try_donate_from_key(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong k,
    void *p_p_o);

/* Get a reference to the pointer in the k-th key field.
 */
void *key_value_pair_reference_from_key(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong k);

/* Get the k-th key field.
 */
ulong key_value_pair_get_key(KeyValuePair *this, KeyValueInfo *info, ulong k);

/* Set the v-th value field to 'value'.
 */
void key_value_pair_set_value(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong v,
    ulong value);

/* Donate a pointer to the v-th value field.
 */
void key_value_pair_donate_to_value(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong v,
    void *p_p_o);

/* Donate a pointer from the v-th value field.
 */
void key_value_pair_donate_from_value(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong v,
    void *p_p_o);

/* Try to donate a pointer from the v-th value field. Returns whether a pointer
 * was found.
 */
bool key_value_pair_try_donate_from_value(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong v,
    void *p_p_o);

/* Get a reference to the pointer in the v-th value field.
 */
void *key_value_pair_reference_from_value(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong v);

/* Get the v-th value field.
 */
ulong key_value_pair_get_value(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong v);

/* Set the f-th field (with key fields followed by value fields) to 'field'.
 * Only used internally and in unit testing.
 */
void key_value_pair_set_field(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong f,
    ulong field);

/* Get the f-th field. Only used internally and in unit testing.
 */
ulong key_value_pair_get_field(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong f);

/* Donate a pointer to the f-th field. Only used internally and in unit
 * testing.
 */
void key_value_pair_donate_to_field(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong f,
    void *p_p_o);

/* Put a reference to the pointer in the f-th field.
 */
void key_value_pair_reference_to_field(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong f,
    void *p);

/* Donate a pointer from the f-th field. Only used internally and in unit
 * testing.
 */
void key_value_pair_donate_from_field(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong f,
    void *p_p_o);

/* Try to donate a pointer from the f-th field. Returns whether a pointer was
 * found.
 */
bool key_value_pair_try_donate_from_field(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong f,
    void *p_p_o);

/* Get a reference to the pointer in the f-th field.
 */
void *key_value_pair_reference_from_field(
    KeyValuePair *this,
    KeyValueInfo *info,
    ulong f);

/* Move a pair. Zeroes the source pair if there is destructor in the
 * KeyValueInfo.
 */
void key_value_pair_move(
    KeyValuePair *src,
    KeyValuePair *dst,
    KeyValueInfo *info);

/* Set the pair to be invalid, according to the method specified in the
 * KeyValueInfo.
 */
void key_value_pair_set_invalid(KeyValuePair *this, KeyValueInfo *info);

/* Whether the pair is invalid, according to the method specified in the
 * KeyValueInfo.
 */
bool key_value_pair_get_invalid(KeyValuePair *this, KeyValueInfo *info);

/* Convert a pair from one KeyValueInfo spec to another. Both pairs must
 * already have been created.
 */
void key_value_pair_convert(
    KeyValuePair *old_pair,
    KeyValueInfo *old_info,
    KeyValuePair *new_pair,
    KeyValueInfo *new_info);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _H */
