#include "NaiveKeyValueCollection.h"

static void print_(NaiveKeyValueCollection *this,
    ulong k,
    KeyValuePair *pair,
    Printer *p);


void naive_key_value_collection_new(
    NaiveKeyValueCollection **p_this,
    KeyValueInfo *info,
    bool key_frequency) {
  ulong k;
  ulong v;
  ulong num_single_keys;
  ulong num_values;
  NaiveKeyValueCollection *this;

  if (info->destroy != null || info->print != null) {
    throw("Cannot store objects in a NaiveKeyValueCollection");
  }

  this = newo(p_this);

  key_value_info_new_copy(&this->_info, info);

  num_single_keys = 1;

  for (k = 0; k < info->numKeys; k++) {
    num_single_keys *= info->_fields[k]->maxValue + 1;
  }

  this->keyFrequency = key_frequency;
  this->numSingleKeys = num_single_keys;
  this->numValues = num_values = info->numValues;
  new(&this->_values, num_values);

  for (v = 0; v < num_values; v++) {
    new(&this->_values[v], num_single_keys);
  }
}

void naive_key_value_collection_delete(NaiveKeyValueCollection **p_this) {
  ulong v;
  NaiveKeyValueCollection *this;

  if ((this = *p_this) == null) {
    return;
  }

  for (v = 0; v < this->numValues; v++) {
    deletev(&this->_values[v]);
  }

  deletev(&this->_values);
  key_value_info_delete(&this->_info);

  deleteo(p_this);
}

void naive_key_value_collection_print(
    NaiveKeyValueCollection *this,
    Printer *p) {
  KeyValuePair *_pair = null;

  sprint(p, "NaiveKeyValueCollection");
  rprint(p, "Number of values: %lu", this->numValues);
  rprint(p, "Number of single keys: %lu", this->numSingleKeys);

  if (this->numValues != 0) {
    sprint(p, "Contents");
    key_value_pair_new(&_pair, this->_info);
    print_(this, 0, _pair, p);
    key_value_pair_delete(&_pair, this->_info);
    fprint(p);
  }

  fprint(p);
}

void naive_key_value_collection_add_values(
    NaiveKeyValueCollection *this,
    KeyValuePair *pair) {
  ulong single_key;
  ulong v;

  single_key = naive_key_value_collection_single_key_from_pair(this, pair);

  for (v = 0; v < this->numValues; v++) {
    this->_values[v][single_key]
      += key_value_pair_get_value(pair, this->_info, v)
      + (this->keyFrequency ? 1 : 0);
  }
}

ulong naive_key_value_collection_get_value(
    NaiveKeyValueCollection *this,
    KeyValuePair *pair,
    ulong v) {
  ulong single_key;

  single_key = naive_key_value_collection_single_key_from_pair(this, pair);
  return this->_values[v][single_key];
}

ulong naive_key_value_collection_single_key_from_pair(
    NaiveKeyValueCollection *this,
    KeyValuePair *pair) {
  char *_e = null;
  ulong single_key;
  ulong k;
  ulong *max_keys;
  KeyValuePairPrintInfo kvppi;

  max_keys = this->_info->maxKeys;
  single_key = 0;

  for (k = 0; k < this->_info->numKeys; k++) {
    single_key *= max_keys[k] + 1;
    single_key += key_value_pair_get_key(pair, this->_info, k);
  }

  if (single_key >= this->numSingleKeys) {
    kvppi.this = pair;
    kvppi.info = this->_info;
    aprint(&_e, 1, key_value_pair, &kvppi);
    string_renew_prepend_f(
        &_e,
        "Impossible: Single key %lu greater than maximum %lu",
        single_key,
        this->numSingleKeys - 1);
    throw(_e);
  }

  return single_key;
}

void naive_key_value_collection_pair_from_single_key(
    NaiveKeyValueCollection *this,
    ulong single_key,
    KeyValuePair *pair) {
  long k;
  ulong n;

  for (k = this->_info->numKeys - 1; k >= 0; k--) {
    n = this->_info->maxKeys[k] + 1;
    key_value_pair_set_key(pair, this->_info, k, single_key % n);
    single_key /= n;
  }
}


static void print_(NaiveKeyValueCollection *this,
    ulong k,
    KeyValuePair *pair,
    Printer *p) {
  ulong key;
  ulong max_key;
  ulong v;
  ulong vv;
  KeyValueInfo *info;

  info = this->_info;

  max_key = info->maxKeys[k];

  for (key = 0; key <= max_key; key++) {
    if (info->numKeys > 1) {
      sprint(p, "Key %lu = %lu", k, key);
    } else {
      sprint(p, "Key = %lu", key);
    }

    key_value_pair_set_key(pair, info, k, key);

    if (k == info->numKeys - 1) {
      for (v = 0; v < info->numValues; v++) {
        vv = naive_key_value_collection_get_value(this, pair, v);

        if (info->numValues > 1) {
          rprint(p, "[%lu] %lu", v, vv);
        } else {
          rprint(p, "%lu", vv);
        }
      }
    } else {
      print_(this, k + 1, pair, p);
    }

    fprint(p);
  }
}
