/* Object class for naive collections of key-value pairs, for unit testing
 * purposes only. Not efficient or scalable. The class simply allocates a
 * flattened multi-dimensional array of ulongs for each value field, where each
 * multi-dimensional key is converted into a "single key" that is an index into
 * the array. Objects cannot be stored inside the collection, since methods for
 * handling them have not been built out.
 */
#ifndef NAIVE_KEY_VALUE_COLLECTION_H
#define NAIVE_KEY_VALUE_COLLECTION_H

#ifdef __cplusplus
extern "C" {
#endif

#include "KeyValuePair.h"

typedef struct {
  bool keyFrequency;
  ulong numSingleKeys;
  ulong numValues;
  ulong **_values;
  KeyValueInfo *_info;
} NaiveKeyValueCollection;

/* Constructor. If key_frequency is true, then the info structure has been
 * created by KeyFrequency, which stores one less than the frequency. We need
 * to take this into account when retrieving or adding.
 */
void naive_key_value_collection_new(
    NaiveKeyValueCollection **p_this,
    KeyValueInfo *info,
    bool key_frequency);

/* Destructor.
 */
void naive_key_value_collection_delete(NaiveKeyValueCollection **p_this);

/* Print.
 */
void naive_key_value_collection_print(
    NaiveKeyValueCollection *this,
    Printer *p);

/* Add a pair to the collection. Looks up the entry from the key, and adds the
 * values to any already in there.
 */
void naive_key_value_collection_add_values(
    NaiveKeyValueCollection *this,
    KeyValuePair *pair);

/* Get the v-th value for the key provided in the pair.
 */
ulong naive_key_value_collection_get_value(
    NaiveKeyValueCollection *this,
    KeyValuePair *pair,
    ulong v);

/* Compute the single key corresponding to the key (fields) of a pair.
 */
ulong naive_key_value_collection_single_key_from_pair(
    NaiveKeyValueCollection *this,
    KeyValuePair *pair);

/* Set the key (fields) of a pair to be those corresponding to the given single
 * key.
 */
void naive_key_value_collection_pair_from_single_key(
    NaiveKeyValueCollection *this,
    ulong single_key,
    KeyValuePair *pair);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _H */
