#include "RandomWeighted.h"

static void node_new_(RandomWeightedNode **p_this,
    void *object,
    double weight,
    RandomWeightedNode *parent);

static void node_delete_(RandomWeightedNode **p_this);

static void node_print_(
    RandomWeightedNode *this,
    Printer *p,
    void print(void *this, Printer *p));

static void remove_(RandomWeightedNode *this,
    RandomWeightedNode *child,
    double weight);

static void veto_(RandomWeightedNode *this,
    RandomWeightedNode *child,
    double weight);

static void unveto_(RandomWeightedNode *this,
    RandomWeightedNode *child,
    double weight);

static void clear_vetoes_(RandomWeightedNode *this);

static void clear_removes_(RandomWeightedNode *this);

static void zero_(RandomWeightedNode *this,
    RandomWeightedNode *child,
    double weight);

static void node_check_(RandomWeighted *this, RandomWeightedNode *node);

static ulong num_(RandomWeightedNode *this, sbyte d);

static double weight_(RandomWeightedNode *this, sbyte d);


void random_weighted_newI(
    RandomWeighted **p_this,
    void print(void *this, Printer *p)) {
  RandomWeighted *this;

  this = newo(p_this);

  random_new(&this->_random);
  this->print = print;
}

void random_weighted_delete(RandomWeighted **p_this) {
  RandomWeighted *this;

  if ((this = *p_this) == null) {
    return;
  }

  node_delete_(&this->_root);
  random_delete(&this->_random);

  deleteo(p_this);
}

void random_weighted_print(RandomWeighted *this, Printer *p) {
  sprint(p, "RandomWeighted");
  rprint(p, "Number: %lu", this->num);

  if (this->num > 0) {
    rprint(p, "Weight: %.16g", this->weight);
  }

  if (this->vetoedNum != 0) {
    rprint(p, "Vetoed number: %lu", this->vetoedNum);
    rprint(p, "Vetoed weight: %.16g", this->vetoedWeight);
  }

  if (this->removedNum != 0) {
    rprint(p, "Removed number: %lu", this->removedNum);
    rprint(p, "Removed weight: %.16g", this->removedWeight);
  }

  if (this->_root != null) {
    sprint(p, "Tree");
    node_print_(this->_root, p, this->print);
    fprint(p);
  }

  fprint(p);
}

RandomWeightedNode *random_weighted_add(
    RandomWeighted *this,
    void *object,
    double weight) {
  byte width;
  ulong flags;
  RandomWeightedNode *node;
  RandomWeightedNode **p_node;

  if (object == null) {
    throw("Object pointer cannot be null");
  }

  if (double_is_infinite_or_nan(weight)) {
    throw("Weight (%.16g) cannot be infinite or NaN", weight);
  }

  if (weight < 0 || double_is_subnormal(weight)) {
    weight = 0;
  }

  this->num++;
  this->weight += weight;
  width = ulong_bit_width(this->num);
  flags = ulong_reverse_bits(this->num << (64 - width));
  flags >>= 1;

  for (p_node = &this->_root, node = null, width--;
      width > 0;
      width--, flags >>= 1) {
    node = *p_node;

    if ((flags & 1) == 0) {
      node->leftNum++;
      node->leftWeight += weight;
      p_node = &node->_left;
    } else {
      node->rightNum++;
      node->rightWeight += weight;
      p_node = &node->_right;
    }
  }

  node_new_(p_node, object, weight, node);

  return *p_node;
}

RandomWeightedNode *random_weighted_get(
    RandomWeighted *this,
    double per_object_weight) {
  double total_weight;

  if (this->vetoedNum + this->removedNum > this->num) {
    throw("Impossible");
  }

  total_weight = this->weight - this->vetoedWeight - this->removedWeight
    + per_object_weight * (this->num - this->vetoedNum - this->removedNum);

  if (total_weight < 0) {
    throw("Impossible: %.16g", total_weight);
  }

  return random_weighted_get_from_weight(
      this,
      per_object_weight,
      random_double_unit_cc(this->_random) * total_weight);
}

RandomWeightedNode *random_weighted_get_from_weight(
    RandomWeighted *this,
    double per_object_weight,
    double weight) {
  double w;
  RandomWeightedNode *node;

  if (random_weighted_is_empty(this)) {
    throw("Cannot get from empty collection");
  }

  if (weight > (w = this->weight - this->vetoedWeight - this->removedWeight
        + per_object_weight * (this->num - this->vetoedNum - this->removedNum))
      || weight < 0) {
    throw("Invalid weight %.16g: must be in [0, %.16g]", weight, w);
  }

  for (node = this->_root;;) {
    node_check_(this, node);

    if (weight <
        (w = weight_(node, -1) + per_object_weight * num_(node, -1))) {
      node = node->_left;
    } else if (weight == w && num_(node, 0) + num_(node, +1) == 0) {
      node = node->_left;
    } else if (weight > (w = weight_(node, -1) + weight_(node, 0)
          + per_object_weight * (num_(node, -1) + num_(node, 0)))) {
      weight -= w;
      node = node->_right;
    } else if (weight == w && num_(node, 0) == 0) {
      if (num_(node, +1) != 0) {
        node = node->_right;
      } else {
        node = node->_left;
      }
    } else {
      return node;
    }
  }

  throw("Impossible");
  return null;
}

void random_weighted_veto(RandomWeighted *this, RandomWeightedNode *node) {
  if (node->thisRemovedNum == 1) {
    return;
  }

  if (node->thisVetoedNum != 0) {
    throw("Object has already been vetoed");
  }

  this->vetoedNum++;
  this->vetoedWeight += node->thisWeight;
  this->vetoedWeight = upper_bound(this->vetoedWeight, this->weight);

  node->thisVetoedNum = 1;
  node->thisVetoedWeight = node->thisWeight;
  veto_(node->parent, node, node->thisWeight);
}

void random_weighted_unveto(RandomWeighted *this, RandomWeightedNode *node) {
  if (node->thisRemovedNum == 1) {
    throw("Object has been removed; cannot be unvetoed");
  }

  if (node->thisVetoedNum != 1) {
    throw("Object is not vetoed");
  }

  node->thisVetoedNum = 0;
  node->thisVetoedWeight = 0;

  this->vetoedNum--;
  this->vetoedWeight -= node->thisWeight;
  this->vetoedWeight = lower_bound(this->vetoedWeight, 0);

  unveto_(node->parent, node, node->thisWeight);
}

void random_weighted_clear_vetoes(RandomWeighted *this) {
  if (this->vetoedNum == 0) {
    return;
  }

  this->vetoedNum = 0;
  this->vetoedWeight = 0;
  clear_vetoes_(this->_root);
}

void random_weighted_remove(RandomWeighted *this, RandomWeightedNode *node) {
  if (node->thisRemovedNum == 1) {
    throw("Object has already been removed");
  }

  if (node->thisVetoedNum == 1) {
    random_weighted_unveto(this, node);
  }

  this->removedNum++;
  this->removedWeight += node->thisWeight;
  this->removedWeight = upper_bound(this->removedWeight, this->weight);

  node->thisRemovedNum = 1;
  node->thisRemovedWeight = node->thisWeight;
  remove_(node->parent, node, node->thisWeight);
}

void random_weighted_clear_removes(RandomWeighted *this) {
  if (this->removedNum == 0) {
    return;
  }

  this->removedNum = 0;
  this->removedWeight = 0;
  clear_removes_(this->_root);
}

void random_weighted_zero(RandomWeighted *this, RandomWeightedNode *node) {
  zero_(node->parent, node, node->thisWeight);
  this->weight -= node->thisWeight;
  this->weight = lower_bound(this->weight, 0);
  node->thisWeight = 0;
}

bool random_weighted_is_empty(RandomWeighted *this) {
  return this->num - this->vetoedNum - this->removedNum == 0;
}


static void node_new_(RandomWeightedNode **p_this,
    void *object,
    double weight,
    RandomWeightedNode *parent) {
  RandomWeightedNode *this;

  this = newo(p_this);

  this->object = object;
  this->thisWeight = weight;
  this->parent = parent;
}

static void node_delete_(RandomWeightedNode **p_this) {
  RandomWeightedNode *this;

  if ((this = *p_this) == null) {
    return;
  }

  node_delete_(&this->_left);
  node_delete_(&this->_right);

  deleteo(p_this);
}

static void node_print_(
    RandomWeightedNode *this,
    Printer *p,
    void print(void *this, Printer *p)) {
  sprint(p, "RandomWeightedNode");

  if (this->_left != null) {
    sprint(p, "Left");
    rprint(p, "Number: %lu", this->leftNum);
    rprint(p, "Weight: %.16g", this->leftWeight);

    if (this->leftVetoedNum != 0) {
      rprint(p, "Vetoed number: %lu", this->leftVetoedNum);
      rprint(p, "Vetoed weight: %.16g", this->leftVetoedWeight);
    }

    if (this->leftRemovedNum != 0) {
      rprint(p, "Removed number: %lu", this->leftRemovedNum);
      rprint(p, "Removed weight: %.16g", this->leftRemovedWeight);
    }

    sprint(p, "Subtree");
    node_print_(this->_left, p, print);
    fprint(p);
    fprint(p);
  }

  sprint(p, "This");
  rprint(p, "Weight: %.16g", this->thisWeight);

  if (this->thisVetoedNum != 0) {
    rprint(p, "Vetoed");
  }

  if (this->thisRemovedNum != 0) {
    rprint(p, "Removed");
  }

  if (this->object != null) {
    if (print != null) {
      sprint(p, "Object");
      print(this->object, p);
      fprint(p);
    } else {
      rprint(p, "Object: %p", this->object);
    }
  }

  fprint(p);

  if (this->_right != null) {
    sprint(p, "Right");
    rprint(p, "Number: %lu", this->rightNum);
    rprint(p, "Weight: %.16g", this->rightWeight);

    if (this->rightVetoedNum != 0) {
      rprint(p, "Vetoed number: %lu", this->rightVetoedNum);
      rprint(p, "Vetoed weight: %.16g", this->rightVetoedWeight);
    }

    if (this->rightRemovedNum != 0) {
      rprint(p, "Removed number: %lu", this->rightRemovedNum);
      rprint(p, "Removed weight: %.16g", this->rightRemovedWeight);
    }

    sprint(p, "Subtree");
    node_print_(this->_right, p, print);
    fprint(p);
    fprint(p);
  }

  fprint(p);
}

static void veto_(RandomWeightedNode *this,
    RandomWeightedNode *child,
    double weight) {
  if (this == null) {
    return;
  }

  if (child == this->_left) {
    this->leftVetoedNum++;
    this->leftVetoedWeight += weight;
    this->leftVetoedWeight = upper_bound(
        this->leftVetoedWeight,
        this->leftWeight);
  } else if (child == this->_right) {
    this->rightVetoedNum++;
    this->rightVetoedWeight += weight;
    this->rightVetoedWeight = upper_bound(
        this->rightVetoedWeight,
        this->rightWeight);
  } else {
    throw("Impossible");
  }

  veto_(this->parent, this, weight);
}

static void unveto_(RandomWeightedNode *this,
    RandomWeightedNode *child,
    double weight) {
  if (this == null) {
    return;
  }

  if (child == this->_left) {
    this->leftVetoedNum--;
    this->leftVetoedWeight -= weight;
    this->leftVetoedWeight = lower_bound(this->leftVetoedWeight, 0);
  } else if (child == this->_right) {
    this->rightVetoedNum--;
    this->rightVetoedWeight -= weight;
    this->rightVetoedWeight = lower_bound(this->rightVetoedWeight, 0);
  } else {
    throw("Impossible");
  }

  unveto_(this->parent, this, weight);
}

static void remove_(RandomWeightedNode *this,
    RandomWeightedNode *child,
    double weight) {
  if (this == null) {
    return;
  }

  if (child == this->_left) {
    this->leftRemovedNum++;
    this->leftRemovedWeight += weight;
    this->leftRemovedWeight = upper_bound(
        this->leftRemovedWeight,
        this->leftWeight);
  } else if (child == this->_right) {
    this->rightRemovedNum++;
    this->rightRemovedWeight += weight;
    this->rightRemovedWeight = upper_bound(
        this->rightRemovedWeight,
        this->rightWeight);
  } else {
    throw("Impossible");
  }

  remove_(this->parent, this, weight);
}

static void clear_vetoes_(RandomWeightedNode *this) {
  if (this->leftVetoedNum > 0) {
    this->leftVetoedNum = 0;
    this->leftVetoedWeight = 0;
    clear_vetoes_(this->_left);
  }

  this->thisVetoedNum = 0;
  this->thisVetoedWeight = 0;

  if (this->rightVetoedNum > 0) {
    this->rightVetoedNum = 0;
    this->rightVetoedWeight = 0;
    clear_vetoes_(this->_right);
  }
}

static void clear_removes_(RandomWeightedNode *this) {
  if (this->leftRemovedNum > 0) {
    this->leftRemovedNum = 0;
    this->leftRemovedWeight = 0;
    clear_removes_(this->_left);
  }

  this->thisRemovedNum = 0;
  this->thisRemovedWeight = 0;

  if (this->rightRemovedNum > 0) {
    this->rightRemovedNum = 0;
    this->rightRemovedWeight = 0;
    clear_removes_(this->_right);
  }
}

static void zero_(RandomWeightedNode *this,
    RandomWeightedNode *child,
    double weight) {
  if (this == null) {
    return;
  }

  if (child == this->_left) {
    this->leftWeight -= weight;
    this->leftWeight = lower_bound(this->leftWeight, 0);
  } else if (child == this->_right) {
    this->rightWeight -= weight;
    this->rightWeight = lower_bound(this->rightWeight, 0);
  } else {
    throw("Impossible");
  }

  zero_(this->parent, this, weight);
}

static void node_check_(RandomWeighted *this, RandomWeightedNode *node) {
  if (node == null) {
    throw("Impossible: node is null");
  }

  num_(node, -1);
  num_(node, 0);
  num_(node, +1);

  weight_(node, -1);
  weight_(node, 0);
  weight_(node, +1);
}

static ulong num_(RandomWeightedNode *this, sbyte d) {
  if (d < 0) {
    return this->leftNum - this->leftVetoedNum - this->leftRemovedNum;
  }

  if (d > 0) {
    return this->rightNum - this->rightVetoedNum - this->rightRemovedNum;
  }

  return 1 - this->thisVetoedNum - this->thisRemovedNum;
}

static double weight_(RandomWeightedNode *this, sbyte d) {
  if (d < 0) {
    return this->leftWeight
      - (this->leftVetoedWeight + this->leftRemovedWeight);
  }

  if (d > 0) {
    return this->rightWeight
      - (this->rightVetoedWeight + this->rightRemovedWeight);
  }

  return this->thisWeight - (this->thisVetoedWeight + this->thisRemovedWeight);
}
