/* Object class for performing a weighted random selection from a set of
 * objects, where we have the following requirements:
 * - The (positive) weight of each object is the sum of a value that is the
 *   same for all objects and an object-specific value.
 * - The value that is the same for all objects can change as we add more
 *   objects, but at any time it is the same value for all objects.
 * - The object-specific value for any object already added does not change as
 *   we add more objects.
 * - If there are n objects already added, we do not want to recompute anything
 *   of O(n) when we add the next object.
 *
 * This class fulfills these requirements by populating a binary search tree,
 * along each level left to right, where when we traverse down each node we add
 * the number and weight of child objects in to the left or right number and
 * weight on that node. The nodes will not be in any particular order, but if
 * we specify a random weight between 0 and the total weight in the tree, then
 * we can use binary search to find which node of the tree has the weight
 * interval that contains that value.
 *
 * This class also allows for objects to be removed from the tree, by setting
 * their weight and count to zero (and propagating that change up all parent
 * nodes), and also for objects to be temporarily vetoed, using a similar
 * mechanism with veto weights and counts propagated upwards, which allows for
 * random weighted sampling without replacement. The vetoes and removes in the
 * set can be cleared when needed.
 */
#ifndef RANDOM_WEIGHTED_H
#define RANDOM_WEIGHTED_H

#ifdef __cplusplus
extern "C" {
#endif

#include "Random.h"

typedef struct RandomWeightedNode {
  double leftRemovedWeight;
  double leftVetoedWeight;
  double leftWeight;
  double rightRemovedWeight;
  double rightVetoedWeight;
  double rightWeight;
  double thisRemovedWeight;
  double thisVetoedWeight;
  double thisWeight;
  ulong leftNum;
  ulong leftRemovedNum;
  ulong leftVetoedNum;
  ulong rightNum;
  ulong rightRemovedNum;
  ulong rightVetoedNum;
  ulong thisRemovedNum : 1;
  ulong thisVetoedNum : 1;
  void *object;
  struct RandomWeightedNode *parent;
  struct RandomWeightedNode *_left;
  struct RandomWeightedNode *_right;
} RandomWeightedNode;

typedef struct {
  ulong num;
  ulong removedNum;
  ulong vetoedNum;
  double removedWeight;
  double vetoedWeight;
  double weight;
  RandomWeightedNode *_root;
  Random *_random;
  void (*print)(void *this, Printer *p);
} RandomWeighted;

/* Constructor. If the print method is null, the pointer value will simply be
 * printed.
 */
#define random_weighted_new(p_this, print) \
  random_weighted_newI( \
      (p_this), \
      (void (*)(void *, Printer *))print)

void random_weighted_newI(
    RandomWeighted **p_this,
    void print(void *this, Printer *p));

/* Destructor.
 */
void random_weighted_delete(RandomWeighted **p_this);

/* Print.
 */
void random_weighted_print(RandomWeighted *this, Printer *p);

/* Add an object. This just stores a pointer to the object; no donation is
 * involved here. The node pointer created to store this object is returned.
 */
RandomWeightedNode *random_weighted_add(
    RandomWeighted *this,
    void *object,
    double weight);

/* Get a weighted randomly-selected object from the set, specifying the weight
 * that is common to all objects. The pointer to the node containing it is
 * returned, which can be used in the remove and veto methods below. The object
 * pointer can be simply extracted from the node. Note: no changes should be
 * made to the node except through the methods here.
 */
RandomWeightedNode *random_weighted_get(
    RandomWeighted *this,
    double per_object_weight);

/* Same, where the (usually random) weight value is supplied (for unit tests).
 */
RandomWeightedNode *random_weighted_get_from_weight(
    RandomWeighted *this,
    double per_object_weight,
    double weight);

/* Veto an object from the set, which removes it temporarily, until the vetoes
 * in the set are cleared using the method below. This allows sampling without
 * replacement, without recomputing the whole tree.
 */
void random_weighted_veto(RandomWeighted *this, RandomWeightedNode *node);

/* Unveto an object from the set. Only used in the remove method; it is more
 * efficient to use the clear method below for external cases.
 */
void random_weighted_unveto(RandomWeighted *this, RandomWeightedNode *node);

/* Clear all vetoes, so that all vetoed nodes are replaced into the set.
 */
void random_weighted_clear_vetoes(RandomWeighted *this);

/* Remove an object from the set. As noted above, this does not delete the
 * node, but simply sets its weight to zero, and propagates that change
 * upwards. Can be cleared with the following method.
 */
void random_weighted_remove(RandomWeighted *this, RandomWeightedNode *node);

/* Clear all removes, so that all removed nodes are replaced into the set.
 */
void random_weighted_clear_removes(RandomWeighted *this);

/* Set the weight of an object to zero. It still can be selected through the
 * common per-object weight.
 */
void random_weighted_zero(RandomWeighted *this, RandomWeightedNode *node);

/* Whether the set is empty, i.e., the number active minus the number vetoed is
 * zero.
 */
bool random_weighted_is_empty(RandomWeighted *this);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _H */
