#include "Ranker.h"
#include <float.h>
#include <limits.h>
#include <math.h>

typedef struct {
  double value;
  ulong freqRank;
} Entry_;

static void check_(Ranker *this, char *action);

static void print_(KeyValuePairPrintInfo *kvppi, Printer *p);

static void merge_(KeyValuePair *acc, KeyValuePair *add, KeyValueInfo *info);

static sbyte compare_(KeyValuePair *a, KeyValuePair *b, KeyValueInfo *info);

static sbyte compare_find_value_(
    KeyValuePair *needle,
    KeyValuePair *hay,
    KeyValueInfo *info);

static sbyte compare_find_rank_(
    KeyValuePair *needle,
    KeyValuePair *hay,
    KeyValueInfo *info);

static void min_max_(Ranker *this);


void ranker_new(Ranker **p_this) {
  ulong max;
  Entry_ entry;
  KeyValueInfo *_info = null;
  Ranker *this;

  this = newo(p_this);

  max = ULONG_MAX;
  key_value_info_new(&_info, 1, &max, 1, &max);
  _info->x = this;
  _info->print = print_;
  _info->merge = merge_;
  _info->compare = compare_;

  key_value_collection_new(&this->_collection, _info);
  key_value_info_delete(&_info);

  entry.value = DBL_MAX;
  entry.freqRank = 0;
  key_value_collection_add(this->_collection, (KeyValuePair *)&entry);
}

void ranker_new_copy(Ranker **p_this, Ranker *other) {
  Ranker *this;

  this = newo(p_this);
  this->totalFrequency = other->totalFrequency;
  this->minValue = other->minValue;
  this->maxValue = other->maxValue;
  key_value_array_new_copy(&this->_array, other->_array);
}

void ranker_new_load(Ranker **p_this, File *f) {
  Ranker *this;

  this = newo(p_this);
  load_label("[Ranker]", f);
  load_value(this->totalFrequency, f);
  load(this->_array, key_value_array, f);
  load_label("[/Ranker]", f);
  min_max_(this);
}

void ranker_delete(Ranker **p_this) {
  Ranker *this;

  if ((this = *p_this) == null) {
    return;
  }

  key_value_collection_delete(&this->_collection);
  key_value_array_delete(&this->_array);
  deleteo(p_this);
}

void ranker_print(Ranker *this, Printer *p) {
  sprint(p, "Ranker");

  rprint(p, "Finalized: %s", bool_string(ranker_is_finalized(this)));
  rprint(p, "Total frequency: %lu", this->totalFrequency);
  rprint(p, "Minimum value: %.16g", this->minValue);
  rprint(p, "Maximum value: %.16g", this->maxValue);

  if (this->_collection != null) {
    key_value_collection_print(this->_collection, p);
  }

  if (this->_array != null) {
    key_value_array_print(this->_array, p);
  }

  fprint(p);
}

void ranker_save(Ranker *this, File *f) {
  check_(this, "save");
  save_label("[Ranker]", f);
  save_value(this->totalFrequency, f);
  save(this->_array, key_value_array, f);
  save_label("[/Ranker]", f);
}

bool ranker_equals(Ranker *this, Ranker *other) {
  check_(this, "test equality (this)");
  check_(other, "test equality (evidence)");

  return other->totalFrequency == this->totalFrequency
    && key_value_array_equals(this->_array, other->_array);
}

void ranker_add(Ranker *this, double value, ulong frequency) {
  Entry_ entry;

  if (frequency == 0) {
    throw("Frequency cannot be zero");
  }

  if (double_is_infinite_or_nan(value)) {
    throw("Cannot add the infinite or NaN value %.16g", value);
  }

  if (double_is_negative_zero_or_subnormal(value)) {
    value = 0;
  }

  entry.value = value;
  entry.freqRank = frequency;
  key_value_collection_add(this->_collection, (KeyValuePair *)&entry);
}

void ranker_finalize(Ranker *this) {
  ulong frequency;
  ulong i;
  ulong n;
  ulong rank;
  ulong total_frequency;
  Entry_ *entry;
  KeyValueArray *array;

  key_value_array_new_from_collection(&this->_array, &this->_collection);
  array = this->_array;

  if ((n = array->numPairs) == 1) {
    throw("Nothing added before finalizing");
  }

  for (i = 0, total_frequency = 0; i < n; i++) {
    entry = (Entry_ *)key_value_array_get(array, i);
    total_frequency += entry->freqRank;
  }

  this->totalFrequency = total_frequency;

  for (i = 0, rank = 0; i < n; i++, rank += frequency) {
    entry = (Entry_ *)key_value_array_get(array, i);
    frequency = entry->freqRank;
    entry->freqRank = rank;
  }

  if (rank != total_frequency) {
    throw("Impossible");
  }

  min_max_(this);
}

bool ranker_is_finalized(Ranker *this) {
  return this->_array != null;
}

long ranker_crank_from_value(Ranker *this, double value) {
  Entry_ needle;
  Entry_ *hay;
  KeyValueArray *array;

  check_(this, "get crank from value");
  array = this->_array;

  if (value < this->minValue) {
    return -(long)this->totalFrequency;
  }

  if (value > this->maxValue) {
    return +(long)this->totalFrequency;
  }

  array->_info->compare = compare_find_value_;
  needle.value = value;
  hay = (Entry_ *)key_value_array_find(array, (KeyValuePair *)&needle);

  if (hay == null) {
    throw("Impossible");
  }

  if (value == hay->value) {
    return (long)hay->freqRank + (long)(hay + 1)->freqRank
      - (long)this->totalFrequency;
  }

  return 2 * (long)hay->freqRank - (long)this->totalFrequency;
}

double ranker_value_from_crank(Ranker *this, double crank) {
  double rank;
  double value;
  Entry_ needle;
  Entry_ *hay;
  KeyValueArray *array;

  check_(this, "get value from crank");
  array = this->_array;

  if (crank <= -(long)this->totalFrequency) {
    return this->minValue;
  }

  if (crank >= +(long)this->totalFrequency) {
    return this->maxValue;
  }

  array->_info->compare = compare_find_rank_;
  rank = (crank + this->totalFrequency) / 2;
  needle.freqRank = (ulong)floor(rank);
  hay = (Entry_ *)key_value_array_find(array, (KeyValuePair *)&needle);

  if (hay == null) {
    throw("Impossible");
  }

  value = hay->value;

  if (rank == needle.freqRank && rank > 0) {
    needle.freqRank--;
    hay = (Entry_ *)key_value_array_find(array, (KeyValuePair *)&needle);

    if (hay == null) {
      throw("Impossible");
    }

    value = (value + hay->value) / 2;
  }

  return value;
}


static void check_(Ranker *this, char *action) {
  if (!ranker_is_finalized(this)) {
    throw("Cannot %s before being finalized", action);
  }
}

static void print_(KeyValuePairPrintInfo *kvppi, Printer *p) {
  Entry_ *entry;
  Ranker *this;

  entry = (Entry_ *)kvppi->this;
  this = kvppi->info->x;

  rprint(p, "Value: %.16g", entry->value);
  rprint(p,
      ranker_is_finalized(this) ? "Rank: %lu" : "Frequency: %lu",
      entry->freqRank);
}

static void merge_(KeyValuePair *acc, KeyValuePair *add, KeyValueInfo *info) {
  Entry_ *acc_entry;
  Entry_ *add_entry;

  acc_entry = (Entry_ *)acc;
  add_entry = (Entry_ *)add;

  acc_entry->freqRank += add_entry->freqRank;
}

static sbyte compare_(KeyValuePair *a, KeyValuePair *b, KeyValueInfo *info) {
  double value_a;
  double value_b;
  Entry_ *entry_a;
  Entry_ *entry_b;

  entry_a = (Entry_ *)a;
  entry_b = (Entry_ *)b;

  value_a = entry_a->value;
  value_b = entry_b->value;

  if (value_a < value_b) {
    return -1;
  }

  if (value_a > value_b) {
    return +1;
  }

  return 0;
}

static sbyte compare_find_value_(
    KeyValuePair *needle,
    KeyValuePair *hay,
    KeyValueInfo *info) {
  double needle_value;
  double hay_value;
  Entry_ *needle_entry;
  Entry_ *hay_entry;

  needle_entry = (Entry_ *)needle;
  hay_entry = (Entry_ *)hay;

  needle_value = needle_entry->value;
  hay_value = hay_entry->value;

  if (needle_value == hay_value) {
    return 0;
  }

  if (needle_value > hay_value) {
    return +1;
  }

  if (needle_value > (hay_entry - 1)->value) {
    return 0;
  }

  return -1;
}

static sbyte compare_find_rank_(
    KeyValuePair *needle,
    KeyValuePair *hay,
    KeyValueInfo *info) {
  ulong needle_rank;
  Entry_ *needle_entry;
  Entry_ *hay_entry;

  needle_entry = (Entry_ *)needle;
  hay_entry = (Entry_ *)hay;

  needle_rank = needle_entry->freqRank;

  if (needle_rank < hay_entry->freqRank) {
    return -1;
  }

  if (needle_rank < (hay_entry + 1)->freqRank) {
    return 0;
  }

  return +1;
}

static void min_max_(Ranker *this) {
  Entry_ *entry;
  KeyValueArray *array;

  array = this->_array;

  entry = (Entry_ *)key_value_array_get(array, 0);
  this->minValue = entry->value;

  entry = (Entry_ *)key_value_array_get(array, array->numPairs - 2);
  this->maxValue = entry->value;
}
