#include "TimeBudget.h"
#include <ctype.h>

#define MILLION_ 1000000

TimeBudget time_budget_parse(char *budget) {
  bool error;
  char c;
  char *_s = null;
  char **_parts = null;
  ulong num_parts;
  double d;
  TimeBudget tb;

  /* Check for empty string.
   */
  if (string_is_empty(budget)) {
    throw("Budget string is empty");
  }

  /* Initialize the split into integral hours, minutes, and seconds.
   */
  tb.hours = 0;
  tb.minutes = 0;
  tb.seconds = 0;

  /* Check if we have one of the last three formats: '<H>h', '<M>m', or '<S>s'.
   */
  if (isalpha(c = budget[string_length(budget) - 1])) {
    /* We do. Remove that character.
     */
    string_new_shorten(&_s, budget, 1);

    /* Parse the number that's left.
     */
    if (!try_parse_double(&d, _s)) {
      throw("Could not parse '%s' as '<H>h', '<M>m', or '<S>s'", budget);
    }

    string_delete(&_s);

    /* Use the character to convert to seconds.
     */
    switch (tolower(c)) {
      case 'h':
        tb.totalSeconds = (uint)round(d * 3600);
        break;
      case 'm':
        tb.totalSeconds = (uint)round(d * 60);
        break;
      case 's':
        tb.totalSeconds = (uint)round(d);
        break;
      default:
        throw("Cannot parse '%s': final character '%c' must be "
            "'h', 'm', or 's'",
            budget,
            c);
        tb.totalSeconds = 0;
        break;
    }

    /* Sanity check.
     */
    if (tb.totalSeconds >= 256 * 3600) {
      throw("Time budget must be less than 256 hours");
    }

    /* Split into integral hours, minutes, and seconds.
     */
    tb.hours = tb.totalSeconds / 3600;
    tb.minutes = tb.totalSeconds % 3600 / 60;
    tb.seconds = tb.totalSeconds % 60;
  } else {
    /* It must be one of the first two forms above: '<H>:<MM>:<SS>' or
     * '<M>:<SS>'. Split on the colon(s).
     */
    num_parts = strings_new_split(&_parts, budget, ':');

    /* Parse and sanity check.
     */
    error = false;

    if (num_parts < 2 || num_parts > 3 || string_length(_parts[1]) != 2
        || (num_parts == 3 && string_length(_parts[2]) != 2)) {
      error = true;
    } else {
      if (num_parts == 2) {
        if (!try_parse_byte(&tb.minutes, _parts[0])
            || !try_parse_byte(&tb.seconds, _parts[1])) {
          error = true;
        }
      } else {
        if (!try_parse_byte(&tb.hours, _parts[0])
            || !try_parse_byte(&tb.minutes, _parts[1])
            || !try_parse_byte(&tb.seconds, _parts[2])) {
          error = true;
        }
      }

      if (tb.minutes >= 60 || tb.seconds >= 60) {
        error = true;
      }

      delete(&_parts, num_parts, string);
    }

    if (error) {
      throw("Time budget ('%s') not in '<H>:<MM>:<SS>' or '<M>:<SS>' format",
          budget);
    }

    /* Compute the total budget in seconds.
     */
    tb.totalSeconds = ((uint)tb.hours * 60 + tb.minutes) * 60 + tb.seconds;
  }

  /* Sanity check.
   */
  if (tb.totalSeconds == 0) {
    throw("Time budget cannot be zero");
  }

  /* Save total us too.
   */
  tb.us = tb.totalSeconds * MILLION_;

  return tb;
}

TimeBudget time_budget_from_us(long us) {
  long seconds;
  TimeBudget tb;

  seconds = (long)round((double)us / MILLION_);

  if (seconds <= 0) {
    throw("Rounded seconds (%ld) from microseconds (%ld) must be positive",
        seconds,
        us);
  }

  tb.totalSeconds = (ulong)seconds;
  tb.us = tb.totalSeconds * MILLION_;
  tb.seconds = tb.totalSeconds % 60;
  tb.minutes = (tb.totalSeconds / 60) % 60;
  tb.hours = tb.totalSeconds / 3600;

  return tb;
}

void string_new_time_budget(char **p_this, TimeBudget tb) {
  string_new_empty(p_this);

  if (tb.hours > 0) {
    string_renew_append_f(
        p_this,
        "%u hour%s",
        tb.hours,
        plural_s(tb.hours));
  }

  if (tb.minutes > 0) {
    if (tb.hours > 0) {
      string_renew_append(p_this, tb.seconds > 0 ? ", " : " and ");
    }

    string_renew_append_f(
        p_this,
        "%u minute%s",
        tb.minutes,
        plural_s(tb.minutes));
  }

  if (tb.seconds > 0) {
    if (tb.hours > 0 || tb.minutes > 0) {
      string_renew_append(
          p_this,
          tb.hours > 0 && tb.minutes > 0 ? ", and " : " and ");
    }

    string_renew_append_f(
        p_this,
        "%u second%s",
        tb.seconds,
        plural_s(tb.seconds));
  }

  if (tb.totalSeconds > 59) {
    string_renew_append_f(p_this, " (%lu seconds)", tb.totalSeconds);
  }
}
