#include "gamma.h"
#include <math.h>

#define ITMAX_ 10000
#define EPS_ 3.0e-7
#define FPMIN_ 1.0e-30
#define CRIT_EPS_ 1e-7

static void check_a_x_(float a, float x);

static void gser_(float *gamser, float a, float x, float *gln);

static void gcf_(float *gammcf, float a, float x, float *gln);


float gammln(float xx) {
  const double cof_[6] = {
    76.18009172947146,
    -86.50532032941677,
    24.01409824083091,
    -1.231739572450155,
    0.1208650973866179e-2,
    -0.5395239384953e-5};
  double x;
  double y;
  double tmp;
  double ser;
  int j;

  y = x = xx;
  tmp = x + 5.5;
  tmp -= (x + 0.5) * log(tmp);
  ser = 1.000000000190015;

  for (j = 0; j <= 5; j++) {
    ser += cof_[j] / ++y;
  }

  return -tmp + log(2.5066282746310005 * ser / x);
}

float gammp(float a, float x) {
  float gamser;
  float gammcf;
  float gln;

  check_a_x_(a, x);

  if (x < a + 1) {
    gser_(&gamser, a, x, &gln);

    return gamser;
  }

  gcf_(&gammcf, a, x, &gln);

  return 1 - gammcf;
}

float gammq(float a, float x) {
  float gamser;
  float gammcf;
  float gln;

  check_a_x_(a, x);

  if (x < a + 1) {
    gser_(&gamser, a, x, &gln);

    return 1 - gamser;
  }

  gcf_(&gammcf, a, x, &gln);

  return gammcf;
}

float erff(float x) {
  return x < 0 ? -gammp(0.5, x * x) : gammp(0.5, x * x);
}

float erffc(float x) {
  return x < 0 ? 1 + gammp(0.5, x * x) : gammq(0.5, x * x);
}

float chisqp(float chisq, ulong dof) {
  return gammp(0.5 * dof, 0.5 * chisq);
}

float chisqq(float chisq, ulong dof) {
  return gammq(0.5 * dof, 0.5 * chisq);
}

float critchisqq(float q, ulong dof) {
  double chisq;
  double max_chisq;
  double min_chisq;

  if (q <= 0 || q >= 1) {
    throw("q (%.16g) must be in (0, 1)", q);
  }

  for (max_chisq = 1; chisqq(max_chisq, dof) > q; max_chisq *= 2) {
  }

  for (min_chisq = max_chisq / 2; chisqq(min_chisq, dof) < q; min_chisq /= 2) {
  }

  chisq = sqrt(min_chisq * max_chisq);

  while (max_chisq / min_chisq - 1 > CRIT_EPS_) {
    if (chisqq(chisq, dof) < q) {
      max_chisq = chisq;
    } else {
      min_chisq = chisq;
    }

    chisq = sqrt(min_chisq * max_chisq);
  }

  return chisq;
}

float approx_critchisqq(float q, double dof) {
  if (dof < 1) {
    throw("dof (%.16g) must be at least 1", dof);
  }

  if (dof > 50) {
    return dof - sqrt(2 * dof) * norminv(q);
  }

  if (floor(dof) != dof) {
    throw("dof (%.16g) must be an integer", dof);
  }

  return critchisqq(q, (ulong)dof);
}

static void check_a_x_(float a, float x) {
  if (x < 0) {
    throw("x cannot be negative: was %.16g", x);
  }

  if (a <= 0) {
    throw("a must be positive: was %.16g", a);
  }
}

static void gser_(float *gamser, float a, float x, float *gln) {
  int n;
  float sum, del, ap;

  *gln = gammln(a);

  if (x <= 0.0) {
    if (x != 0.0) {
      throw("Impossible: x = %.16g", x);
    }

    *gamser = 0.0;

    return;
  }

  ap = a;
  del = sum = 1 / a;

  for (n = 1; n <= ITMAX_; n++) {
    ap++;
    del *= x / ap;
    sum += del;

    if (fabs(del) < fabs(sum) * EPS_) {
      *gamser = sum * exp(-x + a * log(x) - *gln);

      return;
    }
  }

  throw("a too large, ITMAX_ too small: "
      "a = %.16g, x = %.16g, del = %.16g, sum = %.16g",
      a,
      x,
      del,
      sum);
}

static void gcf_(float *gammcf, float a, float x, float *gln) {
  int i;
  float an;
  float b;
  float c;
  float d;
  float del;
  float h;

  *gln = gammln(a);
  b = x + 1 - a;
  c = 1 / FPMIN_;
  d = 1 / b;
  h = d;

  for (i = 1; i <= ITMAX_; i++) {
    an = -i * (i - a);
    b += 2;
    d = an * d + b;

    if (fabs(d) < FPMIN_) {
      d=FPMIN_;
    }

    c = b + an / c;

    if (fabs(c) < FPMIN_) {
      c = FPMIN_;
    }

    d = 1 / d;
    del = d * c;
    h *= del;

    if (fabs(del - 1) < EPS_) {
      break;
    }
  }

  if (i > ITMAX_) {
    throw("a too large, ITMAX_ too small");
  }

  *gammcf = exp(-x + a * log(x) - (*gln)) * h;
}
