/* Static library of functions related to the gamma function, from Numerical
 * Recipes in C. Single-precision only. Includes single-precision versions of
 * erf and erfc, which are given an extra 'f' in their names to disambiguate
 * from those in the erf library.
 */
#ifndef GAMMA_H
#define GAMMA_H

#ifdef __cplusplus
extern "C" {
#endif

#include "normal.h"
#include "numbers.h"

/* Returns the value ln(Gamma(x)) for x > 0. Full accuracy for x > 1. For
 * 0 < x < 1, first use the reflection formula
 *
 *   Gamma(1 - z) = pi / (Gamma(z) sin(pi z)) = pi z / (Gamma(1 + z) sin(pi z))
 */
float gammln(float x);

/* Returns the incomplete gamma function P(a, x).
 */
float gammp(float a, float x);

/* Returns the incomplete gamma function Q(a, x) = 1 - P(a, x).
 */
float gammq(float a, float x);

/* Returns the error function erf(x), with float precision (hence the extra
 * 'f', as noted above).
 */
float erff(float x);

/* Returns the complementary error function erfc(x), with float precision
 * (hence the extra 'f', as noted above).
 */
float erffc(float x);

/* Returns the chi-square probability P(\chi^2, dof) for dof degrees of
 * freedom.
 */
float chisqp(float chisq, ulong dof);

/* Returns the complementary chi-square probability Q(\chi^2, dof).
 */
float chisqq(float chisq, ulong dof);

/* Compute the critical chi-square value corresponding to a given value of q
 * for the given number of degrees of freedom. Only used for dof <= 50 in the
 * following.
 */
float critchisqq(float q, ulong dof);

/* Compute an approximate critical chi-square value corresponding to a given
 * value of q for the given number of degrees of freedom, which uses the normal
 * approximation for dof > 50. A double value is used for 'dof' because Bear
 * can easily overflow a ulong with its (virtual) degrees of freedom.
 */
float approx_critchisqq(float q, double dof);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _H */
