/* Static library providing convenience methods around pthreads.
 */
#ifndef PARALLEL_H
#define PARALLEL_H

#ifdef __cplusplus
extern "C" {
#endif

#include "processors.h"
#include <limits.h>

/* The minimum number of operations before this library will spawn another
 * thread. Fairly arbitrarily but empirically determined at some point in time.
 */
#define PARALLEL_MIN_OPS_PER_THREAD 100000

/* Structure that will be passed to the run function.
 */
typedef struct {
  ulong numThreads;
  ulong t;
  ulong start;
  ulong end;
  void *x;
} Parallel;

/* Parallel 'for' loop. It is conceptually the same as
 *
 *   for (i = 0; i < n; i++) {
 *     run(...);
 *   }
 *
 * Under the hood, the loop is distributed across a number of threads, in such
 * a way that one processor is always left free, and that nested calls do not
 * spawn more than the maximum number of threads available. 'it_ops' specifies
 * the typical number of operations that each iteration of 'i' carries out,
 * where the definition of "operation" is left a little vague but could be a
 * memory access or an arithmetic operation. The class decides on a minimum
 * number of such iterations to allocate to each thread. A run() function will
 * typically do something like
 *
 *   Pass_ *pass;
 *
 *   pass = par->x;
 *
 *   for (i = par->start; i < par->end; i++) {
 *     [process]
 *   }
 *
 * where 'Pass_' is often an internal struct used to pass through information.
 */
void parallel_for(ulong n, ulong it_ops, void run(Parallel *par), void *x);

/* Same, where we split all into parallel threads if possible.
 */
#define parallel_for_all(n, run, x) parallel_for((n), ULONG_MAX, (run), (x))

/* Same, where there is just a single index 0 that we are calling on.
 */
#define parallel_single(run, x) parallel_for_all(1, (run), (x))

/* Compute the number of threads that a call to parallel_for() above would use
 * at the current time for the given number of operations per iteration. This
 * can be used by multi-pass algorithms that partition the data in a certain
 * way and then merge the results. After setting up with this information, the
 * actual call to parallel_for() can specify num_threads by calling
 * parallel_for_specify().
 */
ulong parallel_num_threads(ulong n, ulong it_ops);

/* Specify the number of threads. Usually only a good idea after calling
 * parallel_num_threads() above. Note that this overrides all other heuristics
 * in the class, such as threads already being in use, so should be used
 * sparingly.
 */
void parallel_for_specify(
    ulong n,
    ulong num_threads,
    void run(Parallel *par),
    void *x);

/* Run in as many parallel threads as are available, without any predfined
 * 'for' loop.
 */
void parallel_do(void run(Parallel *par),void *x);

/* Signal that this thread is suspending. Called by Lock and Condition.
 */
void parallel_suspend(void);

/* Specify that this thread is resuming.
 */
void parallel_resume(void);

/* Override the maximum number of threads, for unit tests.
 */
void parallel_override_max_num_threads(ulong num_threads);

/* Simulate thread resource constraints, for unit testing: one out of every
 * denominator thread creation calls will be randomly simulated to fail.
 */
void parallel_simulate_resource_constraints(ulong denominator);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _H */
