#define PAW_C
#include "paw.h"
#include <math.h>

paw g_PAW_min_ = {0x0100};
paw g_PAW_max_ = {0x7eff};
paw g_PAW_zero_ = {0x0000};
paw g_PAW_one_ = {0x3f00};
paw g_PAW_two_ = {0x4000};
paw g_PAW_half_ = {0x3e00};
paw g_PAW_posinf_ = {0x7f00};
paw g_PAW_neginf_ = {0xff00};
paw g_PAW_nan_ = {0x7f01};


typedef union {
  paw p;
  struct {
    ushort m : 8;
    ushort e : 7;
    ushort n : 1;
  } a;
} C_;

typedef union {
  double d;
  struct {
    ulong m : 52;
    ulong e : 11;
    ulong n : 1;
  } a;
} D_;


paw paw_make(double d) {
  return paw_make_quantized(d, 8);
}

paw paw_make_quantized(double d, byte bits) {
  bool n;
  ushort e;
  ushort l;
  ushort m;
  ulong dm;
  ulong r;
  ulong s;
  C_ c;
  D_ dc;

  if (bits > 8) {
    throw("Paws cannot have more than 8 bits of mantissa; you specified %u",
        bits);
  }

  dc.d = d;

  if ((e = (ushort)dc.a.e) == 0) {
    return PAW_ZERO;
  }

  dm = dc.a.m;
  n = dc.a.n == 1;

  if (e == 0x7ff) {
    if (dm == 0) {
      return n ? PAW_NEGINF : PAW_POSINF;
    } else {
      return PAW_NAN;
    }
  }

  m = (ushort)(dm >> (52 - bits) << (8 - bits));
  r = dm & (0xfffffffffffff >> bits);
  s = 0x8000000000000 >> bits;
  l = 1 << (8 - bits);

  if (r > s || (r == s && (m & l) == l)) {
    m += 1 << (8 - bits);

    if (m == 0x100) {
      e++;
    }
  }

  if (e > 1086) {
    return n ? PAW_NEGINF : PAW_POSINF;
  }

  if (e < 961) {
    return PAW_ZERO;
  }

  c.a.e = e - 960;
  c.a.n = n;
  c.a.m = m;

  return c.p;
}

paw paw_make_relative_quantized(double d, double max_abs, byte bits) {
  if (max_abs < 0) {
    throw("max_abs (%.16g) cannot be negative", max_abs);
  }

  if (fabs(d) > max_abs) {
    throw("abs(d) (%.16g) > max_abs (%.16g)", fabs(d), max_abs);
  }

  if (d == 0) {
    return PAW_ZERO;
  }

  if (d < 0) {
    return paw_negate(paw_make_relative_quantized(-d, max_abs, bits));
  }

  return paw_subtract(
      paw_make_quantized(max_abs + d, bits),
      paw_make_quantized(max_abs, bits));
}

paw paw_from_parts(bool negative, byte exponent, byte mantissa) {
  C_ c;

  c.a.n = negative;
  c.a.e = exponent;
  c.a.m = mantissa;

  return c.p;
}

void paw_print(paw *this, Printer *p) {
  rprint(p, "Paw: %.4g", double_from_paw(*this));
}

bool paw_equals(paw this, paw other) {
  return other.u == this.u;
}

bool paw_gt(paw this, paw other) {
  return double_from_paw(this) > double_from_paw(other);
}

bool paw_lt(paw this, paw other) {
  return double_from_paw(this) < double_from_paw(other);
}

bool paw_ge(paw this, paw other) {
  return double_from_paw(this) >= double_from_paw(other);
}

bool paw_le(paw this, paw other) {
  return double_from_paw(this) <= double_from_paw(other);
}

double double_from_paw(paw p) {
  C_ c;
  D_ d;

  if (paw_is_zero(p)) {
    return 0;
  }

  if (paw_is_positive_infinity(p)) {
    return DBL_POSINF;
  }

  if (paw_is_negative_infinity(p)) {
    return DBL_NEGINF;
  }

  if (paw_is_nan(p)) {
    return DBL_NAN;
  }

  c.p = p;
  d.a.n = c.a.n;
  d.a.e = c.a.e + 960;
  d.a.m = ((ulong)c.a.m) << 44;

  return d.d;
}

double pawize_double(double d) {
  return double_from_paw(paw_make(d));
}

paw paw_zero(void) {
  paw p = {0};

  return p;
}

paw paw_one(void) {
  return paw_make(1);
}

paw paw_half(void) {
  return paw_make(0.5);
}

paw paw_positive_infinity(void) {
  C_ c;

  c.a.n = 0;
  c.a.e = 0x7f;
  c.a.m = 0;

  return c.p;
}

paw paw_negative_infinity(void) {
  C_ c;

  c.a.n = 1;
  c.a.e = 0x7f;
  c.a.m = 0x0;

  return c.p;
}

paw paw_nan(void) {
  C_ c;

  c.a.n = 0;
  c.a.e = 0x7f;
  c.a.m = 0x1;

  return c.p;
}

bool paw_is_zero(paw p) {
  return p.u == 0;
}

bool paw_is_positive(paw p) {
  return p.u != 0 && !paw_is_nan(p) && (p.u & 0x8000) == 0;
}

bool paw_is_negative(paw p) {
  return !paw_is_nan(p) && (p.u & 0x8000) == 0x8000;
}

bool paw_is_positive_infinity(paw p) {
  C_ c;

  c.p = p;

  return c.a.e == 0x7f && c.a.m == 0x0 && c.a.n == 0;
}

bool paw_is_negative_infinity(paw p) {
  C_ c;

  c.p = p;

  return c.a.e == 0x7f && c.a.m == 0x0 && c.a.n == 1;
}

bool paw_is_nan(paw p) {
  C_ c;

  c.p = p;

  return c.a.e == 0x7f && c.a.m != 0x0;
}

bool paw_is_infinite_or_nan(paw p) {
  C_ c;

  c.p = p;

  return c.a.e == 0x7f;
}

bool double_pawized_is_infinite_or_nan(double d) {
  return paw_is_infinite_or_nan(paw_make(d));
}

paw paw_negate(paw p) {
  if (paw_is_zero(p) || paw_is_nan(p)) {
    return p;
  }

  p.u ^= 0x8000;

  return p;
}

paw paw_abs(paw p) {
  p.u &= 0x7fff;

  return p;
}

paw paw_add(paw a, paw b) {
  return paw_make(double_from_paw(a) + double_from_paw(b));
}

paw paw_subtract(paw a, paw b) {
  return paw_make(double_from_paw(a) - double_from_paw(b));
}

paw paw_multiply(paw a, paw b) {
  return paw_make(double_from_paw(a) * double_from_paw(b));
}

paw paw_divide(paw a, paw b) {
  return paw_make(double_from_paw(a) / double_from_paw(b));
}

paw paw_bound(paw x, paw min, paw max) {
  return paw_make(bound(
        double_from_paw(x),
        double_from_paw(min),
        double_from_paw(max)));
}

paw paw_lower_bound(paw x, paw min) {
  return paw_make(lower_bound(double_from_paw(x), double_from_paw(min)));
}

paw paw_upper_bound(paw x, paw max) {
  return paw_make(upper_bound(double_from_paw(x), double_from_paw(max)));
}

paw paw_min(paw a, paw b) {
  return paw_make(minimum(double_from_paw(a), double_from_paw(b)));
}

paw paw_max(paw a, paw b) {
  return paw_make(maximum(double_from_paw(a), double_from_paw(b)));
}

sbyte paw_compare(paw *a, paw *b, void *x) {
  double da;
  double db;

  da = double_from_paw(*a);
  db = double_from_paw(*b);

  return da < db ? -1 : da > db ? +1 : 0;
}
