/* Value class for floating point numbers built specifically for use by Bear.
 * Very similar to Google Brain's bfloat16 but with 7 exponent bits and 8
 * mantissa bits, rather than Google's 8 and 7: I don't need numbers between
 * 10^19 and 10^38, I have no legacy 'float' implementation, and I am not
 * optimizing matrix multiplication in hardware like they are, so I can keep an
 * extra bit of mantissa and still keep 16-bit cardinality efficiency.
 *
 * I could backronym "PAW" as Precision Adjusted Workingvalue, but it's really
 * just something that a bear uses.
 */
#ifndef PAW_H
#define PAW_H

#ifdef __cplusplus
extern "C" {
#endif

#include "Printer.h"

#define PAW_MIN g_PAW_min_
#define PAW_MAX g_PAW_max_
#define PAW_ZERO g_PAW_zero_
#define PAW_ONE g_PAW_one_
#define PAW_TWO g_PAW_two_
#define PAW_HALF g_PAW_half_
#define PAW_POSINF g_PAW_posinf_
#define PAW_NEGINF g_PAW_neginf_
#define PAW_NAN g_PAW_nan_

typedef struct paw {
  ushort u;
} paw;

/* Make a paw from a double or float.
 */
paw paw_make(double d);

/* Make a paw that is quantized to the given number of mantissa bits.
 */
paw paw_make_quantized(double d, byte bits);

/* Make a paw that is quantized to the given number of mantissa bits, relative
 * to the maximum absolute value provided.
 */
paw paw_make_relative_quantized(double d, double max_abs, byte bits);

/* Get a paw from explicit negative, exponent, and mantissa values.
 */
paw paw_from_parts(bool negative, byte exponent, byte mantissa);

/* Print. Note that this takes a pointer to a paw, for consistency with all
 * other print methods.
 */
void paw_print(paw *this, struct Printer *p);

/* Equals.
 */
bool paw_equals(paw this, paw other);

/* Convert a paw to a double.
 */
double double_from_paw(paw p);

/* Pawize a double, i.e., convert it to a paw and back to a double.
 */
double pawize_double(double d);

/* Whether a paw is zero.
 */
bool paw_is_zero(paw p);

/* Whether a paw is positive.
 */
bool paw_is_positive(paw p);

/* Whether a paw is negative.
 */
bool paw_is_negative(paw p);

/* Whether a paw is positive infinity.
 */
bool paw_is_positive_infinity(paw p);

/* Whether a paw is negative infinity.
 */
bool paw_is_negative_infinity(paw p);

/* Whether a paw is not a number.
 */
bool paw_is_nan(paw p);

/* Whether a paw is positive or negative infinity or not a number. This method
 * is more efficient than calling the three methods individually.
 */
bool paw_is_infinite_or_nan(paw p);

/* Whether a pawized double is infinite or not a number.
 */
bool double_pawized_is_infinite_or_nan(double d);

/* Negate.
 */
paw paw_negate(paw p);

/* Absolute value.
 */
paw paw_abs(paw p);

/* Add.
 */
paw paw_add(paw a, paw b);

/* Subtract.
 */
paw paw_subtract(paw a, paw b);

/* Multiply.
 */
paw paw_multiply(paw a, paw b);

/* Divide.
 */
paw paw_divide(paw a, paw b);

/* Bound.
 */
paw paw_bound(paw x, paw min, paw max);

/* Lower-bound.
 */
paw paw_lower_bound(paw x, paw min);

/* Upper-bound.
 */
paw paw_upper_bound(paw x, paw max);

/* Greater than.
 */
bool paw_gt(paw this, paw other);

/* Less than.
 */
bool paw_lt(paw this, paw other);

/* Greater than or equal to.
 */
bool paw_ge(paw this, paw other);

/* Less than or equal to.
 */
bool paw_le(paw this, paw other);

/* Minimum.
 */
paw paw_min(paw a, paw b);

/* Maximum.
 */
paw paw_max(paw a, paw b);

/* Comparator.
 */
sbyte paw_compare(paw *a, paw *b, void *x);

#ifndef PAW_C
extern paw g_PAW_min_;
extern paw g_PAW_max_;
extern paw g_PAW_zero_;
extern paw g_PAW_one_;
extern paw g_PAW_two_;
extern paw g_PAW_half_;
extern paw g_PAW_posinf_;
extern paw g_PAW_neginf_;
extern paw g_PAW_nan_;
#endif

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* _H */
