#include "prog_bear.h"

static void create_column_name_(
    char **p_this,
    char *prefix,
    char separator,
    char *suffix,
    ulong num_column_names,
    char **column_names);


void prog_bear_args_1(char *prog) {
  bool feature = false;
  char *article = "";
  char *frequency = "";
  char *header = "";
  char *jindexes = "";
  char *jnames = "";
  char *multi_article = "";
  char *skip = "";
  char *types = "";
  char *zero = "";

  if (string_equals(prog, PROG_BEAR_MEMORY)) {
    article = "the";
    frequency = " (if --has-frequency-column)";
    header = " (if\f--has-header-row)";
    jindexes = " (in quotes), e.g., '[0, 3, 17]'";
    jnames = " (in quotes), e.g., '[\"cat\", \"dog\"]'";
    multi_article = "the ";
    types = ": 'csv', 'tsv', 'hsv', 'nsv', or 'txt' (default\f=\fauto-detect)";
    zero = " (zero-offset)";
  } else if (string_equals(prog, PROG_BEAR_PREDICT)) {
    article = "a";
    skip = " to skip";
  } else if (string_equals(prog, PROG_BEAR_FEATURES)) {
    article = "a";
    feature = true;
  } else {
    throw("Unrecognized program enum '%s'", prog);
  }

  args_optional_integer(
      "label-column",
      'l',
      ARGS_INTEGER_NO_DEFAULT,
      "Index%s of %s label column%s",
      zero,
      article,
      skip);
  args_optional_string(
      "multi-label-columns",
      'm',
      ARGS_STRING_NO_DEFAULT,
      "Indexes of %slabel columns%s, as JSON list%s",
      multi_article,
      skip,
      jindexes);
  args_optional_integer(
      "id-column",
      'i',
      ARGS_INTEGER_NO_DEFAULT,
      "Index of an ID column to pass through");
  args_optional_string(
      "multi-id-columns",
      'j',
      ARGS_STRING_NO_DEFAULT,
      "Indexes of ID columns to pass through");
  args_flag("has-frequency-column",
      'f',
      "One column is a frequency column%s",
      skip);
  args_optional_integer(
      "frequency-column",
      'c',
      ARGS_INTEGER_NO_DEFAULT,
      "Frequency column%s",
      frequency);

  if (!feature) {
    args_flag("has-header-row",
        'H',
        "The first row is a header row");
    args_flag("confirm-header-row",
        'y',
        "Confirm that a header row of numbers is correct");
  }

  args_optional_string(
      "label-name",
      'L',
      ARGS_STRING_NO_DEFAULT,
      "Name of %s label column%s%s",
      article,
      skip,
      header);
  args_optional_string(
      "multi-label-names",
      'M',
      ARGS_STRING_NO_DEFAULT,
      "Names of %slabel columns%s%s, as JSON list%s",
      multi_article,
      skip,
      header,
      jnames);
  args_optional_string(
      "id-name",
      'I',
      ARGS_STRING_NO_DEFAULT,
      "Name of an ID column to pass through");
  args_optional_string(
      "multi-id-names",
      'J',
      ARGS_STRING_NO_DEFAULT,
      "Names of ID columns to pass through");
  args_optional_string(
      "frequency-name",
      'C',
      ARGS_STRING_NO_DEFAULT,
      "Name of the frequency column%s",
      skip);

  if (feature) {
    args_optional_string(
        "filetype",
        't',
        null,
        "Type of the file");
    args_flag("confirm-header-row",
        'y',
        "Confirm that a header row of numbers is correct");
  } else {
    args_optional_string(
        "missing-string",
        'e',
        "",
        "String for a missing value");
    args_optional_string(
        "input-filetype",
        't',
        null,
        "Type of the input file%s",
        types);
    args_optional_string(
        "output-filetype",
        'T',
        null,
        "Type of the output file%s",
        types);
  }
}

void prog_bear_args_2(void) {
  args_flag("no-features-out", 'x', "Don't write features to the output file");
  args_flag_sole("version", 'v', "Print the Bear version number and exit");
  args_flag("quiet", 'q', "Run quietly, with no progress logging");
}

void prog_bear_version(void) {
  if (args_set("version")) {
    printf("Bear Version %s\n", BEAR_VERSION);

    exit(0);
  }
}

void prog_bear_parse_args(
    bool has_header_row,
    bool has_frequency_column,
    ulong *p_num_labels,
    ulong **p_label_columns,
    char ***p_label_column_names,
    ulong *p_num_ids,
    ulong **p_id_columns,
    char ***p_id_column_names,
    ulong *p_frequency_column,
    char **p_frequency_column_name,
    bool predict) {
  char **id_column_names;
  char **label_column_names;
  ulong l;
  ulong *id_columns;
  ulong *label_columns;

  /* Get the label column(s) or their name(s). In the latter case we extract
   * the actual column indexes later.
   */
  if (args_set_veto_until_null(
        "label-column",
        "label-name",
        "multi-label-columns",
        "multi-label-names",
        null)) {
    /* Have specified --label-column. Easiest.
     */
    *p_num_labels = 1;
    label_columns = new(p_label_columns, 1);
    label_columns[0] = args_get_nonnegative_integer("label-column");
  } else if (args_set_veto_until_null(
        "label-name",
        "multi-label-columns",
        "multi-label-names",
        null)) {
    /* Have specified --label-name. Extract the name.
     */
    if (!has_header_row) {
      throw("Cannot specify --label-name without --has-header-row");
    }

    *p_num_labels = 1;
    label_column_names = new(p_label_column_names, 1);
    string_new_copy(&label_column_names[0], args_get_string("label-name"));
  } else if (args_set_veto("multi-label-columns", "multi-label-names")) {
    /* Have specified --multi-label-columns. Extract them.
     */
    label_columns = ulong_array_new_from_json_entire(
        p_label_columns,
        p_num_labels,
        args_get_string("multi-label-columns"));

    if (*p_num_labels == 0) {
      throw("There must be at least one label column");
    }

    /* Sort into ascending order.
     */
    qsortx(label_columns, *p_num_labels, ulong_compare, false, null);

    /* Sanity check.
     */
    for (l = 0; l < *p_num_labels - 1; l++) {
      if (label_columns[l] == label_columns[l + 1]) {
        throw("Repeated label column %lu", label_columns[l]);
      }
    }
  } else if (args_set("multi-label-names")) {
    /* Have specified --multi-label-names. Extract the names.
     */
    if (!has_header_row) {
      throw("Cannot specify --label-name without --has-header-row");
    }

    label_column_names = string_array_new_from_json_entire(
        p_label_column_names,
        p_num_labels,
        args_get_string("multi-label-names"));

    if (*p_num_labels == 0) {
      throw("There must be at least one label column");
    }

    /* Sort into alphabetical order, for the following sanity check.
     */
    qsortx(label_column_names,
        *p_num_labels,
        string_pointer_compare,
        false,
        null);

    /* Sanity checks.
     */
    for (l = 0; l < *p_num_labels; l++) {
      if (string_is_empty(label_column_names[l])) {
          throw("Label column name cannot be the empty string");
      }

      if (l < *p_num_labels - 1 && string_equals(
            label_column_names[l],
            label_column_names[l + 1])) {
        throw("Repeated label column name '%s'", label_column_names[l]);
      }
    }
  } else {
    /* No label(s) specified. This is only acceptable in predict mode.
     */
    if (!predict) {
      throw("Must specify one of --label-column, --label-name, "
          "--multi-label-columns, or --multi-label-names");
    }

    /* Otherwise, we set the number of labels to zero. The calling code will
     * use the number of labels in the Bear model.
     */
     *p_num_labels = 0;
  }

  /* If specified, get the ID column(s) or their name(s). In the latter case we
   * extract the actual column indexes later.
   */
  if (args_set_veto_until_null(
        "id-column",
        "id-name",
        "multi-id-columns",
        "multi-id-names",
        null)) {
    /* Have specified --id-column. Easiest.
     */
    *p_num_ids = 1;
    id_columns = new(p_id_columns, 1);
    id_columns[0] = args_get_nonnegative_integer("id-column");
  } else if (args_set_veto_until_null(
        "id-name",
        "multi-id-columns",
        "multi-id-names",
        null)) {
    /* Have specified --id-name. Extract the name.
     */
    if (!has_header_row) {
      throw("Cannot specify --id-name without --has-header-row");
    }

    *p_num_ids = 1;
    id_column_names = new(p_id_column_names, 1);
    string_new_copy(&id_column_names[0], args_get_string("id-name"));
  } else if (args_set_veto("multi-id-columns", "multi-id-names")) {
    /* Have specified --multi-id-columns. Extract them.
     */
    id_columns = ulong_array_new_from_json_entire(
        p_id_columns,
        p_num_ids,
        args_get_string("multi-id-columns"));

    if (*p_num_ids == 0) {
      throw("At least one ID column must be specified");
    }

    /* Sort into ascending order.
     */
    qsortx(id_columns, *p_num_ids, ulong_compare, false, null);

    /* Sanity check.
     */
    for (l = 0; l < *p_num_ids - 1; l++) {
      if (id_columns[l] == id_columns[l + 1]) {
        throw("Repeated ID column %lu", id_columns[l]);
      }
    }
  } else if (args_set("multi-id-names")) {
    /* Have specified --multi-id-names. Extract the names.
     */
    if (!has_header_row) {
      throw("Cannot specify --id-name without --has-header-row");
    }

    string_array_new_from_json_entire(
        p_id_column_names,
        p_num_ids,
        args_get_string("multi-id-names"));
    id_column_names = *p_id_column_names;

    if (*p_num_ids == 0) {
      throw("At least one ID column must be specified");
    }

    /* Sort into alphabetical order, for the following sanity check.
     */
    qsortx(id_column_names,
        *p_num_ids,
        string_pointer_compare,
        false,
        null);

    /* Sanity checks.
     */
    for (l = 0; l < *p_num_ids; l++) {
      if (string_is_empty(id_column_names[l])) {
          throw("Label column name cannot be the empty string");
      }

      if (l < *p_num_ids - 1 && string_equals(
            id_column_names[l],
            id_column_names[l + 1])) {
        throw("Repeated ID column name '%s'", id_column_names[l]);
      }
    }
  } else {
    *p_num_ids = 0;
  }

  /* If specified, extract the frequency column or its name. In the latter case
   * the actual column index will be determined later.
   */
  *p_frequency_column = ULONG_MAX;
  *p_frequency_column_name = null;

  if (has_frequency_column) {
    if (args_set_veto("frequency-column", "frequency-name")) {
      /* Have specified --frequency-column. Easy.
       */
      *p_frequency_column = args_get_integer("frequency-column");
    } else if (args_set("frequency-name")) {
      /* Have specified --frequency-name. Extract the name.
       */
      if (!has_header_row) {
        throw("Cannot specify --frequency-name without --has-header-row");
      }

      *p_frequency_column_name = args_get_string("frequency-name");
    } else {
      throw("Must specify either --frequency-column or --frequency-name "
          "when --has-frequency-column is specified");
    }
  } else {
    /* Sanity checks.
     */
    if (args_set("frequency-column")) {
      throw("Cannot specify --frequency-column without "
          "--has-frequency-column");

    }

    if (args_set("frequency-name")) {
      throw("Cannot specify --frequency-name without --has-frequency-column");
    }
  }
}

void prog_bear_feedback_1(
    bool has_header_row,
    bool has_frequency_column,
    char **label_column_names,
    ulong num_labels,
    ulong *label_columns,
    ulong frequency_column,
    char *frequency_column_name,
    char *missing_string,
    Logger *_l) {
  ulong l;

  rlog(_l, "Has header row: %s", bool_string(has_header_row));

  if (label_columns != null) {
    if (num_labels == 1) {
      rlog(_l, "Label column: %lu", label_columns[0]);
    } else {
      slog(_l, "Label columns");

      for (l = 0; l < num_labels; l++) {
        rlog(_l, "[%lu] %lu", l, label_columns[l]);
      }

      flog(_l);
    }
  } else {
    if (num_labels == 1) {
      rlog(_l, "Label column name: '%s'", label_column_names[0]);
    } else {
      slog(_l, "Label column names");

      for (l = 0; l < num_labels; l++) {
        rlog(_l, "[%lu] '%s'", l, label_column_names[l]);
      }

      flog(_l);
    }
  }

  rlog(_l, "Has frequency column: %s", bool_string(has_frequency_column));

  if (has_frequency_column) {
    if (frequency_column != ULONG_MAX) {
      rlog(_l, "Frequency column: %lu", frequency_column);
    } else {
      rlog(_l, "Frequency column name: '%s'", frequency_column_name);
    }
  }

  rlog(_l, "Missing value string: '%s'", missing_string);
}

void prog_bear_read_header(
    bool has_frequency_column,
    bool confirm_header_row,
    ulong num_labels,
    ulong num_ids,
    char *missing_string,
    char ***p_label_column_names,
    char ***p_id_column_names,
    char *frequency_column_name,
    ulong *p_frequency_column,
    TextFile *in,
    Logger *_l,
    ulong *p_num_column_names,
    char ***p_column_names,
    ulong **p_label_columns,
    ulong **p_id_columns,
    bool print_column_names) {
  char **column_names;
  char **id_column_names;
  char **label_column_names;
  char **_name_refs = null;
  ulong l;
  ulong n;
  ulong num_column_names;
  ulong *id_columns;
  ulong *label_columns;
  double d;

  /* Try to read the header row.
   */
  if (!text_file_try_read_copy(
        in,
        p_num_column_names,
        p_column_names)) {
    throw("Couldn't read header row");
  }

  num_column_names = *p_num_column_names;
  column_names = *p_column_names;

  /* Trim any whitespace in the column names.
   */
  for (n = 0; n < num_column_names; n++) {
    string_renew_trim_whitespace(&column_names[n]);
  }

  /* Check if there are any duplicates by creating a set of references to
   * those column names, sorting them, and looking for adjacent names that
   * are equal.
   */
  new(&_name_refs, num_column_names);

  for (n = 0; n < num_column_names; n++) {
    _name_refs[n] = column_names[n];
  }

  qsortx(_name_refs, num_column_names, string_pointer_compare, false, null);

  for (n = 0; n < num_column_names - 1; n++) {
    if (string_equals(_name_refs[n], _name_refs[n + 1])) {
      throw("Duplicate header column name '%s'", _name_refs[n]);
    }
  }

  deletev(&_name_refs);

  /* Check if it has not been confirmed that the header row is correct.
   */
  if (!confirm_header_row) {
    /* It has not. Perform some sanity checks for the user, which they can
     * override using the above flag, as to whether the header row actually
     * looks like a row of data rather than a header row. Firstly, check if
     * any column name is equal to the missing string (by default the empty
     * string), which should not occur. Secondly, check if all of the column
     * names can be parsed as doubles, which suggests a row of data.
     */
    for (n = 0; n < num_column_names; n++) {
      if (string_equals(column_names[n], missing_string)) {
        throw("Column name %lu in the header is equal to the missing string: "
            "are you sure that your file has a header row? "
            "Use --confirm-header-row if you are sure that it is correct.",
            n);
      }

      if (!try_parse_double(&d, column_names[n])) {
        break;
      }
    }

    /* If we got through all the names, then they can all be parsed as
     * doubles.
     */
    if (n == num_column_names) {
      throw("All of the column names in your header row are numbers: "
          "are you sure that your file has a header row? "
          "Use --confirm-header-row if you are sure that it is correct.");
    }
  }

  /* Print the column names that we found.
   */
  if (print_column_names) {
    slog(_l, "Column names");

    for (n = 0; n < num_column_names; n++) {
      rlog(_l, "[%lu] '%s'", n, column_names[n]);
    }

    flog(_l);
  }

  /* Check if label column names were specified.
   */
  if ((label_column_names = *p_label_column_names) != null) {
    /* They were. Find them and print them too.
     */
    slog(_l, "Label columns");
    label_columns = new(p_label_columns, num_labels);

    /* Walk through the label column names.
     */
    for (l = 0; l < num_labels; l++) {
      /* Look for this label column name in the column names.
       */
      for (n = 0; n < num_column_names; n++) {
        if (string_equals(label_column_names[l], column_names[n])) {
          label_columns[l] = n;
          rlog(_l, "'%s' -> Column %lu", label_column_names[l], n);

          break;
        }
      }

      /* Sanity check.
       */
      if (n == num_column_names) {
        throw("Could not find the label column name '%s' in the header row",
            label_column_names[l]);
      }
    }

    flog(_l);

    /* Sort the label columns into ascending order.
     */
    qsortx(label_columns, num_labels, ulong_compare, false, null);
  } else if ((label_columns = *p_label_columns) != null) {
    /* Label column names were not specified, but label columns were. Find
     * their names.
     */
    slog(_l, "Label columns");
    label_column_names = new(p_label_column_names, num_labels);
    label_columns = *p_label_columns;

    /* Walk through the label columns.
     */
    for (l = 0; l < num_labels; l++) {
      /* Copy the column name.
       */
      string_new_copy(&label_column_names[l], column_names[label_columns[l]]);
      rlog(_l, "Column %lu -> '%s'", label_columns[l], label_column_names[l]);
    }

    flog(_l);
  }

  /* Check if ID column names were specified.
   */
  if ((id_column_names = *p_id_column_names) != null) {
    /* They were. Find them and print them too.
     */
    slog(_l, "ID columns");
    id_columns = new(p_id_columns, num_ids);

    /* Walk through the ID column names.
     */
    for (l = 0; l < num_ids; l++) {
      /* Look for this ID column name in the column names.
       */
      for (n = 0; n < num_column_names; n++) {
        if (string_equals(id_column_names[l], column_names[n])) {
          id_columns[l] = n;
          rlog(_l, "'%s' -> Column %lu", id_column_names[l], n);

          break;
        }
      }

      /* Sanity check.
       */
      if (n == num_column_names) {
        throw("Could not find the ID column name '%s' in the header row",
            id_column_names[l]);
      }
    }

    flog(_l);

    /* Sort the ID columns into ascending order.
     */
    qsortx(id_columns, num_ids, ulong_compare, false, null);
  } else if ((id_columns = *p_id_columns) != null) {
    /* Label column names were not specified, but ID columns were. Find their
     * names.
     */
    slog(_l, "Label columns");
    id_column_names = new(p_id_column_names, num_ids);
    id_columns = *p_id_columns;

    /* Walk through the ID columns.
     */
    for (l = 0; l < num_ids; l++) {
      /* Copy the column name.
       */
      string_new_copy(&id_column_names[l], column_names[id_columns[l]]);
      rlog(_l, "Column %lu -> '%s'", id_columns[l], id_column_names[l]);
    }

    flog(_l);
  }

  /* If we specified a frequency column name, find it, and print it.
   */
  if (frequency_column_name != null) {
    /* Look for the frequency column name in the column names.
     */
    for (n = 0; n < num_column_names; n++) {
      if (string_equals(frequency_column_name, column_names[n])) {
        *p_frequency_column = n;
        rlog(_l,
            "Frequency column: '%s' -> Column %lu",
            frequency_column_name,
            n);

        break;
      }
    }

    /* Sanity check.
     */
    if (n == num_column_names) {
      throw("Could not find the frequency column name '%s' in the header row",
          frequency_column_name);
    }
  }
}

void prog_bear_check_columns(
    ulong num_labels,
    ulong *label_columns,
    ulong num_ids,
    ulong *id_columns,
    bool has_frequency_column,
    ulong frequency_column) {
  ulong c;
  ulong i;
  ulong l;
  ulong num_columns;
  ulong *_columns = null;

  /* Compute the number of specified columns.
   */
  num_columns = num_labels + num_ids + (has_frequency_column ? 1 : 0);

  /* Create an array for them.
   */
  new(&_columns, num_columns);

  /* Add all the labels.
   */
  for (l = 0, c = 0; l < num_labels; l++, c++) {
    _columns[c] = label_columns[l];
  }

  /* Add all the IDs.
   */
  for (i = 0; i < num_ids; i++, c++) {
    _columns[c] = id_columns[i];
  }

  /* Add the frequency column, if it exists.
   */
  if (has_frequency_column) {
    _columns[c++] = frequency_column;
  }

  /* Sanity check.
   */
  if (c != num_columns) {
    throw("Impossible");
  }

  /* Sort the columns into ascending order.
   */
  qsortx(_columns, num_columns, ulong_compare, false, null);

  /* Check that there are no duplicates.
   */
  for (c = 0; c < num_columns - 1; c++) {
    if (_columns[c] == _columns[c + 1]) {
      throw("Repeated column %lu specified", _columns[c]);
    }
  }

  /* Clean up.
   */
  deletev(&_columns);
}

void prog_bear_write_header(
    bool predict,
    bool debug,
    bool no_features_out,
    ulong num_ids,
    ulong *id_columns,
    ulong num_features,
    ulong *feature_columns,
    ulong num_labels,
    ulong *label_columns,
    ulong num_columns,
    ulong *p_num_column_names,
    char ***p_column_names,
    TextFile *out) {
  byte it;
  char ch;
  char column_name_separators[] = "_- :";
  char *column_name;
  char *p_ch;
  char *prefix;
  char *_s = null;
  char **column_names;
  ulong c;
  ulong f;
  ulong i;
  ulong l;
  ulong n;
  ulong num_column_names;

  /* Sanity check.
   */
  if (debug && predict) {
    throw("Impossible");
  }

  /* If we are in debug mode, create column names for any columns without
   * them in the input file.
   */
  if (debug) {
    /* Try to match the separator character used in the supplied column
     * names. Look for any of the obvious ones.
     */
    for (n = 0, ch = '\0'; n < *p_num_column_names; n++) {
      if (string_contains_chars(
            (*p_column_names)[n],
            column_name_separators)) {
        for (p_ch = column_name_separators; (ch = *p_ch) != '\0'; p_ch++) {
          if (string_contains_char((*p_column_names)[n], ch)) {
            break;
          }
        }
      }

      if (ch != '\0') {
        break;
      }
    }

    /* If we didn't find any, use the first in our list.
     */
    if (ch == '\0') {
      ch = column_name_separators[0];
    }

    /* Walk through any columns without names and create unique ones.
     */
    for (; *p_num_column_names < num_columns; (*p_num_column_names)++) {
      string_new_f(&_s, "%lu", *p_num_column_names);
      renewup1(p_column_names, *p_num_column_names);
      create_column_name_(
          &(*p_column_names)[*p_num_column_names],
          "column",
          ch,
          _s,
          *p_num_column_names,
          *p_column_names);
      string_delete(&_s);
    }
  }

  /* Dereference.
   */
  num_column_names = *p_num_column_names;
  column_names = *p_column_names;

  /* Write out all the column names for the IDs. Walk through the IDs.
   */
  for (i = 0; i < num_ids; i++) {
    /* Extract the column for this ID.
     */
    c = id_columns[i];

    /* Extract its name. If it wasn't supplied, use an empty string.
     */
    column_name = c >= num_column_names ? "" : column_names[c];

    /* Write it out.
     */
    text_file_write(out, column_name);
  }

  /* Don't write out features if we don't want them.
   */
  if (!no_features_out) {
    /* Write out all the column names for the features. Walk through the
     * features.
     */
    for (f = 0; f < num_features; f++) {
      /* Extract the column for this feature.
       */
      c = feature_columns[f];

      /* Extract its name. If it wasn't supplied, use an empty string.
       */
      column_name = c >= num_column_names ? "" : column_names[c];

      /* Write it out.
       */
      text_file_write(out, column_name);
    }
  }

  /* Switch on whether this is the predict program.
   */
  if (predict) {
    /* The predict program. We need to construct and write out column names for
     * all the predicted labels, followed by all the true labels, if supplied.
     * Check if label columns have been supplied.
     */
    if (label_columns != null) {
      /* They have. Walk through the iterations.
       */
      for (it = 0; it < 2; it++) {
        /* Set the prefix.
         */
        prefix = it == 0 ? "predicted" : "actual";

        for (l = 0; l < num_labels; l++) {
          /* Extract the column for this label.
           */
          c = label_columns[l];

          /* Construct a name.
           */
          if (c >= num_column_names || string_is_empty(column_names[c])) {
            string_new_f(&_s, "%s_l%lu", prefix, l);
          } else {
            string_new_f(&_s, "%s_%s", prefix, column_names[c]);
          }

          /* Write it out.
           */
          text_file_write(out, _s);

          /* Clean up.
           */
          string_delete(&_s);
        }
      }
    } else {
      /* Label columns have not been supplied. We just write out generic
       * columns names.
       */
      for (l = 0; l < num_labels; l++) {
        string_new_f(&_s, "predicted_l%lu", l);
        text_file_write(out, _s);
        string_delete(&_s);
      }
    }
  } else {
    /* Not the predict program. Write out all the column names for the labels.
     * Walk through the labels.
     */
    for (l = 0; l < num_labels; l++) {
      /* Extract the column for this label.
       */
      c = label_columns[l];

      /* Extract its name. If it wasn't supplied, use an empty string.
       */
      column_name = c >= num_column_names ? "" : column_names[c];

      /* Write it out.
       */
      text_file_write(out, column_name);
    }
  }

  /* If we are in debug mode, create and write out column names for the label
   * prediction and residual columns.
   */
  if (debug) {
    /* Create and write out the label prediction column names. Walk through the
     * labels.
     */
    for (l = 0; l < num_labels; l++) {
      /* Extract the column for this label.
       */
      c = label_columns[l];

      /* Extract its name. If it wasn't supplied, use an empty string.
       */
      column_name = column_names[c];

      /* Create the label prediction column name.
       */
      create_column_name_(
          &_s,
          "predicted",
          ch,
          column_name,
          num_column_names,
          column_names);

      /* Write it out.
       */
      text_file_write(out, _s);

      /* Delete the label prediction column name.
       */
      string_delete(&_s);
    }

    /* Create and write out the label residual column names. Walk through the
     * labels.
     */
    for (l = 0; l < num_labels; l++) {
      /* Extract the column for this label.
       */
      c = label_columns[l];

      /* Extract its name. If it wasn't supplied, use an empty string.
       */
      column_name = column_names[c];

      /* Create the label residual column name.
       */
      create_column_name_(
          &_s,
          "residual",
          ch,
          column_name,
          num_column_names,
          column_names);

      /* Write it out.
       */
      text_file_write(out, _s);

      /* Delete the label residual column name.
       */
      string_delete(&_s);
    }
  }

  text_file_end_record(out);
}


/* Create a column name created from the desired prefix, separator character,
 * and suffix, adding separator characters at the end if the created name
 * clashes with any of the existing column names.
 */
static void create_column_name_(
    char **p_this,
    char *prefix,
    char separator,
    char *suffix,
    ulong num_column_names,
    char **column_names) {
  bool changed;
  ulong n;

  string_new_f(p_this, "%s%c%s", prefix, separator, suffix);

  do {
    for (n = 0, changed = false; n < num_column_names; n++) {
      if (string_equals(*p_this, column_names[n])) {
        string_renew_append_char(p_this, separator);
        changed = true;

        break;
      }
    }
  } while (changed);
}

void prog_bear_parse_features(
    ulong num_features,
    ulong *feature_columns,
    ulong num_fields,
    char **fields,
    char *missing_string,
    double *features) {
  char *_field = null;
  ulong c;
  ulong f;
  double d;

  /* Walk through all the features.
   */
  for (f = 0; f < num_features; f++) {
    /* Extract the column for this feature.
     */
    c = feature_columns[f];

    /* Trim whitespace from the field.
     */
    if (c >= num_fields) {
      string_new_copy(&_field, missing_string);
    } else {
      string_new_trim_whitespace(&_field, fields[c]);
    }

    /* If the feature is missing, Bear requires that we encode it as infinite
     * or NaN; use NaN. Otherwise, parse it as a double.
     */
    if (string_equals(_field, missing_string)) {
      d = DBL_NAN;
    } else {
      d = parse_double(_field);
    }

    /* Store the result.
     */
    features[f] = d;

    /* Clean up.
     */
    string_delete(&_field);
  }
}
