#include "bear_features_list.h"

/* List the names of all the features in an input file. Useful for
 * understanding Bear's final models.
 */
int main(int argc, char *argv[]) {
  bool confirm_header_row;
  bool has_frequency_column;
  char *filename;
  char *filetype;
  char *frequency_column_name;
  char **_column_names = null;
  char **_id_column_names = null;
  char **_label_column_names = null;
  ulong c;
  ulong f;
  ulong i;
  ulong l;
  ulong frequency_column;
  ulong num_columns;
  ulong num_features;
  ulong num_ids;
  ulong num_labels;
  ulong *_feature_columns = null;
  ulong *_id_columns = null;
  ulong *_label_columns = null;
  Logger *_l = null;
  TextFile *_tf = null;

  /* Command-line arguments.
   */
  args_description(
      "List the feature names in a Bear input file");
  args_string("filename", "Name of the input data file");
  prog_bear_args_1(PROG_BEAR_FEATURES);
  args_parse(argc, argv);

  /* Parse arguments.
   */
  filename = args_get_string("filename");
  filetype = args_get_string("filetype");
  has_frequency_column = args_get_flag("has-frequency-column");
  confirm_header_row = args_get_flag("confirm-header-row");

  /* Start logging with the supplied or implied arguments.
   */
  logger_new(&_l, stdout);
  rlog(_l, "Welcome to Bear Machine Learning!");
  rlog(_l, "Filename: '%s'", filename);
  rlog(_l,
      "Filetype: %s%s%s",
      filetype == null ? "" : "'",
      filetype == null ? "(auto-detect)" : filetype,
      filetype == null ? "" : "'");

  /* Parse more of the args.
   */
  prog_bear_parse_args(
    true,
    has_frequency_column,
    &num_labels,
    &_label_columns,
    &_label_column_names,
    &num_ids,
    &_id_columns,
    &_id_column_names,
    &frequency_column,
    &frequency_column_name,
    true);

  /* Open the file.
   */
  text_file_new_type_auto(&_tf, filename, "r", filetype);

  /* Read the header row.
   */
  prog_bear_read_header(
      has_frequency_column,
      confirm_header_row,
      num_labels,
      num_ids,
      "",
      &_label_column_names,
      &_id_column_names,
      frequency_column_name,
      &frequency_column,
      _tf,
      _l,
      &num_columns,
      &_column_names,
      &_label_columns,
      &_id_columns,
      false);

  /* Compute the number of features.
   */
  num_features = num_columns - num_labels - num_ids
    - (has_frequency_column ? 1 : 0);

  /* Construct the set of feature columns.
   */
  new(&_feature_columns, num_features);

  for (c = 0, f = 0, i = 0, l = 0; f < num_features && c < num_columns; c++) {
    if (has_frequency_column && c == frequency_column) {
      rlog(_l, "Frequency column: Column %lu -> '%s'", c, _column_names[c]);

      continue;
    }

    if (_label_columns != null && l < num_labels && c == _label_columns[l]) {
      l++;

      continue;
    }

    if (_id_columns != null && i < num_ids && c == _id_columns[i]) {
      i++;

      continue;
    }

    _feature_columns[f++] = c;
  }

  if (f != num_features) {
    throw("Impossible");
  }

  /* Print the names of the feature columns.
   */
  slog(_l, "Features");

  for (f = 0; f < num_features; f++) {
    rlog(_l,
        "[%lu] Column %lu -> '%s'",
        f,
        _feature_columns[f],
        _column_names[_feature_columns[f]]);
  }

  flog(_l);

  /* Clean up.
   */
  text_file_delete(&_tf);
  deletev(&_feature_columns);
  delete(&_id_column_names, num_labels, string);
  deletev(&_id_columns);
  delete(&_label_column_names, num_labels, string);
  deletev(&_label_columns);
  delete(&_column_names, num_columns, string);

  /* Final feedback.
   */
  rlog(_l, "Thank you for using Bear Machine Learning!");
  logger_delete(&_l);

  return 0;
}
