#include "bear_predict.h"

/* Production program for streaming an arbitrarily large number of prediction
 * examples through a pre-trained Bear model.
 */
int main(int argc, char *argv[]) {
  bool quiet;
  bool confirm_header_row;
  bool has_frequency_column;
  bool has_header_row;
  bool no_features_out;
  char *frequency_column_name;
  char *input_filename;
  char *input_filetype;
  char *missing_string;
  char *output_filename;
  char *output_filetype;
  char *_bear_filename = null;
  char *_s = null;
  char **fields;
  char **_column_names = null;
  char **_id_column_names = null;
  char **_label_column_names = null;
  ulong c;
  ulong f;
  ulong frequency_column;
  ulong i;
  ulong l;
  ulong num_column_names;
  ulong num_columns;
  ulong num_features;
  ulong num_fields;
  ulong num_ids;
  ulong num_labels;
  ulong *_feature_columns = null;
  ulong *_id_columns = null;
  ulong *_label_columns = null;
  long start_time_us;
  double feature;
  double *_features = null;
  double *_labels = null;
  Logger *_l = null;
  Progress *_p = null;
  TextFile *_in = null;
  TextFile *_out = null;
  Bear *_bear = null;

  /* Start the clock.
   */
  start_time_us = unixtime_current_us();

  /* Command-line arguments.
   */
  args_description(
      "Predict by streaming examples through a pre-trained Bear model");
  args_string(
      "input-filename",
      "Name of the input data file, or '-' for stdin");
  args_string(
      "bear-filename",
      "Name of the Bear model file saved by memory_bear");
  args_string(
      "output-filename",
      "Name of the desired output file, or '-' for stdout");
  prog_bear_args_1(PROG_BEAR_PREDICT);
  prog_bear_args_2();
  args_parse(argc, argv);

  /* Check for version request.
   */
  prog_bear_version();

  /* Parse arguments.
   */
  input_filename = args_get_string("input-filename");
  output_filename = args_get_string("output-filename");
  input_filetype = args_get_string("input-filetype");
  output_filetype = args_get_string("output-filetype");
  string_new_copy(&_bear_filename, args_get_string("bear-filename"));
  string_renew_ensure_ends_with(&_bear_filename, ".bear.gz", true);
  has_header_row = args_get_flag("has-header-row");
  confirm_header_row = args_get_flag("confirm-header-row");
  has_frequency_column = args_get_flag("has-frequency-column");
  missing_string = args_get_string("missing-string");
  no_features_out = args_get_flag("no-features-out");
  quiet = args_get_flag("quiet");

  /* Parse more of the args.
   */
  prog_bear_parse_args(
    has_header_row,
    has_frequency_column,
    &num_labels,
    &_label_columns,
    &_label_column_names,
    &num_ids,
    &_id_columns,
    &_id_column_names,
    &frequency_column,
    &frequency_column_name,
    true);

  /* Start logging with the supplied or implied arguments.
   */
  logger_new(&_l,
      quiet ? null : string_equals(output_filename, "-") ? stderr : stdout);

  rlog(_l, "Welcome to Bear Machine Learning!");
  rlog(_l, "Input filename: '%s'", input_filename);
  rlog(_l,
      "Input filetype: %s%s%s",
      input_filetype == null ? "" : "'",
      input_filetype == null ? "(auto-detect)" : input_filetype,
      input_filetype == null ? "" : "'");
  rlog(_l, "Bear model filename: '%s'", _bear_filename);
  rlog(_l, "Output filename: '%s'", output_filename);
  rlog(_l,
      "Output filetype: %s%s%s",
      output_filetype == null ? "" : "'",
      output_filetype == null ? "(auto-detect)" : output_filetype,
      output_filetype == null ? "" : "'");

  /* Open the input and output files.
   */
  text_file_new_type_auto(&_in, input_filename, "r", input_filetype);
  text_file_new_type_auto(&_out, output_filename, "w", output_filetype);

  /* Load the Bear model.
   */
  load_storage(_bear, bear, _bear_filename);

  /* Check that the number of labels specified (if any) agrees with the saved
   * model.
   */
  if (num_labels == 0) {
    num_labels = _bear->numLabels;
  } else if (_bear->numLabels != num_labels) {
    throw("Number of labels specified (%lu) does not agree with that "
        "in the saved Bear model '%s' (%lu)",
        num_labels,
        _bear_filename,
        _bear->numLabels);
  }

  /* Extract information from the model.
   */
  num_features = _bear->numFeatures;
  num_columns = num_features + num_labels + num_ids
    + (has_frequency_column ? 1 : 0);

  /* If we have a header row, read it from the input file and process it.
   */
  if (has_header_row) {
    prog_bear_read_header(
        has_frequency_column,
        confirm_header_row,
        num_labels,
        num_ids,
        missing_string,
        &_label_column_names,
        &_id_column_names,
        frequency_column_name,
        &frequency_column,
        _in,
        _l,
        &num_column_names,
        &_column_names,
        &_label_columns,
        &_id_columns,
        true);
  } else {
    num_column_names = 0;
  }

  /* Check that any specified label columns don't leave any columns unused.
   */
  if (_label_columns != null) {
    for (l = 0; l < num_labels; l++) {
      if (_label_columns[l] >= num_columns) {
        throw("Label column %lu (%lu) exceeds the maximum possible (%lu); "
            "unused columns are not permitted",
            l,
            _label_columns[l],
            num_columns - 1);
      }
    }
  }

  /* Construct the set of feature columns.
   */
  new(&_feature_columns, num_features);

  for (c = 0, f = 0, l = 0, i = 0; f < num_features && c < num_columns; c++) {
    if (has_frequency_column && c == frequency_column) {
      continue;
    }

    if (_label_columns != null && l < num_labels && c == _label_columns[l]) {
      l++;

      continue;
    }

    if (_id_columns != null && i < num_ids && c == _id_columns[i]) {
      i++;

      continue;
    }

    _feature_columns[f++] = c;
  }

  if (f != num_features) {
    throw("Impossible");
  }

  /* If we have a header row, write it to the output file.
   */
  if (has_header_row) {
    prog_bear_write_header(
        true,
        false,
        no_features_out,
        num_ids,
        _id_columns,
        num_features,
        _feature_columns,
        num_labels,
        _label_columns,
        num_columns,
        &num_column_names,
        &_column_names,
        _out);
  }

  /* Progress for the data rows: we don't know how many there will be.
   */
  progress_new_unbounded(
      &_p,
      _l,
      has_header_row ? "data row processed" : "row processed",
      has_header_row ? "data rows processed" : "rows processed");

  /* Create scratch arrays for all the feature values and label values for one
   * row.
   */
  new(&_features, num_features);
  new(&_labels, num_labels);

  /* Continue while we continue to read rows from the input file.
   */
  while (text_file_try_read(_in, &num_fields, &fields)) {
    /* Extract the feature values.
     */
    prog_bear_parse_features(
        num_features,
        _feature_columns,
        num_fields,
        fields,
        missing_string,
        _features);

    /* Write out any ID columns.
     */
    for (i = 0; i < num_ids; i++) {
      c = _id_columns[i];
      text_file_write(_out, c < num_fields ? fields[c] : missing_string);
    }

    /* Write out all the feature values, or the missing string for missing
     * values, unless specified otherwise.
     */
    if (!no_features_out) {
      for (f = 0; f < num_features; f++) {
        if (double_pawized_is_infinite_or_nan(feature = _features[f])) {
          text_file_write(_out, missing_string);
        } else {
          text_file_write_double(_out, feature);
        }
      }
    }

    /* Use our Bear model to predict all the labels for this row of features.
     */
    bear_predict(_bear, _features, _labels);

    /* Write out all the predicted label values.
     */
    for (l = 0; l < num_labels; l++) {
      text_file_write_double(_out, _labels[l]);
    }

    /* If there are label columns specified, write out the true labels.
     */
    if (_label_columns != null) {
      for (l = 0; l < num_labels; l++) {
        c = _label_columns[l];
        text_file_write(_out, c < num_fields ? fields[c] : missing_string);
      }
    }

    /* End the output row.
     */
    text_file_end_record(_out);

    /* Bump progress.
     */
    progress_bump(_p);
  }

  /* Clean up.
   */
  progress_delete(&_p);
  bear_delete(&_bear);
  text_file_delete(&_out);
  text_file_delete(&_in);
  string_delete(&_bear_filename);
  deletev(&_labels);
  deletev(&_features);
  deletev(&_feature_columns);
  delete(&_label_column_names, num_labels, string);
  deletev(&_label_columns);
  delete(&_id_column_names, num_ids, string);
  deletev(&_id_columns);
  delete(&_column_names, num_column_names, string);

  /* Final feedback.
   */
  string_new_time_since(&_s, &start_time_us);
  rlog(_l, "Program execution time: %s", _s);
  string_delete(&_s);
  rlog(_l, "Thank you for using Bear Machine Learning!");
  logger_delete(&_l);

  return 0;
}
