#include "compare_data.h"
#include <math.h>

#define EPSILON_ 1e-15

int main(int argc, char *argv[]) {
  bool quiet;
  char *filename_1;
  char *filename_2;
  char **fields_1;
  char **fields_2;
  ulong f;
  ulong num_fields_1;
  ulong num_fields_2;
  ulong r;
  double d1;
  double d2;
  Logger *_l = null;
  Progress *_p = null;
  TextFile *_tf1 = null;
  TextFile *_tf2 = null;

  args_description("Compare the data in two text files for value equality");
  args_string("filename-1", "Filename of the first data file");
  args_string("filename-2", "Filename of the second data file");
  args_flag("quiet", 'q', "Run quietly, with no progress logging");
  args_parse(argc, argv);

  filename_1 = args_get_string("filename-1");
  filename_2 = args_get_string("filename-2");
  quiet = args_get_flag("quiet");

  logger_new(&_l, quiet ? null : stdout);
  rlog(_l, "File 1: '%s'", filename_1);
  rlog(_l, "File 2: '%s'", filename_2);

  text_file_new(&_tf1, filename_1, "r");
  text_file_new(&_tf2, filename_2, "r");

  slog(_l, "Compare rows");
  progress_new_unbounded(&_p, _l, "row compared", "rows compared");

  for (r = 0; text_file_try_read(_tf1, &num_fields_1, &fields_1); r++) {
    if (!text_file_try_read(_tf2, &num_fields_2, &fields_2)) {
      throw("'%s' had more rows than '%s'", filename_1, filename_2);
    }

    if (num_fields_1 != num_fields_2) {
      throw("Row %lu: '%s' had %lu fields but '%s' had %lu",
          r,
          filename_1,
          num_fields_1,
          filename_2,
          num_fields_2);
    }

    for (f = 0; f < num_fields_1; f++) {
      if (try_parse_double(&d1, fields_1[f])) {
        if (!try_parse_double(&d2, fields_2[f])) {
          throw("Row %lu: field %lu: '%s' field '%s' was parsable as double "
              "%.16g but '%s' field '%s' was not",
              r,
              f,
              filename_1,
              fields_1[f],
              d1,
              filename_2,
              fields_2[f]);
        }

        if (fabs(d2 - d1)
            / maximum(EPSILON_, maximum(fabs(d2), fabs(d1))) > EPSILON_) {
          throw("Row %lu: field %lu: '%s' field '%s' = %.16g but '%s' "
              "field '%s' = %.16g",
              r,
              f,
              filename_1,
              fields_1[f],
              d1,
              filename_2,
              fields_2[f],
              d2);
        }
      } else {
        if (try_parse_double(&d2, fields_2[f])) {
          throw("Row %lu: field %lu: '%s' field '%s' was parsable as double "
              "%.16g but '%s' field '%s' was not",
              r,
              f,
              filename_2,
              fields_2[f],
              d2,
              filename_1,
              fields_1[f]);
        }

        if (!string_equals(fields_1[f], fields_2[f])) {
          throw("Row %lu: field %lu: '%s' field '%s' is different from "
              "'%s' field '%s'",
              r,
              f,
              filename_1,
              fields_1[f],
              filename_2,
              fields_2[f]);
        }
      }
    }

    progress_bump(_p);
  }

  if (text_file_try_read(_tf2, &num_fields_2, &fields_2)) {
    throw("'%s' had more rows than '%s'", filename_2, filename_1);
  }

  text_file_delete(&_tf2);
  text_file_delete(&_tf1);

  progress_delete(&_p);
  flog(_l);
  rlog(_l, "Files compared equal!");
  logger_delete(&_l);

  return 0;
}
