#include "intermediate_bear_tutorial_data.h"
#include <math.h>

#define SIDE_LENGTH_ 7
#define OUTER_R_ 2.5
#define INNER_R_ 1.5
#define HEIGHT_ 2.5
#define DEFAULT_NOISE_STDEV_ 0.125


/* Create a trivariate dataset for the intermediate Bear tutorial.
 */
int main(int argc, char *argv[]) {
  char *filename;
  char *production_filename;
  char *x_filename;
  char *x_production_filename;
  ulong i;
  ulong j;
  ulong samples_per_side;
  ulong x_samples_per_side;
  double noise_stdev;
  double r;
  double scale;
  double x;
  double x_scale;
  double y;
  double z;
  Logger *_l = null;
  Random *_random = null;
  TextFile *_tf = null;
  TextFile *_tf_x = null;

  /* Arguments. The defaults will produce a file like that supplied on the page
   * https://johncostella.com/bear/intermediate/.
   */
  args_description(
      "Create a trivariate dataset for the intermediate Bear tutorial "
      "representing the height of a piece of brass tube sitting on the ground "
      "plus random noise, and its bivariate (x, z) projection");
  args_string(
      "filename",
      "Filename of the full trivariate (x,\fy,\fz) data file");
  args_integer(
      "samples-per-side",
      "Number of samples on each side of the base, in [2, 255]");
  args_string(
      "x-filename",
      "Filename of the bivariate (x,\fz) projection data file");
  args_integer(
      "x-samples-per-side",
      "Number of samples on each side of the base for the bivariate data, "
      "(minimum\f2)");
  args_optional_string(
      "production-filename",
      'p',
      ARGS_STRING_NO_DEFAULT,
      "Write the prediction (x,\fy) values to this filename, "
      "rather than FILENAME");
  args_optional_string(
      "x-production-filename",
      'P',
      ARGS_STRING_NO_DEFAULT,
      "Write the bivariate prediction x values to this filename, "
      "rather than X_FILENAME");
  args_optional_real(
      "noise-stdev",
      'e',
      DEFAULT_NOISE_STDEV_,
      "Standard deviation of the noise to be added to the linear correlation");
  args_optional_integer(
      "random-seed",
      'r',
      ARGS_INTEGER_NO_DEFAULT,
      "Seed for the random number generator (default\f=\fautomatic)");
  args_parse(argc, argv);

  logger_new(&_l, stdout);
  filename = args_get_string("filename");
  production_filename = args_get_string("production-filename");
  samples_per_side = args_get_bounded_integer("samples-per-side", 2, 255);
  x_filename = args_get_string("x-filename");
  x_production_filename = args_get_string("x-production-filename");
  x_samples_per_side = args_get_bounded_integer(
      "x-samples-per-side",
      2,
      ARGS_INTEGER_NO_DEFAULT);
  noise_stdev = args_get_real("noise-stdev");

  rlog(_l, "Filename: '%s'", filename);
  rlog(_l, "Samples per side: %lu", samples_per_side);
  rlog(_l, "X-projection filename: '%s'", x_filename);
  rlog(_l, "X-projection samples per side: %lu", x_samples_per_side);
  rlog(_l, "Noise standard deviation: %.16g", noise_stdev);

  if (args_set("random-seed")) {
    rlog(_l, "Random seed: %lu", args_get_positive_integer("random-seed"));
  }

  scale = (double)SIDE_LENGTH_ / (samples_per_side - 1);
  x_scale = (double)SIDE_LENGTH_ / (x_samples_per_side - 1);
  text_file_new_type(&_tf, args_get_string("filename"), "w", "csv");
  text_file_new_type(&_tf_x, args_get_string("x-filename"), "w", "csv");

  if (args_set("random-seed")) {
    random_new_seed(&_random, args_get_positive_integer("random-seed"));
  } else {
    random_new(&_random);
  }

  /* Write out the full trivariate training data.
   */
  for (j = 0; j < samples_per_side; j++) {
    y = ((double)j - 0.5 * (samples_per_side - 1)) * scale;

    for (i = 0; i < samples_per_side; i++) {
      x = ((double)i - 0.5 * (samples_per_side - 1)) * scale;

      r = sqrt(x * x + y * y);
      z = (r <= OUTER_R_ && r >= INNER_R_ ? HEIGHT_ : 0)
        + noise_stdev * random_normal(_random);

      text_file_write_double(_tf, x);
      text_file_write_double(_tf, y);
      text_file_write_double(_tf, z);
      text_file_end_record(_tf);
    }
  }

  /* Write out the bivariate training data, dithering the values in the x
   * direction so that we can see them in Excel.
   */
  for (j = 0; j < x_samples_per_side; j++) {
    y = ((double)j - 0.5 * (x_samples_per_side - 1)) * x_scale;

    for (i = 0; i < x_samples_per_side; i++) {
      x = ((double)i - 0.5 * (x_samples_per_side - 1)) * x_scale;

      r = sqrt(x * x + y * y);
      z = (r <= OUTER_R_ && r >= INNER_R_ ? HEIGHT_ : 0)
        + noise_stdev * random_normal(_random);

      text_file_write_double(
          _tf_x,
          x + (random_double_unit_oo(_random) - 0.5) * x_scale);
      text_file_write_double(_tf_x, z);
      text_file_end_record(_tf_x);
    }
  }

  if (production_filename != null) {
    text_file_delete(&_tf);
    text_file_new(&_tf, production_filename, "w");
  }

  if (x_production_filename != null) {
    text_file_delete(&_tf_x);
    text_file_new(&_tf_x, x_production_filename, "w");
  }

  /* Write out the full trivariate testing data.
   */
  for (j = 0; j < samples_per_side; j++) {
    y = ((double)j - 0.5 * (samples_per_side - 1)) * scale;

    for (i = 0; i < samples_per_side; i++) {
      x = ((double)i - 0.5 * (samples_per_side - 1)) * scale;

      text_file_write_double(_tf, x);
      text_file_write_double(_tf, y);
      text_file_end_record(_tf);
    }
  }

  /* Write out the bivariate testing data rows for a limited number of values
   * of j, to give us enough for graphing without overkill, again dithering the
   * values in the x direction.
   */
  for (j = 0;
      j < maximum(1, minimum(x_samples_per_side, 900 / x_samples_per_side));
      j++) {
    for (i = 0; i < x_samples_per_side; i++) {
      x = ((double)i - 0.5 * (x_samples_per_side - 1)) * x_scale;

      text_file_write_double(
          _tf_x,
          x + (random_double_unit_oo(_random) - 0.5) * x_scale);
      text_file_end_record(_tf_x);
    }
  }

  random_delete(&_random);
  text_file_delete(&_tf_x);
  text_file_delete(&_tf);
  logger_delete(&_l);

  return 0;
}
