#include "memory_bear.h"

/* Program for running a fully in-memory implementation of Bear on a single
 * separated-values text file.
 */
int main(int argc, char *argv[]) {
  bool confirm_header_row;
  bool debug;
  bool has_frequency_column;
  bool has_header_row;
  bool missing_label;
  bool no_features_out;
  bool quiet;
  bool train;
  char *input_filename;
  char *input_filetype;
  char *missing_string;
  char *output_filename;
  char *output_filetype;
  char *time_budget;
  char *frequency_column_name;
  ulong *_feature_columns = null;
  char *_field = null;
  char *_s = null;
  char *_save_filename = null;
  char **fields;
  char **_column_names = null;
  char **_id_column_names = null;
  char **_label_column_names = null;
  ulong c;
  ulong f;
  ulong frequency_column;
  ulong i;
  ulong id_column;
  ulong l;
  ulong num_column_names;
  ulong num_columns;
  ulong num_features;
  ulong num_fields;
  ulong num_ids;
  ulong num_labels;
  ulong num_predict_rows;
  ulong num_rows;
  ulong num_train_rows;
  ulong r;
  ulong t;
  ulong u;
  ulong *_id_columns = null;
  ulong *_label_columns = null;
  ulong *_frequencies = null;
  SwappableUlongs *_frequencies_s = null;
  long max_time_us;
  long start_time_us;
  double d;
  double feature;
  double gib;
  double label;
  double *_features = null;
  double *_predict_features_in = null;
  double *_predict_labels_out = null;
  double **_predict_features = null;
  SwappableDoubles **_predict_features_s = null;
  double **_train_features = null;
  SwappableDoubles **_train_features_s = null;
  double **_train_labels = null;
  SwappableDoubles **_train_labels_s = null;
  LossFunction *loss_function;
  TimeBudget budget;
  Logger *_l = null;
  Progress *_p = null;
  TextFile *ids;
  TextFile *_predict_ids = null;
  TextFile *_train_ids = null;
  TextFile *_in = null;
  TextFile *_out = null;
  Bear *_bear = null;

  /* The user will specify a time budget, so start the clock now.
   */
  start_time_us = unixtime_current_us();

  /* Command-line arguments.
   */
  args_description("In-memory implementation of Bear");
  args_string(
      "input-filename",
      "Name of the input data file; rows with empty labels will be predicted "
      "and written to the output file");
  args_string(
      "time-budget",
      "Maximum amount of wall time for training models: formats %s",
      TIME_BUDGET_FORMATS);
  args_optional_string(
      "save-filename",
      's',
      ARGS_STRING_NO_DEFAULT,
      "Save the final Bear model to this filename");
  args_optional_string(
      "output-filename",
      'o',
      ARGS_STRING_NO_DEFAULT,
      "Write the predictions to this filename");
  prog_bear_args_1(PROG_BEAR_MEMORY);

  string_new_copy(&_s, "The loss function, which must be one of:");

  for (f = 0; f < LOSS_FUNCTIONS_NUM; f++) {
    string_renew_append_f(
        &_s,
        "\n\t%s: %s",
        LOSS_FUNCTIONS[2 * f],
        LOSS_FUNCTIONS[2 * f + 1]);
  }

  args_optional_string("loss-function", 'n', LOSS_FUNCTION_MSE, _s);
  string_delete(&_s);

  prog_swappable_args();
  args_optional_string(
      "details-filename",
      'D',
      ARGS_STRING_NO_DEFAULT,
      "Write details of the final model to this filename");
  args_flag("debug", 'd', "Also write training rows to the output file");
  prog_bear_args_2();
  args_parse(argc, argv);

  /* Check for version request.
   */
  prog_bear_version();

  /* Get filenames.
   */
  input_filename = args_get_string("input-filename");

  if (args_set("save-filename")) {
    string_new_copy(&_save_filename, args_get_string("save-filename"));
    string_renew_ensure_ends_with(&_save_filename, ".bear.gz", true);
  }

  output_filename = args_get_string("output-filename");

  /* Sanity check.
   */
  if (output_filename != null
      && string_equals(input_filename, output_filename)) {
    throw("Input and output filenames cannot both be '%s'", input_filename);
  }

  /* More arguments.
   */
  input_filetype = args_get_string("input-filetype");
  output_filetype = args_get_string("output-filetype");
  time_budget = args_get_string("time-budget");
  has_header_row = args_get_flag("has-header-row");
  confirm_header_row = args_get_flag("confirm-header-row");
  has_frequency_column = args_get_flag("has-frequency-column");
  missing_string = args_get_string("missing-string");
  no_features_out = args_get_flag("no-features-out");
  quiet = args_get_flag("quiet");
  debug = args_get_flag("debug");
  gib = args_get_positive_real("gibibytes-limit");
  loss_function = loss_function_get(args_get_string("loss-function"));

  /* Parse the time budget string.
   */
  budget = time_budget_parse(time_budget);

  /* Compute the maximum unixtime before we have to exit.
   */
  max_time_us = start_time_us + budget.us;

  /* Parse more of the args.
   */
  prog_bear_parse_args(
    has_header_row,
    has_frequency_column,
    &num_labels,
    &_label_columns,
    &_label_column_names,
    &num_ids,
    &_id_columns,
    &_id_column_names,
    &frequency_column,
    &frequency_column_name,
    false);

  /* Sanity checks.
   */
  if (output_filename != null && !file_is_writable(output_filename)) {
    throw("Output file '%s' is not writable", output_filename);
  }

  if (_save_filename != null && !file_is_writable(_save_filename)) {
    throw("Save file '%s' is not writable", _save_filename);
  }

  if (output_filename == null && _save_filename == null) {
    throw("At least one of --save-filename or --output-filename must be "
        "specified");
  }

  if (debug && output_filename == null) {
    throw("Output filename must be specified for debug mode");
  }

  /* Start logging with the supplied or implied arguments.
   */
  logger_new(&_l, quiet ? null : stdout);

  rlog(_l, "Welcome to Bear Machine Learning!");
  rlog(_l, "Input filename: '%s'", input_filename);
  rlog(_l,
      "Input filetype: %s%s%s",
      input_filetype == null ? "" : "'",
      input_filetype == null ? "(auto-detect)" : input_filetype,
      input_filetype == null ? "" : "'");

  if (output_filename != null) {
    rlog(_l, "Output filename: '%s'", output_filename);
    rlog(_l,
        "Output filetype: %s%s%s",
        output_filetype == null ? "" : "'",
        output_filetype == null ? "(auto-detect)" : output_filetype,
        output_filetype == null ? "" : "'");
  }

  if (_save_filename != null) {
    rlog(_l, "Model save filename: '%s'", _save_filename);
  }

  rlog(_l, "Loss function: %s", loss_function->description);

  string_new_time_budget(&_s, budget);
  rlog(_l, "Time budget: %s", _s);
  string_delete(&_s);

  if (gib != ARGS_REAL_NO_DEFAULT) {
    rlog(_l, "Swappable memory limit: %.16g GiB", gib);
    swappable_set_in_memory_limit(gibibytes_to_bytes(gib));
  }

  /* More feedback.
   */
  prog_bear_feedback_1(
    has_header_row,
    has_frequency_column,
    _label_column_names,
    num_labels,
    _label_columns,
    frequency_column,
    frequency_column_name,
    missing_string,
    _l);

  /* Open the input and output files.
   */
  if (output_filename != null) {
    text_file_new_type_auto(&_out, output_filename, "w", output_filetype);
  }

  text_file_new_type_auto(&_in, input_filename, "r", input_filetype);

  /* First pass of reading the input file. We check the header row (if present)
   * and then read all the data rows, checking that each field is either the
   * null string or can be parsed as a double. This also lets us count the
   * number of data rows.
   */
  slog(_l, "Read the input file '%s' in a first pass", input_filename);

  /* Time budget check.
   */
  if (unixtime_current_us() > max_time_us) {
    throw("Time budget exhausted before we even read the input file!");
  }

  /* Check if we are supposed to have a header row.
   */
  if (has_header_row) {
    /* We are. Read and process it.
     */
    prog_bear_read_header(
        has_frequency_column,
        confirm_header_row,
        num_labels,
        num_ids,
        missing_string,
        &_label_column_names,
        &_id_column_names,
        frequency_column_name,
        &frequency_column,
        _in,
        _l,
        &num_column_names,
        &_column_names,
        &_label_columns,
        &_id_columns,
        true);
  }

  /* Check that no specified columns have been repeated. We can't do this until
   * we have potentially processed the header row above.
   */
  prog_bear_check_columns(
      num_labels,
      _label_columns,
      num_ids,
      _id_columns,
      has_frequency_column,
      frequency_column);

  /* Progress for the data rows: we don't yet know how many there are.
   */
  progress_new_unbounded(
      &_p,
      _l,
      has_header_row ? "data row read" : "row read",
      has_header_row ? "data rows read" : "rows read");

  /* Read data rows until there are no more. We start with no columns, and
   * expand it as we find rows with more columns.
   */
  for (num_rows = 0, num_columns = 0;
      text_file_try_read(_in, &num_fields, &fields);
      num_rows++) {
    /* Time budget check.
     */
    if (unixtime_current_us() > max_time_us) {
      throw("Time budget exhausted!");
    }

    /* Update the maximum number of columns that we have found in any row.
     */
    num_columns = maximum(num_columns, num_fields);

    /* Walk through all of the read fields in this row.
     */
    for (c = 0, i = 0; c < num_fields; c++) {
      /* If this is an ID column, skip it.
       */
      if (i < num_ids && _id_columns[i] == c) {
        i++;

        continue;
      }

      /* Trim whitespace from the field.
       */
      string_new_trim_whitespace(&_field, fields[c]);

      /* Null fields are fine.
       */
      if (string_equals(_field, missing_string)) {
        string_delete(&_field);

        continue;
      }

      /* Otherwise, it must be able to be parsed as a double.
       */
      if (!try_parse_double(&d, _field)) {
        /* If we find a non-number in the first row and a header row was not
         * specified, it's likely that this is actually a header row; reflect
         * that in the error message. Otherwise, note that Bear currently does
         * not handle categorical data.
         */
        if (!has_header_row && num_rows == 0) {
          throw("Found text '%s' in column %lu of the first row: "
              "use --has-header-row if your file has a header row "
              "(sorry, Bear currently only handles ordinal numerical data)",
              fields[c],
              c);
        } else {
          throw("Found text '%s' in column %lu of data row %lu "
              "(sorry, Bear currently only handles ordinal numerical data)",
              _field,
              c,
              num_rows);
        }
      }

      /* Clean up.
       */
      string_delete(&_field);
    }

    progress_bump(_p);
  }

  /* Finished the first pass.
   */
  progress_delete(&_p);
  flog(_l);

  /* Sanity check.
   */
  if (num_rows == 0) {
    throw("Number of data rows cannot be zero");
  }

  /* Print the number of columns that we found in the data rows of the file.
   */
  rlog(_l, "Number of columns: %lu", num_columns);

  /* Check that all label column indexes are valid.
   */
  for (l = 0; l < num_labels; l++) {
    if (_label_columns[l] >= num_columns) {
      throw("Label column %lu is invalid: max %lu",
          _label_columns[l],
          num_columns - 1);
    }
  }

  /* Check that all ID column indexes are valid.
   */
  for (i = 0; i < num_ids; i++) {
    if (_id_columns[i] >= num_columns) {
      throw("ID column %lu is invalid: max %lu",
          _id_columns[i],
          num_columns - 1);
    }
  }

  /* Check that the frequency column, if specified, is valid.
   */
  if (has_frequency_column && frequency_column >= num_columns) {
    throw("Frequency column %lu is invalid: max %lu",
        frequency_column,
        num_columns - 1);
  }

  /* Compute and print the number of features, which is the number of columns
   * that are not a label column, an ID column, or the frequency column.
   */
  num_features = num_columns - num_labels - num_ids
    - (has_frequency_column ? 1 : 0);
  rlog(_l, "Number of features: %lu", num_features);

  /* Create an array of the feature column indexes. Create the array.
   */
  new(&_feature_columns, num_features);

  /* Walk through the columns.
   */
  for (c = 0, f = 0, l = 0, i = 0; c < num_columns; c++) {
    if (has_frequency_column && c == frequency_column) {
      /* Skip the frequency column.
       */
    } else if (l < num_labels && _label_columns[l] == c) {
      /* If this is a label column, skip it, and move to the next label column.
       * Note that this works because they are sorted into ascending order.
       */
      l++;
    } else if (i < num_ids && _id_columns[i] == c) {
      /* If this is an ID column, skip it, and move to the next ID column. Note
       * that this works because they are sorted into ascending order.
       */
      i++;
    } else {
      /* Sanity check.
       */
      if (f >= num_features) {
        throw("Impossible: f = %lu, num_features = %lu", f, num_features);
      }

      /* Store the column index for this feature.
       */
      _feature_columns[f++] = c;
    }
  }

  /* Sanity checks.
   */
  if (c != num_columns
      || f != num_features
      || l != num_labels
      || i != num_ids) {
    throw("Impossible");
  }

  /* Print the feature column(s).
   */
  if (num_features == 1) {
    if (has_header_row && _feature_columns[0] < num_column_names) {
      rlog(_l,
          "Feature column: %lu -> '%s'",
          _feature_columns[0],
          _column_names[_feature_columns[0]]);
    } else {
      rlog(_l, "Feature column: %lu", _feature_columns[0]);
    }
  } else if (num_features >= 2) {
    slog(_l, "Feature columns");

    for (f = 0; f < num_features; f++) {
      if (has_header_row && _feature_columns[f] < num_column_names) {
        rlog(_l,
            "[%lu] Column %lu -> '%s'",
            f,
            _feature_columns[f],
            _column_names[_feature_columns[f]]);
      } else {
        rlog(_l, "[%lu] Column %lu", f, _feature_columns[f]);
      }
    }

    flog(_l);
  }

  /* Second pass of reading the input file. We check whether each row is a
   * training row, with nonnull label(s), or a prediction row, with null
   * label(s). We also check all training rows that the frequency column, if
   * specified, exists and is a valid positive integer.
   */
  slog(_l, "Read the input file '%s' in a second pass", input_filename);
  progress_new(&_p, _l, num_rows);
  text_file_rewind(_in);

  /* If we have a header row, skip it.
   */
  if (has_header_row) {
    text_file_read_discard(_in);
  }

  /* Walk through the rows.
   */
  for (r = 0, num_train_rows = 0, num_predict_rows = 0; r < num_rows; r++) {
    /* Time budget check.
     */
    if (unixtime_current_us() > max_time_us) {
      throw("Time budget exhausted!");
    }

    /* Read the row.
     */
    num_fields = text_file_read(_in, &fields);

    /* Walk through all the label columns. We start by assuming that the row is
     * a training row, i.e., has nonnull label(s).
     */
    for (l = 0, train = true; l < num_labels; l++) {
      /* This label is null if it either doesn't exist in the read row or if it
       * exists but is equal to the null string.
       */
      missing_label = _label_columns[l] >= num_fields
        || string_equals(fields[_label_columns[l]], missing_string);

      /* Check for the case where there are both null and nonnull labels in the
       * same row.
       */
      if ((missing_label && train && l > 0) || (!missing_label && !train)) {
          throw("Row %lu has both null and nonnull labels: "
              "they must all be one or the other",
              r);
      }

      /* If the label is null then this is not a training row: it is a
       * prediction row.
       */
      if (missing_label) {
        train = false;
      }
    }

    /* Switch on whether this is a training or a prediction row.
     */
    if (train) {
      /* Training row. If a frequency column has been specified, check that it
       * exists, is not the null string, and is a positive integer.
       */
      if (has_frequency_column) {
        if (frequency_column >= num_fields) {
          throw("Frequency column %lu does not exist for row %lu "
              "(training row %lu), which only has %lu fields",
              frequency_column,
              r,
              num_train_rows,
              num_fields);
        }

        if (string_equals(fields[frequency_column], missing_string)) {
          throw("Frequency cell for row %lu (training row %lu) is empty, "
              "which is not allowed",
              r,
              num_train_rows);
        }

        if (!try_parse_ulong(&u, fields[frequency_column])) {
          throw("Frequency cell '%s' for row %lu (training row %lu) "
              "is a number but not a positive integer, which is not allowed",
              fields[frequency_column],
              r,
              num_train_rows);
        }

        if (u == 0) {
          throw("Frequency for row %lu (training row %lu) is zero, "
              "which is not allowed",
              r,
              num_train_rows);
        }
      }

      /* Increment the number of training rows.
       */
      num_train_rows++;
    } else {
      /* Prediction row. Increment the number of them.
       */
      num_predict_rows++;
    }

    progress_bump(_p);
  }

  /* Sanity checks.
   */
  if (num_train_rows == 0) {
    throw("No training rows (i.e., with label%s specified) found",
        plural_s(num_labels));
  }

  if (!debug && (num_predict_rows == 0 && output_filename != null)) {
    throw("No prediction rows (i.e., with label%s not specified) found, but "
        "an --output-filename of '%s' was specified",
        plural_s(num_labels),
        output_filename);
  }

  if (num_predict_rows != 0 && output_filename == null) {
    throw("%lu prediction row%s (i.e., with label%s not specified) w%s found, "
        "but no --output-filename was specified",
        num_predict_rows,
        plural_s(num_predict_rows),
        plural_s(num_labels),
        plural(num_predict_rows, "as", "ere"));
  }

  /* Finished the second pass.
   */
  progress_delete(&_p);
  flog(_l);

  /* Print the numbers of training and predictions rows.
   */
  rlog(_l, "Number of training rows: %lu", num_train_rows);
  rlog(_l, "Number of prediction rows: %lu", num_predict_rows);

  /* Create and access swappable matrixes for the training and prediction
   * features and labels.
   */
  swappable_doubles_newm_access(
      &_train_features_s,
      num_features,
      num_train_rows,
      &_train_features);
  swappable_doubles_newm_access(
      &_train_labels_s,
      num_labels,
      num_train_rows,
      &_train_labels);
  swappable_doubles_newm_access(
      &_predict_features_s,
      num_features,
      num_predict_rows,
      &_predict_features);

  /* If we have a frequency column, create and access a swappable array for it.
   * Note that only training rows have frequencies.
   */
  if (has_frequency_column) {
    swappable_ulongs_new_access(
        &_frequencies_s,
        num_train_rows,
        &_frequencies);
  }

  /* Third pass of reading the input file. This time we read in all the data.
   */
  slog(_l, "Read the input file '%s' in a third pass", input_filename);
  progress_new(&_p, _l, num_rows);
  text_file_rewind(_in);

  /* Skip any header row.
   */
  if (has_header_row) {
    text_file_read_discard(_in);
  }

  /* Create a scratch array of all the feature values for one row.
   */
  new(&_features, num_features);

  /* If we have any IDs, create MemoryFiles for them.
   */
  if (num_ids > 0) {
    text_file_new_memory(&_train_ids, "nsv");
    text_file_new_memory(&_predict_ids, "nsv");
  }

  /* Walk through all the rows.
   */
  for (r = 0, t = 0, u = 0; r < num_rows; r++) {
    /* Time budget check.
     */
    if (unixtime_current_us() > max_time_us) {
      throw("Time budget exhausted!");
    }

    /* Read the row.
     */
    num_fields = text_file_read(_in, &fields);

    /* Figure out if this is a training row or a prediction row.
     */
    for (l = 0, train = true; l < num_labels; l++) {
      if (_label_columns[l] >= num_fields
          || string_equals(fields[_label_columns[l]], missing_string)) {
        train = false;

        break;
      }
    }

    /* If this is a training row and we have a frequency column, parse and
     * store the frequency.
     */
    if (train && has_frequency_column) {
      _frequencies[t] = parse_ulong(fields[frequency_column]);
    }

    /* If this is a training row, parse and store the label(s), which cannot be
     * infinite or not a number.
     */
    if (train) {
      for (l = 0; l < num_labels; l++) {
        d = parse_double(fields[_label_columns[l]]);

        if (double_pawized_is_infinite_or_nan(d)) {
          if (double_is_infinite_or_nan(d)) {
            throw("Label %lu ('%s') for data row %lu is infinite or NaN, "
                "which is not allowed",
                l,
                fields[_label_columns[l]],
                r);
          }

          throw("Label %lu ('%s') for data row %lu is out of range for a paw, "
              "which is not allowed",
              l,
              fields[_label_columns[l]],
              r);
        }

        _train_labels[l][t] = d;
      }
    }

    /* For all rows, parse and store all the features.
     */
    prog_bear_parse_features(
        num_features,
        _feature_columns,
        num_fields,
        fields,
        missing_string,
        _features);

    /* Store the features in the appropriate matrix.
     */
    for (f = 0; f < num_features; f++) {
      d = _features[f];

      if (train) {
        _train_features[f][t] = d;
      } else {
        _predict_features[f][u] = d;
      }
    }

    /* If we have IDs, extract and store them.
     */
    if (num_ids > 0) {
      /* Get a reference to the text file we will use.
       */
      ids = train ? _train_ids : _predict_ids;

      /* Go through the ID columns, and write out the strings from the row.
       */
      for (i = 0; i < num_ids; i++) {
        if ((id_column = _id_columns[i]) >= num_fields) {
          text_file_write_empty_field(ids);
        } else {
          text_file_write(ids, fields[id_column]);
        }
      }

      /* End the record.
       */
      text_file_end_record(ids);
    }

    /* Increment the training row or prediction row index as appropriate.
     */
    if (train) {
      t++;
    } else {
      u++;
    }

    progress_bump(_p);
  }

  /* Finished the third pass.
   */
  deletev(&_features);
  progress_delete(&_p);
  text_file_delete(&_in);
  flog(_l);
  rlog(_l,
      "Extracted data for %lu training row%s and %lu prediction row%s.",
      num_train_rows,
      plural_s(num_train_rows),
      num_predict_rows,
      plural_s(num_predict_rows));

  /* Time budget check.
   */
  if (unixtime_current_us() > max_time_us) {
    throw("Time budget exhausted before we even got to launch Bear!");
  }

  /* Release the swappables so that they can be swapped out by Bear if needed.
   */
  swappable_values_releasem(_train_features_s, num_features, &_train_features);
  swappable_values_releasem(_train_labels_s, num_labels, &_train_labels);
  swappable_values_releasem(
      _predict_features_s,
      num_features,
      &_predict_features);

  if (has_frequency_column) {
    swappable_values_release(_frequencies_s, &_frequencies);
  }

  /* If we have any IDs, rewind and release their files.
   */
  if (num_ids > 0) {
    text_file_rewind(_train_ids);
    text_file_release(_train_ids);
    text_file_rewind(_predict_ids);
    text_file_release(_predict_ids);
  }

  /* Launch Bear to model the data. The time budget is the specified budget
   * minus whatever time we have used up to this point.
   */
  rlog(_l,
      "Launch Bear to model %lu label%s with %lu feature%s "
      "using %lu example%s.",
      num_labels,
      plural_s(num_labels),
      num_features,
      plural_s(num_features),
      num_train_rows,
      plural_s(num_train_rows));

  bear_new(&_bear,
      num_features,
      num_labels,
      num_train_rows,
      _train_features_s,
      _train_labels_s,
      _frequencies_s,
      loss_function,
      start_time_us,
      max_time_us,
      _l);

  /* If we have a save filename specified, save the Bear model and verify the
   * integrity of that saved model.
   */
  if (_save_filename != null) {
    slog(_l, "Save the final Bear model to '%s'", _save_filename);
    save_storage_verify(_bear, bear, Bear, _save_filename);
    flog(_l);
  }

  /* If we have an output filename specified, write out the predictions. If we
   * are in debug mode, also first write back out the training rows, together
   * with predictions for them and resulting residuals.
   */
  if (output_filename != null) {
    slog(_l, "Write the output file '%s'", output_filename);

    /* If we have a header row, write it out.
     */
    if (has_header_row) {
      prog_bear_write_header(
          false,
          debug,
          no_features_out,
          num_ids,
          _id_columns,
          num_features,
          _feature_columns,
          num_labels,
          _label_columns,
          num_columns,
          &num_column_names,
          &_column_names,
          _out);
    }

    /* Create arrays for the features and labels for each example.
     */
    new(&_predict_features_in, num_features);
    new(&_predict_labels_out, num_labels);

    /* If we are in debug mode, write out all the training rows. together with
     * predictions for them and resulting residuals.
     */
    if (debug) {
      /* Access the required swappables.
       */
      swappable_values_accessm(
          _train_features_s,
          num_features,
          &_train_features);
      swappable_values_accessm(_train_labels_s, num_labels, &_train_labels);

      if (num_ids > 0) {
        text_file_reacquire(_train_ids);
      }

      /* Walk through all the training rows.
       */
      for (t = 0; t < num_train_rows; t++) {
        /* Write out any IDs.
         */
        if (num_ids > 0) {
          if ((num_fields = text_file_read(_train_ids, &fields)) != num_ids) {
            throw("Impossible");
          }

          for (i = 0; i < num_ids; i++) {
            text_file_write(_out, fields[i]);
          }
        }

        /* Walk through the features.
         */
        for (f = 0; f < num_features; f++) {
          /* Get the feature value.
           */
          feature = _train_features[f][t];

          /* Copy it to our prediction input array.
           */
          _predict_features_in[f] = feature;

          /* Only write it out if we want it.
           */
          if (!no_features_out) {
            /* Check if the feature value is missing.
             */
            if (double_pawized_is_infinite_or_nan(feature)) {
              /* It is missing. Write out the missing string.
               */
              text_file_write(_out, missing_string);
            } else {
              /* It's not missing. Write it out.
               */
              text_file_write_double(_out, feature);
            }
          }
        }

        /* Walk through the labels.
         */
        for (l = 0; l < num_labels; l++) {
          /* Get the label for this row.
           */
          label = _train_labels[l][t];

          /* Write it out.
           */
          text_file_write_double(_out, label);
        }

        /* Use our Bear model to predict all the labels for this row of
         * features.
         */
        bear_predict(_bear, _predict_features_in, _predict_labels_out);

        /* Write out all the predicted label values.
         */
        for (l = 0; l < num_labels; l++) {
          text_file_write_double(_out, _predict_labels_out[l]);
        }

        /* Compute and write out all the residuals.
         */
        for (l = 0; l < num_labels; l++) {
          text_file_write_double(
              _out,
              _train_labels[l][t] - _predict_labels_out[l]);
        }

        /* End the record.
         */
        text_file_end_record(_out);
      }

      /* Release the swappables.
       */
      swappable_values_releasem(
          _train_features_s,
          num_features,
          &_train_features);
      swappable_values_releasem(_train_labels_s, num_labels, &_train_labels);

      if (num_ids > 0) {
        text_file_release(_train_ids);
      }
    }

    /* Access the predictions input data.
     */
    swappable_values_accessm(
        _predict_features_s,
        num_features,
        &_predict_features);

    if (num_ids > 0) {
      text_file_reacquire(_predict_ids);
    }

    /* Write out all the prediction rows, using our Bear to make all of the
     * required predictions. Walk through all the prediction rows.
     */
    for (u = 0; u < num_predict_rows; u++) {
      /* Write out any IDs.
       */
      if (num_ids > 0) {
        if ((num_fields = text_file_read(_predict_ids, &fields)) != num_ids) {
          throw("Impossible");
        }

        for (i = 0; i < num_ids; i++) {
          text_file_write(_out, fields[i]);
        }
      }

      /* Walk through the features.
       */
      for (f = 0; f < num_features; f++) {
        /* Extract the feature value.
         */
        feature = _predict_features[f][u];

        /* Store it.
         */
        _predict_features_in[f] = feature;

        /* Only write it out if we want it.
         */
        if (!no_features_out) {
          /* Write it out, or the missing string if it is missing.
           */
          if (double_pawized_is_infinite_or_nan(feature)) {
            text_file_write(_out, missing_string);
          } else {
            text_file_write_double(_out, feature);
          }
        }
      }

      /* Use our Bear model to predict all the labels for this row of features.
       */
      bear_predict(_bear, _predict_features_in, _predict_labels_out);

      /* Write out all the predicted label values.
       */
      for (l = 0; l < num_labels; l++) {
        text_file_write_double(_out, _predict_labels_out[l]);
      }

      /* End the record.
       */
      text_file_end_record(_out);
    }

    flog(_l);

    /* Clean up.
     */
    deletev(&_predict_features_in);
    deletev(&_predict_labels_out);
    swappable_values_releasem(
        _predict_features_s,
        num_features,
        &_predict_features);

    if (num_ids > 0) {
      text_file_release(_predict_ids);
    }
  }

  /* If we've been asked for a prediction models file, write it.
   */
  if (args_set("details-filename")) {
    slog(_l,
        "Create final model details file '%s'",
        args_get_string("details-filename"));
    bear_write_model_summary(_bear, args_get_string("details-filename"));
    flog(_l);
  }

  /* Clean up.
   */
  if (num_ids > 0) {
    text_file_reacquire(_train_ids);
    text_file_delete(&_train_ids);
    text_file_reacquire(_predict_ids);
    text_file_delete(&_predict_ids);
  }

  bear_delete(&_bear);
  text_file_delete(&_out);
  delete(&_column_names, num_column_names, string);
  swappable_values_delete(&_frequencies_s);
  swappable_values_deletem(&_predict_features_s, num_features);
  swappable_values_deletem(&_train_labels_s, num_labels);
  swappable_values_deletem(&_train_features_s, num_features);
  deletev(&_feature_columns);
  delete(&_label_column_names, num_labels, string);
  deletev(&_label_columns);
  delete(&_id_column_names, num_ids, string);
  deletev(&_id_columns);
  string_delete(&_save_filename);

  /* Final feedback.
   */
  string_new_time_since(&_s, &start_time_us);
  rlog(_l, "Program execution time: %s", _s);
  string_delete(&_s);
  rlog(_l, "Thank you for using Bear Machine Learning!");
  logger_delete(&_l);

  return 0;
}
