#include "simple_bear_tutorial_data.h"
#include <math.h>

#define DEFAULT_NUM_TRAIN_ROWS 50
#define DEFAULT_NUM_PREDICT_ROWS 250
#define DEFAULT_X_MIN_ -2.5
#define DEFAULT_X_MAX_ 2.5
#define DEFAULT_BIAS_ 10
#define DEFAULT_WEIGHT_ 3
#define DEFAULT_QUADRATIC_ 0
#define DEFAULT_NOISE_STDEV_ 1.5
#define DEFAULT_NULL_PERCENTAGE_ 0
#define DEFAULT_NULL_BIAS_ 100
#define DEFAULT_NULL_STRING_ ""
#define DEFAULT_X_VALUE_COLUMN_NAME_ "x-value"
#define DEFAULT_Y_VALUE_COLUMN_NAME_ "y-value"
#define DEFAULT_FREQUENCY_COLUMN_ 0
#define DEFAULT_FREQUENCY_COLUMN_NAME_ "frequency"
#define DEFAULT_FREQUENCY_MEAN_ 2

/* Create a dataset for the simple Bear tutorial.
 */
int main(int argc, char *argv[]) {
  bool did_missing;
  bool do_frequencies;
  bool has_header_row;
  char *filename;
  char *frequency_column_name;
  char *missing_string;
  char *production_filename;
  ulong f;
  ulong frequency_column;
  ulong num_predict_rows;
  ulong num_train_rows;
  ulong row;
  double b;
  double l;
  double max_x;
  double min_x;
  double noise_stdev;
  double p;
  double q;
  double w;
  double x;
  double x_max;
  double x_min;
  double missing_bias;
  double missing_fraction;
  double y;
  Random *_random = null;
  TextFile *_tf = null;
  Logger *_l = null;
  Progress *_p = null;

  /* Arguments. The defaults will produce a file like that supplied on the page
   * for the simple tutorial, https://johncostella.com/bear/simple/.
   */
  args_description(
      "Create a sample bivariate linearly correlated dataset for the simple "
      "Bear tutorial, with various practical variations for testing, "
      "including equally-spaced prediction examples");
  args_string("filename", "Filename of the file to be created");
  args_optional_integer(
      "num-train-rows",
      't',
      DEFAULT_NUM_TRAIN_ROWS,
      "Number of rows of training data to create");
  args_optional_integer(
      "num-predict-rows",
      'p',
      DEFAULT_NUM_PREDICT_ROWS,
      "Number of rows of prediction x-values to create");
  args_optional_string(
      "production-filename",
      'P',
      ARGS_STRING_NO_DEFAULT,
      "Write the prediction x-values to this filename, rather than FILENAME");
  args_optional_real(
      "x-min",
      'm',
      DEFAULT_X_MIN_,
      "Minimum value of the x uniform distribution");
  args_optional_real(
      "x-max",
      'M',
      DEFAULT_X_MAX_,
      "Maximum value of the x uniform distribution");
  args_optional_real(
      "weight",
      'w',
      DEFAULT_WEIGHT_,
      "Weight constant w of the linear correlation, where y = w x + b");
  args_optional_real(
      "bias",
      'b',
      DEFAULT_BIAS_,
      "Bias constant b of the linear correlation, where y = w x + b");
  args_optional_real(
      "noise-stdev",
      'e',
      DEFAULT_NOISE_STDEV_,
      "Standard deviation of the noise to be added to the linear correlation");
  args_optional_real(
      "missing-percentage",
      'n',
      DEFAULT_NULL_PERCENTAGE_,
      "Percentage of rows for which x will be (randomly) missing");
  args_optional_real(
      "missing-bias",
      'B',
      DEFAULT_NULL_BIAS_,
      "Bias added to rows for which x is missing");
  args_flag("header-row",
      'H',
      "Make the first row a header row");
  args_optional_string(
      "x-column-name",
      'x',
      DEFAULT_X_VALUE_COLUMN_NAME_,
      "Column name for the x column, if --header-row is set");
  args_optional_string(
      "y-column-name",
      'y',
      DEFAULT_Y_VALUE_COLUMN_NAME_,
      "Column name for the y column, if --header-row is set");
  args_flag("include-frequency-column",
      'f',
      "Add a third column that is a frequency field, and populate it");
  args_optional_integer(
      "frequency-column",
      'g',
      DEFAULT_FREQUENCY_COLUMN_,
      "Column index (zero-offset) for the frequency column, if "
      "--include-frequency-column is set");
  args_optional_string(
      "frequency-column-name",
      'G',
      DEFAULT_FREQUENCY_COLUMN_NAME_,
      "Column name for the frequency column, if --header-row and "
      "--include-frequency-column are set");
  args_optional_real(
      "mean-freq",
      'l',
      DEFAULT_FREQUENCY_MEAN_,
      "Mean frequency, if --frequency-column is set");
  args_optional_string(
      "missing-string",
      'N',
      DEFAULT_NULL_STRING_,
      "String to be used to indicate a missing cell");
  args_optional_real(
      "quadratic",
      'q',
      DEFAULT_QUADRATIC_,
      "Quadratic constant q where y = q x^2 + w x + b");
  args_optional_integer(
      "random-seed",
      'r',
      ARGS_INTEGER_NO_DEFAULT,
      "Seed for the random number generator (default\f=\fautomatic)");
  args_parse(argc, argv);

  logger_new(&_l, stdout);
  filename = args_get_string("filename");
  production_filename = args_get_string("production-filename");
  has_header_row = args_get_flag("header-row");
  num_train_rows = args_get_positive_integer("num-train-rows");
  num_predict_rows = args_get_nonnegative_integer("num-predict-rows");
  x_min = args_get_real("x-min");
  x_max = args_get_real("x-max");
  q = args_get_real("quadratic");
  w = args_get_real("weight");
  b = args_get_real("bias");
  noise_stdev = args_get_real("noise-stdev");
  missing_fraction = args_get_bounded_real("missing-percentage", 0, 100) / 100;
  missing_bias = args_get_real("missing-bias");
  missing_string = args_get_string("missing-string");
  do_frequencies = args_get_flag("include-frequency-column");

  if (args_set("random-seed")) {
    random_new_seed(&_random, args_get_positive_integer("random-seed"));
  } else {
    random_new(&_random);
  }

  if (missing_fraction > 0 && missing_fraction * num_train_rows < 0.1) {
    throw("Null percentage of %.16g%% is not practical for only "
        "%lu training rows",
        missing_fraction * 100,
        num_train_rows);
  }

  if (do_frequencies) {
    frequency_column = args_get_bounded_integer("frequency-column", 0, 2);

    if (has_header_row) {
      frequency_column_name = args_get_string("frequency-column-name");
    }
  } else {
    if (args_set("frequency-column")) {
      throw("Cannot specify --frequency-column without "
          "--include-frequency-column");
    }

    if (args_set("frequency-column-name")) {
      throw("Cannot specify --frequency-column-name without "
          "--include-frequency-column");
    }
  }

  if (do_frequencies) {
    l = exp(-(args_get_bounded_real("mean-freq", 1, ARGS_REAL_NO_MAX) - 1));
  }

  do {
    slog(_l,
        "Writing %lu training row%s",
        num_train_rows,
        plural_s(num_train_rows));
    progress_new(&_p, _l, num_train_rows);

    if (_tf != null) {
      text_file_delete(&_tf);
      file_remove(filename);
    }

    text_file_new_type(&_tf, filename, "w", "csv");

    if (has_header_row) {
      if (do_frequencies && frequency_column == 0) {
        text_file_write(_tf, frequency_column_name);
      }

      text_file_write(_tf, args_get_nonempty_string("x-column-name"));

      if (do_frequencies && frequency_column == 1) {
        text_file_write(_tf, frequency_column_name);
      }

      text_file_write(_tf, args_get_nonempty_string("y-column-name"));

      if (do_frequencies && frequency_column == 2) {
        text_file_write(_tf, frequency_column_name);
      }

      text_file_end_record(_tf);
    }

    for (row = 0, min_x = DBL_MAX, max_x = -DBL_MAX, did_missing = false;
        row < num_train_rows;
        row++) {
      x = x_min + random_double_unit_cc(_random) * (x_max - x_min);
      y = b + random_normal(_random) * noise_stdev;

      if (do_frequencies) {
        f = 0;
        p = 1;

        do {
          f++;
          p *= random_double_unit_cc(_random);
        } while (p > l);
      }

      if (do_frequencies && frequency_column == 0) {
        text_file_write_ulong(_tf, f);
      }

      if (random_double_unit_co(_random) < missing_fraction) {
        text_file_write(_tf, missing_string);
        y += missing_bias;
        did_missing = true;
      } else {
        y += q * x * x + w * x;
        text_file_write_double(_tf, x);
        min_x = minimum(min_x, x);
        max_x = maximum(max_x, x);
      }

      if (do_frequencies && frequency_column == 1) {
        text_file_write_ulong(_tf, f);
      }

      text_file_write_double(_tf, y);

      if (do_frequencies && frequency_column == 2) {
        text_file_write_ulong(_tf, f);
      }

      text_file_end_record(_tf);
      progress_bump(_p);
    }

    progress_delete(&_p);
    flog(_l);

    if (min_x == DBL_MAX) {
      rlog(_l, "No non-missing feature rows written; retrying.");
    }

    if (missing_fraction > 0 && !did_missing) {
      rlog(_l,
          "No missing feature rows written despite missing percentage "
          "being %.16g%%; retrying.",
          missing_fraction * 100);
    }
  } while (min_x == DBL_MAX || (missing_fraction > 0 && !did_missing));

  random_delete(&_random);

  if (production_filename != null) {
    text_file_delete(&_tf);
    text_file_new(&_tf, production_filename, "w");
  }

  if (num_predict_rows > 0) {
    slog(_l, "Writing %lu prediction rows", num_predict_rows);
    progress_new(&_p, _l, num_predict_rows);

    if (did_missing) {
      text_file_write_empty_record(_tf);
    }

    for (row = 0; row < num_predict_rows; row++) {
      x = min_x + (max_x - min_x) * row / (num_predict_rows - 1);
      text_file_write_double(_tf, x);
      text_file_end_record(_tf);
      progress_bump(_p);
    }

    progress_delete(&_p);
    flog(_l);
  }

  text_file_delete(&_tf);
  logger_delete(&_l);

  return 0;
}
