#include "HypercubeTest.h"

static Lock lock_ = LOCK_INITIALIZER;

static void double_sum_(Hypercube *this,
    bool *walk_dimensions,
    ulong *indexes,
    ulong o,
    void *x);

static void ulong_sum_(UlongHypercube *this,
    bool *walk_dimensions,
    ulong *indexes,
    ulong o,
    void *x);


START

TEST(DoubleHypercube, ZeroDimensional) {
  double d1;
  double d2;
  DoubleHypercube *_ = null;

  double_hypercube_new(&_, 0, null);
  d1 = -24.0966;
  hypercube_set(_, 0, &d1);
  EXPECT_DEQ(-24.0966, d1);
  hypercube_get(_, 0, &d2);
  EXPECT_DEQ(-24.0966, d2);
  hypercube_delete(&_);
}

TEST(DoubleHypercube, Simple) {
  ulong column;
  ulong o;
  ulong row;
  ulong _dimensions[] = {2, 3};
  ulong _indexes[2];
  double d;
  double _ps[6];
  DoubleHypercube *_ = null;
  DoubleHypercube *_copy = null;

  _ps[0] = 655;
  _ps[1] = 35;
  _ps[2] = -24;
  _ps[3] = 9;
  _ps[4] = 19.7;
  _ps[5] = 197;

  double_hypercube_new(&_, 2, _dimensions);

  for (o = 0; o < 6; o++) {
    hypercube_set(_, o, &_ps[o]);
  }

  for (o = 0; o < 6; o++) {
    hypercube_get(_, o, &d);
    EXPECT_DEQ(_ps[o], d);
  }

  test_save_load(Hypercube, hypercube, _);

  for (row = 0; row < _dimensions[0]; row++) {
    _indexes[0] = row;

    for (column = 0; column < _dimensions[1]; column++) {
      _indexes[1] = column;
      hypercube_get_from_indexes(_, _indexes, &d);
      EXPECT_DEQ(_ps[row * _dimensions[1] + column], d);
    }
  }

  hypercube_new_copy(&_copy, _);
  test_save_load(Hypercube, hypercube, _copy);
  hypercube_delete(&_);

  for (o = 0; o < 6; o++) {
    hypercube_get(_copy, o, &d);
    EXPECT_DEQ(_ps[o], d);
  }

  for (row = 0; row < _dimensions[0]; row++) {
    _indexes[0] = row;

    for (column = 0; column < _dimensions[1]; column++) {
      _indexes[1] = column;
      hypercube_get_from_indexes(_copy, _indexes, &d);
      EXPECT_DEQ(_ps[row * _dimensions[1] + column], d);
    }
  }

  hypercube_delete(&_copy);
}

TEST(DoubleHypercube, ForEach) {
  bool _walks[4];
  ulong n;
  ulong o;
  ulong _dimensions[] = {2, 3, 4, 5};
  ulong _is[4];
  double d;
  double sum;
  DoubleHypercube *_ = null;

  /* Test one dimension without walking it.
   */
  double_hypercube_new(&_, 1, _dimensions);
  d = 42;
  hypercube_set(_, 0, &d);
  d = -6.75;
  hypercube_set(_, 1, &d);
  _walks[0] = false;
  sum = 0;
  hypercube_foreach(_, _walks, null, double_sum_, &sum, 1);
  EXPECT_DEQ(42, sum);

  /* Test one dimension, walking it, default specified index = 0.
   */
  _walks[0] = true;
  sum = 0;
  hypercube_foreach(_, _walks, null, double_sum_, &sum, 1);
  EXPECT_DEQ(35.25, sum);
  hypercube_delete(&_);

  /* Test two dimensions, not walking either, defaults.
   */
  double_hypercube_new(&_, 2, _dimensions);
  n = _dimensions[0] * _dimensions[1];

  for (o = 0, d = 10.5; o < n; o++, d += 1) {
    hypercube_set(_, o, &d);
  }

  _walks[0] = false;
  _walks[1] = false;
  sum = 0;
  hypercube_foreach(_, _walks, null, double_sum_, &sum, 1);
  EXPECT_DEQ(10.5, sum);

  /* Test two dimensions, walking 0, default 1.
   */
  _walks[0] = true;
  _walks[1] = false;
  sum = 0;
  hypercube_foreach(_, _walks, null, double_sum_, &sum, 1);
  EXPECT_DEQ(24, sum);

  /* Test two dimensions, walking 0, specifying 1.
   */
  _is[0] = 0;
  _is[1] = 2;
  sum = 0;
  hypercube_foreach(_, _walks, _is, double_sum_, &sum, 1);
  EXPECT_DEQ(28, sum);

  /* Test two dimensions, walking 1, default 0.
   */
  _walks[0] = false;
  _walks[1] = true;
  sum = 0;
  hypercube_foreach(_, _walks, null, double_sum_, &sum, 1);
  EXPECT_DEQ(34.5, sum);

  /* Test two dimensions, walking both.
   */
  _walks[0] = true;
  _walks[1] = true;
  sum = 0;
  hypercube_foreach(_, _walks, null, double_sum_, &sum, 1);
  EXPECT_DEQ(78, sum);

  hypercube_delete(&_);

  /* Test three dimensions, walking all.
   */
  double_hypercube_new(&_, 3, _dimensions);
  n = _dimensions[0] * _dimensions[1] * _dimensions[2];

  for (o = 0, d = 10.5; o < n; o++, d += 1) {
    hypercube_set(_, o, &d);
  }

  _walks[0] = true;
  _walks[1] = true;
  _walks[2] = true;
  sum = 0;
  hypercube_foreach(_, _walks, null, double_sum_, &sum, 1);
  EXPECT_DEQ(528, sum);

  hypercube_delete(&_);

  /* Test four dimensions, walking all.
   */
  double_hypercube_new(&_, 4, _dimensions);
  n = _dimensions[0] * _dimensions[1] * _dimensions[2] * _dimensions[3];

  for (o = 0, d = 1; o < n; o++, d += 1) {
    hypercube_set(_, o, &d);
  }

  _walks[0] = true;
  _walks[1] = true;
  _walks[2] = true;
  _walks[3] = true;
  sum = 0;
  hypercube_foreach(_, _walks, null, double_sum_, &sum, 1);
  EXPECT_DEQ(7260, sum);

  /* Test four dimensions, walking 1 and 3, specifying 0 and 2.
   */
  _walks[0] = false;
  _walks[1] = true;
  _walks[2] = false;
  _walks[3] = true;
  _is[0] = 1;
  _is[1] = 0;
  _is[2] = 2;
  _is[3] = 0;
  sum = 0;
  hypercube_foreach(_, _walks, _is, double_sum_, &sum, 1);
  EXPECT_DEQ(1395, sum);

  hypercube_delete(&_);
}

TEST(UlongHypercube, ZeroDimensional) {
  ulong u1;
  ulong u2;
  UlongHypercube *_ = null;

  u1 = 240966;
  ulong_hypercube_new(&_, 0, null);
  hypercube_set(_, 0, &u1);
  hypercube_get(_, 0, &u2);
  EXPECT_ULEQ(240966, u2);
  hypercube_delete(&_);
}

TEST(UlongHypercube, Simple) {
  ulong column;
  ulong o;
  ulong row;
  ulong u;
  ulong _dimensions[] = {2, 3};
  ulong _indexes[2];
  ulong _us[] = {65535, 24, 9, 1966, 0, 1968};
  UlongHypercube *_ = null;
  UlongHypercube *_copy = null;

  ulong_hypercube_new(&_, 2, _dimensions);

  for (o = 0; o < 6; o++) {
    hypercube_set(_, o, &_us[o]);
  }

  for (o = 0; o < 6; o++) {
    hypercube_get(_, o, &u);
    EXPECT_ULEQ(_us[o], u);
  }

  test_save_load(Hypercube, hypercube, _);

  for (row = 0; row < _dimensions[0]; row++) {
    _indexes[0] = row;

    for (column = 0; column < _dimensions[1]; column++) {
      _indexes[1] = column;
      hypercube_get_from_indexes(_, _indexes, &u);
      EXPECT_ULEQ(_us[row * _dimensions[1] + column], u);
    }
  }

  hypercube_new_copy(&_copy, _);
  test_save_load(Hypercube, hypercube, _copy);
  hypercube_delete(&_);

  for (o = 0; o < 6; o++) {
    hypercube_get(_copy, o, &u);
    EXPECT_ULEQ(_us[o], u);
  }

  for (row = 0; row < _dimensions[0]; row++) {
    _indexes[0] = row;

    for (column = 0; column < _dimensions[1]; column++) {
      _indexes[1] = column;
      hypercube_get_from_indexes(_copy, _indexes, &u);
      EXPECT_ULEQ(_us[row * _dimensions[1] + column], u);
    }
  }

  hypercube_delete(&_copy);
}

TEST(UlongHypercube, ForEach) {
  bool _walks[4];
  ulong n;
  ulong o;
  ulong _dimensions[] = {2, 3, 4, 5};
  ulong _is[4];
  ulong u;
  ulong sum;
  UlongHypercube *_ = null;

  /* Test one dimension without walking it.
   */
  ulong_hypercube_new(&_, 1, _dimensions);
  u = 42;
  hypercube_set(_, 0, &u);
  u = 66;
  hypercube_set(_, 1, &u);
  _walks[0] = false;
  sum = 0;
  hypercube_foreach(_, _walks, null, ulong_sum_, &sum, 1);
  EXPECT_ULEQ(42, sum);

  /* Test one dimension, walking it, default specified index = 0.
   */
  _walks[0] = true;
  sum = 0;
  hypercube_foreach(_, _walks, null, ulong_sum_, &sum, 1);
  EXPECT_ULEQ(108, sum);
  hypercube_delete(&_);

  /* Test two dimensions, not walking either, defaults.
   */
  ulong_hypercube_new(&_, 2, _dimensions);
  n = _dimensions[0] * _dimensions[1];

  for (o = 0, u = 105; o < n; o++, u++) {
    hypercube_set(_, o, &u);
  }

  _walks[0] = false;
  _walks[1] = false;
  sum = 0;
  hypercube_foreach(_, _walks, null, ulong_sum_, &sum, 1);
  EXPECT_ULEQ(105, sum);

  /* Test two dimensions, walking 0, default 1.
   */
  _walks[0] = true;
  _walks[1] = false;
  sum = 0;
  hypercube_foreach(_, _walks, null, ulong_sum_, &sum, 1);
  EXPECT_ULEQ(213, sum);

  /* Test two dimensions, walking 0, specifying 1.
   */
  _is[0] = 0;
  _is[1] = 2;
  sum = 0;
  hypercube_foreach(_, _walks, _is, ulong_sum_, &sum, 1);
  EXPECT_ULEQ(217, sum);

  /* Test two dimensions, walking 1, default 0.
   */
  _walks[0] = false;
  _walks[1] = true;
  sum = 0;
  hypercube_foreach(_, _walks, null, ulong_sum_, &sum, 1);
  EXPECT_ULEQ(318, sum);

  /* Test two dimensions, walking both.
   */
  _walks[0] = true;
  _walks[1] = true;
  sum = 0;
  hypercube_foreach(_, _walks, null, ulong_sum_, &sum, 1);
  EXPECT_ULEQ(645, sum);

  hypercube_delete(&_);

  /* Test three dimensions, walking all.
   */
  ulong_hypercube_new(&_, 3, _dimensions);
  n = _dimensions[0] * _dimensions[1] * _dimensions[2];

  for (o = 0, u = 105; o < n; o++, u++) {
    hypercube_set(_, o, &u);
  }

  _walks[0] = true;
  _walks[1] = true;
  _walks[2] = true;
  sum = 0;
  hypercube_foreach(_, _walks, null, ulong_sum_, &sum, 1);
  EXPECT_ULEQ(2796, sum);

  hypercube_delete(&_);

  /* Test four dimensions, walking all.
   */
  ulong_hypercube_new(&_, 4, _dimensions);
  n = _dimensions[0] * _dimensions[1] * _dimensions[2] * _dimensions[3];

  for (o = 0, u = 105; o < n; o++, u++) {
    hypercube_set(_, o, &u);
  }

  _walks[0] = true;
  _walks[1] = true;
  _walks[2] = true;
  _walks[3] = true;
  sum = 0;
  hypercube_foreach(_, _walks, null, ulong_sum_, &sum, 1);
  EXPECT_ULEQ(19740, sum);

  /* Test four dimensions, walking 1 and 3, specifying 0 and 2.
   */
  _walks[0] = false;
  _walks[1] = true;
  _walks[2] = false;
  _walks[3] = true;
  _is[0] = 1;
  _is[1] = 0;
  _is[2] = 2;
  _is[3] = 0;
  sum = 0;
  hypercube_foreach(_, _walks, _is, ulong_sum_, &sum, 1);
  EXPECT_ULEQ(2955, sum);

  hypercube_delete(&_);
}

FINISH


static void double_sum_(Hypercube *this,
    bool *walk_dimensions,
    ulong *indexes,
    ulong o,
    void *x) {
  ulong d;
  ulong oo;
  double v;
  double *sum;

  for (d = 0, oo = 0; d < this->numDimensions; d++) {
    oo += this->_strides[d] * indexes[d];
  }

  EXPECT_ULEQ(oo, o);

  sum = x;
  hypercube_get(this, o, &v);
  lock(&lock_);
  *sum += v;
  unlock(&lock_);
}

static void ulong_sum_(UlongHypercube *this,
    bool *walk_dimensions,
    ulong *indexes,
    ulong o,
    void *x) {
  ulong u;
  ulong *sum;

  sum = x;
  hypercube_get(this, o, &u);
  lock(&lock_);
  *sum += u;
  unlock(&lock_);
}
