#include "KeyValueCollectionTest.h"
#include <math.h>

#define N_ 100
#define M_ 100000
#define P_ 1000000
#define CHUNK_SIZE_ 1024
#define DUMMY_ 24
#define SCALE_1_ 2
#define SCALE_2_ 3
#define SCALE_2_2_ 5
#define TR_SCALE_1_ 3
#define TR_SCALE_2_ 2
#define TRANSFORM_SCALE_ 2

typedef struct {
  byte scale_1;
  byte scale_2;
} Pass_;

typedef struct {
  byte scale;
} Pass_1_;

typedef struct {
  ulong dummy;
  byte scale;
} Pass_2_;

typedef struct {
  char *s1;
  char *s2;
  ushort dummy;
} Object_;

typedef struct {
  ushort dummy1;
  char *s1;
  byte dummy2;
  char *s2;
} Object_1_;

typedef struct {
  uint dummy;
  char *s;
} Object_2_1_;

typedef struct {
  byte dummy;
  char *s;
} Object_2_2_;

static void construct_(
    KeyValuePair *pair,
    KeyValueInfo *info,
    byte k0,
    byte k1,
    byte v0);

static void string_(char **p_this, byte len);

static void destroy_(KeyValuePair *pair, KeyValueInfo *info);

static void print_(KeyValuePairPrintInfo *kvppi, Printer *p);

static void merge_(KeyValuePair *acc, KeyValuePair *add, KeyValueInfo *info);

static sbyte compare_(KeyValuePair *a, KeyValuePair *b, KeyValueInfo *info);

static bool get_invalid_(KeyValuePair *pair, KeyValueInfo *info);

static void set_invalid_(KeyValuePair *pair, KeyValueInfo *info);


START

TEST(KeyValueCollection, New) {
  ulong max = 4;
  KeyValueInfo *_info = null;
  KeyValueCollection *_collection = null;

  key_value_collection_set_chunk_size(2);

  set_info_new(&_info, 1, &max);
  key_value_collection_new(&_collection, _info);
  key_value_collection_delete(&_collection);
  key_value_info_delete(&_info);
}

TEST(KeyValueCollection, ChunkSize) {
  ulong current;

  current = key_value_collection_get_chunk_size();
  EXPECT_ULEQ(current, key_value_collection_set_chunk_size(1234));
  EXPECT_ULEQ(1234, key_value_collection_get_chunk_size());
  EXPECT_ULEQ(1234, key_value_collection_set_chunk_size(current));
  EXPECT_ULEQ(current, key_value_collection_get_chunk_size());
}

TEST(KeyValueCollection, Empty) {
  ulong max = 7;
  KeyValueInfo *_info = null;
  KeyValueArray *_array = null;
  KeyValueCollection *_collection = null;

  set_info_new(&_info, 1, &max);
  key_value_collection_new(&_collection, _info);
  key_value_array_new_from_collection(&_array, &_collection);
  EXPECT_ULEQ(0, _array->maxNumPairs);
  EXPECT_ULEQ(0, _array->numPairs);
  EXPECT_TRUE(_array->finalized);
  key_value_array_delete(&_array);
  key_value_info_delete(&_info);
}

TEST(KeyValueCollection, Simple) {
  ulong max = 7;
  KeyValueInfo *_info = null;
  KeyValuePair *_pair = null;
  KeyValueCollection *_collection = null;

  set_info_new(&_info, 1, &max);
  key_value_collection_new(&_collection, _info);
  key_value_pair_new(&_pair, _info);
  key_value_pair_set_key(_pair, _info, 0, 5);

  /* Add key 5 to the chunk.
   */
  EXPECT_FALSE(key_value_collection_donate(_collection, _pair));

  /* Add key 3 to the chunk, which extracts it into the tree.
   */
  key_value_pair_set_key(_pair, _info, 0, 3);
  EXPECT_FALSE(key_value_collection_donate(_collection, _pair));

  /* Without objects, pairs are not zeroed after donation.
   */
  EXPECT_ULEQ(3, key_value_pair_get_key(_pair, _info, 0));

  /* Should now find key 3 in the tree. Pair is not cleared as there is no
   * destructor.
   */
  key_value_pair_set_key(_pair, _info, 0, 3);
  EXPECT_TRUE(key_value_collection_donate(_collection, _pair));
  EXPECT_ULEQ(3, key_value_pair_get_key(_pair, _info, 0));

  /* Add new key 1 to the chunk.
   */
  key_value_pair_set_key(_pair, _info, 0, 1);
  EXPECT_FALSE(key_value_collection_donate(_collection, _pair));
  EXPECT_ULEQ(1, key_value_pair_get_key(_pair, _info, 0));

  /* Add key 1 again to the chunk. It is not in the tree when this is called.
   * But it fills the chunk, which is extracted into the tree, and nodes 0 and
   * 1, being the complete level 1, are merged into the left child of a
   * newly-created root node at level 2.
   */
  key_value_pair_set_key(_pair, _info, 0, 1);
  EXPECT_ULEQ(1, key_value_pair_get_key(_pair, _info, 0));
  EXPECT_FALSE(key_value_collection_donate(_collection, _pair));
  EXPECT_ULEQ(1, key_value_pair_get_key(_pair, _info, 0));

  /* Should now find key 1 in the tree.
   */
  key_value_pair_set_key(_pair, _info, 0, 1);
  EXPECT_TRUE(key_value_collection_donate(_collection, _pair));
  EXPECT_ULEQ(1, key_value_pair_get_key(_pair, _info, 0));

  /* And should find key 3 in the tree too.
   */
  key_value_pair_set_key(_pair, _info, 0, 3);
  EXPECT_TRUE(key_value_collection_donate(_collection, _pair));
  EXPECT_ULEQ(3, key_value_pair_get_key(_pair, _info, 0));

  /* Add new key 6 to the chunk.
   */
  key_value_pair_set_key(_pair, _info, 0, 6);
  EXPECT_FALSE(key_value_collection_donate(_collection, _pair));

  /* Add key 6 to the chunk again. It is not in the tree when this is called.
   * This fills the chunk, which is extracted to the right-then-left child of
   * the tree.
   */
  EXPECT_FALSE(key_value_collection_donate(_collection, _pair));

  /* Should now find key 6 in the tree.
   */
  EXPECT_TRUE(key_value_collection_donate(_collection, _pair));

  /* And should find key 5 in the tree too.
   */
  key_value_pair_set_key(_pair, _info, 0, 5);
  EXPECT_TRUE(key_value_collection_donate(_collection, _pair));

  key_value_collection_delete(&_collection);
  key_value_pair_delete(&_pair, _info);
  key_value_info_delete(&_info);
}

TEST(KeyValueCollection, Objects) {
  byte k0;
  byte k1;
  byte v1;
  ulong f;
  ulong max_keys[] = {7, 9, ULONG_MAX};
  ulong max_values[] = {ULONG_MAX, 12, 1};
  Pass_ pass;
  KeyValueInfo *_info = null;
  KeyValuePair *_pair = null;
  KeyValueCollection *_collection = null;

  /* Use the same test objects as in KeyValueArrayTest (KeyValueArray.Objects).
   */
  key_value_info_new(&_info, 3, max_keys, 3, max_values);
  _info->x = &pass;
  _info->destroy = destroy_;
  _info->print = print_;
  _info->merge = merge_;
  _info->compare = compare_;
  _info->getInvalid = get_invalid_;
  _info->setInvalid = set_invalid_;

  pass.scale_1 = SCALE_1_;
  pass.scale_2 = SCALE_2_;

  /* Use a chunk size of 2 pairs.
   */
  key_value_collection_new(&_collection, _info);
  key_value_pair_new(&_pair, _info);

  /* Add a first pair.
   */
  k0 = 5;
  k1 = 3;
  v1 = 2;
  construct_(_pair, _info, k0, k1, v1);
  EXPECT_FALSE(key_value_collection_donate(_collection, _pair));

  /* Verify donation.
   */
  for (f = 0; f < 6; f++) {
    EXPECT_ULEQ(0, key_value_pair_get_field(_pair, _info, f));
  }

  /* Add a second pair.
   */
  k0 = 2;
  k1 = 9;
  v1 = 11;
  construct_(_pair, _info, k0, k1, v1);
  EXPECT_FALSE(key_value_collection_donate(_collection, _pair));

  /* The chunk has now been filled and added to the tree. Adding in the first
   * key should merge in with the tree.
   */
  k0 = 5;
  k1 = 3;
  v1 = 3;
  construct_(_pair, _info, k0, k1, v1);
  EXPECT_TRUE(key_value_collection_donate(_collection, _pair));

  /* Verify donation.
   */
  for (f = 0; f < 6; f++) {
    EXPECT_ULEQ(0, key_value_pair_get_field(_pair, _info, f));
  }

  /* Adding a third pair should go into the chunk.
   */
  k0 = 2;
  k1 = 7;
  v1 = 1;
  construct_(_pair, _info, k0, k1, v1);
  EXPECT_FALSE(key_value_collection_donate(_collection, _pair));

  /* Adding the same pair again should also go into the chunk, which will be
   * full and so will be finalized and added to the tree.
   */
  k0 = 2;
  k1 = 7;
  v1 = 4;
  construct_(_pair, _info, k0, k1, v1);
  EXPECT_FALSE(key_value_collection_donate(_collection, _pair));

  /* Adding the same pair yet again should now be able to merged into the tree.
   */
  k0 = 2;
  k1 = 7;
  v1 = 4;
  construct_(_pair, _info, k0, k1, v1);
  EXPECT_TRUE(key_value_collection_donate(_collection, _pair));

  /* Add another pair so that we have a chunk again.
   */
  k0 = 6;
  k1 = 1;
  v1 = 4;
  construct_(_pair, _info, k0, k1, v1);
  EXPECT_FALSE(key_value_collection_donate(_collection, _pair));

  key_value_collection_delete(&_collection);
  key_value_pair_delete(&_pair, _info);
  key_value_info_delete(&_info);
}
TEST(KeyValueCollection, Print) {
  ulong max = 7;
  KeyValueInfo *_info = null;
  KeyValuePair *_pair = null;
  KeyValueCollection *_collection = null;
  KeyValueArray *_array = null;

  /* Simple case of an empty collection with chunk size 3.
   */
  key_value_collection_set_chunk_size(3);
  set_info_new(&_info, 1, &max);
  key_value_collection_new(&_collection, _info);
  key_value_collection_delete(&_collection);
  key_value_info_delete(&_info);

  /* Repeat the construction above for a nontrivial case, adding an extra pair
   * so that there is a chunk.
   */
  key_value_collection_set_chunk_size(2);
  set_info_new(&_info, 1, &max);
  key_value_collection_new(&_collection, _info);
  key_value_pair_new(&_pair, _info);
  key_value_pair_set_key(_pair, _info, 0, 5);
  key_value_collection_donate(_collection, _pair);
  key_value_pair_set_key(_pair, _info, 0, 3);
  key_value_collection_donate(_collection, _pair);
  key_value_pair_set_key(_pair, _info, 0, 1);
  key_value_collection_donate(_collection, _pair);
  key_value_pair_set_key(_pair, _info, 0, 1);
  key_value_collection_donate(_collection, _pair);
  key_value_pair_set_key(_pair, _info, 0, 6);
  key_value_collection_donate(_collection, _pair);
  key_value_pair_set_key(_pair, _info, 0, 6);
  key_value_collection_donate(_collection, _pair);
  key_value_pair_set_key(_pair, _info, 0, 7);
  key_value_collection_donate(_collection, _pair);

  /* Extract the array.
   */
  key_value_array_new_from_collection(&_array, &_collection);
  key_value_array_delete(&_array);
  key_value_pair_delete(&_pair, _info);
  key_value_info_delete(&_info);
}

TEST(KeyValueCollection, Comprehensive) {
  ulong expected_num_finalized_records;
  ulong frequency;
  ulong i;
  ulong j;
  ulong k;
  ulong key;
  ulong max_frequency;
  ulong num_keys;
  ulong num_pairs;
  ulong single_key;
  ulong *_max_keys = null;
  KeyValueInfo *info;
  KeyValueInfo *_info = null;
  KeyValueArray *_array = null;
  NaiveKeyValueCollection *_ncollection = null;
  KeyValueCollection *_collection = null;
  KeyValuePair *_pair = null;
  KeyValuePair *pairj;
  KeyValuePair *_pairf = null;
  KeyValuePair *pair_found;
  Random *_random = null;

  key_value_collection_set_chunk_size(CHUNK_SIZE_);
  random_new(&_random);

  for (i = 0; i < N_; i++) {
    num_pairs = random_modulo(_random, M_);
    num_keys = random_modulo(_random, 5);
    new(&_max_keys, num_keys);

    for (k = 0; k < num_keys; k++) {
      _max_keys[k] = random_modulo(
          _random,
          maximum((uint)pow(num_pairs, 1.0 / num_keys), 1));
    }

    max_frequency = random_modulo(_random, P_) + 1;

    key_frequency_info_new(
        &_info,
        num_keys,
        _max_keys,
        num_pairs * max_frequency);

    key_value_collection_new(&_collection, _info);

    naive_key_value_collection_new(&_ncollection, _info, true);
    key_value_pair_new(&_pair, _info);
    key_value_pair_new(&_pairf, _info);
    expected_num_finalized_records = 0;

    for (j = 0; j < num_pairs; j++) {
      for (k = 0; k < num_keys; k++) {
        key = random_modulo(_random, _max_keys[k] + 1);
        key_value_pair_set_key(_pair, _info, k, key);
      }

      frequency = random_modulo(_random, max_frequency) + 1;
      key_frequency_pair_set_frequency(_pair, _info, frequency);

      if (naive_key_value_collection_get_value(_ncollection, _pair, 0) == 0) {
        expected_num_finalized_records++;
      }

      naive_key_value_collection_add_values(_ncollection, _pair);
      key_value_collection_donate(_collection, _pair);
    }

    key_value_array_new_from_collection(&_array, &_collection);
    key_value_array_verify_finalized(_array);

    key_value_info_delete(&_info);
    info = _array->_info;

    EXPECT_TRUE(key_value_array_is_full(_array));
    EXPECT_TRUE(_array->finalized);
    EXPECT_ULEQ(expected_num_finalized_records, _array->numPairs);

    for (j = 0, single_key = 0; j < _array->numPairs; j++, single_key++) {
      while (_ncollection->_values[0][single_key] == 0) {
        naive_key_value_collection_pair_from_single_key(
            _ncollection,
            single_key,
            _pairf);
        pair_found = key_value_array_find(_array, _pairf);
        EXPECT_NULL(pair_found);

        single_key++;
      }

      pairj = key_value_array_get(_array, j);
      frequency = key_frequency_pair_get_frequency(pairj, info);
      EXPECT_ULEQ(_ncollection->_values[0][single_key], frequency);
      naive_key_value_collection_pair_from_single_key(
          _ncollection,
          single_key,
          _pairf);
      pair_found = key_value_array_find(_array, _pairf);
      EXPECT_NONNULL(pair_found);
      EXPECT_ULEQ(0, info->compare(pairj, pair_found, info));
    }

    key_value_pair_delete(&_pair, info);
    key_value_pair_delete(&_pairf, info);
    key_value_array_delete(&_array);
    naive_key_value_collection_delete(&_ncollection);
    key_value_collection_delete(&_collection);
    deletev(&_max_keys);
  }

  random_delete(&_random);
}

FINISH


static void construct_(
    KeyValuePair *pair,
    KeyValueInfo *info,
    byte k0,
    byte k1,
    byte v1) {
  char *_s = null;
  Pass_ *pass;
  Object_ *_o = null;

  key_value_pair_set_key(pair, info, 0, k0);
  key_value_pair_set_key(pair, info, 1, k1);

  pass = info->x;
  newo(&_o);
  _o->dummy = DUMMY_;
  string_(&_o->s1, k0);
  string_(&_o->s2, k1);
  key_value_pair_donate_to_key(pair, info, 2, &_o);
  string_(&_s, v1 * pass->scale_1);
  key_value_pair_donate_to_value(pair, info, 0, &_s);
  key_value_pair_set_value(pair, info, 1, v1);
}

static void string_(char **p_this, byte len) {
  byte u;
  char *this;

  new(p_this, len + 1);
  this = *p_this;

  this[len] = '\0';

  for (u = 0; u < len; u++) {
    this[u] = 'a' + modulo(u, 26);
  }
}

static void destroy_(KeyValuePair *pair, KeyValueInfo *info) {
  char *_s = null;
  Pass_ *pass;
  Object_ *_o = null;

  pass = info->x;

  if (!(pass->scale_1 == SCALE_1_ && pass->scale_2 == SCALE_2_)
      && !(pass->scale_1 == TR_SCALE_1_ && pass->scale_2 == TR_SCALE_2_)) {
    throw("Impossible");
  }

  if (!key_value_pair_try_donate_from_key(pair, info, 2, &_o)) {
    return;
  }

  if (_o->dummy != DUMMY_) {
    throw("Impossible");
  }

  string_delete(&_o->s1);
  string_delete(&_o->s2);
  key_value_pair_donate_from_value(pair, info, 0, &_s);
  string_delete(&_s);

  deleteo(&_o);
}

static void print_(KeyValuePairPrintInfo *kvppi, Printer *p) {
  KeyValuePair *pair;
  KeyValueInfo *info;
  Object_ *o;

  pair = kvppi->this;
  info = kvppi->info;
  rprint(p, "Key 0: %lu", key_value_pair_get_key(pair, info, 0));
  rprint(p, "Key 1: %lu", key_value_pair_get_key(pair, info, 1));
  sprint(p, "Key 2");
  o = key_value_pair_reference_from_key(pair, info, 2);
  rprint(p, "String 1: '%s'", o->s1);
  rprint(p, "String 2: '%s'", o->s2);
  rprint(p,
      "Value 0: '%s'",
      key_value_pair_reference_from_value(pair, info, 0));
  rprint(p, "Value 1: %lu", key_value_pair_get_value(pair, info, 1));
  rprint(p,
      "Value 2: Invalid %s",
      bool_string(key_value_pair_get_value(pair, info, 2)));
  fprint(p);
}

static void merge_(KeyValuePair *acc, KeyValuePair *add, KeyValueInfo *info) {
  byte v1acc;
  byte v1add;
  byte v1;
  char *_s = null;
  Pass_ *pass;
  Object_ *_o = null;

  pass = info->x;

  /* See the comments in KeyValueArrayTest for this same method.
   */
  key_value_pair_donate_from_key(add, info, 2, &_o);
  string_delete(&_o->s1);
  string_delete(&_o->s2);
  deleteo(&_o);

  key_value_pair_donate_from_value(add, info, 0, &_s);
  string_delete(&_s);

  key_value_pair_donate_from_value(acc, info, 0, &_s);
  string_delete(&_s);
  v1acc = key_value_pair_get_value(acc, info, 1);
  v1add = key_value_pair_get_value(add, info, 1);
  v1 = v1acc + v1add;
  string_(&_s, v1 * pass->scale_1);
  key_value_pair_donate_to_value(acc, info, 0, &_s);
  key_value_pair_set_value(acc, info, 1, v1);
}

static sbyte compare_(KeyValuePair *a, KeyValuePair *b, KeyValueInfo *info) {
  ulong k0a;
  ulong k0b;
  ulong k1a;
  ulong k1b;

  k0a = key_value_pair_get_key(a, info, 0);
  k1a = key_value_pair_get_key(a, info, 1);
  k0b = key_value_pair_get_key(b, info, 0);
  k1b = key_value_pair_get_key(b, info, 1);

  if (k0a < k0b) {
    return -1;
  }

  if (k0a > k0b) {
    return +1;
  }

  if (k1a < k1b) {
    return -1;
  }

  if (k1a > k1b) {
    return +1;
  }

  return 0;
}

static bool get_invalid_(KeyValuePair *pair, KeyValueInfo *info) {
  return key_value_pair_get_value(pair, info, 2) == 1;
}

static void set_invalid_(KeyValuePair *pair, KeyValueInfo *info) {
  key_value_pair_set_value(pair, info, 2, 1);
}
