#include "SetInfoTest.h"
#include <math.h>

#define N_ 50
#define M_ 100000
#define P_ 1000000
#define CHUNK_SIZE_ 1024

static void destroy_(KeyValuePair *pair, KeyValueInfo *info);

static void print_(KeyValuePairPrintInfo *kvppi, Printer *p);

static void merge_(KeyValuePair *acc, KeyValuePair *add, KeyValueInfo *info);

static sbyte compare_(KeyValuePair *a, KeyValuePair *b, KeyValueInfo *info);


START

TEST(SetInfo, New) {
  int a;
  ulong max_keys[] = {6, 24};
  KeyValueInfo *_info = null;

  set_info_new(&_info, 2, max_keys);
  _info->x = &a;
  _info->destroy = destroy_;
  _info->print = print_;
  _info->merge = merge_;
  _info->compare = compare_;
  EXPECT_ULEQ(2, _info->numKeys);
  EXPECT_ULEQ(0, _info->numValues);
  EXPECT_ULEQ(2, _info->numFields);
  EXPECT_ULEQ(6, _info->maxKeys[0]);
  EXPECT_ULEQ(24, _info->maxKeys[1]);
  EXPECT_PEQ(&a, _info->x);
  EXPECT_PEQ(destroy_, _info->destroy);
  EXPECT_PEQ(print_, _info->print);
  EXPECT_PEQ(merge_, _info->merge);
  EXPECT_PEQ(compare_, _info->compare);
  key_value_info_delete(&_info);
}

TEST(SetInfo, NewValues) {
  int a;
  ulong max_keys[] = {6, 24};
  KeyValueInfo *_info = null;

  set_info_new(&_info, 2, max_keys);
  _info->x = &a;
  _info->print = print_;
  _info->compare = compare_;
  EXPECT_ULEQ(2, _info->numKeys);
  EXPECT_ULEQ(0, _info->numValues);
  EXPECT_ULEQ(2, _info->numFields);
  EXPECT_ULEQ(6, _info->maxKeys[0]);
  EXPECT_ULEQ(24, _info->maxKeys[1]);
  EXPECT_PEQ(&a, _info->x);
  EXPECT_PEQ(print_, _info->print);
  EXPECT_PEQ(compare_, _info->compare);
  key_value_info_delete(&_info);
}

TEST(SetInfo, NewSimple) {
  ulong max_keys[] = {6, 24};
  KeyValueInfo *_info = null;

  set_info_new(&_info, 2, max_keys);
  EXPECT_ULEQ(2, _info->numKeys);
  EXPECT_ULEQ(0, _info->numValues);
  EXPECT_ULEQ(2, _info->numFields);
  EXPECT_ULEQ(6, _info->maxKeys[0]);
  EXPECT_ULEQ(24, _info->maxKeys[1]);
  key_value_info_delete(&_info);
}

TEST(SetCollection, Comprehensive) {
  ulong i;
  ulong j;
  ulong k;
  ulong num_keys;
  ulong num_pairs;
  ulong key;
  ulong single_key;
  ulong expected_num_finalized_records;
  ulong *_max_keys = null;
  KeyValueInfo *_info = null;
  KeyValueInfo *_linfo = null;
  KeyValueArray *_array = null;
  NaiveKeyValueCollection *_ncollection = null;
  KeyValueCollection *_collection = null;
  KeyValuePair *_pair = null;
  KeyValuePair *_lpair = null;
  KeyValuePair *pairj;
  KeyValuePair *_pairf = null;
  KeyValuePair *_lpairf = null;
  KeyValuePair *pair_found;
  Random *_random = null;
  sbyte (*compare)(KeyValuePair *a, KeyValuePair *b, KeyValueInfo *info);

  key_value_collection_set_chunk_size(CHUNK_SIZE_);
  random_new(&_random);

  for (i = 0; i < N_; i++) {
    num_pairs = random_modulo(_random, M_);
    num_keys = 1 + random_modulo(_random, 5);
    new(&_max_keys, num_keys);

    for (k = 0; k < num_keys; k++) {
      _max_keys[k] = random_modulo(
          _random,
          (uint)pow(num_pairs, 1.0 / num_keys)) + 1;
    }

    set_info_new(&_info, num_keys, _max_keys);
    compare = _info->compare;
    key_frequency_info_new(
        &_linfo,
        num_keys,
        _max_keys,
        1);
    key_value_collection_new(&_collection, _info);
    naive_key_value_collection_new(&_ncollection, _linfo, true);
    key_value_pair_new(&_pair, _info);
    key_value_pair_new(&_lpair, _linfo);
    key_value_pair_new(&_pairf, _info);
    key_value_pair_new(&_lpairf, _linfo);
    expected_num_finalized_records = 0;

    for (j = 0; j < num_pairs; j++) {
      for (k = 0; k < num_keys; k++) {
        key = random_modulo(_random, _max_keys[k] + 1);
        key_value_pair_set_key(_pair, _info, k, key);
        key_value_pair_set_key(_lpair, _linfo, k, key);
      }

      key_frequency_pair_set_frequency(_lpair, _linfo, 1);

      if (naive_key_value_collection_get_value(_ncollection, _lpair, 0) == 0) {
        expected_num_finalized_records++;
        naive_key_value_collection_add_values(_ncollection, _lpair);
      }

      key_value_collection_donate(_collection, _pair);
    }

    key_value_array_new_from_collection(&_array, &_collection);
    EXPECT_TRUE(key_value_array_is_full(_array));
    EXPECT_TRUE(_array->finalized);
    EXPECT_ULEQ(expected_num_finalized_records, _array->numPairs);

    for (j = 0, single_key = 0; j < _array->numPairs; j++, single_key++) {
      while (_ncollection->_values[0][single_key] == 0) {
        naive_key_value_collection_pair_from_single_key(
            _ncollection,
            single_key,
            _lpairf);

        for (k = 0; k < num_keys; k++) {
          key_value_pair_set_key(
              _pairf,
              _info,
              k,
              key_value_pair_get_key(_lpairf, _linfo, k));
        }

        pair_found = key_value_array_find(_array, _pairf);
        EXPECT_NULL(pair_found);
        single_key++;
      }

      pairj = key_value_array_get(_array, j);
      EXPECT_NONNULL(pairj);
      naive_key_value_collection_pair_from_single_key(
          _ncollection,
          single_key,
          _lpairf);

      for (k = 0; k < num_keys; k++) {
        key_value_pair_set_key(
            _pairf,
            _info,
            k,
            key_value_pair_get_key(_lpairf, _linfo, k));
      }

      pair_found = key_value_array_find(_array, _pairf);
      EXPECT_NONNULL(pair_found);
      EXPECT_ULEQ(0, compare(pairj, pair_found, _info));
    }

    key_value_pair_delete(&_lpair, _linfo);
    key_value_pair_delete(&_pair, _info);
    key_value_pair_delete(&_lpairf, _linfo);
    key_value_pair_delete(&_pairf, _info);
    key_value_array_delete(&_array);
    naive_key_value_collection_delete(&_ncollection);
    key_value_collection_delete(&_collection);
    key_value_info_delete(&_linfo);
    key_value_info_delete(&_info);
    deletev(&_max_keys);
  }

  random_delete(&_random);
}

FINISH

static void destroy_(KeyValuePair *pair, KeyValueInfo *info) {
}

static void print_(KeyValuePairPrintInfo *kvppi, Printer *p) {
}

static void merge_(KeyValuePair *acc, KeyValuePair *add, KeyValueInfo *info) {
}

static sbyte compare_(KeyValuePair *a, KeyValuePair *b, KeyValueInfo *info) {
  return 0;
}
