#include "paw_test.h"

static double c_(double d);

static double q_(double d, byte bits);

static double r_(double d, double m, byte bits);


START

TEST(Paw, Size) {
  ASSERT_ULEQ(2, sizeof(paw));
}

TEST(Paw, SpecialValues) {
  EXPECT_ULEQ(0x0100, PAW_MIN.u);
  EXPECT_NEAR(2.168404344971009e-19, double_from_paw(PAW_MIN), 1e-38);

  EXPECT_ULEQ(0x7eff, PAW_MAX.u);
  EXPECT_NEAR(1.841071527669059e19, double_from_paw(PAW_MAX), 1e4);

  EXPECT_ULEQ(0x0000, PAW_ZERO.u);
  EXPECT_DEQ(0, double_from_paw(PAW_ZERO));

  EXPECT_ULEQ(0x3f00, PAW_ONE.u);
  EXPECT_DEQ(1, double_from_paw(PAW_ONE));

  EXPECT_ULEQ(0x4000, PAW_TWO.u);
  EXPECT_DEQ(2, double_from_paw(PAW_TWO));

  EXPECT_ULEQ(0x3e00, PAW_HALF.u);
  EXPECT_DEQ(0.5, double_from_paw(PAW_HALF));

  EXPECT_ULEQ(0x7f00, PAW_POSINF.u);
  EXPECT_ULEQ(0xff00, PAW_NEGINF.u);
  EXPECT_ULEQ(0x7f01, PAW_NAN.u);
}

TEST(Paw, FromParts) {
  paw p;

  p = paw_from_parts(0, 0, 0);
  EXPECT_DEQ(0, double_from_paw(p));

  p = paw_from_parts(0, 0x3f, 0x0);
  EXPECT_DEQ(1, double_from_paw(p));

  p = paw_from_parts(0, 0x3f, 0x80);
  EXPECT_DEQ(1.5, double_from_paw(p));

  p = paw_from_parts(0, 0x3f, 0xc0);
  EXPECT_DEQ(1.75, double_from_paw(p));

  p = paw_from_parts(1, 0x3f, 0x0);
  EXPECT_DEQ(-1, double_from_paw(p));

  p = paw_from_parts(1, 0x3f, 0x80);
  EXPECT_DEQ(-1.5, double_from_paw(p));

  p = paw_from_parts(1, 0x3f, 0xc0);
  EXPECT_DEQ(-1.75, double_from_paw(p));

  p = paw_from_parts(0, 0x40, 0x0);
  EXPECT_DEQ(2, double_from_paw(p));

  p = paw_from_parts(0, 0x40, 0x80);
  EXPECT_DEQ(3, double_from_paw(p));

  p = paw_from_parts(0, 0x40, 0xc0);
  EXPECT_DEQ(3.5, double_from_paw(p));

  p = paw_from_parts(1, 0x40, 0x0);
  EXPECT_DEQ(-2, double_from_paw(p));

  p = paw_from_parts(1, 0x40, 0x80);
  EXPECT_DEQ(-3, double_from_paw(p));

  p = paw_from_parts(1, 0x40, 0xc0);
  EXPECT_DEQ(-3.5, double_from_paw(p));

  p = paw_from_parts(0, 0x3e, 0x0);
  EXPECT_DEQ(0.5, double_from_paw(p));

  p = paw_from_parts(0, 0x3e, 0x80);
  EXPECT_DEQ(0.75, double_from_paw(p));

  p = paw_from_parts(0, 0x3e, 0xc0);
  EXPECT_DEQ(0.875, double_from_paw(p));

  p = paw_from_parts(1, 0x3e, 0x0);
  EXPECT_DEQ(-0.5, double_from_paw(p));

  p = paw_from_parts(1, 0x3e, 0x80);
  EXPECT_DEQ(-0.75, double_from_paw(p));

  p = paw_from_parts(1, 0x3e, 0xc0);
  EXPECT_DEQ(-0.875, double_from_paw(p));
}

TEST(Double, Pawize) {
  EXPECT_DEQ(0, pawize_double(0));
  EXPECT_DEQ(1, pawize_double(1));
  EXPECT_DEQ(-1, pawize_double(-1));
  EXPECT_DEQ(0.5, pawize_double(0.5));
  EXPECT_DEQ(-0.5, pawize_double(-0.5));
  EXPECT_DEQ(DBL_POSINF, pawize_double(1e20));
  EXPECT_DEQ(DBL_NEGINF, pawize_double(-1e20));
  EXPECT_DEQ(0, pawize_double(1e-20));
  EXPECT_DEQ(0, pawize_double(-1e-20));
  EXPECT_DEQ(241152, pawize_double(240966));
  EXPECT_DEQ(-19660800, pawize_double(-19680707));
}

TEST(Paw, Comprehensive) {
  byte bits;
  byte e;
  byte n;
  ushort m;
  paw p;
  paw p2;
  double d;
  double d2;

  for (bits = 0; bits <= 8; bits++) {
    for (n = 0; n < 2; n++) {
      for (e = 0x1; e <= 0x7e; e++) {
        for (m = 0; m < 0x100; m += 1 << (8 - bits)) {
          p = paw_from_parts(n, e, m);
          d = double_from_paw(p);
          p2 = paw_make_quantized(d, bits);
          EXPECT_PAWEQ(p, p2);
          d2 = double_from_paw(p2);
          EXPECT_DEQ(d, d2);
        }
      }
    }
  }
}

TEST(Paw, ComparisonsRelational) {
  paw p1;
  paw p2;

  EXPECT_PAWEQ(PAW_ZERO, PAW_ZERO);
  EXPECT_FALSE(paw_gt(PAW_ZERO, PAW_ZERO));
  EXPECT_FALSE(paw_lt(PAW_ZERO, PAW_ZERO));
  EXPECT_TRUE(paw_ge(PAW_ZERO, PAW_ZERO));
  EXPECT_TRUE(paw_le(PAW_ZERO, PAW_ZERO));

  p1 = paw_make(12.3);
  p2 = paw_make(-3.45e-12);

  EXPECT_PAWEQ(p1, p1);
  EXPECT_FALSE(paw_gt(p1, p1));
  EXPECT_FALSE(paw_lt(p1, p1));
  EXPECT_TRUE(paw_ge(p1, p1));
  EXPECT_TRUE(paw_le(p1, p1));

  EXPECT_FALSE(paw_equals(p1, p2));
  EXPECT_TRUE(paw_gt(p1, p2));
  EXPECT_FALSE(paw_lt(p1, p2));
  EXPECT_TRUE(paw_ge(p1, p2));
  EXPECT_FALSE(paw_le(p1, p2));

  EXPECT_FALSE(paw_equals(p2, p1));
  EXPECT_FALSE(paw_gt(p2, p1));
  EXPECT_TRUE(paw_lt(p2, p1));
  EXPECT_FALSE(paw_ge(p2, p1));
  EXPECT_TRUE(paw_le(p2, p1));

  EXPECT_PAWEQ(p2, p2);
  EXPECT_FALSE(paw_gt(p2, p2));
  EXPECT_FALSE(paw_lt(p2, p2));
  EXPECT_TRUE(paw_ge(p2, p2));
  EXPECT_TRUE(paw_le(p2, p2));

  EXPECT_PAWEQ(p1, paw_max(p1, p1));
  EXPECT_PAWEQ(p1, paw_min(p1, p1));

  EXPECT_PAWEQ(p1, paw_max(p1, p2));
  EXPECT_PAWEQ(p2, paw_min(p1, p2));

  EXPECT_PAWEQ(p1, paw_max(p2, p1));
  EXPECT_PAWEQ(p2, paw_min(p2, p1));

  EXPECT_PAWEQ(p2, paw_max(p2, p2));
  EXPECT_PAWEQ(p2, paw_min(p2, p2));

  EXPECT_FALSE(paw_gt(PAW_ONE, PAW_NAN));
  EXPECT_FALSE(paw_lt(PAW_ONE, PAW_NAN));
  EXPECT_FALSE(paw_ge(PAW_ONE, PAW_NAN));
  EXPECT_FALSE(paw_le(PAW_ONE, PAW_NAN));

  EXPECT_FALSE(paw_gt(PAW_ONE, PAW_POSINF));
  EXPECT_TRUE(paw_lt(PAW_ONE, PAW_POSINF));
  EXPECT_FALSE(paw_ge(PAW_ONE, PAW_POSINF));
  EXPECT_TRUE(paw_le(PAW_ONE, PAW_POSINF));

  EXPECT_TRUE(paw_gt(PAW_ONE, PAW_NEGINF));
  EXPECT_FALSE(paw_lt(PAW_ONE, PAW_NEGINF));
  EXPECT_TRUE(paw_ge(PAW_ONE, PAW_NEGINF));
  EXPECT_FALSE(paw_le(PAW_ONE, PAW_NEGINF));
}

TEST(Paw, ExactIntegers) {
  EXPECT_DEQ(0, c_(0));

  EXPECT_DEQ(1, c_(1));
  EXPECT_DEQ(-1, c_(-1));

  EXPECT_DEQ(2, c_(2));
  EXPECT_DEQ(-2, c_(-2));

  EXPECT_DEQ(3, c_(3));
  EXPECT_DEQ(-3, c_(-3));

  EXPECT_DEQ(4, c_(4));
  EXPECT_DEQ(-4, c_(-4));

  EXPECT_DEQ(7, c_(7));
  EXPECT_DEQ(-7, c_(-7));
  EXPECT_DEQ(8, c_(8));
  EXPECT_DEQ(-8, c_(-8));
  EXPECT_DEQ(9, c_(9));
  EXPECT_DEQ(-9, c_(-9));

  EXPECT_DEQ(15, c_(15));
  EXPECT_DEQ(-15, c_(-15));
  EXPECT_DEQ(16, c_(16));
  EXPECT_DEQ(-16, c_(-16));
  EXPECT_DEQ(17, c_(17));
  EXPECT_DEQ(-17, c_(-17));

  EXPECT_DEQ(31, c_(31));
  EXPECT_DEQ(-31, c_(-31));
  EXPECT_DEQ(32, c_(32));
  EXPECT_DEQ(-32, c_(-32));
  EXPECT_DEQ(33, c_(33));
  EXPECT_DEQ(-33, c_(-33));

  EXPECT_DEQ(63, c_(63));
  EXPECT_DEQ(-63, c_(-63));
  EXPECT_DEQ(64, c_(64));
  EXPECT_DEQ(-64, c_(-64));
  EXPECT_DEQ(65, c_(65));
  EXPECT_DEQ(-65, c_(-65));

  EXPECT_DEQ(127, c_(127));
  EXPECT_DEQ(-127, c_(-127));
  EXPECT_DEQ(128, c_(128));
  EXPECT_DEQ(-128, c_(-128));
  EXPECT_DEQ(129, c_(129));
  EXPECT_DEQ(-129, c_(-129));

  EXPECT_DEQ(254, c_(254));
  EXPECT_DEQ(-254, c_(-254));
  EXPECT_DEQ(255, c_(255));
  EXPECT_DEQ(-255, c_(-255));
  EXPECT_DEQ(256, c_(256));
  EXPECT_DEQ(-256, c_(-256));
  EXPECT_DEQ(257, c_(257));
  EXPECT_DEQ(-257, c_(-257));
  EXPECT_DEQ(258, c_(258));
  EXPECT_DEQ(-258, c_(-258));

  EXPECT_DEQ(510, c_(510));
  EXPECT_DEQ(-510, c_(-510));
  EXPECT_DEQ(511, c_(511));
  EXPECT_DEQ(-511, c_(-511));
  EXPECT_DEQ(512, c_(512));
  EXPECT_DEQ(-512, c_(-512));
}

TEST(Paw, RoundedIntegers) {
  EXPECT_DEQ(512, c_(513));
  EXPECT_DEQ(-512, c_(-513));
  EXPECT_DEQ(514, c_(514));
  EXPECT_DEQ(-514, c_(-514));
  EXPECT_DEQ(516, c_(515));
  EXPECT_DEQ(-516, c_(-515));
  EXPECT_DEQ(516, c_(516));
  EXPECT_DEQ(-516, c_(-516));
  EXPECT_DEQ(516, c_(517));
  EXPECT_DEQ(-516, c_(-517));
  EXPECT_DEQ(518, c_(518));
  EXPECT_DEQ(-518, c_(-518));
  EXPECT_DEQ(520, c_(519));
  EXPECT_DEQ(-520, c_(-519));
  EXPECT_DEQ(520, c_(520));
  EXPECT_DEQ(-520, c_(-520));
  EXPECT_DEQ(520, c_(521));
  EXPECT_DEQ(-520, c_(-521));
  EXPECT_DEQ(522, c_(522));
  EXPECT_DEQ(-522, c_(-522));
  EXPECT_DEQ(524, c_(523));
  EXPECT_DEQ(-524, c_(-523));

  EXPECT_DEQ(1020, c_(1020));
  EXPECT_DEQ(-1020, c_(-1020));
  EXPECT_DEQ(1020, c_(1021));
  EXPECT_DEQ(-1020, c_(-1021));
  EXPECT_DEQ(1022, c_(1022));
  EXPECT_DEQ(-1022, c_(-1022));
  EXPECT_DEQ(1024, c_(1023));
  EXPECT_DEQ(-1024, c_(-1023));
  EXPECT_DEQ(1024, c_(1024));
  EXPECT_DEQ(-1024, c_(-1024));
  EXPECT_DEQ(1024, c_(1025));
  EXPECT_DEQ(-1024, c_(-1025));
  EXPECT_DEQ(1024, c_(1026));
  EXPECT_DEQ(-1024, c_(-1026));
  EXPECT_DEQ(1028, c_(1027));
  EXPECT_DEQ(-1028, c_(-1027));
  EXPECT_DEQ(1028, c_(1028));
  EXPECT_DEQ(-1028, c_(-1028));
  EXPECT_DEQ(1028, c_(1029));
  EXPECT_DEQ(-1028, c_(-1029));
  EXPECT_DEQ(1032, c_(1030));
  EXPECT_DEQ(-1032, c_(-1030));
  EXPECT_DEQ(1032, c_(1031));
  EXPECT_DEQ(-1032, c_(-1031));
  EXPECT_DEQ(1032, c_(1032));
  EXPECT_DEQ(-1032, c_(-1032));
  EXPECT_DEQ(1032, c_(1033));
  EXPECT_DEQ(-1032, c_(-1033));
  EXPECT_DEQ(1032, c_(1034));
  EXPECT_DEQ(-1032, c_(-1034));
  EXPECT_DEQ(1036, c_(1035));
  EXPECT_DEQ(-1036, c_(-1035));
}

TEST(Paw, MaxOverflow) {
  double d;

  d = double_from_parts(0, 1086, 0xff00000000000);
  EXPECT_DEQ(d, c_(d));
  d = double_from_parts(1, 1086, 0xff00000000000);
  EXPECT_DEQ(d, c_(d));

  EXPECT_TRUE(double_is_positive_infinity(
        c_(double_from_parts(0, 1086, 0xff80000000000))));
  EXPECT_TRUE(double_is_negative_infinity(
        c_(double_from_parts(1, 1086, 0xff80000000000))));
  EXPECT_TRUE(double_is_positive_infinity(
        c_(double_from_parts(0, 1087, 0x0))));
  EXPECT_TRUE(double_is_negative_infinity(
        c_(double_from_parts(1, 1087, 0x0))));
}

TEST(Paw, MinUnderflow) {
  double d;

  d = double_from_parts(0, 961, 0x0);
  EXPECT_DEQ(d, c_(d));
  d = double_from_parts(1, 1086, 0xff00000000000);
  EXPECT_DEQ(d, c_(d));

  EXPECT_DEQ(0, c_(double_from_parts(0, 960, 0xff7ffffffffff)));
  EXPECT_DEQ(0, c_(double_from_parts(1, 960, 0xff7ffffffffff)));
}

TEST(Paw, ExactFractions) {
  EXPECT_DEQ(1.0 / 2, c_(1.0 / 2));
  EXPECT_DEQ(-1.0 / 2, c_(-1.0 / 2));

  EXPECT_DEQ(3.0 / 4, c_(3.0 / 4));
  EXPECT_DEQ(-3.0 / 4, c_(-3.0 / 4));

  EXPECT_DEQ(1.0 / 4, c_(1.0 / 4));
  EXPECT_DEQ(-1.0 / 4, c_(-1.0 / 4));

  EXPECT_DEQ(7.0 / 8, c_(7.0 / 8));
  EXPECT_DEQ(-7.0 / 8, c_(-7.0 / 8));
  EXPECT_DEQ(1.0 / 8, c_(1.0 / 8));
  EXPECT_DEQ(-1.0 / 8, c_(-1.0 / 8));
}

TEST(Paw, Quantized) {
  EXPECT_DEQ(0, q_(0, 0));

  EXPECT_DEQ(1, q_(1, 0));
  EXPECT_DEQ(-1, q_(-1, 0));

  EXPECT_DEQ(2, q_(2, 0));
  EXPECT_DEQ(-2, q_(-2, 0));
  EXPECT_DEQ(2, q_(3, 0));
  EXPECT_DEQ(-2, q_(-3, 0));

  EXPECT_DEQ(4, q_(4, 0));
  EXPECT_DEQ(4, q_(5, 0));
  EXPECT_DEQ(4, q_(6, 0));
  EXPECT_DEQ(8, q_(7, 0));

  EXPECT_DEQ(8, q_(8, 0));
  EXPECT_DEQ(8, q_(9, 0));
  EXPECT_DEQ(8, q_(12, 0));
  EXPECT_DEQ(16, q_(13, 0));
  EXPECT_DEQ(16, q_(15, 0));

  EXPECT_DEQ(16, q_(16, 0));
  EXPECT_DEQ(16, q_(17, 0));
  EXPECT_DEQ(16, q_(24, 0));
  EXPECT_DEQ(32, q_(31, 0));

  EXPECT_DEQ(32, q_(32, 0));
  EXPECT_DEQ(32, q_(33, 0));
  EXPECT_DEQ(32, q_(48, 0));
  EXPECT_DEQ(64, q_(63, 0));

  EXPECT_DEQ(0, q_(0, 1));

  EXPECT_DEQ(1, q_(1, 1));
  EXPECT_DEQ(-1, q_(-1, 1));

  EXPECT_DEQ(2, q_(2, 1));
  EXPECT_DEQ(-2, q_(-2, 1));
  EXPECT_DEQ(3, q_(3, 1));
  EXPECT_DEQ(-3, q_(-3, 1));

  EXPECT_DEQ(4, q_(4, 1));
  EXPECT_DEQ(-4, q_(-4, 1));
  EXPECT_DEQ(4, q_(5, 1));
  EXPECT_DEQ(-4, q_(-5, 1));
  EXPECT_DEQ(6, q_(6, 1));
  EXPECT_DEQ(-6, q_(-6, 1));
  EXPECT_DEQ(8, q_(7, 1));
  EXPECT_DEQ(-8, q_(-7, 1));

  EXPECT_DEQ(8, q_(8, 1));
  EXPECT_DEQ(-8, q_(-8, 1));
  EXPECT_DEQ(8, q_(9, 1));
  EXPECT_DEQ(-8, q_(-9, 1));
  EXPECT_DEQ(8, q_(10, 1));
  EXPECT_DEQ(-8, q_(-10, 1));
  EXPECT_DEQ(12, q_(11, 1));
  EXPECT_DEQ(-12, q_(-11, 1));
  EXPECT_DEQ(12, q_(12, 1));
  EXPECT_DEQ(-12, q_(-12, 1));
  EXPECT_DEQ(12, q_(13, 1));
  EXPECT_DEQ(-12, q_(-13, 1));
  EXPECT_DEQ(16, q_(14, 1));
  EXPECT_DEQ(-16, q_(-14, 1));
  EXPECT_DEQ(16, q_(15, 1));
  EXPECT_DEQ(-16, q_(-15, 1));

  EXPECT_DEQ(16, q_(16, 1));
  EXPECT_DEQ(16, q_(17, 1));
  EXPECT_DEQ(16, q_(18, 1));
  EXPECT_DEQ(16, q_(19, 1));
  EXPECT_DEQ(16, q_(20, 1));
  EXPECT_DEQ(24, q_(21, 1));
  EXPECT_DEQ(24, q_(22, 1));
  EXPECT_DEQ(24, q_(23, 1));
  EXPECT_DEQ(24, q_(24, 1));
  EXPECT_DEQ(24, q_(25, 1));
  EXPECT_DEQ(24, q_(26, 1));
  EXPECT_DEQ(24, q_(27, 1));
  EXPECT_DEQ(32, q_(28, 1));
  EXPECT_DEQ(32, q_(29, 1));
  EXPECT_DEQ(32, q_(30, 1));
  EXPECT_DEQ(32, q_(31, 1));
  EXPECT_DEQ(32, q_(32, 1));

  EXPECT_DEQ(0, q_(0, 2));

  EXPECT_DEQ(1, q_(1, 2));
  EXPECT_DEQ(-1, q_(-1, 2));

  EXPECT_DEQ(2, q_(2, 2));
  EXPECT_DEQ(3, q_(3, 2));

  EXPECT_DEQ(4, q_(4, 2));
  EXPECT_DEQ(5, q_(5, 2));
  EXPECT_DEQ(6, q_(6, 2));
  EXPECT_DEQ(7, q_(7, 2));

  EXPECT_DEQ(8, q_(8, 2));
  EXPECT_DEQ(8, q_(9, 2));
  EXPECT_DEQ(10, q_(10, 2));
  EXPECT_DEQ(12, q_(11, 2));
  EXPECT_DEQ(12, q_(12, 2));
  EXPECT_DEQ(12, q_(13, 2));
  EXPECT_DEQ(14, q_(14, 2));
  EXPECT_DEQ(16, q_(15, 2));

  EXPECT_DEQ(16, q_(16, 2));
  EXPECT_DEQ(16, q_(17, 2));
  EXPECT_DEQ(16, q_(18, 2));
  EXPECT_DEQ(20, q_(19, 2));
  EXPECT_DEQ(20, q_(20, 2));
  EXPECT_DEQ(20, q_(21, 2));
  EXPECT_DEQ(24, q_(22, 2));
  EXPECT_DEQ(24, q_(23, 2));
  EXPECT_DEQ(24, q_(24, 2));
  EXPECT_DEQ(24, q_(25, 2));
  EXPECT_DEQ(24, q_(26, 2));
  EXPECT_DEQ(28, q_(27, 2));
  EXPECT_DEQ(28, q_(28, 2));
  EXPECT_DEQ(28, q_(29, 2));
  EXPECT_DEQ(32, q_(30, 2));
  EXPECT_DEQ(32, q_(31, 2));

  EXPECT_DEQ(32, q_(32, 2));
  EXPECT_DEQ(-32, q_(-32, 2));

  EXPECT_DEQ(0, q_(0, 7));

  EXPECT_DEQ(1, q_(1, 7));
  EXPECT_DEQ(-1, q_(-1, 7));

  EXPECT_DEQ(255, q_(255, 7));
  EXPECT_DEQ(-255, q_(-255, 7));

  EXPECT_DEQ(256, q_(256, 7));
  EXPECT_DEQ(256, q_(257, 7));
  EXPECT_DEQ(258, q_(258, 7));
  EXPECT_DEQ(260, q_(259, 7));
  EXPECT_DEQ(260, q_(260, 7));
  EXPECT_DEQ(260, q_(261, 7));
  EXPECT_DEQ(262, q_(262, 7));
  EXPECT_DEQ(264, q_(263, 7));
  EXPECT_DEQ(264, q_(264, 7));
}

TEST(Paw, RelativeQuantized) {
  EXPECT_DEQ(0, r_(0, 0, 0));
  EXPECT_DEQ(0, r_(0, 0, 1));
  EXPECT_DEQ(0, r_(0, 0, 2));
  EXPECT_DEQ(0, r_(0, 0, 3));
  EXPECT_DEQ(0, r_(0, 0, 4));
  EXPECT_DEQ(0, r_(0, 0, 5));
  EXPECT_DEQ(0, r_(0, 0, 6));
  EXPECT_DEQ(0, r_(0, 0, 7));
  EXPECT_DEQ(0, r_(0, 0, 8));

  EXPECT_DEQ(0, r_(0, 256, 8));
  EXPECT_DEQ(1, r_(1, 256, 8));
  EXPECT_DEQ(-1, r_(-1, 256, 8));
  EXPECT_DEQ(2, r_(2, 256, 8));
  EXPECT_DEQ(-2, r_(-2, 256, 8));
  EXPECT_DEQ(3, r_(3, 256, 8));
  EXPECT_DEQ(-3, r_(-3, 256, 8));

  EXPECT_DEQ(0, r_(0, 512, 8));
  EXPECT_DEQ(0, r_(1, 512, 8));
  EXPECT_DEQ(0, r_(-1, 512, 8));
  EXPECT_DEQ(2, r_(2, 512, 8));
  EXPECT_DEQ(-2, r_(-2, 512, 8));
  EXPECT_DEQ(4, r_(3, 512, 8));
  EXPECT_DEQ(-4, r_(-3, 512, 8));
  EXPECT_DEQ(4, r_(4, 512, 8));
  EXPECT_DEQ(-4, r_(-4, 512, 8));
  EXPECT_DEQ(4, r_(5, 512, 8));
  EXPECT_DEQ(-4, r_(-5, 512, 8));
  EXPECT_DEQ(6, r_(6, 512, 8));
  EXPECT_DEQ(-6, r_(-6, 512, 8));
  EXPECT_DEQ(8, r_(7, 512, 8));
  EXPECT_DEQ(-8, r_(-7, 512, 8));
  EXPECT_DEQ(8, r_(8, 512, 8));
  EXPECT_DEQ(-8, r_(-8, 512, 8));
  EXPECT_DEQ(8, r_(9, 512, 8));
  EXPECT_DEQ(-8, r_(-9, 512, 8));
  EXPECT_DEQ(10, r_(10, 512, 8));
  EXPECT_DEQ(-10, r_(-10, 512, 8));
  EXPECT_DEQ(12, r_(11, 512, 8));
  EXPECT_DEQ(-12, r_(-11, 512, 8));
  EXPECT_DEQ(12, r_(12, 512, 8));
  EXPECT_DEQ(-12, r_(-12, 512, 8));
  EXPECT_DEQ(12, r_(13, 512, 8));
  EXPECT_DEQ(-12, r_(-13, 512, 8));

  EXPECT_DEQ(0, r_(0, 128, 7));
  EXPECT_DEQ(1, r_(1, 128, 7));
  EXPECT_DEQ(-1, r_(-1, 128, 7));
  EXPECT_DEQ(2, r_(2, 128, 7));
  EXPECT_DEQ(-2, r_(-2, 128, 7));
  EXPECT_DEQ(3, r_(3, 128, 7));
  EXPECT_DEQ(-3, r_(-3, 128, 7));

  EXPECT_DEQ(0, r_(0, 256,7));
  EXPECT_DEQ(0, r_(1, 256,7));
  EXPECT_DEQ(0, r_(-1, 256,7));
  EXPECT_DEQ(2, r_(2, 256,7));
  EXPECT_DEQ(-2, r_(-2, 256,7));
  EXPECT_DEQ(4, r_(3, 256,7));
  EXPECT_DEQ(-4, r_(-3, 256,7));
  EXPECT_DEQ(4, r_(4, 256,7));
  EXPECT_DEQ(-4, r_(-4, 256,7));
  EXPECT_DEQ(4, r_(5, 256,7));
  EXPECT_DEQ(-4, r_(-5, 256,7));
  EXPECT_DEQ(6, r_(6, 256,7));
  EXPECT_DEQ(-6, r_(-6, 256,7));
  EXPECT_DEQ(8, r_(7, 256,7));
  EXPECT_DEQ(-8, r_(-7, 256,7));
  EXPECT_DEQ(8, r_(8, 256,7));
  EXPECT_DEQ(-8, r_(-8, 256,7));
  EXPECT_DEQ(8, r_(9, 256,7));
  EXPECT_DEQ(-8, r_(-9, 256,7));
  EXPECT_DEQ(10, r_(10, 256,7));
  EXPECT_DEQ(-10, r_(-10, 256,7));
  EXPECT_DEQ(12, r_(11, 256,7));
  EXPECT_DEQ(-12, r_(-11, 256,7));
  EXPECT_DEQ(12, r_(12, 256,7));
  EXPECT_DEQ(-12, r_(-12, 256,7));
  EXPECT_DEQ(12, r_(13, 256,7));
  EXPECT_DEQ(-12, r_(-13, 256,7));

  EXPECT_DEQ(256, r_(256, 256, 8));
  EXPECT_DEQ(-256, r_(-256, 256, 8));

  EXPECT_DEQ(512, r_(512, 512, 8));
  EXPECT_DEQ(-512, r_(-512, 512, 8));

  EXPECT_DEQ(1024, r_(1024, 1024, 8));
  EXPECT_DEQ(-1024, r_(-1024, 1024, 8));

  EXPECT_DEQ(127, r_(127, 127, 7));
  EXPECT_DEQ(-127, r_(-127, 127, 7));

  EXPECT_DEQ(254, r_(254, 254, 7));
  EXPECT_DEQ(-254, r_(-254, 254, 7));

  EXPECT_DEQ(63, r_(63, 63, 6));
  EXPECT_DEQ(-63, r_(-63, 63, 6));

  EXPECT_DEQ(126, r_(126, 126, 6));
  EXPECT_DEQ(-126, r_(-126, 126, 6));


  EXPECT_DEQ(256, r_(256, 256, 0));
  EXPECT_DEQ(-256, r_(-256, 256, 0));
  EXPECT_DEQ(256, r_(255, 256, 0));
  EXPECT_DEQ(-256, r_(-255, 256, 0));
  EXPECT_DEQ(256, r_(129, 256, 0));
  EXPECT_DEQ(-256, r_(-129, 256, 0));
  EXPECT_DEQ(0, r_(128, 256, 0));
  EXPECT_DEQ(0, r_(-128, 256, 0));
  EXPECT_DEQ(0, r_(1, 256, 0));
  EXPECT_DEQ(0, r_(-1, 256, 0));

  EXPECT_DEQ(256, r_(255, 255, 0));
  EXPECT_DEQ(-256, r_(-255, 255, 0));
  EXPECT_DEQ(256, r_(130, 255, 0));
  EXPECT_DEQ(-256, r_(-130, 255, 0));
  EXPECT_DEQ(0, r_(129, 255, 0));
  EXPECT_DEQ(0, r_(-129, 255, 0));

  EXPECT_DEQ(128, r_(128, 128, 0));
  EXPECT_DEQ(128, r_(65, 128, 0));
  EXPECT_DEQ(0, r_(64, 128, 0));

  EXPECT_DEQ(128, r_(127, 127, 0));
  EXPECT_DEQ(128, r_(66, 127, 0));
  EXPECT_DEQ(0, r_(65, 127, 0));

  EXPECT_DEQ(128, r_(126, 126, 0));
  EXPECT_DEQ(128, r_(67, 126, 0));
  EXPECT_DEQ(0, r_(66, 126, 0));

  EXPECT_DEQ(256, r_(193, 193, 0));
  EXPECT_DEQ(128, r_(192, 192, 0));
  EXPECT_DEQ(128, r_(165, 165, 0));

  EXPECT_DEQ(128, r_(164, 165, 0));
  EXPECT_DEQ(128, r_(28, 165, 0));
  EXPECT_DEQ(0, r_(27, 165, 0));

  EXPECT_DEQ(192, r_(165, 165, 1));
  EXPECT_DEQ(192, r_(156, 165, 1));
  EXPECT_DEQ(64, r_(155, 165, 1));
  EXPECT_DEQ(64, r_(59, 165, 1));
  EXPECT_DEQ(0, r_(58, 165, 1));
  EXPECT_DEQ(0, r_(0, 165, 1));
}

TEST(Paw, Special) {
  paw p;

  EXPECT_DEQ(0, c_(DBL_SUBNORMAL));
  EXPECT_TRUE(double_is_positive_infinity(c_(DBL_POSINF)));
  EXPECT_TRUE(double_is_negative_infinity(c_(DBL_NEGINF)));
  EXPECT_TRUE(double_is_nan(c_(DBL_NAN)));

  p = PAW_ZERO;
  EXPECT_TRUE(paw_is_zero(p));
  EXPECT_FALSE(paw_is_positive(p));
  EXPECT_FALSE(paw_is_negative(p));
  EXPECT_FALSE(paw_is_positive_infinity(p));
  EXPECT_FALSE(paw_is_negative_infinity(p));
  EXPECT_FALSE(paw_is_nan(p));
  EXPECT_FALSE(paw_is_infinite_or_nan(p));

  p = paw_make(0);
  EXPECT_TRUE(paw_is_zero(p));
  EXPECT_FALSE(paw_is_positive(p));
  EXPECT_FALSE(paw_is_negative(p));
  EXPECT_FALSE(paw_is_positive_infinity(p));
  EXPECT_FALSE(paw_is_negative_infinity(p));
  EXPECT_FALSE(paw_is_nan(p));
  EXPECT_FALSE(paw_is_infinite_or_nan(p));

  p = paw_make(1);
  EXPECT_FALSE(paw_is_zero(p));
  EXPECT_TRUE(paw_is_positive(p));
  EXPECT_FALSE(paw_is_negative(p));
  EXPECT_FALSE(paw_is_positive_infinity(p));
  EXPECT_FALSE(paw_is_negative_infinity(p));
  EXPECT_FALSE(paw_is_nan(p));
  EXPECT_FALSE(paw_is_infinite_or_nan(p));

  p = paw_make(-1);
  EXPECT_FALSE(paw_is_zero(p));
  EXPECT_FALSE(paw_is_positive(p));
  EXPECT_TRUE(paw_is_negative(p));
  EXPECT_FALSE(paw_is_positive_infinity(p));
  EXPECT_FALSE(paw_is_negative_infinity(p));
  EXPECT_FALSE(paw_is_nan(p));
  EXPECT_FALSE(paw_is_infinite_or_nan(p));

  p = paw_make(DBL_POSINF);
  EXPECT_FALSE(paw_is_zero(p));
  EXPECT_TRUE(paw_is_positive(p));
  EXPECT_FALSE(paw_is_negative(p));
  EXPECT_TRUE(paw_is_positive_infinity(p));
  EXPECT_FALSE(paw_is_negative_infinity(p));
  EXPECT_FALSE(paw_is_nan(p));
  EXPECT_TRUE(paw_is_infinite_or_nan(p));

  p = paw_make(DBL_NEGINF);
  EXPECT_FALSE(paw_is_zero(p));
  EXPECT_FALSE(paw_is_positive(p));
  EXPECT_TRUE(paw_is_negative(p));
  EXPECT_FALSE(paw_is_positive_infinity(p));
  EXPECT_TRUE(paw_is_negative_infinity(p));
  EXPECT_FALSE(paw_is_nan(p));
  EXPECT_TRUE(paw_is_infinite_or_nan(p));

  p = paw_make(DBL_NAN);
  EXPECT_FALSE(paw_is_zero(p));
  EXPECT_FALSE(paw_is_positive(p));
  EXPECT_FALSE(paw_is_negative(p));
  EXPECT_FALSE(paw_is_positive_infinity(p));
  EXPECT_FALSE(paw_is_negative_infinity(p));
  EXPECT_TRUE(paw_is_nan(p));
  EXPECT_TRUE(paw_is_infinite_or_nan(p));

  p = PAW_POSINF;
  EXPECT_FALSE(paw_is_zero(p));
  EXPECT_TRUE(paw_is_positive(p));
  EXPECT_FALSE(paw_is_negative(p));
  EXPECT_TRUE(paw_is_positive_infinity(p));
  EXPECT_FALSE(paw_is_negative_infinity(p));
  EXPECT_FALSE(paw_is_nan(p));
  EXPECT_TRUE(paw_is_infinite_or_nan(p));

  p = PAW_NEGINF;
  EXPECT_FALSE(paw_is_zero(p));
  EXPECT_FALSE(paw_is_positive(p));
  EXPECT_TRUE(paw_is_negative(p));
  EXPECT_FALSE(paw_is_positive_infinity(p));
  EXPECT_TRUE(paw_is_negative_infinity(p));
  EXPECT_FALSE(paw_is_nan(p));
  EXPECT_TRUE(paw_is_infinite_or_nan(p));

  p = PAW_NAN;
  EXPECT_FALSE(paw_is_zero(p));
  EXPECT_FALSE(paw_is_positive(p));
  EXPECT_FALSE(paw_is_negative(p));
  EXPECT_FALSE(paw_is_positive_infinity(p));
  EXPECT_FALSE(paw_is_negative_infinity(p));
  EXPECT_TRUE(paw_is_nan(p));
  EXPECT_TRUE(paw_is_infinite_or_nan(p));

  p = paw_make(DBL_POSINF);
  EXPECT_FALSE(paw_is_zero(p));
  EXPECT_TRUE(paw_is_positive(p));
  EXPECT_FALSE(paw_is_negative(p));
  EXPECT_TRUE(paw_is_positive_infinity(p));
  EXPECT_FALSE(paw_is_negative_infinity(p));
  EXPECT_FALSE(paw_is_nan(p));
  EXPECT_TRUE(paw_is_infinite_or_nan(p));

  p = paw_make(DBL_NEGINF);
  EXPECT_FALSE(paw_is_zero(p));
  EXPECT_FALSE(paw_is_positive(p));
  EXPECT_TRUE(paw_is_negative(p));
  EXPECT_FALSE(paw_is_positive_infinity(p));
  EXPECT_TRUE(paw_is_negative_infinity(p));
  EXPECT_FALSE(paw_is_nan(p));
  EXPECT_TRUE(paw_is_infinite_or_nan(p));

  p = paw_make(DBL_NAN);
  EXPECT_FALSE(paw_is_zero(p));
  EXPECT_FALSE(paw_is_positive(p));
  EXPECT_FALSE(paw_is_negative(p));
  EXPECT_FALSE(paw_is_positive_infinity(p));
  EXPECT_FALSE(paw_is_negative_infinity(p));
  EXPECT_TRUE(paw_is_nan(p));
  EXPECT_TRUE(paw_is_infinite_or_nan(p));
}

TEST(Double, PawizedIsInfiniteOrNan) {
  EXPECT_FALSE(double_pawized_is_infinite_or_nan(0));
  EXPECT_FALSE(double_pawized_is_infinite_or_nan(double_from_paw(PAW_MAX)));
  EXPECT_FALSE(double_pawized_is_infinite_or_nan(-double_from_paw(PAW_MAX)));
  EXPECT_TRUE(
      double_pawized_is_infinite_or_nan(double_from_paw(PAW_MAX) * 1.001));
  EXPECT_TRUE(
      double_pawized_is_infinite_or_nan(-double_from_paw(PAW_MAX) * 1.001));
  EXPECT_TRUE(double_pawized_is_infinite_or_nan(DBL_POSINF));
  EXPECT_TRUE(double_pawized_is_infinite_or_nan(DBL_NEGINF));
  EXPECT_TRUE(double_pawized_is_infinite_or_nan(DBL_NAN));
}

TEST(Paw, Arithmetic) {
  paw p1;
  paw p2;
  paw p3;

  p1 = paw_make(7.5);
  p2 = paw_make(-3.25);
  p3 = paw_make(-1.5);

  EXPECT_PAWEQ(PAW_ZERO, paw_negate(PAW_ZERO));
  EXPECT_PAWEQ(PAW_NAN, paw_negate(PAW_NAN));
  EXPECT_PAWEQ(paw_make(-7.5), paw_negate(p1));
  EXPECT_PAWEQ(paw_make(3.25), paw_negate(p2));
  EXPECT_PAWEQ(paw_make(1.5), paw_negate(p3));
  EXPECT_PAWEQ(PAW_MAX, paw_negate(paw_negate(PAW_MAX)));
  EXPECT_PAWEQ(PAW_MIN, paw_negate(paw_negate(PAW_MIN)));

  EXPECT_PAWEQ(PAW_ZERO, paw_abs(PAW_ZERO));
  EXPECT_PAWEQ(PAW_ONE, paw_abs(PAW_ONE));
  EXPECT_PAWEQ(PAW_ONE, paw_abs(paw_negate(PAW_ONE)));
  EXPECT_PAWEQ(PAW_TWO, paw_abs(PAW_TWO));
  EXPECT_PAWEQ(PAW_TWO, paw_abs(paw_negate(PAW_TWO)));
  EXPECT_PAWEQ(PAW_HALF, paw_abs(PAW_HALF));
  EXPECT_PAWEQ(PAW_HALF, paw_abs(paw_negate(PAW_HALF)));
  EXPECT_PAWEQ(paw_make(3.14), paw_abs(paw_make(3.14)));
  EXPECT_PAWEQ(paw_make(3.14), paw_abs(paw_make(-3.14)));
  EXPECT_PAWEQ(PAW_POSINF, paw_abs(PAW_POSINF));
  EXPECT_PAWEQ(PAW_POSINF, paw_abs(PAW_NEGINF));

  EXPECT_PAWEQ(paw_make(4.25), paw_add(p1, p2));

  EXPECT_PAWEQ(paw_make(10.75), paw_subtract(p1, p2));
  EXPECT_PAWEQ(paw_make(-10.75), paw_subtract(p2, p1));

  EXPECT_PAWEQ(paw_make(2.25), paw_multiply(p3, p3));
  EXPECT_PAWEQ(paw_make(-6.5), paw_multiply(p2, paw_make(2)));
  EXPECT_PAWEQ(p2, paw_multiply(p2, PAW_ONE));

  EXPECT_PAWEQ(paw_make(-1.625), paw_divide(p2, paw_make(2)));

  EXPECT_PAWEQ(p2, paw_bound(p1, p2, p2));
  EXPECT_PAWEQ(p2, paw_bound(p1, paw_make(-10), p2));
  EXPECT_PAWEQ(p1, paw_bound(p2, p1, p1));
  EXPECT_PAWEQ(p1, paw_bound(p2, p1, paw_make(10)));
}

FINISH


static double c_(double d) {
  return double_from_paw(paw_make(d));
}

static double q_(double d, byte bits) {
  return double_from_paw(paw_make_quantized(d, bits));
}

static double r_(double d, double m, byte bits) {
  return double_from_paw(paw_make_relative_quantized(d, m, bits));
}
